/**
 * @file aes.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Шифрование по алгоритмам AES128/192/256 в том числе для режимов CBC, CTR, CCM.
 * @version 0.1
 * @date 2024-04-25
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef AES_H
#define AES_H

#include <cdefs.h>
#include <stdint.h>
#include <stddef.h>

__BEGIN_DECLS

// #define TEST_AES

#define AES_BLOCK_SIZE  16               ///< Размер блока данныхх AES

///////////////////////////////////////////////////////////////////////////////////////////////
// AES
///////////////////////////////////////////////////////////////////////////////////////////////

/**
 * @brief Установка ключа. Должна быть выполнена перед любыми операциями шифрования/дешифрования.
 * @param key Ключ, должен быть размером 128, 192 или 256 бит.
 * @param w Контекст шифрования
 * @param keysize Битовая длина ключа, должна быть 128, 192 или 256.
 */
void aes_key_setup(const uint8_t key[], uint32_t w[], int keysize);

/**
 * @brief Шифрование AES.
 * @param in Незашифрованные данные (16 байт).
 * @param out Зашифрованные данные (16 байт).
 * @param key Контекст шифрования (результат aes_key_setup).
 * @param keysize Битовая длина ключа, должна быть 128, 192 или 256.
 */
void aes_encrypt(const uint8_t in[], uint8_t out[], const uint32_t key[], int keysize);

/**
 * @brief Дешифрование AES.
 * @param in Зашифрованные данные (16 байт).
 * @param out Расшифрованные данные (16 байт).
 * @param key Контекст шифрования (результат aes_key_setup).
 * @param keysize Битовая длина ключа, должна быть 128, 192 или 256.
 */
void aes_decrypt(const uint8_t in[], uint8_t out[], const uint32_t key[], int keysize);

///////////////////////////////////////////////////////////////////////////////////////////////
// AES в режиме CBC
///////////////////////////////////////////////////////////////////////////////////////////////

/**
 * @brief Шифрование AES в режиме CBC.
 * @param in Незашифрованные данные (размер кратен 16 байтам).
 * @param in_len Размер входных данных (размер кратен 16 байтам).
 * @param out Зашифрованные данные.
 * @param key Контекст шифрования (результат aes_key_setup).
 * @param keysize Битовая длина ключа, должна быть 128, 192 или 256.
 * @param iv Вектор инициализации шифрования, должен быть размером AES_BLOCK_SIZE (16 байт).
 * @return int 
 */
int aes_encrypt_cbc(const uint8_t in[], size_t in_len, uint8_t out[], const uint32_t key[], int keysize, const uint8_t iv[]);

/**
 * @brief Расчёт значения CBC-MAC входных данных.
 * 
 * @param in Незашифрованные данные (размер кратен 16 байтам).
 * @param in_len Размер входных данных (размер кратен 16 байтам).
 * @param out Вычисленное значение CBC-MAC
 * @param key Контекст шифрования (результат aes_key_setup).
 * @param keysize Битовая длина ключа, должна быть 128, 192 или 256.
 * @param iv Вектор инициализации шифрования, должен быть размером AES_BLOCK_SIZE (16 байт).
 * @return int 
 */
int aes_encrypt_cbc_mac(const uint8_t in[], size_t in_len, uint8_t out[], const uint32_t key[], int keysize, const uint8_t iv[]);

///////////////////////////////////////////////////////////////////////////////////////////////
// AES в режиме CTR
///////////////////////////////////////////////////////////////////////////////////////////////

/**
 * @brief 
 * 
 * @param iv Вектор инициализации шифрования, размер должен быть кратен AES_BLOCK_SIZE (16 байт).
 * @param counter_size Bytes of the IV used for counting (low end).
 */
void increment_iv(uint8_t iv[], int counter_size);

/**
 * @brief Шифрование AES-CTR.
 * @param in Незашифрованные данные.
 * @param in_len Размер входных данных (любой размер).
 * @param out Зашифрованные данные.
 * @param key Контекст шифрования (результат aes_key_setup).
 * @param keysize Битовая длина ключа, должна быть 128, 192 или 256.
 * @param iv Вектор инициализации шифрования, должен быть размером AES_BLOCK_SIZE (16 байт).
 */
void aes_encrypt_ctr(const uint8_t in[], size_t in_len, uint8_t out[], const uint32_t key[], int keysize, const uint8_t iv[]);

/**
 * @brief Расшифровка AES-CTR.
 * @param in Зашифрованные данные.
 * @param in_len Размер входных данных (любой размер).
 * @param out Расшифрованные данные.
 * @param key Контекст шифрования (результат aes_key_setup).
 * @param keysize Битовая длина ключа, должна быть 128, 192 или 256.
 * @param iv Вектор инициализации шифрования, должен быть размером AES_BLOCK_SIZE (16 байт).
 */
void aes_decrypt_ctr(const uint8_t in[], size_t in_len, uint8_t out[], const uint32_t key[], int keysize, const uint8_t iv[]);

///////////////////////////////////////////////////////////////////////////////////////////////
// AES в режиме CCM
///////////////////////////////////////////////////////////////////////////////////////////////

/**
 * @brief Шифрование AES-CCM.
 * Возвращает true, если входные данные корректны.
 * @param plaintext IN - Незашифрованные данные.
 * @param plaintext_len IN - Размер входных данных (любой размер).
 * @param associated_data IN - Связанные данные: включаются в аутентификацию, но не в шифрование.
 * @param associated_data_len IN - Размер связанных данных.
 * @param nonce IN - Одноразовый номер, который будет использоваться для шифрования.
 * @param nonce_len IN - Размер одноразового номера.
 * @param ciphertext OUT - Зашифрованное сообщение: объединение сообщения и MAC.
 * @param ciphertext_len OUT - Размер зашифрованного сообщения: plaintext_len + mac_len.
 * @param mac_len IN - Необходимый размер MAC, должен быть 4, 6, 8, 10, 12, 14 или 16.
 * @param key IN - Ключ для шифрования.
 * @param keysize IN - Битовая длина ключа, должна быть 128, 192 или 256.
 * @return int 
 */
int aes_encrypt_ccm(const uint8_t plaintext[], uint32_t plaintext_len, const uint8_t associated_data[], unsigned short associated_data_len, const uint8_t nonce[], unsigned short nonce_len, uint8_t ciphertext[], uint32_t *ciphertext_len, uint32_t mac_len, const uint8_t key[], int keysize);

/**
 * @brief Расшифровка AES-CCM.
 * Возвращает true, если входные данные корректны.
 * Используйте mac_auth, чтобы убедиться, что расшифровка/проверка подлинности были выполнены правильно.
 * Если аутентификация не удалась, открытый текст обнуляется. Чтобы переопределить это, вызовите с mac_auth = NULL.
 * Правильная процедура заключается в расшифровке с включенной аутентификацией (mac_auth != NULL) и выполнении второго вызова,
 * который явно игнорирует аутентификацию, если первый вызов завершился неудачей.
 * @param ciphertext IN - Зашифрованный текст - объединение зашифрованного открытого текста и MAC.
 * @param ciphertext_len IN - Размер ciphertext в байтах.
 * @param assoc IN - Связанные данные: включаются в аутентификацию, но не в шифрование.
 * @param assoc_len IN - Размер связанных данных.
 * @param nonce IN - Одноразовый номер, такой же, как при шифровании.
 * @param nonce_len IN - Размер одноразового номера.
 * @param plaintext OUT - Расшифрованное сообщение. Размер должен вместить ciphertext_len - mac_len.
 * @param plaintext_len OUT - Размер расшифрованного сообщения.
 * @param mac_len IN - Длина вычисленного значения MAC.
 * @param mac_auth OUT - TRUE, если аутентификация прошла успешно или FALSE, если нет. Нулевой указатель проигнорирует аутентификацию.
 * @param key IN - Ключ для расшифровки.
 * @param keysize IN - Битовая длина ключа, должна быть 128, 192 или 256.
 * @return int 
 */
int aes_decrypt_ccm(const uint8_t ciphertext[], uint32_t ciphertext_len, const uint8_t assoc[], unsigned short assoc_len, const uint8_t nonce[], unsigned short nonce_len, uint8_t plaintext[], uint32_t *plaintext_len, uint32_t mac_len, int *mac_auth, const uint8_t key[], int keysize);

#ifdef TEST_AES

// Тестирование
int test_aes();

#endif

__END_DECLS

#endif
