/**
 * @file cjson.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с JSON.
 * @version 0.1
 * @date 2024-05-16
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 * Библиотека разработана на основе оригинальных исходных текстов 
 * библиотеки cJSON (https://github.com/DaveGamble/cJSON).
 * @license MIT
 * @copyright Copyright (c) 2009-2017 Dave Gamble and cJSON contributors
 * @copyright Copyright (c) 2015-2016 The Khronos Group Inc.
 * @copyright Copyright (c) 2015-2016 Valve Corporation
 * @copyright Copyright (c) 2015-2016 LunarG, Inc.
 * 
 * @example 1. Пример разбора строки JSON
    char json_str[] = "{ \
        \"Str\": \"str1\",      \
        \"Num\":  1.2345,       \
        \"Arr\":  [1,2,3,4],    \
        \"Bool\": true,         \
        \"N\": null,            \
        \"Obj1\":{ \
            \"Str\": \"str2\",  \
            \"Num\":  6.789,    \
            \"Arr\":  [1.1,2.1,3.1,4.1], \
            \"Bool\": false, \
            \"N\": null \
        }, \
        \"Obj2\":{ \
            \"Str\": \"str3\",  \
            \"Num\":  3.14,     \
            \"Arr\":  [7,8,9,0], \
            \"Bool\": false, \
            \"N\": null \
        } \
    }";
    //
    // разбор строки с текстовым представлением JSON
    cJSON *json = cJSON_Parse(json_str);
    if(json) {
        char empty[] = "not found";
        //
        printf("JSON parse OK!\n");
        //
        // Доступ к полям JSON
        printf("Str: %s\n", cJSON_GetStringByPath(json, "Str", empty));             // "str1"
        printf("Num: %f\n", cJSON_GetDoubleByPath(json, "Num", -1));                // 1.2345
        printf("Arr: %d,%d,%d,%d\n",                                                // 1,2,3,4
            cJSON_GetIntByPath(json, "Arr/0", -1),
            cJSON_GetIntByPath(json, "Arr/1", -1),
            cJSON_GetIntByPath(json, "Arr/2", -1),
            cJSON_GetIntByPath(json, "Arr/3", -1));
        printf("Bool: %d\n", cJSON_GetBoolByPath(json, "Bool", -1));                // 1
        //
        printf("Obj1/Str: %s\n", cJSON_GetStringByPath(json, "Obj1/Str", empty));   // "str2"
        printf("Obj1/Num: %f\n", cJSON_GetDoubleByPath(json, "Obj1/Num", -1));      // 6.789
        printf("Obj1/Arr: %.1f,%.1f,%.1f,%.1f\n",                                   // 1.1,1.2,1.3,1.4
            cJSON_GetDoubleByPath(json, "Obj1/Arr/0", -1),
            cJSON_GetDoubleByPath(json, "Obj1/Arr/1", -1),
            cJSON_GetDoubleByPath(json, "Obj1/Arr/2", -1),
            cJSON_GetDoubleByPath(json, "Obj1/Arr/3", -1));
        printf("Obj1/Bool: %d\n", cJSON_GetBoolByPath(json, "Obj1/Bool", -1));      // 0
        //
        printf("Obj2/Str: %s\n", cJSON_GetStringByPath(json, "Obj2/Str", empty));   // "str3"
        printf("Obj2/Num: %f\n", cJSON_GetDoubleByPath(json, "Obj2/Num", -1));      // 3.14
        printf("Obj2/Arr: %d,%d,%d,%d\n",                                           // 7,8,9,0
            cJSON_GetIntByPath(json, "Obj2/Arr/0", -1),
            cJSON_GetIntByPath(json, "Obj2/Arr/1", -1),
            cJSON_GetIntByPath(json, "Obj2/Arr/2", -1),
            cJSON_GetIntByPath(json, "Obj2/Arr/3", -1));
        printf("Obj2/Bool: %d\n", cJSON_GetBoolByPath(json, "Obj2/Bool", -1));      // 0
        //
        // освобождаем память 
        cJSON_Delete(json);
    } else {
        //
        // при разборе строки возникла ошибка
        printf("Error parse JSON:\n");
        printf(HRED "%s" CRESET, cJSON_GetErrorPtr());
    }
 * 
 * @example 2. Конструирование JSON
    int intarr[4]   = {1, 2, 3, 4};
    float fltarr[4] = {1.1, 2.2, 3.3, 4.4};
    //
    // создаём новый объект
    cJSON *json = cJSON_CreateObject();
    // добавляем в него поля
    cJSON_AddStringToObject(json, "name", "Иванов");
    cJSON_AddNumberToObject(json, "age", 30);
    cJSON_AddBoolToObject(json, "f", 0);
    cJSON_AddNullToObject(json, "n");
    cJSON_AddItemToObject(json, "arri", cJSON_CreateIntArray(intarr, 4));
    // создаём дочерний объект
    cJSON *json_child = cJSON_CreateObject();
    cJSON_AddStringToObject(json_child, "name", "Петров");
    cJSON_AddNumberToObject(json_child, "age", 31);
    cJSON_AddBoolToObject(json_child, "t", 1);
    cJSON_AddNullToObject(json_child, "n");
    cJSON_AddItemToObject(json_child, "arrf", cJSON_CreateFloatArray(fltarr, 4));
    // добавляем дочерний объект к основному
    cJSON_AddItemToObject(json, "obj", json_child);
    //
    // преобразуем в строку (память строки потом необходимо освободить)
    char *str = cJSON_PrintUnformatted(json);
    printf("%s\n", str);
    //
    // освобождаем память
    free(str);
    cJSON_Delete(json);
 * 
 */

#ifndef CJSON_H
#define CJSON_H

#include <cdefs.h>

__BEGIN_DECLS

/**
 * @brief Типы данных JSON
 */
enum cJSON_Type {
    cJSON_False  =          0,              ///< Тип JSON: boolean false
    cJSON_True   =          1,              ///< Тип JSON: boolean true
    cJSON_NULL   =          2,              ///< Тип JSON: null
    cJSON_Number =          3,              ///< Тип JSON: число
    cJSON_String =          4,              ///< Тип JSON: строка
    cJSON_Array  =          5,              ///< Тип JSON: массив
    cJSON_Object =          6,              ///< Тип JSON: объект
    cJSON_IsReference   =   256,            ///< Тип JSON: ссылка
    cJSON_StringIsConst =   512
};

/**
 * @brief Структура описания элемента JSON.
 */
typedef struct cJSON {
    struct cJSON *next, *prev;              ///< Следующий/предыдущий элемент массива или объекта
    struct cJSON *child;                    ///< Дочерний элемент.
    int type;                               ///< Тип элемента: перечисление cJSON_Type
    char *valuestring;                      ///< Строковое значение элемента, если тип элемента cJSON_String
    int valueint;                           ///< Числовое (целое) значение элемента, если тип элемента cJSON_Number
    double valuedouble;                     ///< Числовое (плавающая точка) значение элемента, если тип элемента cJSON_Number
    char *string;                           ///< Имя элемента, если этот элемент является дочерним по отношению к объекту или находится в списке подпунктов объекта.
} cJSON;

/**
 * @brief Разбор строки с выражением JSON в структуры cJSON.
 * Функция выполняет разбор строкового представления JSON, выделяет необходимую память и создаёт в ней связанную структуру элементов cJSON.
 * После окончания работы для освобождения памяти необходимо вызвать функцию cJSON_Delete().
 * При ошибке разбора функция возвращает NULL, для анализа причины можно использовать функцию cJSON_GetErrorPtr().
 * @param value Указатель на строку.
 * @return cJSON* 
 */
cJSON *cJSON_Parse(const char *value);

/**
 * @brief Освобождает память выделенную для элемента cJSON и всех его подэлементов.
 * @param c Указатель на структуру cJSON.
 */
void cJSON_Delete(cJSON *c);

/**
 * @brief Возвращает указатель на участок строки, который вызвал ошибку при разборе строки JSON (вызове cJSON_Parse).
 * Обычно причина ошибки кроется в первых нескольких символах строки.
 * @return const char* 
 */
const char *cJSON_GetErrorPtr(void);

/**
 * @brief Поиск элемента JSON по текcтовому пути.
 * Путь представляет собой строку, вида: "xxx/yyy/nnn", где xxx,yyy,nnn наименования элементов объекта или индексы элементов массива.
 * Функция возвращает указатель на элемент по указанному пути. Если элемент отсутствует, то функция возвращает NULL.
 * @param item Элемент с которого начинается поиск.
 * @param path Путь к элементу.
 * @return cJSON* 
 */
cJSON *cJSON_ByPath(cJSON *item, char *path);

/**
 * @brief Получение значения типа "int" для элемента JSON путь к которому указан в параметре path.
 * Если указанный элемент присутствует и его тип соответствует требуемому, то функция возвращает его значение.
 * В противном случае функция возвращает значение defval.
 * Если параметр path пустая строка или NULL, то функция возвращает значение элемента item.
 * @param item Указатель на структуру cJSON.
 * @param path Путь к элементу JSON или NULL.
 * @param defval Значение по умолчанию.
 * @return int 
 */
int cJSON_GetIntByPath(cJSON *item, char *path, int defval);

/**
 * @brief Получение значения типа "bool" для элемента JSON путь к которому указан в параметре path.
 * Если указанный элемент присутствует и его тип соответствует требуемому, то функция возвращает его значение: 1-true, 0-false.
 * В противном случае функция возвращает значение defval.
 * Если параметр path пустая строка или NULL, то функция возвращает значение элемента item.
 * @param item Указатель на структуру cJSON.
 * @param path Путь к элементу JSON или NULL.
 * @param defval Значение по умолчанию.
 * @return int 
 */
int cJSON_GetBoolByPath(cJSON *item, char *path, int defval);

/**
 * @brief Получение значения типа "double" для элемента JSON путь к которому указан в параметре path.
 * Если указанный элемент присутствует и его тип соответствует требуемому, то функция возвращает его значение.
 * В противном случае функция возвращает значение defval.
 * Если параметр path пустая строка или NULL, то функция возвращает значение элемента item.
 * @param item Указатель на структуру cJSON.
 * @param path Путь к элементу JSON или NULL.
 * @param defval Значение по умолчанию.
 * @return int 
 */
double cJSON_GetDoubleByPath(cJSON *item, char *path, double defval);

/**
 * @brief Получение значения типа "float" для элемента JSON путь к которому указан в параметре path.
 * Если указанный элемент присутствует и его тип соответствует требуемому, то функция возвращает его значение.
 * В противном случае функция возвращает значение defval.
 * Если параметр path пустая строка или NULL, то функция возвращает значение элемента item.
 * @param item Указатель на структуру cJSON.
 * @param path Путь к элементу JSON или NULL.
 * @param defval Значение по умолчанию.
 * @return int 
 */
float cJSON_GetFloatByPath(cJSON *item, char *path, float defval);

/**
 * @brief Получение значения типа "string" для элемента JSON путь к которому указан в параметре path.
 * Если указанный элемент присутствует и его тип соответствует требуемому, то функция возвращает его значение.
 * В противном случае функция возвращает значение defval.
 * Если параметр path пустая строка или NULL, то функция возвращает значение элемента item.
 * @param item Указатель на структуру cJSON.
 * @param path Путь к элементу JSON или NULL.
 * @param defval Значение по умолчанию.
 * @return char*
 */
char *cJSON_GetStringByPath(cJSON *item, char *path, char *defval);

/**
 * @brief Преобразует объект cJSON в форматированную (отступы, CR/LF и т.п.) текстовую строку.
 * Функция выделяет пямять для текстовой строки, которая по окончании работы должна быть
 * освобождена функцией free().
 * @param item Указатель на структуру cJSON.
 * @return char* Строка с текстовым представлением.
 */
char *cJSON_Print(cJSON *item);

/**
 * @brief Преобразует объект cJSON в неформатированную текстовую строку.
 * Функция выделяет пямять для текстовой строки, которая по окончании работы должна быть
 * освобождена функцией free().
 * @param item Указатель на структуру cJSON.
 * @return char* Строка с текстовым представлением.
 */
char *cJSON_PrintUnformatted(cJSON *item);

/**
 * @brief Преобразование объекта cJSON в текст, используя буферизацию.
 * prebuffer - это предположение о конечном размере текста. Правильное угадывание сокращает количество перевыделения памяти.
 * @param item Указатель на структуру cJSON.
 * @param prebuffer Предполагаемый размер текста.
 * @param fmt 0-не использовать форматирование,1-использовать форматирование
 * @return char* 
 */
char *cJSON_PrintBuffered(cJSON *item, int prebuffer, int fmt);

/**
 * @brief Возвращает количество элементов в массиве или объекте.
 * @param array Указатель на структуру cJSON (массив или объект).
 * @return int Количество элементов.
 */
int cJSON_GetArraySize(cJSON *array);

/**
 * @brief Возвращает требуемый элемент из массива.
 * Если запрошенный элемент массива отсутствует функция возвращает NULL.
 * @param array Указатель на структуру cJSON (массив).
 * @param item Индекс элемента массива.
 * @return cJSON* 
 */
cJSON *cJSON_GetArrayItem(cJSON *array, int item);

/**
 * @brief Возвращает элемент объекта JSON по имени.
 * Если требуемого элемента не найдено, то функция возвращает NULL.
 * @param object Указатель на структуру cJSON (объект).
 * @param name Имя элемента в объекте.
 * @return cJSON* 
 */
cJSON *cJSON_GetObjectItem(cJSON *object, const char *name);

/**
 * @brief Создание нового элемента JSON с типом cJSON_NULL.
 * @return cJSON* 
 */
cJSON *cJSON_CreateNull(void);

/**
 * @brief Создание нового элемента JSON с типом cJSON_True или cJSON_False.
 * @param b Значение элемента: 1-true,0-false
 * @return cJSON* 
 */
cJSON *cJSON_CreateBool(int b);

/**
 * @brief Создание нового элемента JSON с числовым типом (cJSON_Number).
 * @param num Значение элемента
 * @return cJSON* 
 */
cJSON *cJSON_CreateNumber(double num);

/**
 * @brief Создание нового элемента JSON со строковым типом (cJSON_String).
 * @param string Указатель на строку
 * @return cJSON* 
 */
cJSON *cJSON_CreateString(const char *string);

/**
 * @brief Создание нового элемента JSON с типом массив (cJSON_Array).
 * @return cJSON* 
 */
cJSON *cJSON_CreateArray(void);

/**
 * @brief Создание нового элемента JSON с типом объект (cJSON_Object).
 * @return cJSON* 
 */
cJSON *cJSON_CreateObject(void);

/**
 * @brief Создание массива указанной длины из элементов числового типа (целые числа).
 * @param numbers Указатель на массив значений создаваемого массива.
 * @param count Количество элементов в массиве.
 * @return cJSON* 
 */
cJSON *cJSON_CreateIntArray(const int *numbers, int count);

/**
 * @brief Создание массива указанной длины из элементов числового типа (числа с плавающей точкой одинарной точности).
 * @param numbers Указатель на массив значений создаваемого массива.
 * @param count Количество элементов в массиве.
 * @return cJSON* 
 */
cJSON *cJSON_CreateFloatArray(const float *numbers, int count);

/**
 * @brief Создание массива указанной длины из элементов числового типа (числа с плавающей точкой двойной точности).
 * @param numbers Указатель на массив значений создаваемого массива.
 * @param count Количество элементов в массиве.
 * @return cJSON* 
 */
cJSON *cJSON_CreateDoubleArray(const double *numbers, int count);

/**
 * @brief Создание массива указанной длины из строк.
 * @param strings Указатель на массив значений создаваемого массива.
 * @param count Количество элементов в массиве.
 * @return cJSON* 
 */
cJSON *cJSON_CreateStringArray(const char **strings, int count);

/**
 * @brief Добавление элемента к массиву JSON.
 * Функция добавляет элемент к массиву JSON.
 * @param array Указатель на структуру cJSON (массив).
 * @param item Добавляемый элемент массива.
 */
void cJSON_AddItemToArray(cJSON *array, cJSON *item);

/**
 * @brief Добавление именованного элемента к объекту JSON.
 * Функция добавляет элемент item к объекту JSON object. Функция выполняет выделение памяти для имени name.
 * @param object Указатель на структуру cJSON (объект).
 * @param name Имя добавляемого элемента.
 * @param item Добавляемый элемент объекта.
 */
void cJSON_AddItemToObject(cJSON *object, const char *name, cJSON *item);

/**
 * @brief Добавление именованного элемента к объекту JSON.
 * В отличии от cJSON_AddItemToObject(), функция не выделяет (и потом не освобождает) память для имени name, при этом 
 * строка имени должна быть строковой константой.
 * @param object Указатель на структуру cJSON (объект).
 * @param name Имя добавляемого элемента.
 * @param item Добавляемый элемент объекта.
 */
void cJSON_AddItemToObjectCS(cJSON *object, const char *name, cJSON *item); 

/**
 * @brief Функция добавляет к массиву JSON новый элемент item как ссылку.
 * В отличии от cJSON_AddItemToArray() массив не становится владельцем элемента item и, соответственно,
 * при уничтожении массива память занимаемая элементом не освобождается.
 * @param array Указатель на структуру cJSON (массив).
 * @param item Добавляемый элемент объекта.
 */
void cJSON_AddItemReferenceToArray(cJSON *array, cJSON *item);

/**
 * @brief Функция добавляет к объекту JSON новый именованный элемент item как ссылку.
 * В отличии от cJSON_AddItemToObject() объект не становится владельцем элемента item и, соответственно,
 * при уничтожении объекта память занимаемая элементом не освобождается.
 * @param object Указатель на структуру cJSON (объект).
 * @param name Имя добавляемого элемента.
 * @param item Добавляемый элемент объекта.
 */
void cJSON_AddItemReferenceToObject(cJSON *object, const char *name, cJSON *item);

/**
 * @brief Вставка нового элемента в массив JSON.
 * Функция вставляет новый элемент в массив в позицию, указанную в параметре which. Уже имеющиеся
 * элементы массива расположенные после вставляемого сдвигаются вправо.
 * @param array Указатель на структуру cJSON (массив).
 * @param which Позиция вствки нового элемента.
 * @param newitem Вставляемый элемент массива.
 */
void cJSON_InsertItemInArray(cJSON *array, int which, cJSON *newitem);

/**
 * @brief Замена существующего элемента в массиве JSON.
 * Функция выполняет замену существующего элемента массива в позиции which на новый элемент newitem.
 * Если массив не содержит элемента в позиции which, то функция ничего не делает.
 * Память занимаемая заменяемым элеметом освобождается.
 * @param array Указатель на структуру cJSON (массив).
 * @param which Позиция заменяемого элемента.
 * @param newitem Новый элемент массива.
 */
void cJSON_ReplaceItemInArray(cJSON *array, int which, cJSON *newitem);

/**
 * @brief Замена существующего элемента в объекте JSON.
 * Функция выполняет замену существующего элемента объекта с именем name на новый элемент newitem.
 * Если объект не содержит элемента с именем name, то функция ничего не делает.
 * Память занимаемая заменяемым элеметом освобождается.
 * @param object Указатель на структуру cJSON (объект).
 * @param name Имя заменяемого элемента.
 * @param newitem Новый элемент массива.
 */
void cJSON_ReplaceItemInObject(cJSON *object, const char *name, cJSON *newitem);

/**
 * @brief Извлечение существующего элемента массива.
 * Функция извлекает из массива элемент с индексом which. Элементы с большими индексами сдвигаются "влево".
 * Функция не освобождает память занимаемую извлекаемым элементом.
 * Функция возвращает указатель на извлечённый элемент.
 * Если массив не содержит элемент с указанным индексом, то функция возвращает NULL.
 * @param array Указатель на структуру cJSON (массив).
 * @param which Индекс элемента массива.
 * @return cJSON* 
 */
cJSON *cJSON_DetachItemFromArray(cJSON *array, int which);

/**
 * @brief Удаление существующего элемента массива.
 * Функция удаляет из массива элемент с индексом which и освобождает занимаемую им память.
 * Элементы с большими индексами сдвигаются "влево".
 * Если массив не содержит элемент с указанным идексом, то функция ничего не делает.
 * @param array Указатель на структуру cJSON (массив).
 * @param which Индекс элемента массива.
 */
void cJSON_DeleteItemFromArray(cJSON *array, int which);

/**
 * @brief Извлечение существующего элемента объекта.
 * Функция извлекает из объекта элемент с именем name.
 * Функция не освобождает память занимаемую извлекаемым элементом.
 * Функция возвращает указатель на извлечённый элемент.
 * Если объект не содержит элемент с указанным именем, то функция возвращает NULL.
 * @param array Указатель на структуру cJSON (массив).
 * @param name Имя извлекаемого элемента
 * @return cJSON* 
 */
cJSON *cJSON_DetachItemFromObject(cJSON *object, const char *name);

/**
 * @brief Удаление существующего элемента объекта.
 * Функция удаляет из объекта элемент с именем name и освобождает занимаемую им память.
 * Если объект не содержит элемент с указанным именем, то функция ничего не делает.
 * @param array Указатель на структуру cJSON (массив).
 * @param name Имя извлекаемого элемента
 */
void cJSON_DeleteItemFromObject(cJSON *object, const char *name);

/**
 * @brief Создание дубликата cJSON.
 * Функция создаёт новый элемент cJSON идентичный элементу item. При создании функция выделяет память, которую потом необходимо освободить.
 * При значении recurse!=0 будут дублироваться все дочерние элементы.
 * Указатели next и prev всегда равны нулю при возврате из cJSON_Duplicate.
 * @param item Дублируемый элемент JSON.
 * @param recurse Рекурсивное создание дочерних элементов.
 * @return cJSON* 
 */
cJSON *cJSON_Duplicate(cJSON *item, int recurse);

/**
 * @brief Минификация строки JSON.
 * Функция удаляет ненужные символы: пробелы, табуляцию, коментарии, CR/LF и тп.
 * Результат помещается на место строки, указанной в параметре, в конце новой строки добаляется символ \0.
 * Функция не выполняет выделение или освобождение памяти.
 * @param json Указатель на строку с текстом JSON.
 */
void cJSON_Minify(char *json);

/**
 * @brief Создание нового объекта cJSON.
 * Только для внутреннего использования.
 * @return cJSON* 
 */
cJSON *cJSON_New_Item(void);

/**
 * @brief Набор макросов для добавления новых элементов к объекту.
 */
#define cJSON_AddNullToObject(object, name) cJSON_AddItemToObject(object, name, cJSON_CreateNull())
#define cJSON_AddTrueToObject(object, name) cJSON_AddItemToObject(object, name, cJSON_CreateBool(1))
#define cJSON_AddFalseToObject(object, name) cJSON_AddItemToObject(object, name, cJSON_CreateBool(0))
#define cJSON_AddBoolToObject(object, name, b) cJSON_AddItemToObject(object, name, cJSON_CreateBool(b))
#define cJSON_AddNumberToObject(object, name, n) cJSON_AddItemToObject(object, name, cJSON_CreateNumber(n))
#define cJSON_AddStringToObject(object, name, s) cJSON_AddItemToObject(object, name, cJSON_CreateString(s))

#define cJSON_SetNumberValue(object, val) ((object) ? (object)->valueint = (object)->valuedouble = (val) : (val))

__END_DECLS

#endif
