/**
 * @file hash.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Вычисление значений хеш функций по алгоритмам MD5, SHA1, SHA256.
 * @version 0.1
 * @date 2024-04-11
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * @example
 * 
 * MD5Context ctx;
 * md5Init(&ctx);
 * ...
 * md5Update(&ctx, input1, input1_size);
 * ...
 * md5Update(&ctx, input2, input2_size);
 * ...
 * md5Update(&ctx, input3, input3_size);
 * ...
 * md5Finalize(&ctx);
 *
 * ctx.digest;                               // Результат вычисления (uint8_t* 16 байт)
 */

#ifndef HASH_H
#define HASH_H

#include <cdefs.h>
#include <stddef.h>

__BEGIN_DECLS

// #define TEST_HASH

#define MD5_BLOCK_SIZE  16              // Размер дайджеста MD5

typedef struct {
    uint64_t size;                      //
    uint32_t buffer[4];                 // Текущее накопление
    uint8_t input[64];                  //
    uint8_t digest[MD5_BLOCK_SIZE];     // Результат вычисления
} MD5Context;

/**
 * @brief Инициализация алгоритма расчёта MD5
 * @param ctx Указатель на структуру MD5Context
 */
void md5Init(MD5Context *ctx);

/**
 * @brief Обновление дайджеста MD5
 * 
 * @param ctx Указатель на структуру MD5Context
 * @param input Входные данные
 * @param input_len Размер входных данных
 */
void md5Update(MD5Context *ctx, const uint8_t *input, size_t input_len);

/**
 * @brief Завершение расчёта MD5
 * @param ctx Указатель на структуру MD5Context
 */
void md5Finalize(MD5Context *ctx);

/**
 * @brief Расчёт MD5 для строки завершённой символом 0.
 * @param input Указатель на строку
 * @param result Указатель на буфер для результата (минимум 16 байт)
 */
void md5String(const char *input, uint8_t *result);

// ---------------------------------------------------------------------------------

#define SHA1_BLOCK_SIZE 20                  // Размер дайджеста SHA1

typedef struct {
    uint32_t state[5];
    uint32_t count[2];
    unsigned char buffer[64];
} SHA1Context;

/**
 * @brief Инициализация алгоритма расчёта SHA1.
 * @param context Указатель на структуру SHA1Context
 */
void SHA1Init(SHA1Context *context);

/**
 * @brief Обновление дайджеста SHA1.
 * 
 * @param context Указатель на структуру SHA1Context.
 * @param data Указатель на буфер данных.
 * @param len Размер данных.
 */
void SHA1Update(SHA1Context *context, const unsigned char *data, uint32_t len);

/**
 * @brief Завершение расчёта SHA1.
 * 
 * @param digest Указатель на буфер для результата (SHA1_BLOCK_SIZE: 20 байт).
 * @param context Указатель на структуру SHA1Context.
 */
void SHA1Final(unsigned char digest[SHA1_BLOCK_SIZE], SHA1Context *context);

/**
 * @brief Вычисление дайджеста для содержимого буфера str.
 * 
 * @param hash_out Указатель на буфер для результата (20 байт).
 * @param str Указатель на буфер данных.
 * @param len Размер буфера данных.
 */
void SHA1(char *hash_out, const char *str, uint32_t len);

// ---------------------------------------------------------------------------------

#define SHA256_BLOCK_SIZE 32                // Размер дайджеста SHA256

typedef struct {
    uint8_t data[64];
    uint32_t datalen;
    unsigned long long bitlen;
    uint32_t state[8];
} SHA256Context;

/**
 * @brief Инициализация алгоритма расчёта SHA256.
 * @param ctx Указатель на структуру SHA256Context
 */
void SHA256Init(SHA256Context *ctx);

/**
 * @brief Обновление дайджеста SHA256.
 * 
 * @param ctx Указатель на структуру SHA256Context.
 * @param data Указатель на буфер данных.
 * @param len Размер данных.
 */
void SHA256Update(SHA256Context *ctx, const uint8_t data[], size_t len);

/**
 * @brief Завершение расчёта SHA256.
 * @param ctx Указатель на структуру SHA256Context.
 * @param hash Буфер для результата (32 байта).
 */
void SHA256Final(SHA256Context *ctx, uint8_t hash[]);

// ---------------------------------------------------------------------------------

#ifdef TEST_HASH

int test_md5();
int test_sha1();
int test_sha256();

#endif

__END_DECLS

#endif
