/**
 * @file http.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции сервера/клиента работы по протоколу HTTP.
 * @version 0.1
 * @date 2024-05-19
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef HTTP_H
#define HTTP_H

#include <cdefs.h>
#include <url.h>

__BEGIN_DECLS

/**
 * @brief Тайм-аут на запросы HTTP.
 * Значение по умолчанию 4000 мс.
 */
extern unsigned int http_timeout;

/**
 * @brief Максимальный размер ответа на запрос HTTP.
 * Значение по умолчанию 16384.
 */
extern unsigned int http_max_response_len;

typedef enum http_content_type {
    httpdUSER           = 0,
    httpdJSON           = 1,
    httpdHTML           = 2,
    httpdTEXT           = 3,
    httpdFORMURLENCODED = 4,
    httpdXML            = 5
} http_content_type_t;

// -- Сервер HTTP --------------------------------------------------------------------------------------------------------------

/**
 * @brief Возвращает количество запросов к HTTP серверу.
 * @return unsigned int Количество запросов.
 */
unsigned int HttpSrvDataCount(void);

/**
 * @brief Функция устанавливает пользовательские данные, которые будет возвращены ПЛК в ответ на HTTP-запрос по URL "/data".
 * Функция создаёт копию строки data в служебной области памяти.
 * @param data Указатель на устанавливаемую строку ответа.
 * @param type Тип данных. Возможные значения: httpdUSER | httpdJSON | httpdHTML | httpdTEXT
 */
void StartHttpSrvData(char * data, http_content_type_t type);
#define SetHttpSrvData(data,type)   StartHttpSrvData(data,type)

/**
 * @brief Функция отменяет использование данных пользователя, установленных вызовом функции StartHttpSrvData.
 */
void StopHttpSrvData(void);
#define ClearHttpSrvData()          StopHttpSrvData()

// -- Запрос HTTP --------------------------------------------------------------------------------------------------------------

typedef enum {
    HTTP_RES_OK                 =  1,      ///< запрос выполнен успешно
    HTTP_RES_BAD_URL            = -1,      ///< ошибка в строке URL
    HTTP_RES_BAD_SOCKET         = -2,      ///< не могу создать новый сокет
    HTTP_RES_NO_CONNECTION      = -3,      ///< не могу подключиться к серверу
    HTTP_RES_RESP_TOO_LARGE     = -4,      ///< размер ответа превышает ограничение http_max_response_len
    HTTP_RES_EMPTY_RESP         = -5,      ///< пустой ответ
    HTTP_RES_BAD_RESP_PROTO     = -6,      ///< в заголовке ответа отсутствует HTTP
    HTTP_RES_BAD_RESP_FORMAT    = -7,      ///< неверный формат ответа
    HTTP_RES_AUTH_TOO_LARGE     = -8,      ///< имя пользователя + пароль слишком длинные
    HTTP_RES_HEADERS_TOO_LARGE  = -9,      ///< слишком большой размер заголовков запроса
    HTTP_RES_EMPTY_POST_DATA    = -10      ///< пустые данные при вызове HttpPOST
} http_result_t;


/**
 * @brief Структура - ответ на HTTP запрос.
 */
typedef struct http_response {
    char *response;                         ///< полный ответ
    int response_len;                       ///< размер полного ответа
    //
    char *response_headers;                 ///< заголовки ответа
    char *status_text;                      ///< текстовый статус
    int status_code;                        ///< код ответа
    char *body;                             ///< тело ответа
} http_response_t;

/**
 * @brief Выполнение HTTP запроса GET.
 * Функция возвращает результат выполнения запроса: элемент перечисления http_result_t.
 * При успешном выполнении функция возвращает HTTP_RES_OK и заполняет элементы структуры hresp.
 * Пользователь должен дополнительно проверить статус выполнения запроса (поле hresp->status_code).
 * Память занимаемая структурой в последствии должна быть освобождена функцией HttpFreeResponse.
 * Формат строки url должен соответствовать описанному в функции HttpParseURL.
 * @param url Строка URL.
 * @param custom_headers Пользовательские заголовки (могут быть NULL)
 * @param hresp Указатель на структуру ответа.
 * @return http_result_t Результат выполнения запроса.
 */
http_result_t HttpGET(char *url, char *custom_headers, http_response_t *hresp);

/**
 * @brief Выполнение HTTP запроса POST.
 * Функция возвращает результат выполнения запроса: элемент перечисления http_result_t.
 * При успешном выполнении функция возвращает HTTP_RES_OK и заполняет элементы структуры hresp.
 * Пользователь должен дополнительно проверить статус выполнения запроса (поле hresp->status_code).
 * Память занимаемая структурой в последствии должна быть освобождена функцией HttpFreeResponse.
 * Формат строки url должен соответствовать описанному в функции HttpParseURL.
 * Параметр post_type определяет содержимое HTTP-заголовка "Content-Type". При выборе httpdUSER заголовок
 * необходимо сформировать самостоятельно через параметр custom_headers.
 * @param url Строка URL.
 * @param custom_headers Пользовательские заголовки (могут быть NULL)
 * @param post_data Записываемые данные POST.
 * @param post_type Тип данных POST: httpdUSER | httpdJSON | httpdHTML | httpdTEXT | httpdFORMURLENCODED | httpdXML
 * @param hresp Указатель на структуру ответа.
 * @return http_result_t Результат выполнения запроса.
 */
http_result_t HttpPOST(char *url, char *custom_headers, char *post_data, http_content_type_t post_type, http_response_t *hresp);

/**
 * @brief Освобождение памяти занимаемой ответом HTTP.
 * Функция должна быть вызвана после успешного выполнения HttpGET() и HttpPost() для освобождения
 * памяти, занимаемой ответом.
 * @param hresp Указатель на структуру ответа.
 */
void HttpFreeResponse(http_response_t *hresp);

__END_DECLS

#endif
