/**
 * @file iconv.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции преобразования кодировок символов.
 * @version 0.1
 * @date 2024-05-07
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef ICONV_H
#define ICONV_H

#include <cdefs.h>

__BEGIN_DECLS

//
// Таблицы перекодировки символов для использования в функции "char_string_to_UTF8".
//

/**
 * @brief Таблица преобразования latin-1 (28591).
 * Языки: Danish, Dutch, English, Faeroese, Finnish, French, German, Icelandic, Irish, Italian, Norwegian, Portuguese, Spanish and Swedish
 */
extern const unsigned short charISO_8859_1[];

/**
 * @brief Таблица преобразования latin-2 (28592).
 * Языки: Albanian, Czech, English, German, Hungarian, Polish, Rumanian, Serbo-Croatian, Slovak, Slovene and Swedish
 */
extern const unsigned short charISO_8859_2[];

/**
 * @brief Таблица преобразования latin-3 (28593).
 * Языки: Afrikaans, Catalan, English, Esperanto, French, Galician, German, Italian, Maltese and Turkish
 */
extern const unsigned short charISO_8859_3[];

/**
 * @brief Таблица преобразования latin-4 (28594).
 * Языки: Danish, English, Estonian, Finnish, German, Greenlandic, Lappish, Latvian, Lithuanian, Norwegian and Swedish
 */
extern const unsigned short charISO_8859_4[];

/**
 * @brief Таблица преобразования CYRILLIC (28595).
 * Языки: Bulgarian, Bielorussian, English, Macedonian, Russian, Serbo-Croatian and Ukrainian
 */
extern const unsigned short charISO_8859_5[];

/**
 * @brief Таблица преобразования ARABIC (28596).
 * Языки: ARABIC
 */
extern const unsigned short charISO_8859_6[];

/**
 * @brief Таблица преобразования GREEK (28597).
 * Языки: Greek
 */
extern const unsigned short charISO_8859_7[];

/**
 * @brief Таблица преобразования HEBREW (28598).
 * Языки: HEBREW
 */
extern const unsigned short charISO_8859_8[];

/**
 * @brief Таблица преобразования latin-5 (28599).
 * Языки: English, Finnish, French, German, Irish, Italian, Norwegian, Portuguese, Spanish, Swedish and Turkish
 */
extern const unsigned short charISO_8859_9[];

/**
 * @brief Таблица преобразования Estonian (28603).
 * Языки: Estonian
 */
extern const unsigned short charISO_8859_13[];

/**
 * @brief Таблица преобразования latin-9 (28605).
 */
extern const unsigned short charISO_8859_15[];

/**
 * @brief Таблица преобразования Eastern European (1250).
 * Языки: Eastern European
 */
extern const unsigned short charCP_1250[];

/**
 * @brief Таблица преобразования Cyrillic (1251).
 * Языки: Russian
 */
extern const unsigned short charCP_1251[];

/**
 * @brief Таблица преобразования Latin-1 (1252).
 * Языки: US, Western Europe
 */
extern const unsigned short charCP_1252[];

/**
 * @brief Таблица преобразования Greek (1253).
 * Языки: Greek
 */
extern const unsigned short charCP_1253[];

/**
 * @brief Таблица преобразования Turkish (1254).
 * Языки: Turkish
 */
extern const unsigned short charCP_1254[];

/**
 * @brief Таблица преобразования Hebrew (1255).
 * Языки: Hebrew
 */
extern const unsigned short charCP_1255[];

/**
 * @brief Таблица преобразования Arabic (1256).
 * Языки: Arabic
 */
extern const unsigned short charCP_1256[];

/**
 * @brief Таблица преобразования Baltic (1257).
 * Языки: Baltic
 */
extern const unsigned short charCP_1257[];

/**
 * @brief Таблица преобразования Vietnamese (1258).
 * Языки: Vietnamese
 */
extern const unsigned short charCP_1258[];

/**
 * @brief Таблица преобразования Cyrillic KOI8R (20866).
 * Языки: Russian
 */
extern const unsigned short charKOI8_R[];

/**
 * @brief Таблица преобразования Czech (Kamenicky) (895).
 * Языки: Czech
 */
extern const unsigned short charCP_895[];

/**
 * @brief Таблица преобразования Eastern European (852).
 */
extern const unsigned short charCP_852[];

/**
 * @brief Преобразование строки из кодировки UTF8 в CP1251.
 * При успешном преобразовании функция возвращает значение 1, при ошибке преобразования - 0.
 * @param utf8 Исходная строка в UTF8.
 * @param cp1251 Буфер для строки в CP1251.
 * @param maxLen Размер буфера dest.
 * @return int 
 */
int utf8_to_cp1251(const char *utf8, char *cp1251, int maxLen);

/**
 * @brief Преобразование строки из кодировки cp1251 в UTF8.
 * Функция возвращает количество байт занятых в буфере utf8 (включая завершающий символ 0).
 * При ошибке функция возвращает значение 0. Ошибка может быть вызвана недостаточным размером буфера utf8 или некорректными значениями параметров.
 * @param cp1251 Исходная строка в кодировке cp1251.
 * @param utf8 Буфер для строки в кодировке UTF8.
 * @param maxLen Размер буфера для строки UTF8.
 * @return int 
 */
int cp1251_to_utf8(char *cp1251, char *utf8, int maxLen);

/**
 * @brief Перекодирование символов строки в кодировку UTF8 при помощи таблицы преобразование.
 * При успешном выполнении функция преобразует кодировку символов строки src в кодировку UTF8 и размещает её в буфере dest.
 * Функция возвращает количество байт занятых в буфере dest (включая завершающий символ 0).
 * При ошибке функция возвращает значение 0. Ошибка может быть вызвана недостаточным размером буфера dest или некорректными значениями параметров.
 * @param cptable Таблица преобразования.
 * @param src Строка в исходной кодировке.
 * @param dest Буфер для строки в кодировке UTF8.
 * @param maxdestlen Размер буфера для строки UTF8.
 * @return int 0-ошибка, >0-количество занятых байт.
 */
int char_string_to_UTF8(const unsigned short *cptable, char *src, char *dest, int maxdestlen);

__END_DECLS

#endif
