/**
 * @file lanmon.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции клиента сервера LanMon.
 * Ограничения:
 * - не поддерживается тип данных STRING;
 * - не поддерживаются массивы;
 * - ограниченная поддержка атрибутов.
 * @version 0.1
 * @date 2024-04-25
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef LANMON_H
#define LANMON_H

#include <cdefs.h>
#include <stdbool.h>
#include <time.h>

__BEGIN_DECLS

/**
 * @brief Состояние клиента LanMon
 */
enum lm_state_enum {
    lmsNONE             =   0,          ///< Клиент не запущен
    lmsCONNECTING       =   1,          ///< Подключение к серверу
    lmsDISCONNECTED     =   2,          ///< Отключены
    lmsAUTH             =   3,          ///< Авторизация на сервере
    lmsOK               =   4           ///< Подключены и авторизованы на сервере
};

/**
 * @brief Значения свойства Quality каналов данных
 */
typedef enum lm_quality {
    lmqOK               =   0,          ///< Ok
    lmqOff              =   1,          ///< Выключен
    lmqUndefined        =   2,          ///< Состояние не определено
    lmqFailure          =   3,          ///< Неисправность датчика
    lmqError            =   4,          ///< Неисправность контроллера
    lmqWrong            =   5,          ///< Некоректные данные
    lmqNotConnected     =   6,          ///< Датчик не подключен
    lmqBadConnection    =   7,          ///< Нет соединения
    lmqRegFailure       =   8           ///< Неисправность регистратора
} lm_quality_t;

/**
 * @brief Значения свойства Type каналов данных
 */
typedef enum lm_type {
    lmtEmpty            =   0,          ///< Нет значения
    lmtI2               =   2,          ///< 2 байта со знаком
    lmtI4               =   3,          ///< 4 байта со знаком
    lmtR4               =   4,          ///< 4 байта с плавающей точкой
    lmtR8               =   5,          ///< 8 байт с плавающей точкой
    lmtDATE             =   7,          ///< Дата время в формате Delphi TDateTime (тип данных R8)
    lmtBOOL             =   11,         ///< 2 байта со знаком (тип данных I2: True=-1, False=0)
    lmtI1               =   16,         ///< 1 байт со знаком
    lmtUI1              =   17,         ///< 1 байт без знака
    lmtUI2              =   18,         ///< 2 байта без знака
    lmtUI4              =   19,         ///< 4 байта без знака
    // lmtSTRING        =   256         ///< Строка, один байт на символ в кодировке windows 1251
} lm_type_t;

/**
 * @brief Значение канала данных
 */
typedef union lm_value {
    TDateTime datetime;
    double r8;
    float r4;
    int i4;
    unsigned int ui4;
    short int i2;
    unsigned short int ui2;
    short int boolean;
    char i1;
    unsigned char ui1;
} lm_value_t;

/**
 * @brief Тип фильтрации изменения значений
 */
typedef enum lm_filter {
    lmfNone             =   0,          ///< Фильтрация не используется
    lmfPercent          =   1,          ///< Фильтрация по относительному изменению значения в процентах (только для lmtR4 и lmtR8)
    lmfAbsolute         =   2           ///< Фильтрация по абсолютному изменению значения
} lm_filter_t;

/**
 * @brief Максимальная длина имени канала + 1
 */
#define LM_MAX_NAME_SIZE    64

#pragma pack(push,4)

/**
 * @brief Описание канала данных сервера LanMon
 */
typedef struct lm_channel {
    char name[LM_MAX_NAME_SIZE];        ///< Имя канала (с-строка)
    char *unit;                         ///< Единицы измерения
    char *cmnts;                        ///< Комментарий
    unsigned int num;                   ///< Номер канала на сервере LanMon
    lm_type_t type;                     ///< Тип канала
    lm_quality_t quality;               ///< Качество канала
    lm_value_t value;                   ///< Значение канала
    float  filter;                      ///< Значение фильтра
    struct lanmon {
        int need_registration : 1;      ///< Канал требует регистрации на сервере
        int registered : 1;             ///< Канал зарегистрирован на сервере
        int control : 1;                ///< Значение канала изменено сервером
        int changed : 1;                ///< Значение канала изменено клиентом
        int active : 1;                 ///< Канал активен
        int valid : 1;                  ///< Значение канала было хотя-бы один раз установлено клиентом
        int rw : 1;                     ///< Канал доступен для записи
        lm_filter_t filter : 2;         ///< Тип фильтра значений: lmfNone | lmfPercent | lmfAbsolute
    } state;
} lm_channel_t;

#pragma pack(pop)

/**
 * @brief Функция обработки получения команды управления (запись канала сервером).
 * @param client_id Числовой идентификатор полученного канала.
 * @param channel Указатель на канал.
 */
typedef void (*lmcontrol_func)(int client_id, lm_channel_t *channel);

/**
 * @brief Функция начинает работу встроенного клиента сервера LanMon.
 * Функция запускает отдельный поток, который устанавливает и поддерживает соединение с сервером.
 * Для завершения работы с сервром необходимо использовать функцию LMStop().
 * Система поддерживает только одно соединение с сервером.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0. Ошибка может быть вызвана некорректным
 * значением параметров, нехваткой системных ресурсов или попыткой повтороного запуска.
 * @param host Имя или IP адрес сервера.
 * @param port Номер TCP-порта сервера.
 * @param timeout Таймаут на операции обмена с сервером, мс
 * @param login Имя учётной записи на сервере.
 * @param password Пароль учётной записи на сервере.
 * @param cb Функция - обработчик получения команды управления от сервера.
 * @return int 1-успех, 0-ошибка.
 */
int LMStart(char *host, unsigned short int port, unsigned int timeout, char *login, char *password, lmcontrol_func cb);

/**
 * @brief Функция останавливает работу встроенного клиента сервера LanMon и освобождает все выделенные ресурсы.
 */
void LMStop(void);

/**
 * @brief Получение статуса подключения к серверу LanMon.
 * При активном подключении к серверу функция возвращает значение 1, в противном случае: 0.
 * @return int 1-есть подключение, 0-нет подключения.
 */
int LMConnected(void);

/**
 * @brief Добавление нового канала.
 * При успешном выполнении функция возвращает числовой идентификатор канала: число большее или равное 0.
 * При ошибке функция возвращает значение -1. Ошибка может быть вызвана некорректным значением параметров или
 * попыткой создания канала с уже существующим именем.
 * @param chname Имя канала.
 * @param type Тип канала: элемент перечисления lm_type.
 * @param unit Единицы измерения. Если NULL, то атрибут не будет создан.
 * @param cmnts Текстовый коментарий. Если NULL, то атрибут не будет создан.
 * @param ctrl Признак возможности управления: 1-возможность записи, 0-только чтение.
 * @return int Идентификатор канала или ошибка (-1).
 */
int LMAddChannel(char *chname, lm_type_t type, char *unit, char *cmnts, int rw);

/**
 * @brief Передача значения канала на сервер.
 * Функция устанавливает свойству канала Quality значение lmqOK. Если новое значение канала отличается от предыдущего, то устанавливается
 * признак необходимости передачи значения канала на сервер.
 * Если предыдущее значение канала совпадает с value, то функция возвращает значение 2.
 * Если канал с числовым идентификатором client_id не найден, то функция возвращает значение 0 (ошибка).
 * @param client_id Числовой идентификатор канала.
 * @param value Значение канала.
 * @return int 1-успех, 2-значение отфильтровано, 0-ошибка.
 */
int LMSendValue(int client_id, lm_value_t *value);

/**
 * @brief Установка свойства Quality для указанного канала.
 * Функция позволяет установить значение Quality отличное от lmqOK, для установки значения lmqOK необходимо  использовать
 * функцию LMSendValue. При успешном выполнении функция возвращает значение 1.
 * Если предыдущее значение свойства совпадает с quality, то функция возвращает значение 2.
 * Если канал с числовым идентификатором client_id не найден или (quality == lmqOK), то функция возвращает значение 0 (ошибка).
 * @param client_id Числовой идентификатор канала.
 * @param quality Значение Quality (элемент перечисления lm_quality).
 * @return int 1-успех, 2-значение отфильтровано, 0-ошибка.
 */
int LMSetQuality(int client_id, lm_quality_t quality);

/**
 * @brief Установка значения фильтра на относительное изменение значения канала.
 * Если значение канала изменилось меньше, чем на указанную относительную величину, то оно передаваться на сервер не будет.
 * Фильтр действует только для каналов с типами lmtR4 и lmtR8.
 * Если канал с указанным client_id не найден или его тип не соответствует lmtR4 или lmtR8, то функция 
 * вернёт значение 0, при успешном выполнении: 1.
 * @param client_id Числовой идентификатор канала.
 * @param value Значение фильтра, проценты
 * @return int 1-успех, 0-ошибка
 */
int LMSetFilterPercent(int client_id, float value);

/**
 * @brief Установка значения фильтра на абсолютное изменение значения канала.
 * Если значение канала изменилось меньше, чем на указанную абсолютную величину, то оно передаваться на сервер не будет.
 * Фильтр действует для каналов всех типов.
 * Если канал с указанным client_id не найден, то функция вернёт значение 0, при успешном выполнении: 1.
 * @param client_id Числовой идентификатор канала.
 * @param value Значение фильтра, абсолютная величина изменения.
 * @return int 1-успех, 0-ошибка
 */
int LMSetFilterAbsolute(int client_id, float value);

/**
 * @brief Отключение фильтрации значений канала.
 * Функция отключает предварительно установленную фильтрацию.
 * Если канал с указанным client_id не найден, то функция вернёт значение 0, при успешном выполнении: 1.
 * @param client_id Числовой идентификатор канала.
 * @return int 1-успех, 0-ошибка
 */
int LMSetFilterNone(int client_id);

/**
 * @brief Получение описания канала.
 * При успешном выполнении функция возвращает значение указателя (lm_channel) на канал с числовым идентификатором client_id.
 * При отсутствии запрошенного канала функция возвращает значение NULL.
 * @param client_id Числовой идентификатор канала.
 * @return lm_channel_t* Указатель на описание канала или NULL.
 */
lm_channel_t *LMGetChannelInfo(int client_id);

__END_DECLS

#endif
