/**
 * @file math.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Математические функции.
 * @version 0.1
 * @date 2024-03-09
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef MATH_H
#define MATH_H

#include <cdefs.h>

#define M_E                 2.7182818284590452354  /* e */
#define M_LOG2E             1.4426950408889634074  /* log_2 e */
#define M_LOG10E            0.4342944819032518276  /* log_10 e */
#define M_LN2               0.6931471805599453094  /* log_e 2 */
#define M_LN10              2.3025850929940456840  /* log_e 10 */
#define M_PI                3.1415926535897932384  /* pi */
#define M_PI_2              1.5707963267948966192  /* pi/2 */
#define M_PI_4              0.7853981633974483096  /* pi/4 */
#define M_1_PI              0.3183098861837906715  /* 1/pi */
#define M_2_PI              0.6366197723675813430  /* 2/pi */
#define M_2_SQRTPI          1.1283791670955125739  /* 2/sqrt(pi) */
#define M_SQRT2             1.4142135623730950488  /* sqrt(2) */
#define M_SQRT1_2           0.7071067811865475244  /* 1/sqrt(2) */

#define M_Eq                2.718281828459045235360287471352662498L  /* e */
#define M_LOG2Eq            1.442695040888963407359924681001892137L  /* log_2 e */
#define M_LOG10Eq           0.434294481903251827651128918916605082L  /* log_10 e */
#define M_LN2q              0.693147180559945309417232121458176568L  /* log_e 2 */
#define M_LN10q             2.302585092994045684017991454684364208L  /* log_e 10 */
#define M_PIq               3.141592653589793238462643383279502884L  /* pi */
#define M_PI_2q             1.570796326794896619231321691639751442L  /* pi/2 */
#define M_PI_4q             0.785398163397448309615660845819875721L  /* pi/4 */
#define M_1_PIq             0.318309886183790671537767526745028724L  /* 1/pi */
#define M_2_PIq             0.636619772367581343075535053490057448L  /* 2/pi */
#define M_2_SQRTPIq         1.128379167095512573896158903121545172L  /* 2/sqrt(pi) */
#define M_SQRT2q            1.414213562373095048801688724209698079L  /* sqrt(2) */
#define M_SQRT1_2q          0.707106781186547524400844362104849039L  /* 1/sqrt(2) */

__BEGIN_DECLS

// Математические функции одинарной точности ------------------------

/**
 * @brief Вычисление значения синуса x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float
 */
float sinf(float x);

/**
 * @brief Вычисление значения косинуса x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float
 */
float cosf(float x);

/**
 * @brief Вычисление значения тангенса x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float tanf(float x);

/**
 * @brief Вычисление значения арксинуса x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float asinf(float x);

/**
 * @brief Вычисление значения арккосинуса x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float acosf(float x);

/**
 * @brief Вычисление значения арктангенса x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float atanf(float x);

/**
 * @brief Вычисление значения гиперболического синуса x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float sinhf(float x);

/**
 * @brief Вычисление значения гиперболического косинуса x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float coshf(float x);

/**
 * @brief Вычисление значения гиперболического тангенса x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float tanhf(float x);

/**
 * @brief Вычисление значения экспоненты x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float expf(float x);

/**
 * @brief Вычисление значения степени по основанию 2.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float exp2f(float x);

/**
 * @brief Вычисление значения функции "exp (x) - 1"
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float expm1f(float x);

/**
 * @brief Вычисление значения натурального логарифма x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float logf(float x);

/**
 * @brief Вычисление значения десятичного логарифма x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float log10f(float x);

/**
 * @brief Вычисление значения эквивалентное "log (1 + x)".
 * Вычисляется методом, выдающим точные значения, даже если значение x близко к нулю.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float log1pf(float x);

/**
 * @brief Вычисление значения логарифма по основанию 2 от аргумента x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float log2f(float x);

/**
 * @brief Вычисление абсолютного значения x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float fabsf(float x);

/**
 * @brief Вычисление квадратного корня x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float sqrtf(float x);

/**
 * @brief Округление до наименьшего целого, не меньшего, чем аргумент x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float ceilf(float x);

/**
 * @brief Округление до наибольшего целого значения, но не большее x.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float floorf(float x);

/**
 * @brief Округление до ближайшего целого в направлении от нуля.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float roundf(float x);

/**
 * @brief Округление до ближайшего целого числа в направлении к нулю.
 * Аргумент и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @return float 
 */
float truncf(float x);

/**
 * @brief Функция возвращает значение x в степени y.
 * Аргументы и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @param y 
 * @return float 
 */
float powf(float x, float y);

/**
 * @brief Функция вычисления остатка от деления (с плавающей точкой).
 * Возращаемое значение: x - n * y - где n есть частное x / y, округленное к нулю до ближайшего целого числа.
 * Аргументы и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @param y 
 * @return float 
 */
float fmodf(float x, float y);

/**
 * @brief Computes the IEEE remainder of the floating point division operation x/y.
 * Аргументы и результат число с плавающей точкой одинарной точности.
 * @param x 
 * @param y 
 * @return float 
 */
float remainderf(float x, float y);


// Математические функции двойной точности-----------------------------------------------------

/**
 * @brief Вычисление значения синуса x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double sin(double x);

/**
 * @brief Вычисление значения косинуса x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double cos(double x);

/**
 * @brief Вычисление значения тангенса x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double tan(double x);

/**
 * @brief Вычисление значения арксинуса x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double asin(double x);

/**
 * @brief Вычисление значения арккосинуса x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double acos(double x);

/**
 * @brief Вычисление значения арктангенса x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double atan(double x);

/**
 * @brief Вычисление значения гиперболического синуса x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double sinh(double x);

/**
 * @brief Вычисление значения гиперболического косинуса x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double cosh(double x);

/**
 * @brief Вычисление значения гиперболического тангенса x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double tanh(double x);

/**
 * @brief Вычисление значения экспоненты x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double exp(double x);

/**
 * @brief Вычисление значения степени по основанию 2.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double exp2(double x);

/**
 * @brief Вычисление значения функции "exp (x) - 1"
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double expm1(double x);

/**
 * @brief Вычисление значения натурального логарифма x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double log(double x);

/**
 * @brief Вычисление значения десятичного логарифма x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double log10(double x);

/**
 * @brief Вычисление значения эквивалентное "log (1 + x)".
 * Вычисляется методом, выдающим точные значения, даже если значение x близко к нулю.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double log1p(double x);

/**
 * @brief Вычисление значения логарифма по основанию 2 от аргумента x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double log2(double x);

/**
 * @brief Вычисление абсолютного значения x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double fabs(double x);

/**
 * @brief Вычисление квадратного корня x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double sqrt(double x);

/**
 * @brief Округление до наименьшего целого, не меньшего, чем аргумент x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double ceil(double x);

/**
 * @brief Округление до наибольшего целого значения, но не большее x.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double floor(double x);

/**
 * @brief Округление до ближайшего целого в направлении от нуля.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double round(double x);

/**
 * @brief Округление до ближайшего целого числа в направлении к нулю.
 * Аргумент и результат число с плавающей точкой двойной точности.
 * @param x 
 * @return double 
 */
double trunc(double x);

/**
 * @brief Функция возвращает значение x в степени y.
 * Аргументы и результат число с плавающей точкой двойной точности.
 * @param x 
 * @param y 
 * @return double 
 */
double pow(double x, double y);

/**
 * @brief Функция вычисления остатка от деления (с плавающей точкой).
 * Возращаемое значение: x - n * y - где n есть частное x / y, округленное к нулю до ближайшего целого числа.
 * Аргументы и результат число с плавающей точкой двойной точности.
 * @param x 
 * @param y 
 * @return double 
 */
double fmod(double x, double y);

/**
 * @brief Computes the IEEE remainder of the floating point division operation x/y.
 * Аргументы и результат число с плавающей точкой двойной точности.
 * @param x 
 * @param y 
 * @return double 
 */
double remainder(double x, double y);

__END_DECLS

#endif
