/**
 * @file merc230.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Чтение данных счётчика электрической энергии Меркурий 230.
 * @version 0.1
 * @date 2024-03-30
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * @example Пример чтения текущих данных:
 * 
 * merc230_current_t current_data;
 * //
 * // открываем порт
 * int handle = SerialOpen(1, B9600, PARNONE, 1);
 * //
 * int res = Merc230_ReadCurrentData(handle, 0, &current_data);
 * if(res) {
 *   printf("Data read successfully\n");
 * } else {
 *   printf("Can''t read data\n");
 * }
 * //
 * SerialClose(handle);
 */

#ifndef MERC230_H
#define MERC230_H

#include <cdefs.h>
#include <stdint.h>
#include <unistd.h>

#define FASE_A      0                   ///< Фаза A (1) для напряжений, токов, cosF и мощностей
#define FASE_B      1                   ///< Фаза B (2) для напряжений, токов, cosF и мощностей
#define FASE_C      2                   ///< Фаза C (3) для напряжений, токов, cosF и мощностей
#define FASE_SUMMA  3                   ///< Суммарное значение по фазам

#define ENERGY_ACTIVE_P     0           ///< Энергия активная положительная (тип энергии)
#define ENERGY_ACTIVE_N     1           ///< Энергия активная отрицательная (тип энергии)
#define ENERGY_REACTIVE_P   2           ///< Энергия реактивная положительная (тип энергии)
#define ENERGY_REACTIVE_N   3           ///< Энергия реактивная отрицательная (тип энергии)

/**
 * @brief Структура описания текущих данных
 */
typedef struct _merc230_current {
    float U[3];                         ///< Фазное напряжение по фазам, В
    float I[3];                         ///< Фазный ток по фазам, А
    float cosf[4];                      ///< Коэффициент мощности (фаза 1, фаза 2, фаза 3, суммарный)
    float P[4];                         ///< Активная мощность по фазам, Вт (фаза 1, фаза 2, фаза 3, суммарная)
    float Q[4];                         ///< Реактивная мощность по фазам, вар (фаза 1, фаза 2, фаза 3, суммарная)
    float S[4];                         ///< Полная мощность по фазам, ВА (фаза 1, фаза 2, фаза 3, суммарная)
} merc230_current_t;

/**
 * @brief Структура описания накопленных данных (интеграторы энергии)
 */
typedef struct _merc230_total {
    double summa[4];                    ///< Суммарная энергия   (активная(+) кВтч, активная(-) кВтч, реактивная(+) кварч, реактивная(-) кварч)
    double tarif1[4];                   ///< Энергия по тарифу 1 (активная(+) кВтч, активная(-) кВтч, реактивная(+) кварч, реактивная(-) кварч)
    double tarif2[4];                   ///< Энергия по тарифу 2 (активная(+) кВтч, активная(-) кВтч, реактивная(+) кварч, реактивная(-) кварч)
    double tarif3[4];                   ///< Энергия по тарифу 3 (активная(+) кВтч, активная(-) кВтч, реактивная(+) кварч, реактивная(-) кварч)
    double tarif4[4];                   ///< Энергия по тарифу 4 (активная(+) кВтч, активная(-) кВтч, реактивная(+) кварч, реактивная(-) кварч)
} merc230_total_t;

__BEGIN_DECLS

/**
 * @brief Чтение текущих данных счётчика электрической энергии Меркурий 230.
 * Для чтения данных используется дескриптор последовательного порта или сетевого сокета.
 * Функция возвращает 1 при успешном чтении данных. При ошибке функция возвращает:
 * 0  - счётчик не отвечает на запрос;
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param handle Дескриптор последовательного порта или сетевого сокета.
 * @param addr Сетевой адрес счётчика: 1..240, 0 - широковещательный адрес.
 * @param data Указатель на структуру текущих данных.
 * @return int 
 */
int Merc230_ReadCurrentData(int handle, uint8_t addr, merc230_current_t *data);

/**
 * @brief Чтение накопленных данных счётчика электрической энергии Меркурий 230.
 * Для чтения данных используется дескриптор последовательного порта или сетевого сокета.
 * Функция выполняет чтение накопленных данных для указанного количества тарифов.
 * Функция возвращает 1 при успешном чтении данных. При ошибке функция возвращает:
 * 0  - счётчик не отвечает на запрос;
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param handle Дескриптор последовательного порта или сетевого сокета.
 * @param addr Сетевой адрес счётчика: 1..240, 0 - широковещательный адрес.
 * @param tarif Количество читаемых тарифов: 0..4, 0 - читать только суммарные данные.
 * @param data Указатель на структуру накопленных данных.
 * @return int 
 */
int Merc230_ReadTotalData(int handle, uint8_t addr, uint8_t tarif, merc230_total_t *data);

__END_DECLS

#endif
