/**
 * @file mqtt.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Клиент MQTT. Поддержка версии протокола v3.1.1.
 * @version 0.1
 * @date 2024-04-01
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 * Библиотека разработана на основе оригинальной версии библиотеки "MQTT-C":
 * https://github.com/LiamBindle/MQTT-C
 * @copyright Copyright(c) 2018 Liam Bindle.
 * MIT License
 */

#ifndef MQTT_H
#define MQTT_H

#include <cdefs.h>

#include <threads.h>
#include <stddef.h>
#include <time.h>
#include <url.h>

__BEGIN_DECLS

typedef int     mqtt_pal_socket_handle;
typedef mtx_t   mqtt_pal_mutex_t;

/** 
 * @brief Макро используется для описания элементов перечисления MQTTErrors и соответствующих им
 * сообщений об ошибках.
 */
#define __ALL_MQTT_ERRORS(MQTT_ERROR)                    \
    MQTT_ERROR(MQTT_ERROR_NULLPTR)                       \
    MQTT_ERROR(MQTT_ERROR_CONTROL_FORBIDDEN_TYPE)        \
    MQTT_ERROR(MQTT_ERROR_CONTROL_INVALID_FLAGS)         \
    MQTT_ERROR(MQTT_ERROR_CONTROL_WRONG_TYPE)            \
    MQTT_ERROR(MQTT_ERROR_CONNECT_CLIENT_ID_REFUSED)     \
    MQTT_ERROR(MQTT_ERROR_CONNECT_NULL_WILL_MESSAGE)     \
    MQTT_ERROR(MQTT_ERROR_CONNECT_FORBIDDEN_WILL_QOS)    \
    MQTT_ERROR(MQTT_ERROR_CONNACK_FORBIDDEN_FLAGS)       \
    MQTT_ERROR(MQTT_ERROR_CONNACK_FORBIDDEN_CODE)        \
    MQTT_ERROR(MQTT_ERROR_PUBLISH_FORBIDDEN_QOS)         \
    MQTT_ERROR(MQTT_ERROR_SUBSCRIBE_TOO_MANY_TOPICS)     \
    MQTT_ERROR(MQTT_ERROR_MALFORMED_RESPONSE)            \
    MQTT_ERROR(MQTT_ERROR_UNSUBSCRIBE_TOO_MANY_TOPICS)   \
    MQTT_ERROR(MQTT_ERROR_RESPONSE_INVALID_CONTROL_TYPE) \
    MQTT_ERROR(MQTT_ERROR_CONNECT_NOT_CALLED)            \
    MQTT_ERROR(MQTT_ERROR_SEND_BUFFER_IS_FULL)           \
    MQTT_ERROR(MQTT_ERROR_SOCKET_ERROR)                  \
    MQTT_ERROR(MQTT_ERROR_MALFORMED_REQUEST)             \
    MQTT_ERROR(MQTT_ERROR_RECV_BUFFER_TOO_SMALL)         \
    MQTT_ERROR(MQTT_ERROR_ACK_OF_UNKNOWN)                \
    MQTT_ERROR(MQTT_ERROR_NOT_IMPLEMENTED)               \
    MQTT_ERROR(MQTT_ERROR_CONNECTION_REFUSED)            \
    MQTT_ERROR(MQTT_ERROR_SUBSCRIBE_FAILED)              \
    MQTT_ERROR(MQTT_ERROR_CONNECTION_CLOSED)             \
    MQTT_ERROR(MQTT_ERROR_INITIAL_RECONNECT)             \
    MQTT_ERROR(MQTT_ERROR_INVALID_REMAINING_LENGTH)      \
    MQTT_ERROR(MQTT_ERROR_CLEAN_SESSION_IS_REQUIRED)     \
    MQTT_ERROR(MQTT_ERROR_RECONNECT_FAILED)              \
    MQTT_ERROR(MQTT_ERROR_RECONNECTING)

/** 
 * @brief Макро для формирования перечисления MQTTErrors
*/
#define GENERATE_ENUM(ENUM) ENUM,

/** 
 * @brief Перечисление кодов ошибок MQTT.
 * Сообщение об ошибке может быть получено функцией mqtt_client_error_str.
 */
enum MQTTErrors {
    MQTT_ERROR_UNKNOWN = INT_MIN,
    __ALL_MQTT_ERRORS(GENERATE_ENUM)
    MQTT_OK = 1
};

/**
 * @brief Описание сообщения, полученного от брокера MQTT.
*/
struct mqtt_response_publish {
    uint8_t dup_flag;                       ///< Флаг DUP: 0-первая попытка брокера отправить сообщение, 1-повторные попытки брокера
    uint8_t qos_level;                      ///< Уровень QOS
    uint8_t retain_flag;                    ///< Флаг RETAIN в принятом сообщении
    uint16_t topic_name_size;               ///< Размер имени топика (кол-во символов)
    const void* topic_name;                 ///< Имя топика
    uint16_t packet_id;                     ///< Идентификатор сообщения
    const void* application_message;        ///< Принятое сообщение
    size_t application_message_size;        ///< Размер принятого сообщения
};

/**
 * @brief Тип пакета MQTT
 */
enum MQTTControlPacketType {
    MQTT_CONTROL_CONNECT    =   1u,
    MQTT_CONTROL_CONNACK    =   2u,
    MQTT_CONTROL_PUBLISH    =   3u,
    MQTT_CONTROL_PUBACK     =   4u,
    MQTT_CONTROL_PUBREC     =   5u,
    MQTT_CONTROL_PUBREL     =   6u,
    MQTT_CONTROL_PUBCOMP    =   7u,
    MQTT_CONTROL_SUBSCRIBE  =   8u,
    MQTT_CONTROL_SUBACK     =   9u,
    MQTT_CONTROL_UNSUBSCRIBE=   10u,
    MQTT_CONTROL_UNSUBACK   =   11u,
    MQTT_CONTROL_PINGREQ    =   12u,
    MQTT_CONTROL_PINGRESP   =   13u,
    MQTT_CONTROL_DISCONNECT =   14u
};

/**
 * @brief Статус сообщения в очереди.
 */
enum MQTTQueuedMessageState {
    MQTT_QUEUED_UNSENT,
    MQTT_QUEUED_AWAITING_ACK,
    MQTT_QUEUED_COMPLETE
};

/**
 * @brief Значения, используемые в параметре publish_flags при передаче сообщения брокеру (функция mqtt_client_publish).
 */
enum MQTTPublishFlags {
    MQTT_PUBLISH_DUP        = 8u,
    MQTT_PUBLISH_QOS_0      = ((0u << 1) & 0x06),           ///< QOS0: издатель один раз отправляет сообщение брокеру и не ждет подтверждения
    MQTT_PUBLISH_QOS_1      = ((1u << 1) & 0x06),           ///< QOS1: гарантирует, что сообщение точно будет доставлено брокеру, но есть вероятность дублирования сообщений
    MQTT_PUBLISH_QOS_2      = ((2u << 1) & 0x06),           ///< QOS2: гарантирует доставку сообщений и исключается возможное дублирование отправленных сообщений
    MQTT_PUBLISH_QOS_MASK   = ((3u << 1) & 0x06),
    MQTT_PUBLISH_RETAIN     = 0x01                          ///< признак необходимости сохранения значения брокером
};

/**
 * @brief Структура: сообщение в очереди сообщений.
 */
struct mqtt_queued_message {
    uint8_t *start;                                         ///< Указатель на начало сообщения
    size_t size;                                            ///< Количество байт в сообщении
    enum MQTTQueuedMessageState state;                      ///< Статус сообщения в очереди
    time_t time_sent;                                       ///< Время отправки сообщения
    enum MQTTControlPacketType control_type;                ///< Тип сообщения
    uint16_t packet_id;                                     ///< Идентификатор сообщения
};

/**
 * @brief Структура: очередь сообщений.
 */
struct mqtt_message_queue {
    void *mem_start;                                        ///< Начало очереди в буфере памяти
    void *mem_end;                                          ///< Конец очереди в буфере памяти
    //
    uint8_t *curr;
    size_t curr_sz;
    struct mqtt_queued_message *queue_tail;
};

/**
 * @brief Прототип функции обратного вызова, которая вызывается при получении сообщения от брокера.
 * Все топики, на которые вы подписаны, будут возвращены брокером в виде сообщений mqtt_response_publish.
 * Все публикации, полученные от брокера, будут переданы в эту функцию.
 * В функцию обратного вызова всегда передается указатель на publish_response_callback_state.
 * Используйте publish_response_callback_state, чтобы отслеживать любую необходимую информацию о состоянии.
 * @param state Указатель на publish_response_callback_state.
 * @param publish Указатель на сообщение.
 */
typedef void (publish_callback_t)(void** state, struct mqtt_response_publish *publish);

/**
 * @brief Элемент подписки MQTT
 */
struct mqtt_subscription_item {
    char* topic_name;                                       ///< Имя или маска имени топика
    int max_qos_level;                                      ///< Уровень QOS с которым блокер должен присылать сообщения
    //
    struct mqtt_subscription_item *next;                    ///< Следующий элемент подписки
};

struct mqtt_params {
    url_t *url;                                             ///< Структура с адресом брокера
    //
    uint8_t* sendbuf;                                       ///< Буфер передачи
    size_t sendbufsz;
    //
    uint8_t* recvbuf;                                       ///< Буфер приёма
    size_t recvbufsz;
    //
    char* will_topic;                                       ///< Наименование WILL топика
    void* will_message;                                     ///< Сообщение WILL
    size_t will_message_size;                               ///< Размер сообщения WILL
    //
    uint16_t keep_alive;                                    ///< Интервал проверки связи в секундах
    uint32_t tcp_timeout;
    //
    struct mqtt_subscription_item *subscriptions;           ///< Список элементов подписки
    //
    thrd_t  thread_id;                                      ///< Идентификатор потока
    volatile uint8_t thread_terminate;                      ///< Признак необходимости завершить поток
    volatile uint8_t thread_finished;                       ///< Признак завершения потока
};

/**
 * @brief Структура описания состояния клиента MQTT.
 */
struct mqtt_client {
    struct mqtt_params params;                              ///< Прощальный топик и список подписки
    //
    mqtt_pal_socket_handle socketfd;                        ///< Сокет для подключения к брокеру
    uint16_t pid_lfsr;                                      ///< Статус LFSR для генерации ID пакетов
    uint16_t keep_alive;                                    ///< Время keep-alive в секундах
    int number_of_keep_alives;                              ///< Счётчик отправленных ping
    int number_of_ping_resp;                                ///< Счётчик принятых подтверждений
    size_t send_offset;                                     ///< Смещение в буфере передачи
    time_t time_of_last_send;                               ///< Время последнего переданного сообщения
    enum MQTTErrors error;                                  ///< Статус ошибки клиента. При норме MQTT_OK, до вызова mqtt_connect значение MQTT_ERROR_CONNECT_NOT_CALLED.
    int response_timeout;                                   ///< Время ожидания ACK после передачи сообщения. По умолчанию 30 секунд, можно менять в любой момент
    int number_of_timeouts;                                 ///< Счётчик тайм-аутов
    float typical_response_time;                            ///< Примерная задержка перед ответом брокера
    publish_callback_t *publish_response_callback;          ///< Функция обработки получения сообщения от брокера
    void* publish_response_callback_state;                  ///< Указатель на информацию о состоянии publish_response_callback. Указатель на этот указатель передаётся при вызове publish_response_callback
    /**
     * @brief A user-specified callback, triggered on each \ref mqtt_sync, allowing the user to perform state inspections (and custom socket error detection) on the client.
     * 
     * This callback is triggered on each call to \ref mqtt_sync.
     * If it returns MQTT_OK then mqtt_sync will continue normally (performing reads and writes).
     * If it returns an error then \ref mqtt_sync will not call reads and writes.
     * 
     * This callback can be used to perform custom error detection, namely platform specific socket error detection, and force the client into an error state.
     * 
     * This member is always initialized to NULL but it can be manually set at any time.
     */
    enum MQTTErrors (*inspector_callback)(struct mqtt_client*);
    void (*reconnect_callback)(struct mqtt_client*, void**);    ///< Функция обработки ошибок. В ней может быть выполнено пересоздание сокета, переустановка соединения и т.п.
    void *reconnect_state;                                      ///< Указатель на состояние. Указатель на этот элемент передается в mqtt_client.reconnect_callback.
    /**
     * @brief Буфер временного хранения входящих данных.
     */
    struct {
        uint8_t *mem_start;                                 ///< Указатель на приёмный буфер
        size_t mem_size;                                    ///< Размер приёмного буфера
        uint8_t *curr;                                      ///< Указатель на следующее доступное для записи местоположение в буфере приема
        size_t curr_sz;                                     ///< Количество байт, которые доступны для записи в данный момент
    } recv_buffer;
    mqtt_pal_mutex_t mutex;                                 ///< Мьютекс
    struct mqtt_message_queue mq;                           ///< Очередь передачи сообщений
};

// -----------------------------------------------------------------------------------------------------------------------------

/**
 * @brief Иницилизация (обнуление) полей структуры mqtt_client.
 * Функция должна быть вызвана первой, перед вызовом остальных функци mqtt_client_...
 * @param client Указатель на структуру mqtt_client.
 */
void mqtt_client_init(struct mqtt_client *client);

/**
 * @brief Устанавливает значение WILL (прощального) сообщения.
 * Функция может быть вызвана не более одного раза после mqtt_client_init(), но до вызова mqtt_client_start().
 * Функция выделяет память, которая освобождается при вызове mqtt_client_stop().
 * Без вызова функции прощальное сообщение не используется.
 * @param client Указатель на структуру mqtt_client.
 * @param will_topic Имя топика.
 * @param will_message Содержимое сообщения.
 * @param will_message_size Размер содержимого сообщения.
 */
void mqtt_client_set_willmessage(struct mqtt_client *client, const char* will_topic, const void* will_message, size_t will_message_size);

/**
 * @brief Добавление элемента в список подписки.
 * Функция может быть вызвана произвольное количество раз, но обязательно после mqtt_client_init().
 * Каждый вызов функции добавляет новый элемент подписки.
 * Функция выделяет память, которая освобождается при вызове mqtt_client_stop().
 * Без вызова функции подписка (и соответственно получение сообщений от брокера) не используется.
 * @param client Указатель на структуру mqtt_client.
 * @param topic_name Имя (или маска имён) топика.
 * @param max_qos_level Уровень сервиса, который должен использоваться брокером при отправке сообщение по этой подписке.
 */
void mqtt_client_add_subscription(struct mqtt_client *client, const char* topic_name, int max_qos_level);

/**
 * @brief Установка нового значения интервала проверки связи (keep_alive).
 * Функция позволяет изменить значение интервала проверки связи с брокером.
 * Значение интервала по умолчанию составляет 300 секунд.
 * @param client Указатель на структуру mqtt_client.
 * @param keep_alive Новое значение интервала в секундах.
 */
void mqtt_client_set_keep_alive(struct mqtt_client *client, uint16_t keep_alive);

/**
 * @brief Установка максимального времени на установку TCP-соединения.
 * Значение по умолчанию 4000мс.
 * @param client Указатель на структуру mqtt_client.
 * @param timeout Новое значение, мс.
 */
void mqtt_client_set_tcp_timeout(struct mqtt_client *client, uint32_t tcp_timeout);

/**
 * @brief Запуск процесса подключения и работы с MQTT брокером.
 * Функция запускает выполнение отдельного потока, который производит подключение к брокеру,
 * производит переподключения при обрывых соединения и выполняет все необходимые операции чтения и записи данных.
 * 
 * Параметр url должен содержать URL c реквизитами подключения к брокеру MQTT.
 * Строка должна содержать:
 * - идентификатор протокола: mqtt или mqtts;
 * - имя хоста или IP-адрес брокера.
 * Строка может содержать:
 * - логин и пароль для подключения к брокеру;
 * - номер используемого порта TCP (по умолчанию: mqtt:1883, mqtts:8883).
 * Примеры:
 * mqtt://test.mosquitto.org
 * mqtt://login:password@test.mosquitto.org:1884
 * mqtts://test.mosquitto.org
 * mqtts://login:password@test.mosquitto.org:8885
 * Размеры буферов для передачи и приёма сообщений зависят от размера передваемых/принимаемых сообщений,
 * используемого уровня QOS и интенсивности обмена данными. Размер sendbuf должен быть достаточно большим,
 * чтобы вместить несколько целых сообщений mqtt. Размер recvbuf должен превышать размер любого ожидаемого сообщения.
 * В параметре publish_callback должен быть указатель на функцию обработки сообщений принимаемых от брокера.
 * @param client Указатель на структуру mqtt_client.
 * @param url Строка адреса MQTT брокера.
 * @param sendbufsz Размер буфера памяти для передачи (например 2048).
 * @param recvbufsz Размер буфера памяти для приёма (например 1024).
 * @param publish_callback Функция обработки приёма сообщений брокера.
 * @return int 1-процесс запущен успешно, 0-ошибка (неверный url или нехватка ресурсов).
 */
int mqtt_client_start(struct mqtt_client *client, char *url, size_t sendbufsz, size_t recvbufsz, publish_callback_t *publish_callback);

/**
 * @brief Остановка процесса работы с MQTT брокером.
 * Функция останавливает работу с MQTT брокером и освобождает все занятые ресурсы.
 * @param client Указатель на структуру mqtt_client.
 */
void mqtt_client_stop(struct mqtt_client *client);

/** 
 * @brief Возвращает текстовое сообщение об ошибке MQTT.
 * @param[in] error Код ошибки.
 * @returns Сообщение об ошибке.
 */
const char* mqtt_client_error_str(enum MQTTErrors error);

/**
 * @brief Передача (публикация) значения топика брокеру MQTT.
 * 
 * @param[in,out] client Клиент MQTT.
 * @param[in] topic_name Имя топика.
 * @param[in] application_message Данные топика.
 * @param[in] application_message_size Размер данных в байтах.
 * @param[in] publish_flags Флаги: уровень QOS (MQTT_PUBLISH_QOS_[0,1,2]) и MQTT_PUBLISH_RETAIN (необходимость сохранения сообщения на брокере).
 * @returns При успехе возвращает MQTT_OK, иначе одно из значений перечисления MQTTErrors.
 */
enum MQTTErrors mqtt_client_publish(struct mqtt_client *client, const char* topic_name, const void* application_message, size_t application_message_size, uint8_t publish_flags);

__END_DECLS

#endif
