/**
 * @file netvars.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с сетевыми переменными.
 * 
 * 1. Передача сетевых переменных.
 * Для передачи сетевых используется функция NetVarsSend. Для отправки функция использует значения 
 * глобальных переменных NetVarsPort (UDP-порт) и NetVarsGroup.
 * 2. Приём (получение) сетевых переменных.
 * Имеется два способа приёма (получения) сетевых переменных:
 * - с использованием функции обратного вызова (callback);
 * - c использованием глобальных массивов идентификаторов и значений.
 * В первом случае необходимо использовать функцию NetVarsStartEx, во стором - NetVarsStart.
 * В обоих случаях для приёма используется UDP-порт номер которого указан в переменной NetVarsPort.
 * Значение переменной NetVarsPort должно быть установлено до вызова функций NetVarsStartEx или NetVarsStart.
 * Для завершения приёма сетевых переменных в обоих случаях используется функция NetVarsStop.
 * @example Приём переменных через callback
 * // функция callback
 * void nvcb(int32_t sender, uint16_t group, uint16_t cnt, net_var_t *vars) {
 *    ...
 * }
 * // запуск приёма
 * int res = NetVarsStartEx(nvcb);
 * 
 * @example Приём переменных через глобальные массивы
 * int nums[2];
 * int vals[2];
 * //
 * nums[0] = 1;
 * nums[1] = 5;
 * //
 * int res= NetVarsStart(nums, vals, NULL, 2);
 * 
 * @example Передача сетевых переменных
 * int nums[2];
 * int vals[2];
 * //
 * nums[0] = 1;
 * nums[1] = 5;
 * //
 * vals[0] = 10;
 * vals[1] = 20;
 * int res = NetVarsSend(nums, vals, 2);
 * 
 * @version 0.1
 * @date 2024-04-21
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef NETVARS_H
#define NETVARS_H

#include <cdefs.h>
#include <stdint.h>

__BEGIN_DECLS

/**
 * @brief UDP-порт используемый для приёма и отправки сетевых переменных.
 * Значение по умолчанию: 3301.
 */
extern uint16_t NetVarsPort;

/**
 * @brief Идентификатор группы сетевых переменных.
 * Значение по умолчанию: 1.
 */
extern uint16_t NetVarsGroup;

/**
 * @brief Счётчик принятых пакетов, содержащих сетевые переменные.
 */
extern unsigned int NetVarsPacketCount;

/**
 * @brief Структура описания сетевой переменной.
 */
typedef struct net_var {
    int32_t         id;                     ///< Идентификатор (номер) переменной
    int32_t         value;                  ///< Значение переменной
} net_var_t;

/**
 * @brief Функция обработки получения сетевых переменных.
 * @param sender Серийный номер отправителя.
 * @param group Идентификатор группы сетевых переменных.
 * @param cnt Количество полученных переменных.
 * @param vars Указатель на массив полученных переменных.
 */
typedef void (*netvars_func)(int32_t sender, uint16_t group, uint16_t cnt, net_var_t *vars);

/**
 * @brief Передать значение одной или нескольких сетевых переменных.
 * При отправке используются значения глобальных переменных NetVarsPort и NetVarsGroup.
 * Функция возвращает 1 при успешной отправке или 0 при ошибке.
 * @param nums Массив идентификаторов (номеров) сетевых переменных.
 * @param vals Массив значений сетевых переменных.
 * @param cnt Количество передаваемых переменных.
 * @return int 1 - успех, 0 - ошибка.
 */
int NetVarsSend(int *nums, int *vals, unsigned short cnt);

/**
 * @brief Запуск работы модуля приёма сетевых переменных.
 * Функция запускает выполнение отдельного потока приёма сетевых переменных.
 * При приёме одной или нескольких сетевых переменных производится вызов функции cb, которая должна выполнить
 * соответствующую обработку полученных значений. Вызов функции выполняется в контексте потока обрабатывающего приём сетевых переменных.
 * При приёме используется значение глобальной переменной NetVarsPort (номер UDP порта).
 * Функция обнуляет значение количества принятых сетевых пакетов NetVarsPacketCount.
 * Функция возвращает 1 при успешном выполнении или 0 при ошибке.
 * @param cb Функция - обработчик получения сетевых переменных.
 * @return int  1 - успех, 0 - ошибка.
 */
int NetVarsStartEx(netvars_func cb);

/**
 * @brief Запуск работы модуля приёма сетевых переменных.
 * Функция запускает выполнение отдельного потока приёма сетевых переменных. Идентификаторы принимаемых переменных
 * должны быть указаны в массиве nums. При приёме сетевой переменной её значение помещается в массив vals.
 * Также, если sns!=NULL, то серийный номер отправителя помещается в массив sns.
 * Массивы nums, vals и sns (если он используется) должны иметь глобальную область видимости или выделены функцией malloc.
 * При приёме используются значения глобальных переменных NetVarsPort (номер UDP порта) и NetVarsGroup (номер группы переменных).
 * Функция обнуляет значение количества принятых сетевых пакетов NetVarsPacketCount.
 * Функция возвращает 1 при успешном выполнении или 0 при ошибке.
 * @param nums Массив идентификаторов (номеров) сетевых переменных.
 * @param vals Массив значений сетевых переменных.
 * @param sns Массив серийных номеров отправителей (может быть NULL).
 * @param cnt Количество сетевых переменных.
 * @return int 1 - успех, 0 - ошибка.
 */
int NetVarsStart(int *nums, int *vals, unsigned int *sns, unsigned short cnt);

/**
 * @brief Завершение работы модуля приёма сетевых переменных.
 * Функция останавливает и дожидается завершения работы потока приёма сетевых 
 * переменных и закрывает используемый при работе модуля сетевой сокет.
 */
void NetVarsStop(void);

__END_DECLS

#endif
