/**
 * @file pgsql.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с СУБД PostgreSQL.
 * @version 0.1
 * @date 2024-04-25
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef PGSQL_H
#define PGSQL_H

#include <cdefs.h>

__BEGIN_DECLS

/**
 * @brief Состояние сервера PostgreSQL
 */
enum pg_state_enum {
    pgsNONE             =   0,          ///< Клиент сервера Postgre SQL не запущен
    pgsCONNECTING       =   1,          ///< Подключение к серверу
    pgsDISCONNECTED     =   2,          ///< Отключены от сервера
    pgsAUTH             =   3,          ///< Авторизация на сервере
    pgsOK               =   4,          ///< Подключены и авторизованы на сервере
    pgsSQL              =   5,          ///< Выполняется SQL запрос
    pgsSQL_OK           =   6,          ///< Сервер успешно выполнил SQL запрос
    pgsSQL_BUFSIZE      =   7,          ///< Недостаточный размер буфера ответа
    pgsSQL_ERROR        =   8,          ///< При выполнении SQL запроса на сервере произошла ошибка
    pgsSQL_TIMEOUT      =   9,          ///< За время таймаута не получен ответ сервера
    pgsTCP_ERROR        =   10          ///< Ошибка в работе TCP сокета
};

/**
 * @brief Установка максимальной продолжительности подключения и выполнение SQL запроса.
 * @param timeout Таймаут, мс.
 */
void PgSetTimeout(unsigned int timeout);

/**
 * @brief Установка кодировки символов для клиента СУБД.
 * Список поддерживаемых кодировок можно узнать из документации СУБД: https://www.postgresql.org/docs/current/multibyte.html.
 * Пример: "UTF8", "WIN1251", "WIN866".
 * По умаолчанию используется кодировка сервера.
 * @param encoding Строка с наименованием кодировки.
 */
void PgSetEncoding(char *encoding);

/**
 * @brief Установка буфера, используемого для ответов СУБД.
 * @param buf Указатель на буфер в котором будут размещён ответ сервера.
 * @param bufsize Размер буфера buf.
 */
void PgSetBuffer(void *buf, unsigned int bufsize);

/**
 * @brief Функция запускает процесс подключения к СУБД PostgreSQL асинхронном (неблокирующем) режиме.
 * При успешном начале подключения к серверу функция возвращает значение 1. При ошибке функция возвращает 0.
 * Ошибка может быть вызвана некорректным значением параметров, отсутствием свободные TCP сокетов или слишком
 * маленьким размером буфера buf.
 * @param host Адрес сервера: имя или IP-адрес.
 * @param port Номер порта TCP.
 * @param database Наименование базы данных.
 * @param login Имя пользователя для подключения к СУБД.
 * @param password Пароль для подключения к СУБД.
 * @return int 1-успех, 0-ошибка.
 */
int PgOpen(char *host, unsigned short int port, char *database, char *login, char *password);

/**
 * @brief Функция выполняет подключение к СУБД PostgreSQL в синхронном (блокирующем) режиме.
 * При успешном подключении к СУБД функция возвращает значение 1. При ошибке функция возвращает 0.
 * Ошибка может быть вызвана некорректным значением параметров, недоступностью СУБД, отсутствием свободные TCP сокетов или 
 * слишком маленьким размером буфера buf.
 * @param host Адрес сервера: имя или IP-адрес.
 * @param port Номер порта TCP.
 * @param database Наименование базы данных.
 * @param login Имя пользователя для подключения к СУБД.
 * @param password Пароль для подключения к СУБД.
 * @return int 1-успех, 0-ошибка.
 */
int PgOpenBlock(char *host, unsigned short int port, char *database, char *login, char *password);

/**
 * @brief Функция выполняет отключение от СУБД PostgreSQL.
 * Вызов функции безопасен, в том числе, если подключение к серверу отсутствует.
 */
void PgClose(void);

/**
 * @brief Возвращает состояние подключение к СУБД PostgreSQL.
 * @return int 0-подключение отсутствует, 1-подключение активно.
 */
int PgConnected(void);

/**
 * @brief Функция заполняет переданный буфер buf пятибайтовой строкой с кодом ошибки выполнение SQL запроса.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * @param buf Указатель на буфер (минимум 5 байт).
 * @return int 1-успех, 0-ошибка.
 */
int PgErrorCode(char *buf);

/**
 * @brief Поучение текста ошибки выполнения запроса.
 * Функция помещает в буфер text сообщение об ошибке вызванной выполнением последнего запроса.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * @param text Указатель на буфер.
 * @param textsize Размер буфера.
 * @return int 1 - успех, 0 - ошибка.
 */
int PgErrorText(char *text, unsigned int textsize);

/**
 * @brief Начать выполнение SQL запроса в неблокирующем (асинхронном) режиме.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * Ошибка может быть вызвана отсутствием подключения к СУБД, повторным вызовом функции (запрос уже выполняется) или некоректным значением аргумента.
 * @param request Строка SQL-запроса.
 * @return int 1-успех, 0-ошибка.
 */
int PgSQL(char *request);

/**
 * @brief Начать выполнение SQL запроса в блокирующем (синхронном) режиме.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * Ошибка может быть вызвана отсутствием подключения к СУБД, повторным вызовом функции (запрос уже выполняется) или некоректным значением аргумента.
 * @param request Строка SQL-запроса.
 * @return int 1-успех, 0-ошибка.
 */
int PgSQLBlock(char *request);

/**
 * @brief Получение состояния клиента СУБД PostgreSQL.
 * Результат функции может принимать значения присутствующие в перечислении pg_state_enum.
 * @return int Состояние клиентского подключения.
 */
int PgState(void);

/**
 * @brief Получение длины ответа СУБД на запрос.
 * Значением можно воспользоваться при возникновении ошибки "Сервер выполнил запрос, но переданный буфер слишком мал для приёма ответа сервера" (pgsSQL_BUFSIZE).
 * Следует проверить значение длины ответа и увеличить размер буфера buf в функции PgSetBuffer или изменить SQL запрос.
 * @return unsigned int 
 */
unsigned int PgDataLen(void);

/**
 * @brief Получение количества колонок в ответе СУБД на SQL запрос, выполненный при помощи функции PgSql.
 * Если при выполнении SQL запроса СУБД вернула пустой ответ или произошла ошибка, то возвращается значение 0.
 * @return unsigned int Количество колонок в ответе СУБД.
 */
unsigned int PgColCount(void);

/**
 * @brief Получение количества строк в ответе СУБД на SQL запрос, выполненный при помощи функции PgSql.
 * Если при выполнении SQL запроса СУБД вернула пустой ответ или произошла ошибка, то возвращается значение 0.
 * @return unsigned int Количество строк в ответе СУБД.
 */
unsigned int PgRowCount(void);

/**
 * @brief Получение имени колонки в ответе СУБД на SQL запрос, выполненный при помощи функции PgSql.
 * При успешном выполнении функция помещает в буфер name имя запрошенной колонки и возвращает знаечение 1.
 * При ошибке функция возвращает 0.
 * @param col Номер колонки для которой запрашивается имя: 0..PgColCount()-1.
 * @param name Буфер для имени колонки.
 * @param namesize Размер буфера name.
 * @return int 1-успех, 0-ошибка.
 */
int PgColName(unsigned int col, char *name, unsigned int namesize);

/**
 * @brief Получение текстового значения ячейки с номером колонки col и номером строки row из ответа СУБД.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * @param col Номер колонки: 0..PgColCount()-1
 * @param row Номер строки: 0..PgRowCount()-1
 * @param cell Указатель на буфер, куда будет записан текст значения указанной ячейки.
 * @param cellsize Размер буфера cell.
 * @return int 1-успех, 0-ошибка.
 */
int PgCell(unsigned int col, unsigned int row, char *cell, unsigned int cellsize);

/**
 * @brief Получение числового (int) значения ячейки с номером колонки col и номером строки row из ответа СУБД.
 * При успешном выполнении функция помещает числовое значение в value и возвращает значение 1, при ошибке - 0.
 * @param col Номер колонки: 0..PgColCount()-1
 * @param row Номер строки: 0..PgRowCount()-1
 * @param value Указатель на значение.
 * @return int 1-успех, 0-ошибка.
 */
int PgCellAsInteger(unsigned int col, unsigned int row, int *value);

/**
 * @brief Получение числового (double) значения ячейки с номером колонки col и номером строки row из ответа СУБД.
 * При успешном выполнении функция помещает числовое значение в value и возвращает значение 1, при ошибке - 0.
 * @param col Номер колонки: 0..PgColCount()-1
 * @param row Номер строки: 0..PgRowCount()-1
 * @param value Указатель на значение.
 * @return int 1-успех, 0-ошибка.
 */
int PgCellAsDouble(unsigned int col, unsigned int row, double *value);

__END_DECLS

#endif
