/**
 * @file rc4.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Шифрование по алгоритму RC4 (потоковый шифр).
 * @version 0.1
 * @date 2024-05-02
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef RC4_H
#define RC4_H

#include <cdefs.h>
#include <stdint.h>
#include <stddef.h>

__BEGIN_DECLS

// #define TEST_RC4

typedef struct {
    uint8_t     se[256], sd[256];
    uint32_t    pose, posd;
    uint8_t     te, td;
} RC4_ctx;

/**
 * @brief Инициализация контекста шифрования.
 * @param ctx Контекст шифрования.
 * @param key Ключ шифрования.
 * @param key_len Размер ключа шифрования (1..256).
 */
void rc4_ks(RC4_ctx *ctx, const uint8_t *key, uint32_t key_len);

/**
 * @brief Шифрование RC4.
 * @param ctx Контекст шифрования.
 * @param src Входные (незашифрованные) данные.
 * @param dst Выходные (зашифрованные) данные.
 * @param len Размер входных данных.
 */
void rc4_encrypt(RC4_ctx *ctx, const uint8_t *src, uint8_t *dst, uint32_t len);

/**
 * @brief Расшифровка RC4.
 * 
 * @param ctx Контекст шифрования.
 * @param src Входные (зашифрованные) данные.
 * @param dst Выходные (расшифрованные) данные.
 * @param len Размер входных данных.
 */
void rc4_decrypt(RC4_ctx *ctx, const uint8_t *src, uint8_t *dst, uint32_t len);

#ifdef TEST_RC4

/**
 * @brief Встроенный тест RC4.
 * При корректном выполнении должен вернуть значение 7.
 * @return int 
 */
int test_rc4();

#endif

__END_DECLS

#endif
