/**
 * @file regexp.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Простая реализация регулярных выражений.
 * @version 0.1
 * @date 2024-04-12
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 * Поддерживаемые токены:
 * ----------------------
 * '.'        Точка, соответствует любому символу
 * '^'        Соответствует началу строки
 * '$'        Соответствует концу строки
 * '*'        Повторяется ноль или более раз (greedy)
 * '+'        Повторяется один или более раз (greedy)
 * '?'        Повторяется ноль или один раз (non-greedy)
 * '[abc]'    Соответствует одному из символов {'a', 'b', 'c'}
 * '[^abc]'   Не соответствует ни одному из символов {'a', 'b', 'c'} (пока не реализовано)
 * '[a-zA-Z]' Соответствие диапазону символов { a-z | A-Z }
 * '\s'       Соответствует пробелу или символам \t \f \r \n \v
 * '\S'       Соответствует символам не являющимся пробелами
 * '\w'       Соответствует символу или цифре [a-zA-Z0-9_]
 * '\W'       Соответствует НЕ символу или цифре
 * '\d'       Соответствие цифре [0-9]
 * '\D'       Соответствие НЕ цифре [0-9]
 *
 * @example
 * // Standard int to hold length of match
 * int match_length;
 * 
 * // Standard null-terminated C-string to search:
 * const char* string_to_search = "ahem.. 'hello world !' ..";
 * 
 * // Compile a simple regular expression using character classes, meta-char and greedy + non-greedy quantifiers:
 * re_t pattern = re_compile("[Hh]ello [Ww]orld\\s*[!]?");
 * 
 * // Check if the regex matches the text:
 * int match_idx = re_matchp(pattern, string_to_search, &match_length);
 * if (match_idx != -1) {
 *   printf("match at idx %i, %i chars long.\n", match_idx, match_length);
 * }
 */

#ifndef _TINY_REGEX_H
#define _TINY_REGEX_H

#include <cdefs.h>

__BEGIN_DECLS

#ifndef RE_DOT_MATCHES_NEWLINE
/* Define to 0 if you DON'T want '.' to match '\r' + '\n' */
#define RE_DOT_MATCHES_NEWLINE 1
#endif

/**
 * @brief Тип данных для хранения скомпилированного регулярного выражения.
 */
typedef struct regex_t* re_t;

/**
 * @brief Компиляция регулярного выражения.
 * @param pattern Строка шаблона.
 * @return re_t 
 */
re_t re_compile(const char* pattern);


/**
 * @brief Поиск совпадения скомпилированного шаблона внутри текста.
 * Результат содержит индекс первого символа в строке, которое соответствует шаблону регулярного выражения.
 * Если соответствие не найдено, то функция возвращает -1.
 * Переданный указатель будет содержать длину совпадения.
 * @param pattern Указатель на скомпилированный шаблон.
 * @param text Текст
 * @param matchlength 
 * @return int 
 */
int re_matchp(re_t pattern, const char* text, int* matchlength);

/**
 * @brief Поиск совпадения шаблона внутри текста.
 * Строка шаблона компилируется автоматически.
 * Результат содержит индекс первого символа в строке, которое соответствует шаблону регулярного выражения.
 * Если соответствие не найдено, то функция возвращает -1.
 * Переданный указатель будет содержать длину совпадения.
 * @param pattern Текст шаблона.
 * @param text Текст.
 * @param matchlength 
 * @return int 
 */
int re_match(const char* pattern, const char* text, int* matchlength);

__END_DECLS

#endif /* ifndef _TINY_REGEX_H */
