/**
 * @file satcrc.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции вычисления контрольной суммы.
 * @version 0.1
 * @date 2024-03-28
 * 
 * @copyright Copyright (c) 1999-2018 Lammert Bies
 * https://github.com/lammertb/libcrc
 * 
 * This file is licensed under the MIT License as stated below
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 */

#ifndef SATCRC_H
#define SATCRC_H

#include <cdefs.h>
#include <stdint.h>
#include <stddef.h>

__BEGIN_DECLS

/*
 * #define CRC_POLY_xxxx
 *
 * Определение полиномов для вычисления CRC
 */
#define     CRC_POLY_16     0xA001
#define     CRC_POLY_32     0xEDB88320ul
#define     CRC_POLY_64     0x42F0E1EBA9EA3693ull
#define     CRC_POLY_CCITT  0x1021
#define     CRC_POLY_DNP    0xA6BC
#define     CRC_POLY_KERMIT 0x8408
#define     CRC_POLY_SICK   0x8005

/*
 * #define CRC_START_xxxx
 *
 * Значения используемые для инициализации алгоритмов CRC.
 */
#define     CRC_START_8             0x00
#define     CRC_START_16            0x0000
#define     CRC_START_MODBUS        0xFFFF
#define     CRC_START_XMODEM        0x0000
#define     CRC_START_CCITT_1D0F    0x1D0F
#define     CRC_START_CCITT_FFFF    0xFFFF
#define     CRC_START_KERMIT        0x0000
#define     CRC_START_SICK          0x0000
#define     CRC_START_DNP           0x0000
#define     CRC_START_32            0xFFFFFFFFul
#define     CRC_START_64_ECMA       0x0000000000000000ull
#define     CRC_START_64_WE         0xFFFFFFFFFFFFFFFFull

/**
 * @brief Вычисление контрольной суммы, используемой в протоколе NMEA.
 * Процедура не пытается проверить саму строку.
 * Начальный символ '$' будет проигнорирован, поскольку этот символ является 
 * частью предложения NMEA, но не частью вычисления контрольной суммы.
 * Вычисление останавливается, когда обнаружен CR, LF, символ '*' или конец строки.
 * Поскольку проверка синтаксиса NMEA не задействована, функция всегда возвращается с 
 * успехом, если в качестве параметра не указан нулевой указатель. 
 * Возвращаемое значение - это указатель на буфер результатов, предоставленный вызывающим 
 * приложением, или NULL в случае ошибки.
 * @param input_str Проверяемая строка
 * @param result Результат вычисления
 * @return unsigned char* 
 */
unsigned char *checksum_NMEA(const unsigned char *input_str, unsigned char *result);

/**
 * @brief Расчёт значения контрольной суммы crc8
 * 
 * @param input_str Указатель на массив данных
 * @param num_bytes Длина массива
 * @return uint8_t 
 */
uint8_t crc_8(const unsigned char *input_str, size_t num_bytes);

/**
 * @brief Расчёт значения контрольной суммы crc16
 * 
 * @param input_str Указатель на массив данных
 * @param num_bytes Длина массива
 * @return uint16_t 
 */
uint16_t crc_16(const unsigned char *input_str, size_t num_bytes);

/**
 * @brief Расчёт значения контрольной суммы crc32
 * 
 * @param input_str Указатель на массив данных
 * @param num_bytes Длина массива
 * @return uint32_t 
 */
uint32_t crc_32(const unsigned char *input_str, size_t num_bytes);

/**
 * @brief Расчёт значения контрольной суммы crc16 по алгоритму, используемому в протоколе "Modbus RTU"
 * 
 * @param input_str Указатель на массив данных
 * @param num_bytes Длина массива
 * @return uint16_t 
 */
uint16_t crc_modbus(const unsigned char *input_str, size_t num_bytes);

/**
 * @brief Синоним для crc_modbus() для совместимости с "C-Script".
 */
#define ModbusCrc16(data, cnt)   crc_modbus(data, cnt)

/**
 * @brief Расчёт значения контрольной суммы crc16 по алгоритму, используемому в датчиках производства компании "Sick".
 * 
 * @param input_str Указатель на массив данных
 * @param num_bytes Длина массива
 * @return uint16_t 
 */
uint16_t crc_sick(const unsigned char *input_str, size_t num_bytes);

/**
 * @brief Побайтное вычисление контрольной суммы по алгоритму crc8.
 * 
 * @param crc Текущее значение контрольной суммы
 * @param c Новый байт
 * @return uint8_t Новое значение контрольной суммы
 */
uint8_t update_crc_8(uint8_t  crc, unsigned char c);

/**
 * @brief Побайтное вычисление контрольной суммы по алгоритму crc16.
 * 
 * @param crc Текущее значение контрольной суммы
 * @param c Новый байт
 * @return uint16_t Новое значение контрольной суммы
 */
uint16_t update_crc_16(uint16_t crc, unsigned char c);

/**
 * @brief Побайтное вычисление контрольной суммы по алгоритму crc32.
 * 
 * @param crc Текущее значение контрольной суммы
 * @param c Новый байт
 * @return uint32_t Новое значение контрольной суммы
 */
uint32_t update_crc_32(uint32_t crc, unsigned char c);

/**
 * @brief Побайтное вычисление контрольной суммы по алгоритму, используемому в 
 * датчиках производства компании "Sick".
 * @param crc Текущее значение контрольной суммы
 * @param c Новый байт
 * @return uint16_t Новое значение контрольной суммы
 */
uint16_t update_crc_sick(uint16_t crc, unsigned char c, unsigned char prev_byte);

/**
 * @brief Функция выполняет однопроходное вычисление контрольной суммы по алгоритму XModem CRC.
 * 
 * @param input_str Указатель на массив байт
 * @param num_bytes Количество байт
 * @return uint16_t 
 */
uint16_t crc_xmodem( const unsigned char *input_str, size_t num_bytes );

/**
 * @brief Функция выполняет однопроходное вычисление контрольной суммы по алгоритму CCITT CRC.
 * Для инициализации используется значение 0x1D0F.
 * @param input_str Указатель на массив байт
 * @param num_bytes Количество байт
 * @return uint16_t 
 */
uint16_t crc_ccitt_1d0f( const unsigned char *input_str, size_t num_bytes );

/**
 * @brief Функция выполняет однопроходное вычисление контрольной суммы по алгоритму CCITT CRC.
 * Для инициализации используется значение 0xFFFF.
 * @param input_str Указатель на массив байт
 * @param num_bytes Количество байт
 * @return uint16_t 
 */
uint16_t crc_ccitt_ffff( const unsigned char *input_str, size_t num_bytes );

/**
 * @brief Побайтное вычисление контрольной суммы по алгоритму CRC-CCITT.
 * 
 * @param crc Текущее значение контрольной суммы
 * @param c Новый байт
 * @return uint16_t Новое значение контрольной суммы
 */
uint16_t update_crc_ccitt( uint16_t crc, unsigned char c );

/**
 * @brief Функция выполняет однопроходное вычисление контрольной суммы по алгоритму CRC-64/ECMA
 * (European Computer Manufacturers Association).
 * @param input_str Указатель на массив байт
 * @param num_bytes Количество байт
 * @return uint64_t
 */
uint64_t crc_64_ecma( const unsigned char *input_str, size_t num_bytes );

/**
 * @brief Функция выполняет однопроходное вычисление контрольной суммы по алгоритму CRC-64/WE.
 * @param input_str Указатель на массив байт
 * @param num_bytes Количество байт
 * @return uint64_t
 */
uint64_t crc_64_we( const unsigned char *input_str, size_t num_bytes );

/**
 * @brief Побайтное вычисление контрольной суммы по алгоритму CRC-64.
 * 
 * @param crc Текущее значение контрольной суммы
 * @param c Новый байт
 * @return uint64_t Новое значение контрольной суммы
 */
uint64_t update_crc_64( uint64_t crc, unsigned char c );

__END_DECLS

#endif
