/**
 * @file satmem.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с BACKUPRAM и EPROM.
 * @version 0.1
 * @date 2024-03-14
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef SATMEM_H
#define SATMEM_H

#include <cdefs.h>
#include <plcmodel.h>
#include <stddef.h>
#include <stdint.h>

__BEGIN_DECLS

#if HW_SIZE_BACKUPRAM>0

/**
 * @brief Чтение байта данных из энергонезависимой памяти BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти 0...(HW_SIZE_BACKUPRAM-1)
 * @return uint8_t
 */
uint8_t BackupRamRead8(uint16_t adr);

/**
 * @brief Чтение uint16_t данных из энергонезависимой памяти BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти 0...(HW_SIZE_BACKUPRAM-2)
 * @return uint16_t
 */
uint16_t BackupRamRead16(uint16_t adr);

/**
 * @brief Чтение uint32_t данных из энергонезависимой памяти BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти 0...(HW_SIZE_BACKUPRAM-4)
 * @return uint32_t 
 */
uint32_t BackupRamRead32(uint16_t adr);

/**
 * @brief Чтение массива данных из энергонезависимой памяти BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти 0...(HW_SIZE_BACKUPRAM-1)
 * @param data Указатель на массив данных
 * @param len Количество читаемых байт
 */
void BackupRamRead(uint16_t adr, unsigned char * data, size_t len);

/**
 * @brief Сохранение байта данных в энергонезависимой памяти BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти 0...(HW_SIZE_BACKUPRAM-1)
 * @param val Записываемое значение
 */
void BackupRamWrite8(uint16_t adr, uint8_t val);

/**
 * @brief Запись uint16_t в энергонезависимую память BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти 0...(HW_SIZE_BACKUPRAM-2)
 * @param val Записываемое значение
 */
void BackupRamWrite16(uint16_t adr, uint16_t val);

/**
 * @brief Запись uint32_t в энергонезависимую память BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти 0...(HW_SIZE_BACKUPRAM-4)
 * @param val Записываемое значение
 */
void BackupRamWrite32(uint16_t adr, uint32_t val);

/**
 * @brief Запись массива данных в энергонезависимую память BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти 0...(HW_SIZE_BACKUPRAM-1)
 * @param data Указатель на массив данных
 * @param len Количество записываемых байт
 */
void BackupRamWrite(uint16_t adr, unsigned char * data, size_t len);

#endif              // HW_SIZE_BACKUPRAM>0

#if HW_SIZE_EPROM>0

/**
 * @brief Чтение данных из энергонезависимой памяти EPROM.
 * 
 * @param adr Адрес 0..(HW_SIZE_EPROM/2-1)
 * @return uint16_t 
 */
uint16_t ReadEprom(uint16_t adr);

/**
 * @brief Сохранение данных в энергонезависимой памяти EPROM.
 * 
 * @param adr Адрес 0..(HW_SIZE_EPROM/2-1)
 * @param data Записываемые данные
 */
void WriteEprom(uint16_t adr, uint16_t data);

#endif              // HW_SIZE_EPROM>0

__END_DECLS

#endif
