/**
 * @file satmenu.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Системное меню.
 * @version 0.1
 * @date 2024-03-28
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef SATMENU_H
#define SATMENU_H

#include <cdefs.h>
#include <sysdefs.h>

__BEGIN_DECLS

/**
 * @brief Типы сигналов для функции SetSigName()
 */
#define SIGNAL_TYPE_DI      0           // Дискретные входы
#define SIGNAL_TYPE_DO      1           // Дискретные выходы
#define SIGNAL_TYPE_AI      2           // Аналоговые входы
#define SIGNAL_TYPE_AO      3           // Аналоговые выходы
#define SIGNAL_TYPE_TI      4           // Входы датчиков температуры

/**
 * @brief Отображение системного меню.
 * При вызове функции будет показано системное меню контроллера.
 */
void SystemMenu(void);

/**
 * @brief Проверка статуса системного меню.
 * Если системное меню отображается, то функция возвращает 1, в противном случае - 0.
 * @return int 
 */
int SystemMenuVisible(void);

/**
 * @brief Добавление пункта в системное меню.
 * Функция добавляет в системное меню пункт связанный с переменной типа "int".
 * @param var Указатель на переменную связанную с пунктом меню
 * @param title Имя пункта меню
 * @param min Минимальное значение переменной
 * @param max Максимальное значение переменной
 * @return sysmenu_item_t* Указатель на новый пункт меню
 */
sysmenu_item_t *AddSystemMenuIntItem(const char *title, int *var, int min, int max);

/**
 * @brief Добавление пункта в системное меню.
 * Функция добавляет в системное меню пункт связанный с переменной типа "float".
 * @param var Указатель на переменную связанную с пунктом меню
 * @param title Имя пункта меню
 * @param min Минимальное значение переменной
 * @param max Максимальное значение переменной
 * @param precision Количество знаков после запятой при отображении значения
 * @return sysmenu_item_t* Указатель на новый пункт меню
 */
sysmenu_item_t *AddSystemMenuFloatItem(const char *title, float *var, float min, float max, int precision);

/**
 * @brief Добавление пункта в системное меню.
 * Функция добавляет в системное меню пункт связанный с переменной типа "int", которая может иметь значения 0 (нет) или 1 (да).
 * @param var Указатель на переменную связанную с пунктом меню
 * @param title Имя пункта меню
 * @return sysmenu_item_t* Указатель на новый пункт меню
 */
sysmenu_item_t *AddSystemMenuBoolItem(const char *title, int *var);

/**
 * @brief Удаление пункта из системного меню.
 * 
 * @param item Указатель на пункт меню.
 */
void DeleteSystemMenuItem(sysmenu_item_t *item);

/**
 * @brief Удаление всех пунктов из системного меню.
 */
void ClearSystemMenu(void);

/**
 * @brief Функция позволяет задать имена сигналов, отображаемые в системных меню.
 * @param type Тип сигнала: 0-дискретные входы,1-дискретные выходы,2-аналоговые входы,3-аналоговые выходы,4-датчики температуры
 * @param index Номер сигнала
 * @param name Наименование сигнала
 * @return signame_item_t* Указатель на новый пункт описания имени сигнала
 */
signame_item_t *SetSigName(int type, int index, const char *name);

__END_DECLS

#endif
