/**
 * @file satpid.h
 * @brief Реализация алгоритма PID-регулятора
 * @version 0.1
 * @date 2024-03-17
 * @copyright Copyright (c) 2020 Philip Salmony
 * @author Philip Salmony
 * https://github.com/pms67/PID
 * 
 * @example
 * #include "PID.h"
 * #define PID_KP  2.0f
 * #define PID_KI  0.5f
 * #define PID_KD  0.25f
 * #define PID_TAU 0.02f
 * 
 * #define PID_LIM_MIN -10.0f
 * #define PID_LIM_MAX  10.0f
 * 
 * #define PID_LIM_MIN_INT -5.0f
 * #define PID_LIM_MAX_INT  5.0f
 * 
 * #define SAMPLE_TIME_S 0.01f
 * 
 * // Maximum run-time of simulation
 * #define SIMULATION_TIME_MAX 4.0f
 * 
 * // Simulated dynamical system (first order)
 * float TestSystem_Update(float inp);
 * 
 * int main()
 * {
 *     // Initialise PID controller
 *     PIDController pid = { PID_KP, PID_KI, PID_KD, PID_TAU, PID_LIM_MIN, PID_LIM_MAX, PID_LIM_MIN_INT, PID_LIM_MAX_INT, SAMPLE_TIME_S };
 *     PIDController_Init(&pid);
 *     // Simulate response using test system
 *     float setpoint = 1.0f;
 *     printf("Time (s)\tSystem Output\tControllerOutput\r\n");
 *     for (float t = 0.0f; t <= SIMULATION_TIME_MAX; t += SAMPLE_TIME_S) {
 *         // Get measurement from system
 *         float measurement = TestSystem_Update(pid.out);
 *         // Compute new control signal
 *         PIDController_Update(&pid, setpoint, measurement);
 *         printf("%f\t%f\t%f\r\n", t, measurement, pid.out);
 *     }
 *     return 0;
 * }
 * float TestSystem_Update(float inp) {
 *     static float output = 0.0f;
 *     static const float alpha = 0.02f;
 *     output = (SAMPLE_TIME_S * inp + output) / (1.0f + alpha * SAMPLE_TIME_S);
 *     return output;
 * }
 */

#ifndef PID_CONTROLLER_H
#define PID_CONTROLLER_H

#include <cdefs.h>

__BEGIN_DECLS

typedef struct {

    /* Controller gains */
    float Kp;                                           ///< Коэффициент P
    float Ki;                                           ///< Коэффициент I
    float Kd;                                           ///< Коэффициент D

    /* Derivative low-pass filter time constant */
    float tau;

    /* Output limits */
    float limMin;                                       ///< Нижний предел выходного значения
    float limMax;                                       ///< Верхний предел выходного значения
    
    /* Integrator limits */
    float limMinInt;                                    ///< Нижний предел выходного значения интегратора
    float limMaxInt;                                    ///< Верхний предел выходного значения интегратора

    /* Sample time (in seconds) */
    float T;

    /* Controller "memory" */
    float integrator;
    float prevError;			/* Required for integrator */
    float differentiator;
    float prevMeasurement;		/* Required for differentiator */

    /* Controller output */
    float out;

} PIDController;

void  PIDController_Init(PIDController *pid);
float PIDController_Update(PIDController *pid, float setpoint, float measurement);

__END_DECLS

#endif
