/**
 * @file satscr.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с экраном контроллера.
 * @version 0.2
 * @date 2024-03-13
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef SATSCR_H
#define SATSCR_H

#include <cdefs.h>
#include <plcmodel.h>
#include <sysdefs.h>

__BEGIN_DECLS

#define SCREEN_WIDTH        HW_SCREEN_WIDTH
#define SCREEN_HEIGHT       HW_SCREEN_HEIGHT

#if SCREEN_WIDTH>0

#define _FONT12x16          0
#define _FONT18x32          1

#define COLOR_RED           0xF800      ///< Красный
#define COLOR_BLUE          0x001F      ///< Синий
#define COLOR_GREEN         0x07E0      ///< Зелёный
#define COLOR_YELLOW        0xFFE0      ///< Жёлтый
#define COLOR_BLACK         0x0000      ///< Чёрный
#define COLOR_WHITE         0xFFFF      ///< Белый
#define COLOR_GRAY          0x8410      ///< Серый
#define COLOR_LIGHTGRAY     0xC618      ///< Светло-серый
#define COLOR_SILVER        0xC618      ///< Серебристый
#define COLOR_CYAN          0x07FF      ///< Голубой
#define COLOR_MAGENTA       0xF81F      ///< Фиолетовый
#define COLOR_VALVE         0xF300      ///< Фиолетовый
#define COLOR_ORANGE        0xF300      ///< Оранжевый

// ---------------------------------------------------------------------------------------------------------

// ESC-последовательности для установки цвета символов

// Цвет текста
#define BLK         "\e[30m"            ///< Чёрный           RGB(0,0,0)
#define RED         "\e[31m"            ///< Красный          RGB(170,0,0)
#define GRN         "\e[32m"            ///< Зелёный          RGB(0,170,0)
#define YEL         "\e[33m"            ///< Жёлтый           RGB(170,85,0)
#define BLU         "\e[34m"            ///< Синий            RGB(0,0,170)
#define MAG         "\e[35m"            ///< Фиолетовый       RGB(170,0,170)
#define CYN         "\e[36m"            ///< Голубой          RGB(0,170,170)
#define WHT         "\e[37m"            ///< Светло-серый     RGB(170,170,170)

// Яркий цвет текста
#define HBLK        "\e[90m"            ///< Светло-серый     RGB(85,85,85)
#define HRED        "\e[91m"            ///< Ярко красный     RGB(255,85,85)
#define HGRN        "\e[92m"            ///< Ярко зелёный     RGB(85,255,85)
#define HYEL        "\e[93m"            ///< Ярко жёлтый      RGB(255,255,85)
#define HBLU        "\e[94m"            ///< Светло-синий     RGB(85,85,255)
#define HMAG        "\e[95m"            ///< Светло-пурпурный RGB(255,85,255)
#define HCYN        "\e[96m"            ///< Светло-голубой   RGB(85,255,255)
#define HWHT        "\e[97m"            ///< Белый            RGB(255,255,255)

// Цвет фона
#define BLKB        "\e[40m"            ///< Чёрный           RGB(0,0,0)
#define REDB        "\e[41m"            ///< Красный          RGB(170,0,0)
#define GRNB        "\e[42m"            ///< Зелёный          RGB(0,170,0)
#define YELB        "\e[43m"            ///< Жёлтый           RGB(170,85,0)
#define BLUB        "\e[44m"            ///< Синий            RGB(0,0,170)
#define MAGB        "\e[45m"            ///< Фиолетовый       RGB(170,0,170)
#define CYNB        "\e[46m"            ///< Голубой          RGB(0,170,170)
#define WHTB        "\e[47m"            ///< Светло-серый     RGB(170,170,170)

// Яркий цвет фона
#define BLKHB       "\e[100m"           ///< Светло-серый     RGB(85,85,85)
#define REDHB       "\e[101m"           ///< Ярко красный     RGB(255,85,85)
#define GRNHB       "\e[102m"           ///< Ярко зелёный     RGB(85,255,85)
#define YELHB       "\e[103m"           ///< Ярко жёлтый      RGB(255,255,85)
#define BLUHB       "\e[104m"           ///< Светло-синий     RGB(85,85,255)
#define MAGHB       "\e[105m"           ///< Светло-пурпурный RGB(255,85,255)
#define CYNHB       "\e[106m"           ///< Светло-голубой   RGB(85,255,255)
#define WHTHB       "\e[107m"           ///< Белый            RGB(255,255,255)

// Сброс на состояние по умолчанию
#define CRESET      "\e[0m"

// Очистить экран
#define CLRSCR      "\e[2m"

/**
 * @brief Глобальный указатель на описания файлов ресурсов.
 */
extern const flash_file FlashFiles[];

/**
 * @brief Структура описания экрана
 */
extern struct _SATPLC_SCREEN    _sat_screen;

/**
 * @brief Цвет кисти рисования на дисплее
 */
#define BrushColor  _sat_screen.BrushColor

/**
 * @brief Цвет шрифта (букв)
 */
#define FontColor   _sat_screen.FontColor

/**
 * @brief Цвет пера (линий) рисования на дисплее
 */
#define PenColor    _sat_screen.PenColor

/**
 * @brief Толщина пера в пикселях
 */
#define PenWidth    _sat_screen.PenWidth
 
/**
 * @brief Крупный шрифт
 */
#define BigFont     _sat_screen.FontIndex
 
/**
 * @brief Очистка экрана и заливка цветом кисти.
 */
void ClrScr();

/**
 * @brief Установка курсора (указателя рисования и вывода текста) в положение x, y на дисплее.
 * 
 * @param x 
 * @param y 
 */
void GotoXY(short int x,short int y);

/**
 * @brief Вывод на дисплей целого числа (только значащие цифры).
 * @param val Значение числа
 */
void PrintInt(int val);

/**
 * @brief Вывод на дисплей целого числа заданным количеством цифр (заполнение нулями слева).
 * 
 * @param val Значение числа
 * @param len Количество цифр
 */
void PrintIntLen(int val,unsigned char len);

/**
 * @brief Получение цвета по заданным значениям составляющих R, G и B.
 * @param r Красный
 * @param g Зелёный
 * @param b Синий
 * @return short int 
 */
#define RGB(r,g,b) ((((r)<<8) & 0xF800) | (((g)<<3) & 0x07E0) | (((b)>>3) & 0x001F))
// short int RGB(unsigned char r,unsigned char g,unsigned char b);

/**
 * @brief Установка цвета пикселя дисплея
 * 
 * @param x Координата X
 * @param y Координата Y
 * @param color Цвет пикселя
 */
void SetPixel(int x, int y, int color);

/**
 * @brief Рисование линии из позиции курсора в позицию x,y при помощи текущей настройки пера.
 * 
 * @param x Координата X
 * @param y Координата Y
 */
void LineTo(int x, int y);

/**
 * @brief Рисование прямоугольника. Контур обводится пером, а внутренняя часть заполняется цветом кисти.
 * 
 * @param left Координата левой границы
 * @param top Координата верхней границы
 * @param right Координата правой границы
 * @param bottom Координата нижней границы
 */
void Rectangle(int left, int top, int right, int bottom);

/**
 * @brief Рисование прямоугольника, заполненного цветом кисти. Перо не используется.
 * @param left Координата левой границы
 * @param top Координата верхней границы
 * @param right Координата правой границы
 * @param bottom Координата нижней границы
 */
void RectangleBrush(int left, int top, int right, int bottom);

/**
 * @brief Рисование контура прямоугольника при помощи пера. Внутренняя часть не заполняется. Кисть не используется.
 * @param left Координата левой границы
 * @param top Координата верхней границы
 * @param right Координата правой границы
 * @param bottom Координата нижней границы
 */
void RectanglePen(int left, int top, int right, int bottom);

/**
 * @brief Рисование эллипса или круга. Контур обводится пером, а внутренняя часть заполняется цветом кисти.
 * 
 * @param left Координата левой границы
 * @param top Координата верхней границы
 * @param right Координата правой границы
 * @param bottom Координата нижней границы
 */
void Ellipse(int left, int top, int right, int bottom);

/**
 * @brief Рисование эллипса или круга пером, внутренняя часть не заполняется. Кисть не используется.
 * 
 * @param left Координата левой границы
 * @param top Координата верхней границы
 * @param right Координата правой границы
 * @param bottom Координата нижней границы
 */
void EllipsePen(int left, int top, int right, int bottom);

/**
 * @brief Рисование картинки в заданном месте дисплея. Картинка может рисоваться с прозрачным цветом.
 * 
 * @param x Координата X
 * @param y Координата Y
 * @param id Указатель на файл картинки
 * @param transparent Использовать прозрачность
 */
void DrawPicture(int x, int y, flash_file *id, char transparent);

/**
 * @brief Завершение рисования экрана (копирование видеообласти).
 */
void ShowScr(void);

/**
 * @brief Возвращает указатель на описание файла с именем name в файловой системе filesystem.
 * Если файл найден, то функция возвращает указатель на структуру описания файла. Если файл не найден, то функция
 * возвращает NULL.
 * @param filesystem Указатель на файловую систему
 * @param name Имя файла
 * @return flash_file* 
 */
flash_file *OpenFile(const flash_file *filesystem, const char *name);

#endif      // SCREEN_WIDTH>0

__END_DECLS

#endif
