/**
 * @file satser.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Работа с последовательным портом и Mosbus RTU.
 * @version 0.1
 * @date 2024-03-27
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef SATSER_H
#define SATSER_H

#include <cdefs.h>
#include <stddef.h>

__BEGIN_DECLS

// константы скорости обмена
#define B110        110
#define B300        300
#define B600        600
#define B1200       1200
#define B2400       2400
#define B4800       4800
#define B9600       9600
#define B19200      19200
#define B38400      38400
#define B56000      56000
#define B57600      57600
#define B115200     115200

// нет контроля чётности/нечётности
#define PARNONE     0

// контроль нечётности
#define PARODD      1

// контроль чётности
#define PARENB      2

/**
 * @brief Открыть последовательный порт
 * При успешном выполнении функция возвращает дескриптор: положительное число - идентификатор последовательного порта, который затем необходимо
 * использовать в функциях чтения и записи данных.
 * При ошибке функция возвращает значение меньшее или равное 0.
 * @param port Номер последовательного порта: 1,2,...
 * @param baudrate Скорость обмена: B110,B300,B600,B1200,B2400,B4800,B9600,B19200,B38400,B56000,B57600,B115200
 * @param parity Контроль чётности: 0-None, 1-Odd, 2-Even
 * @param stopbits Количество стоп-бит: 1,2
 * @return int 
 */
int SerialOpen(int port, int baudrate, int parity, int stopbits);

/**
 * @brief Проверка состояния последовательного порта.
 * Функция возвращает значение: 0 - порт закрыт, 1 - порт открыт.
 * @param port Номер последовательного порта: 1,2,...
 * @return int 
 */
int SerialIsOpen(int port);

/**
 * @brief Проверка дескриптора порта на корректность.
 * Возвращает 1, если handle явлается дескриптором открытого последовательного порта.
 * @param handle Дескриптор последовательного порта
 * @return int 
 */
int SerialHandleIsValid(int handle);

/**
 * @brief Закрыть последовательный порт
 * 
 * @param handle Указатель на дескриптор последовательного порта
 */
void SerialClose(int *handle);

/**
 * @brief Запись данных в последовательный порт
 * При успешном выполнении функция возвращает значение 1, при ошибке - значение 0
 * @param handle Дескриптор порта
 * @param data Указатель на массив данных
 * @param len Количество записываемых данных
 * @return int 
 */
int SerialWrite(int handle, char *data, size_t len);

/**
 * @brief Получить количество байтов, доступных для чтения из приёмного буфера последовательного порта.
 * 
 * @param handle Дескриптор порта
 * @return int 
 */
int SerialReadLength(int handle);

/**
 * @brief Прочитать данные из приёмного буфера последовательного порта.
 * Возвращаемое значение: количество прочитанных данных.
 * @param handle Дескриптор порта
 * @param data Указатель на массив данных
 * @param len Количество читаемых данных
 * @return int 
 */
int SerialRead(int handle, char *data, size_t len);

/**
 * @brief Прочитать блок данных указанной длины, в течении не более чем указанного времени.
 * Чтение выполняется в течении не более чем timeout мс.
 * При успешном выполнении возвращает 1, если данные не приняты - 0.
 * @param handle Дескриптор порта
 * @param data Указатель на массив данных
 * @param len Количество читаемых данных
 * @param timeout Максимальное время чтения данных, мс
 * @return int 
 */
int SerialReadBlock(int handle, char *data, size_t len, int timeout);

/**
 * @brief Очистка приёмного буфера последовательного порта.
 * 
 * @param handle  Дескриптор порта
 */
void SerialPurge(int handle);

/**
 * @brief Возвращает количество байт утерянных из-за переполнения приёмного буфера соединения.
 * Вызов функции обнуляет счётчик потерянных байт.
 * @param handle Дескриптор порта
 * @return int Количество потерянных байт или отрицательное число при ошибке.
 */
int SerialLost(int handle);

__END_DECLS

#endif
