/**
 * @file satsock.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с сетью: клиент TCP/IP, UDP/IP.
 * @version 0.1
 * @date 2024-04-04
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef SATSOCK_H
#define SATSOCK_H

#include <cdefs.h>
#include <limits.h>
#include <stddef.h>
#include <stdint.h>

__BEGIN_DECLS

/**
 * @brief Тип (стиль) сокета
 */
typedef enum _socket_style {
    SOCK_STREAM         =   1,              ///< Тип сокета: TCP
    SOCK_DGRAM          =   2               ///< Тип сокета: UDP
} socket_style_t;

/**
 * @brief Статус установки соединения с хостом
 */
enum socket_state_enum {
    sockBUSY             =  -1,             ///< Операция подключения к хосту выполняется и ещё не завершена 
    sockOK               =   0,             ///< Ошибок нет - есть подключение к хосту
    sockNOADR            =   1,             ///< Произошла ошибка - не задан адрес хоста
    sockDNS              =   2,             ///< Не удалось получить IP адрес хоста
    sockBADDNS           =   3,             ///< Произошла ошибка при выполнении DNS запроса
    sockNOCONNECTION     =   4,             ///< Произошла ошибка при выполнении подключения
    sockSERVER           =   5,             ///< Хост разорвал подключение
    sockNOSERVER         =   6,             ///< Не удалось подключиться к хосту
    sockSTART            =   7,             ///< Не удалось начать подключение к хосту
    sockWRITE            =   8,             ///< Произошла ошибка записи в TCP соединение
    sockBADSOCKET        =   9,             ///< Некорректное значение дескриптора сокета
    sockNOLINK           =   10             ///< Нет подключения интерфейса Ethernet
};

/**
 * @brief Статус состояния интерфейса Ethernet
 */
enum eth_state_enum {
    ETH_FAIL             =  -1,             ///< Неисправность PHY
    ETH_NO_LINK          =   0,             ///< Нет подключения по Ethernet
    ETH_LINK_10_HALF     =   1,             ///< Подключение 10 Half Duplex
    ETH_LINK_10_FULL     =   2,             ///< Подключение 10 Full Duplex
    ETH_LINK_100_HALF    =   3,             ///< Подключение 100 Half Duplex
    ETH_LINK_100_FULL    =   4              ///< Подключение 100 Full Duplex
};

/**
 * @brief Структура описания адреса IPv4
 */
struct in_addr
#define s_addr                    S_un.S_addr
#define s_host                    S_un.S_un_b.s_b2
#define s_net                     S_un.S_un_b.s_b1
#define s_imp                     S_un.S_un_w.s_w2
#define s_impno                   S_un.S_un_b.s_b4
#define s_lh                      S_un.S_un_b.s_b3
{
    union {
        uint8_t b[4];
        struct {
            uint8_t s_b1;
            uint8_t s_b2;
            uint8_t s_b3;
            uint8_t s_b4;
        } S_un_b;
        struct {
            uint16_t s_w1;
            uint16_t s_w2;
        } S_un_w;
        uint32_t S_addr;
    } S_un;
};
typedef struct in_addr IN_ADDR, *PIN_ADDR, *LPIN_ADDR;

#define INADDR_ANY          (uint32_t)(0)
#define INADDR_LOOPBACK     0x7F000001                      // 127.0.0.1
#define INADDR_BROADCAST    (uint32_t)(0xFFFFFFFF)          // 255.255.255.255
#define INADDR_NONE         0xFFFFFFFF

#define INADDR_LOOPBACK_S   "127.0.0.1"
#define INADDR_BROADCAST_S  "255.255.255.255"

#define MAX_IPv4_STR_LEN    16

/**
 * @brief Структура описания адреса сокета: порт и адрес IPv4
 */
struct sockaddr_in {
    uint16_t        sin_family;                             // AF_INET:2 не используется
    uint16_t        sin_port;                               // номер порта
    struct in_addr  sin_addr;                               // адрес IPv4
    char            sin_zero[8];                            // резерв
};
typedef struct sockaddr_in SOCKADDR_IN, *PSOCKADDR_IN, *LPSOCKADDR_IN;

/**
 * @brief Функция преобразует интернет-адрес IPv4 из строки со стандартной записью в виде цифр и точек в 
 * двоичные данные и сохраняет его в структуре in_addr, на которую указывает addr.
 * Функция возвращает ненулевое значение, если адрес корректный или 0 в случае ошибки.
 * @param cp Указатель на строку с IPv4 адресом
 * @param addr Указатель на структуру in_addr.
 * @return int 
 */
int inet_atoh(const char *cp, struct in_addr *addr);

/**
 * @brief Функция преобразует адрес addr в строку стандартной формы в виде цифр и точек.
 * Функция помещает строку по адресу, указанному в параметре buf.
 * Размер буфера buf должен быть не меньше чем MAX_IPv4_STR_LEN байт!
 * @param addr Указатель на структуру in_addr.
 * @param buf Указатель на буфер строки с адресом.
 * @return const char* 
 */
const char *inet_htop(struct in_addr *addr, char *buf);

/**
 * @brief Преобразует беззнаковое короткое целое число hostshort из порядка байтов хоста в сетевой порядок байтов.
 * @param v 
 * @return uint16_t 
 */
uint16_t htons(uint16_t v);

/**
 * @brief Преобразует целое число без знака v из порядка байтов хоста в сетевой порядок байтов.
 * @param v 
 * @return uint32_t 
 */
uint32_t htonl(uint32_t v);

/**
 * @brief Создание сетевого сокета указанного типа.
 * Функция создаёт сетевой сокет указанного типа: SOCK_STREAM (TCP/IP) или SOCK_DGRAM (UDP/IP).
 * При успешном выполнении возвращает дескриптор созданного сокета, при ошибке - значение 0.
 * Ошибка может быть вызвана тем, что указан неправильный тип или в системе нет свободных сокетов.
 * @param style SOCK_STREAM | SOCK_DGRAM
 * @return int Дескриптор сокета или 0 при ошибке.
 */
int CreateSocket(socket_style_t style);

/**
 * @brief Функция закрывает сетевое соединение (если оно установлено) и освобождает ресурсы системы
 * связанные с указанным сокетом.
 * @param socket Указатель на дескриптор сокета.
 */
void CloseSocket(int *socket);

/**
 * @brief Проверка дескриптора сокета на корректность.
 * Возвращает 1, если handle явлается дескриптором сетевого сокета или в противном случае 0.
 * @param socket Дескриптор сокета.
 * @return int 
 */
int SocketHandleIsValid(int socket);

/**
 * @brief Привязка сетевого сокета к локальному TCP или UDP порту.
 * Функция делает попытку привязки сокета к локальному порту с указанным номером.
 * Если функцию не использовать, то привязка выполнится автоматически к порту со случайным номером.
 * Если функция используется, то она должна быть вызвана до начала установки соединения.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * Ошибка может быть вызвана некорретным значением сокета или тем, что указанный порт уже
 * используется в системе.
 * @param socket Дескриптор сокета.
 * @param port Номер TCP или UDP порта.
 * @return int При успешном выполнении 1 или 0 при ошибке.
 */
int SocketBind(int socket, uint16_t port);

/**
 * @brief Установка размера приёмного буфера.
 * По умолчанию размер приёмного буфера 1024 байт. Размер приёмного буфера не может превышать 32768.
 * и не может быть менее 256 байт
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * @param socket Дескриптор сокета.
 * @param size Новый размер приёмного буфера.
 * @return int 
 */
int SocketSetBufSize(int socket, size_t size);

/**
 * @brief Получить значение адреса локального сокета.
 * При отсутствии ошибок функция заполняет поля структуры addr и возвращает значение 1.
 * При ошибке функция возвращает значение 0.
 * @param socket Дескриптор сокета.
 * @param addr Указатель на структуру адреса сокета.
 * @return int 
 */
int SocketGetLocalAddr(int socket, struct sockaddr_in *addr);

/**
 * @brief Получить значение адреса удалённого сокета.
 * При отсутствии ошибок функция заполняет поля структуры addr и возвращает значение 1.
 * При ошибке функция возвращает значение 0.
 * @param socket Дескриптор сокета.
 * @param addr Указатель на структуру адреса сокета.
 * @return int 
 */
int SocketGetRemoteAddr(int socket, struct sockaddr_in *addr);

/**
 * @brief Асинхронная (неблокирующая) установка соединения с хостом с указанным IP-адресом и портом.
 * Для сокетов с типом SOCK_STREAM функция начинает асинхронную процедуру установки соединения с указанным хостом.
 * Параметр timeout содержит максимальное время (мс) в течении которого будет выполняться попытка установки соединения.
 * Процесс установки соединения необходимо контролировать вызовами функции SocketState.
 * Для сокетов с типом SOCK_DGRAM функция при необходимости выполняет запрос к серверу DNS и заполняет внутренние структуры
 * указанными значениями адреса и номера порта.
 * В обоих случаях функция возвращает значение без задержки.
 * Сокет должен быть предварительно создан функцией CreateSocket().
 * Функция возвращает 1, если процесс подключения успешно запущен или 0 в случае ошибки.
 * Ошибка может быть вызвана некорректными значениями параметров или нехваткой системных ресурсов.
 * @param socket Дескриптор сокета.
 * @param host Строка именем или IP-адресом хоста.
 * @param port Номер порта (0..65535).
 * @param timeout Максимальное время для установки соединения, мс.
 * @return int При успешном выполнении 1 или 0 при ошибке.
 */
int SocketConnect(int socket, char *host, uint16_t port, uint32_t timeout);

/**
 * @brief Установка соединения с хостом с указанным IP-адресом и портом.
 * Для сокетов с типом SOCK_STREAM функция выполняет процедуру установки соединения с указанным хостом.
 * Параметр timeout содержит максимальное время (мс) в течении которого будет выполняться попытка установки соединения.
 * Для сокетов с типом SOCK_DGRAM функция заполняет внутренние структуры указанными значениями адреса и номера порта.
 * Сокет должен быть предварительно создан функцией CreateSocket().
 * Функция возвращает 1, если подключение успешно выполнено или 0 в случае ошибки.
 * Ошибка может быть вызвана некорректными значениями параметров, нехваткой системных ресурсов или невозможностью
 * подключения к указанному хосту.
 * @param socket Дескриптор сокета.
 * @param host Строка именем или IP-адресом хоста.
 * @param port Номер порта (0..65535).
 * @param timeout Максимальное время для установки соединения, мс.
 * @return int При успешном выполнении 1 или 0 при ошибке.
 */
int SocketConnectBlock(int socket, char *host, uint16_t port, uint32_t timeout);

/**
 * @brief Проверка статуса установки соединения с хостом.
 * Функция возвращает значение из перечисления socket_state_enum соответствующее текущему состоянию указанного сокета.
 * 
 * @param socket Дескриптор сокета.
 * @return int 
 */
int SocketState(int socket);

/**
 * @brief Проверка статуса TCP-соединения.
 * Для сокетов с типом SOCK_STREAM функция проверяет текущий статус TCP-соединения.
 * Функция возвращает 1,если соединение установлено или 0, если соединение отсутствует.
 * Для сокетов с типом SOCK_DGRAM функция всегда возвращает 1.
 * Если параметр socket содержит некорректное значение дескриптора, то функция возвращает 0.
 * @param socket Дескриптор сокета.
 * @return int При установленном соединении: 1, при отсутствии соединения или ошибке: 0.
 */
int SocketConnected(int socket);

/**
 * @brief Запись данных в указанный сокет.
 * Используемый сокет должен быть предварительно создан функцией CreateSocket и назначен
 * адрес удалённого хоста (для SOCK_DGRAM) или установлено соединение (для SOCK_STREAM)
 * функцией SocketConnect.
 * При успешном выполнении функция возвращает количество записанных байт данных, при ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * Ошибка может быть вызвана некорректным значением параметров или отсутствием соединения (для SOCK_STREAM).
 * @param socket Дескриптор сокета.
 * @param data Указатель на массив данных.
 * @param len Количество записываемых данных.
 * @return int Количество записанных данных или отрицательное значение при ошибке.
 */
int SocketWrite(int socket, void *data, size_t len);

/**
 * @brief Передача данных на указанный адрес.
 * Используется только с сокетами типа SOCK_DGRAM.
 * При успешном выполнении функция возвращает количество записанных байт данных, при ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета.
 * @param data Указатель на массив данных.
 * @param len Количество записываемых данных.
 * @param addr Указатель на структуру описывающую адрес назначения.
 * @return int 
 */
int SocketWriteTo(int socket, void *data, size_t len, struct sockaddr_in *addr);

/**
 * @brief Возвращает значение количества принятых и доступных для чтения данных.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета.
 * @return int Количество доступных для чтения байт в приёмном буфере или отрицательное значение при ошибке.
 */
int SocketReadLength(int socket);

/**
 * @brief Прочитать не более size байт данных из приёмного буфера сокета.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета.
 * @param data Указатель на массив данных.
 * @param size Количество читаймых байт данных.
 * @return int Количество прочитанных данных или отрицательное значение при ошибке.
 */
int SocketRead(int socket, void *data, size_t size);

/**
 * @brief Прочитать не более size байт данных из приёмного буфера сокета.
 * При использовании с сокетами типа SOCK_DGRAM функция присваивает адрес отправителя значению параметра addr.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета.
 * @param data Указатель на массив данных.
 * @param size Количество читаймых байт данных.
 * @return int Количество прочитанных данных или отрицательное значение при ошибке.
 */
int SocketReadFrom(int socket, void *data, size_t size, struct sockaddr_in *addr);

/**
 * @brief Прочитать блок данных указанной длины, в течении не более чем указанного времени.
 * Чтение выполняется в течении не более чем timeout мс.
 * При успешном выполнении возвращает 1, если данные не приняты - 0.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета
 * @param data Указатель на массив данных
 * @param len Количество читаемых данных
 * @param timeout Максимальное время чтения данных, мс
 * @return int 
 */
int SocketReadBlock(int socket, void *data, size_t len, int timeout);

/**
 * @brief Очистка приёмного буфера сокета.
 * @param socket Дескриптор сокета.
 */
void SocketClear(int socket);

/**
 * @brief Возвращает количество байт утерянных из-за переполнения приёмного буфера соединения.
 * Вызов функции обнуляет счётчик потерянных байт.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета.
 * @return int Количество потерянных байт или отрицательное число при ошибке.
 */
int SocketLost(int socket);

/**
 * @brief Получение текущего состояния интерфейса Ethernet.
 * Функция возвращает значение из перечисления eth_state_enum.
 * @return int 
 */
int EthernetLink(void);

/**
 * @brief Разрешение или запрет перезапуска интерфейса Ethernet при отсутствии подключения.
 * В случае разрешения (enabled=1) и отсутствия линка периодически будет выполняться инициализация интерфейса ethernet,
 * которая выполняется несколько секунд, в течение которых все функции контроллера приостанавливаются.
 * Если программе не требуется работа с IP сетью, то рекомендуется установить переменную в значение 0.
 * По умолчанию при старте программы перезапуск интерфейса ethernet запрещён.
 * @param enabled Запрет перезапуска: 0, разрешение 1.
 */
void EthernetRestart(int enabled);

__END_DECLS

#endif
