/**
 * @file satsun.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Астрономические функции.
 * @version 0.1
 * @date 2024-03-17
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef SATSUN_H
#define SATSUN_H

#include <cdefs.h>
#include <stdint.h>
#include <time.h>

__BEGIN_DECLS

/**
 * @brief Получение значения времени восхода солнца в определённой географической точке в указанный день.
 * 
 * @param latitude Широта географической точки
 * @param longitude Долгота географической точки
 * @param day День месяца (1-31)
 * @param month Номер месяца(1-12)
 * @param year Год (2022 ...)
 * @return unsigned int Количество секунд от начала суток или 0 при ошибке
 */
unsigned int TimeOfSunrise(float latitude,float longitude, uint8_t day, uint8_t month, uint16_t year);

/**
 * @brief Получение значения времени захода солнца в определённой географической точке в указанный день.
 * 
 * @param latitude Широта географической точки
 * @param longitude Долгота географической точки
 * @param day День месяца (1-31)
 * @param month Номер месяца(1-12)
 * @param year Год (2022 ...)
 * @return unsigned int Количество секунд от начала суток или 0 при ошибке
 */
unsigned int TimeOfSunset(float latitude,float longitude, uint8_t day, uint8_t month, uint16_t year);

/**
 * @brief Расчёт угла возвышения солнца над горизонтом
 * This is a function that calculates the sun position (altitude) at a given location (latitude and longitude) and time (Unix timestamp).
 * It uses the solar coordinate algorithm from the U.S. Naval Observatory Astronomical Applications Department.
 * @param lat Широта
 * @param lon Долгота
 * @param unix Текущее время
 * @return float Угол над горизонтом в градусах. Положительное значение - солнце над горизонтом, отрицательное значение - солнце под горизонтом.
 */
float SunPosition(float lat, float lon, time_t unix);

__END_DECLS

#endif
