/**
 * @file satthreads.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Управление потоками "ISO C threads".
 * @version 0.1
 * @date 2024-03-19
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef SATTHREADS_H
#define SATTHREADS_H

#include <cdefs.h>
#include <time.h>

__BEGIN_DECLS

/**
 * @brief Размер стека для потока
 */
#define THREAD_STACK_SIZE   4096

/**
 * @brief Результаты выполнения функций управления потоками
 */
enum {
    thrd_success = 0,           ///< Операция выполнена успешно
    thrd_nomem,                 ///< Запрошенная операция не удалась, поскольку не удалось выделить достаточно памяти
    thrd_timedout,              ///< Указанное время было достигнуто без получения запрошенного ресурса, обычно мьютекса или условной переменной
    thrd_busy,                  ///< Запрошенная операция не удалась, поскольку запрошенный ресурс уже используется
    thrd_error                  ///< Запрошенная операция не удалась
};

/**
 * @brief Состояние потока
 * Расширение стандарта ANSI C
 */
enum {
    thrd_stopped  = 0,          ///< Поток остановлен
    thrd_finished = 1,          ///< Поток завершил работу
    thrd_running  = 2,          ///< Поток выполняется
    thrd_sleeping = 3           ///< Поток выполняется, но сейчас в состоянии сна
};

/**
 * @brief Тип мьютекса
 * Используется при создании в функции mtx_init().
 * Возможные комбинации:
 * mtx_plain,
 * mtx_timed,
 * mtx_plain | mtx_recursive,
 * mtx_timed | mtx_recursive
 */
enum {
    mtx_plain       = 0x01,     ///< Простой мьютек
    mtx_recursive   = 0x02,     ///< Поддержка рекурсивной блокировки: поток-владелец может заблокировать его более одного раза, не вызывая взаимоблокировки.
    mtx_timed       = 0x04      ///< Мьютекс поддерживает тайм-ауты, ничем не отличается от mtx_plain, оставлен только для совместимости
};

/**
 * @brief Тип - идентификатор потока
 */
typedef int thrd_t;

/**
 * @brief Состояние мьютекса
 */
typedef enum {
    mtx_unlocked    = 0,        ///< Мьютек разблокирован
    mtx_locked      = 1         ///< Мьютек заблокирован
} mtx_state_t;

/**
 * @brief Тип - идентификатор мьютекса
 */
typedef struct {
    mtx_state_t state;          ///< Состояние мьютекса
    int type;                   ///< Тип мьютекса
    int owner;                  ///< Владелец мьютекса (идентификатор потока)
} mtx_t;

/**
 * @brief Идентификатор циклической блокировки.
 * Спин-блокировки являются аналогами мьютексов, позволяющими тратить меньше времени на процедуру блокировки потока,
 * поскольку не требуется переводить поток в заблокированное состояние.
 */
typedef int spin_t;

/**
 * @brief Тип флага, используемого call_once().
 */
typedef spin_t once_flag;

/**
 * @brief Значение используемое для инициализации once_flag.
 */
#define ONCE_FLAG_INIT 0

/**
 * @brief Функция потока
 */
typedef int (*thrd_start_t)(void *);

/**
 * @brief Атомарный вход в критическую секцию
 * @param addr 
 */
void __critical_section_enter(void *addr);

/**
 * @brief Атомарный выход из критической секции
 * @param addr 
 */
void __critical_section_leave(void *addr);

/**
 * @brief Атомарный обмен значениями
 * Функция атомарно записывает значение val по адресу addr, функция возвращает предыдущее значение по адресу addr.
 * @param addr Адрес памяти
 * @param val Записываемое значение
 * @return long Предыдущее значение addr
 */
long __atomic_swap(void *addr, long val);


/**
 * @brief Создание нового потока
 * 
 * @param thr 
 * @param func Функция потока
 * @param arg Аргументы для функции потока
 * @return int 
 */
int thrd_create(thrd_t *thr, thrd_start_t func, void *arg);

/**
 * @brief Функция возвращает идентификатор потока из которого она вызвана.
 * 
 * @return thrd_t 
 */
thrd_t thrd_current(void);

/**
 * @brief Проверка того, что lhs и rhs ссылаются на один и тот же поток.
 * Если lhs и rhs являются разными потоками, то функция возвращает 0, иначе - ненулевое значение.
 * @param lhs Идентификатор потока 1
 * @param rhs Идентификатор потока 2
 * @return int 
 */
int thrd_equal(thrd_t lhs, thrd_t rhs);

/**
 * @brief Приостановка выполнения потока на указанное количество миллисекунд.
 * Обратите внимание: в отличии от стандарта, функция имеет только один параметр.
 * Функция не может быть использована для приостановки выполнения основного потока!
 * @param msec Величина задержки в миллисекундах
 * @return int Остаток времени
 */
int thrd_sleep(uint32_t msec);

/**
 * @brief Передаёт управление другому потоку.
 * Функция формирует сообщение планировщику при возможности передать управление другому потоку.
 */
void thrd_yield (void);

/**
 * @brief Завершение выполнения текущего потока.
 * Функция не возвращает управление.
 * @param res Код возврата
 */
void thrd_exit(int res) __attribute__ ((noreturn));

/**
 * @brief Принудительное завершение выполнения потока thr.
 * Вызов освобождает память выделенную для стека завершаемого потока.
 * Если завершаемый поток в ходе своей работы запрашивал выделение памяти (malloc) или других ресурсов, то их освобождение не выполняется!
 * Вызов выполненный в отношении потока, который уже ранее завершил выполнение приведёт к возврату значения thrd_error.
 * Функция не может быть выполнена в отношении основного потока (вернёт значение thrd_error).
 * Функция не может быть выполнена в отношении текущего потока (вернёт значение thrd_error), используйте для этого thrd_exit().
 * Функция может вернуть значения thrd_success или thrd_error.
 * 
 * @param thr Идентификатор завершаемого потока
 * @return int 
 */
int thrd_detach(thrd_t thr);

/**
 * @brief Блокирует выполнение текущего потока до тех пор, пока выполнение потока thr не будет завершено.
 * Если значение res не равно NULL, то результирующий код потока thr помещается по указателю res.
 * Функция возвращает значение, когда поток thr завершит свою работу.
 * Функция может вернуть значения thrd_success или thrd_error.
 * @param thr Указатель на поток, завершения работы которого ожидается.
 * @param res Результат завершения потока.
 * @return int 
 */
int thrd_join(thrd_t thr, int *res);

/**
 * @brief Вызывает функцию func ровно один раз, даже если она вызывается из нескольких потоков.
 * Завершение функции func синхронизируется - со всеми предыдущими или последующими вызовами call_once с той же переменной флага.
 * @param flag Указатель на флаг.
 * @param func Указтель на функцию.
 */
void call_once(once_flag *flag, void (*func) (void));

/**
 * @brief Инициализация циклической блокировки
 * @example
 * spin_t crit_section;
 * spin_init(&crit_section);
 * ...
 * // начало критической секции
 * spin_lock(&crit_section);
 * ...
 * ...
 * // конец критической секции
 * spin_unlock(&crit_section);
 * 
 * @param p Указатель на идентификатор блокировки
 */
void spin_init(spin_t *p);

/**
 * @brief Вход в циклическую блокировку (критическая секция).
 * Спин-блокировки являются аналогами мьютексов, позволяющими тратить меньше времени на процедуру блокировки потока,
 * поскольку не требуется переводить поток в заблокированное состояние.
 * @param p Указатель на идентификатор блоировки
 */
void spin_lock(spin_t *p);

/**
 * @brief Выход из циклической блокировки (критическая секция).
 * 
 * @param p Указатель на идентификатор блокировки
 */
void spin_unlock(spin_t *p);

/**
 * @brief Инициализация нового мьютекса указанного типа.
 * В качестве типа создаваемого мьютекса могут быть указаны следующие значения: mtx_plain, mtx_timed, mtx_plain | mtx_recursive, mtx_timed | mtx_recursive.
 * Отличий в реализации между mtx_plain и mtx_timed нет.
 * Функция может вернуть значения thrd_success или thrd_error.
 * @param mutex Указатель на идентификатор создаваемого мьютекса
 * @param type Тип создаваемого мьютекса
 * @return int 
 */
int	mtx_init(mtx_t *mutex, int type);

/**
 * @brief Уничтожение мьютекса
 * Если есть какие-либо потоки, ожидающие мьютекса, поведение не определено.
 * @param mutex Указатель на идентификатор мьютекса
 */
void mtx_destroy(mtx_t *mutex);

/**
 * @brief Блокировка указанного мьютекса.
 * Блокирует текущий поток до тех пор, пока указанный мьютекс является заблокированным.
 * Если текущий поток уже заблокировал мьютекс и мьютекс является рекурсивным, то функция возвращает thrd_success.
 * Если текущий поток уже заблокировал мьютекс и мьютекс не является рекурсивным, то функция возвращает thrd_error.
 * Функция может вернуть значения thrd_success или thrd_error.
 * @param mutex Указатель на идентификатор мьютекса
 * @return int 
 */
int	mtx_lock(mtx_t *mutex);

/**
 * @brief Попытка блокировки указанного мьютекса без блокировки текущего потока.
 * Если указанный мьютекс заблокирован, то функция возвращает значение thrd_busy.
 * В противном случе функция блокирует мьютекс и возвращает значение thrd_success.
 * Функция может вернуть значения thrd_success, thrd_busy или thrd_error.
 * @param mutex Указатель на идентификатор мьютекса
 * @return int 
 */
int	mtx_trylock(mtx_t *mutex);

/**
 * @brief Функция разблокирует указанный мьютекс.
 * Функция может вернуть значения thrd_success или thrd_error.
 * @param mutex Указатель на идентификатор мьютекса
 * @return int 
 */
int	mtx_unlock(mtx_t *mutex);

/**
 * @brief Блокировка указанного мьютекса.
 * Блокирует текущий поток до тех пор, пока указанный мьютекс является заблокированным или текущеее время (GMT) превысило значение timespec.
 * Если текущий поток уже заблокировал мьютекс и мьютекс является рекурсивным, то функция возвращает thrd_success.
 * Если текущий поток уже заблокировал мьютекс и мьютекс не является рекурсивным, то функция возвращает thrd_error.
 * Если текущее время превысило значение timespec, а ресурс мьютекс остался заблокированным, то функция возвращает значение thrd_timedout.
 * Обратите внимание: параметр timespec содержит абсолютное время, если вам необходим интервал, то его необходимо вычислить самостоятельно.
 * В отличии от стандарта ANSI C тип параметра timespec объявлен как time_t.
 * Функция может вернуть значения thrd_success, thrd_timedout или thrd_error.
 * @param mutex Указатель на идентификатор мьютекса
 * @param timespec Абсолютное время GMT
 * @return int 
 */
int	mtx_timedlock(mtx_t *mutex, const time_t timespec);

__END_DECLS

#endif
