/**
 * @file stdio.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC"
 * @version 0.1
 * @date 2024-03-09
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef STDIO_H
#define STDIO_H

#include <cdefs.h>
#include <stddef.h>
#include <stdarg.h>

__BEGIN_DECLS

#define CHAR_SET_UTF8       0
#define CHAR_SET_CP1251     1

#define BUFSIZ              (512)

/**
 * @brief Установка кодировки символов используемой в программе.
 * 
 * @param char_set Используемая кодировка: CHAR_SET_UTF8 или CHAR_SET_CP1251
 */
void charset(int char_set);

/**
 * @brief Вывод на экран одного символа
 * 
 * @param c Код символа
 * @return int 
 */
int putchar(int c);

/**
 * @brief Вывод на экран строки символов
 * 
 * @param s Указатель на строку символов
 * @return int 
 */
int puts(const char *s);

/**
 * @brief Функция выводит на экран значения аргументов из списка arg-list в соответствии со строкой параметров.
 * Строка параметров - это строка символов из переменных параметров. Эта строка может состоять из нуля или более
 * параметров: обычных символов (кроме %), которые копируются в поток выхода и спецификаторов преобразования,
 * за каждым из которых следует один или более параметров.
 * Каждый спецификатор преобразований предваряется символом % и оканчивается спецификатором преобразования.
 * Между ними могут находиться (в определённом порядке) ноль или более флагов, необязательная минимальная 
 * ширина поля, необязательная точность и необязательные модификаторы длины.
 * Параметры должны точно соответствовать (после преобразования типа) спецификаторам преобразований.
 * По умолчанию параметры используются в порядке поступления, где каждая `*' и каждый спецификатор преобразования
 * требуют последующего параметра (если указано недостаточно параметров, то это приведёт к ошибке).
 * В общем виде синтаксис спецификатор преобразования имеет следующий вид:\n
 * %[flags][width][.precision][length]type\n
 * Поддерживаемые значения "type":\n
 * d или i : Целое число со знаком.\n
 * u       : Целое число без знака.\n
 * b       : Двоичное представление целого числа без знака.\n
 * o       : Восьмеричное представление целого числа без знака.\n
 * x       : Шестадцатиричное представление целого числа без знака (прописные символы).\n
 * X       : Шестадцатиричное представление целого числа без знака (заглавные символы).\n
 * f или F : Десятичное представление числа с плавающей точкой.\n
 * e или E : Экспоненциальная запись числа с плавающей точкой.\n
 * g или G : Десятичная или экспоненциальная запись числа с плавающей точкой.\n
 * c       : Один символ.\n
 * s       : Строка символов.\n
 * p       : Значение указателя.\n
 * %       : Два символа подряд соответствуют выводу одного символа "%".\n
 * Поддерживаемые значения "flags":\n
 * -       : Выравнивание по левому краю в пределах заданной ширины поля (по умолчанию выравнивание по правому краю).\n
 * +       : Принудительный вывод символа знака числа ("+" или "-"), по умолчанию выводится только "-".\n
 * (пробел): Если знак отсутствует, то перед значением вставляется пробел.\n
 * #       : Для типов (o,b,x,X) выводит дополнительный префикс "0","0b","0x" или "0X" для значений отличных от 0.
 *           Для типов (f,F) принудительно выводит десятичную точку, даже если после неё нет цифр.\n
 * 0       : Дополнение значения до указанной ширины с левой стороны символами "0".\n
 * Поддерживаемые значения "width":\n
 * (число) : Минимальное количество символов в выводимом значении.\n
 * *       : Ширина указывается не в строке формата, а как дополнительный аргумент целочисленного значения,
 *         : предшествующий аргументу, который необходимо форматировать.\n
 * Поддерживаемые значения "precision":\n
 * .(число): Для целочисленных типов (d,i,o,u,x,X): точность определяет минимальное количество записываемых цифр.\n
 *           Если записываемое значение короче этого числа, результат дополняется ведущими нулями.
 *           Значение не усекается, даже если результат длиннее. Precision 0 означает, что для значения 0 не записываются никакие символы.
 *           Для типов (f,F): количество цифр, которые должны быть выведены после десятичной точки. По умолчанию это 6, максимум 9.
 *           Для типа (s): максимальное количество печатаемых символов. По умолчанию все символы печатаются до тех пор, пока не встретится
 *           конечный нулевой символ.\n
 * .*      : Точность указывается не в строке формата, а как дополнительный аргумент целочисленного значения,
 *           предшествующий аргументу, который необходимо форматировать.\n
 * Поддерживаемые значения "length":\n
 * hh      : "char" или "unsigned char"\n
 * h       : "short int" или "unsigned short int"\n
 * l       : "long int" или "unsigned long int"\n
 * ll      : "long long int" или "unsigned long long int"\n
 * j       : "intmax_t" или "uintmax_t"\n
 * z       : size_t
 * @param format Строка параметров
 * @param ... 
 * @return int Число записанных символов (не включая завершающий 0)
 */
int printf(const char* format, ...);

/**
 * @brief Функция записывает в буфер значения аргументов из списка arg-list в соответствии со строкой формата
 * 
 * @param buffer Адрес буфера
 * @param format Строка формата (см.описание printf)
 * @param ... 
 * @return int Число записанных символов (не включая завершающий 0)
 */
int sprintf(char* buffer, const char* format, ...);

/**
 * @brief Простая реализация snprintf
 * 
 * @param buffer Адрес буфера
 * @param count Размер буфера
 * @param format Строка формата (см.описание printf)
 * @param ... 
 * @return int Число записанных символов (не включая завершающий 0)
 */
int snprintf(char* buffer, size_t count, const char* format, ...);

/**
 * @brief Простая реализация vprintf
 * 
 * @param format Строка формата (см.описание printf)
 * @param va Список аргументов
 * @return int 
 */
int vprintf(const char* format, va_list va);

/**
 * @brief Простая реализация vsnprintf
 * 
 * @param buffer Адрес буфера
 * @param count Размер буфера
 * @param format Строка формата (см.описание printf)
 * @param ... 
 * @return int Число записанных символов (не включая завершающий 0)
 */
int vsnprintf(char* buffer, size_t count, const char* format, va_list va);

/**
 * @brief Простая реализация printf с функцией callback
 * 
 * @param out An output function which takes one character and an argument pointer
 * @param arg An argument pointer for user data passed to output function
 * @param format Строка формата (см.описание printf)
 * @param ... 
 * @return int Число записанных символов (не включая завершающий 0)
 */
int fctprintf(void (*out)(char character, void* arg), void* arg, const char* format, ...);

/**
 * @brief Сканирование строки str в соответствии с форматом format.
 * Простейшая реализация функции sscanf: поддерживается только спецификатор "%d".
 * Возвращаемое значение — это количество символов из входной строки, которые были успешно отсканированы, а 
 * не количество совпавших элементов формата, как в стандартном sscanf.
 * @param str Сканируемая строка.
 * @param format Строка формата.
 * @param ... 
 * @return int 
 */
int sscanf(const char *str, const char *format, ...);

__END_DECLS

#endif
