/**
 * @file stdlib.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC"
 * @version 0.1
 * @date 2024-03-09
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef STDLIB_H
#define STDLIB_H

#include <cdefs.h>
#include <errno.h>
#include <stddef.h>

__BEGIN_DECLS

/**
 * @brief Максимальное значение возвращаемое функцией rand()
 */
#define RAND_MAX 0x7fffffff

/**
 * @brief Преобразование строки ASCII в число с плавающей точкой двойной точности
 * 
 * @param s Указатель на строку
 * @return double 
 */
double atof(const char *s);

/**
 * @brief Преобразование строки в целое число
 * 
 * @param str Указатель на строку
 * @return int 
 */
int atoi(const char *str);

/**
 * @brief Преобразование строки в целое число
 * 
 * @param str Указатель на строку
 * @return int 
 */
#define atol(s) atoi(s)

/**
 * @brief Преобразование строки в длинное целое число со знаком.
 * Функция конвертирует начальную часть строки nptr в длинное целое в соответствии с указанным base, 
 * которое должно находиться в диапазоне от 2-х до 36-х включительно или быть равным нулю.
 * При ошибках устанавливает значение переменной errno.
 * @param nptr Указатель на строку
 * @param endptr Указатель на первое некорректное значение в строке
 * @param base Основание системы счисления
 * @return long int 
 */
long int strtol(const char *nptr, char **endptr, int base);

/**
 * @brief Преобразование строки в длинное целое число без знака.
 * Функция конвертирует начальную часть строки nptr в длинное целое в соответствии с указанным base, 
 * которое должно находиться в диапазоне от 2-х до 36-х включительно или быть равным нулю.
 * При ошибках устанавливает значение переменной errno.
 * @param nptr Указатель на строку
 * @param endptr Указатель на первое некорректное значение в строке
 * @param base Основание системы счисления
 * @return unsigned long
 */
unsigned long strtoul(const char *nptr, char **endptr, int base);

/**
 * @brief Преобразование строки ASCII в число с плавающей точкой двойной точности
 * 
 * @param str Указатель на строку.
 * @param ptr Если ptr не равно NULL, то указатель на символ, следующий за последним обработанным символом.
 * @return double 
 */
double strtod(const char *str, char **ptr);

/**
 * @brief Возвращает абсолютное значение числа x
 * 
 * @param x 
 * @return int 
 */
int abs(int x);

#define labs(x) abs(x)

/**
 * @brief Возвращает абсолютное значение числа x
 * 
 * @param x 
 * @return long long
 */
long long llabs(long long x);

/**
 * @brief Генератор целых псевдослучайных чисел в диапазоне 0..RAND_MAX
 * @return int 
 */
int rand();

/**
 * @brief Аппаратный генератор случайных чисел (RNG) в диапазоне 0..2^32-1
 * @return unsigned int
 */
unsigned int hw_rand();

/**
 * @brief Инициализация генератора псевдослучайных чисел.
 * По умолчанию генератор ПСЧ использует начальное значение 1.
 * @param seed 
 */
void srand(unsigned int seed);

/**
 * @brief Сортировка массива
 * 
 * @param aa Указатель на начало массива
 * @param n Количество элементов в массиве
 * @param es Размер одного элемента массива
 * @param cmp Функция сравнения двух элементов
 */
void qsort(void *aa, size_t n, size_t es, int (*cmp)(const void *, const void *));

/**
 * @brief Двоичный поиск в отсортированном массиве 
 * Функция выполняет поиск в массиве из nmemb объектов (на начальный элемент его, размером size, указывает аргумент base) элемент, совпадающий с объектом, обозначенным key.
 * Содержимое массива должно быть расположено по возрастанию в соответствии с функцией compar, производящей сравнение.
 * Функция compar имеет два параметра (в первый передается указатель на объект key, во второй - элемент массива) и возвращает целое число больше нуля,
 * если key больше элемента массива; меньше нуля, если меньше, и ноль, если они равны.
 * @param key Искомый элемент
 * @param base0 Указатель на начальный элемент массива
 * @param nmemb Количество элементов в массиве
 * @param size Размер элемента массива
 * @param compar Функция сравнения двух элементов
 * @return void* 
 */
void *bsearch (const void *key, const void *base0, size_t nmemb, size_t size, int (*compar)(const void *, const void *));

// Прототипы malloc.h (stdlib.h)----------------------------------------------------------

#define CPU_MEM_PNT_MARK    0xc0000000
#define CPU_MEM_PNT_MASK    0x3FFFFFFF
//
#define CPU_MEM_ROM_START   0x00000000
#define CPU_MEM_RAM_START   0x40000000

#define CPU_MEM_IS_ROM(adr) (((uint32_t)(adr) & CPU_MEM_PNT_MARK) == CPU_MEM_ROM_START)
#define CPU_MEM_IS_RAM(adr) (((uint32_t)(adr) & CPU_MEM_PNT_MARK) == CPU_MEM_RAM_START)

#if MALLOC_DEBUG == 1
    void _heap_dump(void);
    #define heap_debug_dump() _heap_dump()
#else
    #define heap_debug_dump()
#endif

/**
 * @brief Выделяет пямять в локальной области видимости (в стеке)
 */
#define alloca(size) __builtin_alloca(size)

/**
 * @brief Выделение памяти
 * 
 * @param size Требуемый размер
 * @return void* Адрес выделенный памяти или NULL при ошибке
 */
void *malloc(size_t size);

/**
 * @brief Выделяет массив в памяти и инициализирует его элементы значением 0
 * 
 * @param num Количество выделяемых элементов
 * @param size Размер одного элемента
 * @return void* Адрес выделенный памяти или NULL при ошибке
 */
void *calloc(size_t num, size_t size);

/**
 * @brief Изменение размера блока памяти
 * 
 * @param mem Указатель на ранее выделенный блок памяти
 * @param newsize Новый размер
 * @return void* Указатель на новый блок памяти
 */
void    *realloc(void *mem, size_t newsize);

/**
 * @brief Освобождение памяти ранее выделенной вызовом malloc()
 * 
 * @param mem Адрес освобождаемой памяти
 */
void    free(void *mem);

__END_DECLS

#endif

