/**
 * @file string.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы со строками и памятью.
 * @version 0.1
 * @date 2024-03-09
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef STRING_H
#define STRING_H

#include <cdefs.h>
#include <stddef.h>

__BEGIN_DECLS

/**
 * @brief Вычисление размера строки (количества байт) завершённой символом '\0'
 * 
 * @param str Указатель на строку
 * @return size_t Размер строки в байтах
 */
size_t  strlen(const char *str);

/**
 * @brief Определяет длину строки фиксированной длины
 * 
 * @param str Указатель на строку
 * @param maxlen Максимальная длина строки
 * @return size_t Размер строки в байтах, но не более maxlen
 */
size_t  strnlen(const char* str, size_t maxlen);

/**
 * @brief Переставляет символы в строке в обратном порядке
 * 
 * @param str Указатель на строку
 */
void    strrev(unsigned char *str);

/**
 * @brief Копирование строки
 * 
 * @param to Место назначения
 * @param from Исходная строка
 * @return char* Указатель на место назначения
 */
char    *strcpy(char *to, const char *from);

/**
 * @brief Копирование строки с ограничение длины
 * 
 * @param to Место назначения
 * @param from Исходная строка
 * @param size Максимальный размер
 * @return char* Указатель на место назначения
 */
char    *strncpy(char *to, const char *from, size_t size);

/**
 * @brief Объединение двух строк
 * Функция добавляет строку from в конец строки to
 * @param to Строка назначения
 * @param from Добавляемая строка
 * @return char* Указатель на строку - результат объединения строк
 */
char    *strcat(char *to, const char *from);

/**
 * @brief Объединение двух строк с ограничением длины
 * Функция добавляет строку не более чем size первых символов строки from в конец строки to
 * @param to Строка назначения
 * @param from Добавляемая строка
 * @param size Максимальный размер строки from
 * @return char* Указатель на строку - результат объединения строк
 */
char    *strncat(char *to, const char *from, size_t size);

/**
 * @brief Посимвольное сравнение двух строк
 * Результат сравнения:
 * <0  : s1 < s2
 * 0   : s1 == s2
 * >0  : s1 > s2
 * 
 * @param s1 Первая строка
 * @param s2 Вторая строка
 * @return int Результат сравнения
 */
int     strcmp(const char *s1, const char *s2);

/**
 * @brief Посимвольное сравнение двух строк с ограниичение длины строки s1
 * Результат сравнения:
 * <0  : s1 < s2
 * 0   : s1 == s2
 * >0  : s1 > s2
 * 
 * @param s1 Первая строка
 * @param s2 Втроая строка
 * @param size Максимальный размер строки s1
 * @return int Результат сравнения
 */
int     strncmp(const char *s1, const char *s2, size_t size);

/**
 * @brief Дублирование строки
 * Функция использует выделение памяти для новой строки при помощи функции malloc().
 * В последствии выделенная память может быть освобождена функцией free().
 * @param str Оригинальная строка
 * @return char* Указатель на новую строку - копию оригинальной
 */
char    *strdup(const char *str);

/**
 * @brief Дублирование строки с ограничение длины
 * 
 * @param s Оригинальная строка
 * @param n Максимальный размер копируемой строки
 * @return char* Указатель на новую строку - копию оригинальной
 */
char    *strndup(const char* s, size_t n);

/**
 * @brief Поиск символа в строке
 * Ищет первое вхождение символа ch в строке
 * 
 * @param p Строка
 * @param ch Искомый символ
 * @return char* Указатель на местонахождение первого совпадения с символом ch в строке
 */
char    *strchr(const char *p, int ch);

/**
 * @brief Поиск символа в строке
 * Ищет последнее вхождение символа ch в строке
 * 
 * @param p Строка
 * @param ch Искомый символ
 * @return char* Указатель на местонахождение последнего совпадения с символом ch в строке
 */
char    *strrchr(const char *p, int ch);

/**
 * @brief Поиск подстроки в строке
 * 
 * @param s Строка в которой выполняется поиск
 * @param find Искомая подстрока
 * @return char* Указатель на начало подстроки, или NULL, если подстрока не найдена
 */
char    *strstr(const char *s, const char *find);

/**
 * @brief Копирование строки с ограничением общей длины.
 * Функция копирует из строки from в буфер to не более чем size - 1 символов и гарантированно устанавливает в конец строки (нулевой символ).
 * @param to Указатель на буфер назначения
 * @param from Указатель на исходную строку
 * @param size Размер буфера назначения
 * @return size_t Размер строки по адресу from
 */
size_t  strlcpy(char *to, const char *from, size_t size);

/**
 * @brief Объединение строк с ограничением общей длины.
 * Добавляет к строке to содержимое строки from, длина результирующей строки не превышает size - 1.
 * @param to Указатель на буфер (строку) назначения
 * @param from Указатель на добавляемую строку
 * @param size Размер буфера назначения
 * @return size_t Суммарный размер строк from и to
 */
size_t  strlcat(char *to, const char *from, size_t size);

/**
 * @brief Поиск набора символов в строке.
 * Определяет максимальную длину начальной подстроки из s1, состоящей исключительно из символов, перечисленных в s2
 * @param s1 Строка
 * @param s2 Набор символов
 * @return size_t Количество символов в начальном сегменте s1, который составляют символы из s2
 */
size_t strspn(const char *s1, const char *s2);

/**
 * @brief Поиск набора символов в строке.
 * Определяет максимальную длину начальной подстроки из s1, состоящей исключительно из символов, не указанных в s2
 * @param s1 Строка
 * @param s2 Набор символов
 * @return size_t Количество символов в начальном сегменте s1, состоящем только из символов, не указанных в s2
 */
size_t strcspn(const char *s1, const char *s2);

/**
 * @brief Поиск строки для любого набора символов.
 * Функция ищет первое совпадение в строке s1 с любым символом из строки s2.
 * Функция возвращает указатель на символ в s1, соответствующий одному из символов в наборе s2, или NULL,
 * если такого символа нет в строке.
 * @param s1 Строка
 * @param s2 Набор символов
 * @return char* 
 */
char *strpbrk(const char *s1, const char *s2);

/**
 * @brief Заполняет буфер указанным значением.
 * 
 * @param dest Указатель на буфер назначения
 * @param val Значение
 * @param len Размер буфера
 * @return void* Указатель на область памяти dest
 */
void *memset(void *dest, int val, size_t len);

/**
 * @brief Копирование области памяти.
 * Копирует len байт из области src в область dest. 
 * @param dest Буфер назначения
 * @param src Исходный буфер
 * @param len Количество копируемых байт
 * @return void* Указатель на dest
 */
void *memmove(void *dest, const void *src, size_t len);

/**
 * @brief Копирование области памяти.
 * Копирует len байт из области src в область dest.
 * Функция полностью аналогична функции memmove.
 * @param dest Буфер назначения
 * @param src Исходный буфер
 * @param len Количество копируемых байт
 * @return void* Указатель на dest
 */
void *memcpy(void *dest, const void *src, size_t len);

/**
 * @brief Сравнение областей памяти.
 * Функция сравнивает первые count байтов областей памяти str1 и str2.
 * Возвращается целое число, меньшее, большее нуля или равное ему, если определено, что s1 меньше, больше, чем s2, или равно ему.
 * @param str1 Указатель на буфер 1
 * @param str2 Указатель на буфер 2
 * @param count Количество сравниваемых байт
 * @return int Результат сравнения
 */
int memcmp(const void *str1, const void *str2, size_t count);

/**
 * @brief Поиск символа в буфере.
 * Функция ищет символ c в первых len байтах области памяти, на которую указывает srcv.
 * Совпадение первого байта с c (представленным как беззнаковый символ) останавливает выполнение операции.
 * @param srcv Указатель на буфер памяти
 * @param c Искомый символ
 * @param len Размер буфера
 * @return void* Указатель на найденный символ или NULL, если он не найден
 */
void *memchr(const void *srcv, int c, size_t len);

/**
 * @brief Обмен содержимого двух областей памяти.
 * 
 * @param m1 Указатель на область памяти 1
 * @param m2 Указатель на область памяти 2
 * @param n Размер областей памяти
 */
void memswap(void *m1, void *m2, size_t n);

/**
 * @brief Поиск лексем в строке.
 * Последовательность вызовов функции разбивает строку на лексемы, разделенные символами разделителями.
 * Функция возвращает указатель на первую найденную лексему в строке. Если лексем не найдено, то возвращается пустой указатель. 
 * @param s Указатель на строку, которую нужно разбить на лексемы
 * @param delim Указатель на строку, содержащую разделители
 * @return char* 
 */
char *strtok(char *s, const char *delim);

/**
 * @brief Реентерабельная версия strtok().
 * @param s Указатель на строку, которую нужно разбить на лексемы.
 * @param delim Указатель на строку, содержащую разделители.
 * @param save_ptr Учёта контекста между последующими вызовами при анализе одной и той же строки.
 * @return char* 
 */
char *strtok_r(char *s, const char *delim, char **save_ptr);

__END_DECLS

#endif
