/**
 * @file time.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с временем.
 * @version 0.1
 * @date 2024-03-09
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef TIME_H
#define TIME_H

#include <cdefs.h>
#include <stdint.h>

__BEGIN_DECLS

/**
 * @brief Наименования текущей TZ.
 * Пример: "RTZ 2"
 */
extern  char *tzname[2];

/**
 * @brief Смещение в секундах текущей TZ относительно GMT.
 * Положительные значения - западнее GMT, отрицательные значения - восточнее GMT.
 * Пример: -10800 = GMT+3
 */
extern  long    timezone;

/**
 * @brief Признак перехода на летнее время.
 * 0 - не используется, !=0 - используется
 */
extern  int     daylight;

/**
 * @brief Определяет количество тактов системных часов в секунду. Используется для пересчета величины, возвращаемой функцией clock(), в секунды.
 */
#define CLOCKS_PER_SEC  100
#define CLK_PER_SEC     CLOCKS_PER_SEC
#define CLK_TCK         CLOCKS_PER_SEC

/**
 * @brief Время, потраченное на выполнение процесса
 * Возвращается функцией clock().
 */
typedef unsigned int    clock_t;

/**
 * @brief Количество секунд, прошедшее с 1-Янв-1970 (Unix Timestamp)
 */
typedef unsigned int    time_t;

#define SIZE_OF_TIME_T  4

/**
 * @brief Дата и время в формате BCC/Delphi
 */
typedef double          TDateTime;

/**
 * @brief Нелинейное, дискретное календарное представление времени
 */
struct tm {
    int tm_sec;         ///< Секунды 0..60 (60 может быть при добавлении дополнительной секунды)
    int tm_min;         ///< Минуты 0..59
    int tm_hour;        ///< Часы 0..23
    int tm_mday;        ///< День месяца 1..31
    int tm_mon;         ///< Месяц 0(Январь)..11
    int tm_year;        ///< Год, начиная с 1900
    int tm_wday;        ///< День недели 0(Воскресенье)..6
    int tm_yday;        ///< День года 0(1 января)..365
    int tm_isdst;       ///< Daylight Savings Time flag
};

/**
 * @brief Возвращает время, затраченное процессором на выполнение программы, или -1, если оно неизвестно.
 * Размерность возвращаемого значения определяется при помощи константы CLOCKS_PER_SEC,
 * которая задаёт количество единиц значения времени в одной секунде.
 * 
 * @return clock_t 
 */
clock_t clock(void);

/**
 * @brief Функция возвращает календарное время или значение (-1), если это время неизвестно.
 * Если указатель tp не равен NULL, то значение времени записывается также и в (*tp).
 * @param timer Указатель на time_t или NULL
 * @return time_t 
 */
time_t time(time_t * timer);

/**
 * @brief Возвращает разность time2-time1, выраженную в секундах.
 * 
 * @param time2 
 * @param time1 
 * @return double Разница между time2 и time1 в секундах
 */
// double difftime(time_t time2, time_t time1);
#define difftime(time2,time1) (double)(time2 - time1)

/**
 * @brief Преобразует местное время, заданное структурой *tp, в календарное
 * 
 * @param tp Местное время
 * @return time_t 
 */
time_t mktime(struct tm *tp);

/**
 * @brief Преобразует календарное время timep в локальное представление времени, определенное во всеобщем скоординированном времени (UTC).
 * 
 * @param timep Указатель на значение текущего календарного времени
 * @return struct tm* 
 */
struct tm *gmtime(const time_t *timep);

/**
 * @brief Преобразует календарное время timep в локальное представление времени, определенное во всеобщем скоординированном времени (UTC).
 * 
 * @param timep Указатель на значение текущего календарного времени
 * @param tm Указатель на структуру tm
 * @return struct tm* 
 */
struct tm *gmtime_r(const time_t *timep, struct tm *tm);

/**
 * @brief Преобразует календарное время timep в локальное представление времени, определенное в локальной временной зоне.
 * 
 * @param timep Указатель на значение текущего календарного времени
 * @return struct tm* 
 */
struct tm *localtime(const time_t *timep);

/**
 * @brief Преобразует календарное время timep в локальное представление времени, определенное в локальной временной зоне.
 * 
 * @param timep Указатель на значение текущего календарного времени
 * @return struct tm* 
 */
struct tm *localtime_r(const time_t *timep, struct tm *result);

/**
 * @brief Инициализирует информацию о преобразованиях времени.
 * Функция tzset() инициализирует переменную tzname из переменной окружения TZ и автоматически вызывается другими функциями,
 * зависящими от временной зоны. Она также устанавливает переменные timezone (секунды западнее GMT) и daylight
 * (если временная зона не использует переход на "летнее" время, то 0, в противном случае ненулевое значение).
 */
void tzset(void);

/**
 * @brief Получение времени, прошедшего с момента запуска контроллера, в миллисекундах.
 * 
 * @return uint32_t 
 */
uint32_t GetTickCount(void);

// Функции работы с датой/временем в стиле C-Script-----------------------------------------

/**
 * @brief Получение текущего календарного времени (GMT).
 * 
 * @param hour Час (0..23)
 * @param min Минута (0..59)
 * @param sec Секунда (0..59)
 */
void GetTime(unsigned char * hour,unsigned char * min,unsigned char * sec);

/**
 * @brief Получение текущего календарного времени (локальная временная зона).
 * 
 * @param hour Час (0..23)
 * @param min Минута (0..59)
 * @param sec Секунда (0..59)
 */
void GetLocalTime(unsigned char * hour,unsigned char * min,unsigned char * sec);

/**
 * @brief Получение текущей календарной даты (GMT).
 * 
 * @param date День месяца (1..31)
 * @param month Месяц (1..12)
 * @param year Год (2024..)
 */
void GetDate(unsigned char * date,unsigned char * month,unsigned short * year);

/**
 * @brief Получение текущей календарной даты (локальная временная зона).
 * 
 * @param date День месяца (1..31)
 * @param month Месяц (1..12)
 * @param year Год (2024..)
 */
void GetLocalDate(unsigned char * date,unsigned char * month,unsigned short * year);

/**
 * @brief Получение текущего календарного времени и текущей календарной даты (GMT).
 * 
 * @param date День месяца (1..31)
 * @param month Месяц (1..12)
 * @param year Год (2024..)
 * @param hour Час (0..23)
 * @param min min Минута (0..59)
 * @param sec sec Секунда (0..59)
 */
void GetDateTime(unsigned char * date,unsigned char * month,unsigned short * year,unsigned char * hour,unsigned char * min,unsigned char * sec);

/**
 * @brief Получение текущего календарного времени и текущей календарной даты (локальная временная зона).
 * 
 * @param date День месяца (1..31)
 * @param month Месяц (1..12)
 * @param year Год (2024..)
 * @param hour Час (0..23)
 * @param min min Минута (0..59)
 * @param sec sec Секунда (0..59)
 */
void GetLocalDateTime(unsigned char * date,unsigned char * month,unsigned short * year,unsigned char * hour,unsigned char * min,unsigned char * sec);

/**
 * @brief Получение значения секунд из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char SecOfTime(time_t t);

/**
 * @brief Получение значения секунд из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char SecOfLocalTime(time_t t);

/**
 * @brief Получение значения минут из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char MinOfTime(time_t t);

/**
 * @brief Получение значения минут из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char MinOfLocalTime(time_t t);

/**
 * @brief Получение значения часа из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char HourOfTime(time_t t);

/**
 * @brief Получение значения часа из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char HourOfLocalTime(time_t t);

/**
 * @brief Получение значения номера дня месяца Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char DayOfTime(time_t t);

/**
 * @brief Получение значения номера дня месяца Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char DayOfLocalTime(time_t t);

/**
 * @brief Получение значения номера месяца из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char MonthOfTime(time_t t);

/**
 * @brief Получение значения номера месяца из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char MonthOfLocalTime(time_t t);

/**
 * @brief Получение значения номера года из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned short int 
 */
unsigned short int YearOfTime(time_t t);

/**
 * @brief Получение значения номера года из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned short int 
 */
unsigned short int YearOfLocalTime(time_t t);

/**
 * @brief Получение значения номера дня недели из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char WeekDayOfTime(time_t t);

/**
 * @brief Получение значения номера дня недели из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned char 
 */
unsigned char WeekDayOfLocalTime(time_t t);

/**
 * @brief Получение значения номера дня года из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned short int 
 */
unsigned short int YearDayOfTime(time_t t);

/**
 * @brief Получение значения номера дня года из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970
 * @return unsigned short int 
 */
unsigned short int YearDayOfLocalTime(time_t t);

/**
 * @brief Получение значения Unix Timestamp из заданных значений времени и даты (GMT).
 * 
 * @param hour Час (0-23)
 * @param min Минута (0-59)
 * @param sec Секунда (0-59)
 * @param day День месяца (1-31)
 * @param month Номер месяца(1-12)
 * @param year Год (1970...)
 * @return time_t Количество секунд с 01-01-1970
 */
time_t MkTime(unsigned char hour,unsigned char min,unsigned char sec,unsigned char day,unsigned char month,unsigned short int year);

/**
 * @brief Получение значения Unix Timestamp из заданных значений времени и даты (локальная временная зона).
 * 
 * @param hour Час (0-23)
 * @param min Минута (0-59)
 * @param sec Секунда (0-59)
 * @param day День месяца (1-31)
 * @param month Номер месяца(1-12)
 * @param year Год (1970...)
 * @return time_t Количество секунд с 01-01-1970
 */
time_t MkLocalTime(unsigned char hour,unsigned char min,unsigned char sec,unsigned char day,unsigned char month,unsigned short int year);

/**
 * @brief Преобразует время в формате TDateTime (Delphi) в time_t.
 * @param dt Время в формате TDateTime.
 * @return time_t 
 */
time_t DateTimeToUnix(TDateTime dt);

/**
 * @brief Преобразует время в формате time_t в формат TDateTime (Delphi).
 * 
 * @param tv Время в формате time_t
 * @return TDateTime 
 */
TDateTime UnixToDateTime(time_t tv);


__END_DECLS

#endif
