/**
 * @file unistd.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Унифицированный обмен данными через сокеты и последовательный порт.
 * @version 0.1
 * @date 2024-04-05
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#ifndef UNISTD_H
#define UNISTD_H

#include <cdefs.h>
#include <stddef.h>

__BEGIN_DECLS

/**
 * @brief Тип дескриптора
 */
typedef enum {
    HDL_INVALID      = 0, 
    HDL_SERIAL       = 1,               // последовательный порт
    HDL_NETWORK      = 2                // сетевое соединение
} handle_t;

/**
 * @brief Прочитать полученные данные из приёмного буфера.
 * Функция возвращает значение количества прочитанных данных или код ошибки:
 * -1 - некорректное значение дескриптора;
 * -2 - соединение не установлено или оборвано.
 * @param fd Дескриптор сокета или последовательного порта
 * @param buf Указатель на массив данных
 * @param count Количество прочитанных данных
 * @return int 
 */
int read(int fd, void *buf, size_t count);

/**
 * @brief Возвращает значение количества принятых и доступных для чтения данных.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param fd Дескриптор сокета или последовательного порта
 * @return int Количество доступных для чтения байт в приёмном буфере или отрицательное значение при ошибке.
 */
int read_length(int fd);

/**
 * @brief Возвращает количество байт утерянных из-за переполнения приёмного буфера соединения.
 * Вызов функции обнуляет счётчик потерянных байт. При ошибке функция возвращает отрицательное значение:
 * -1 - некорректное значение дескриптора;
 * -2 - соединение не установлено или оборвано.
 * @param fd Дескриптор сокета или последовательного порта
 * @return int Количество потерянных байт или отрицательное число при ошибке.
 */
int lost_bytes(int fd);

/**
 * @brief Прочитать блок данных указанной длины, в течении не более чем указанного времени.
 * Чтение выполняется в течении не более чем timeout мс.
 * При успешном выполнении возвращает 1, если данные не приняты - 0.
 * При ошибке функция возвращает код:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param fd Дескриптор сокета или последовательного порта.
 * @param data Указатель на массив данных.
 * @param count Количество читаемых данных.
 * @param timeout Максимальное время чтения данных, мс.
 * @return int 
 */
int read_block(int fd, void *data, size_t count, int timeout);

/**
 * @brief Очистка приёмного буфера сокета или последовательного порта.
 * @param fd Дескриптор сокета или последовательного порта.
 */
void purge(int fd);

/**
 * @brief Запись данных в последовательный порт или сокет.
 * При успешном выполнении функция возвращает количество записанных данных, при ошибке код:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param fd Дескриптор сокета или последовательного порта.
 * @param buf Указатель на массив данных.
 * @param count Количество записываемых данных.
 * @return int 
 */
int write(int fd, void *buf, size_t count);

/**
 * @brief Функция закрывает и освобождает ресурсы занятые сокетом или последовательным портом.
 * @param fd Указатель на дескриптор сокета или последовательного порта.
 */
void close(int *fd);

/**
 * @brief Проверка дескриптора сокета или последовательного порта.
 * Функция возвращает результат:
 * HDL_SERIAL  - fd: корректный дескриптор последовательного порта;
 * HDL_NETWORK - fd: корректный дескриптор сетевого сокета.
 * В ином случае функция возвращает HDL_INVALID.
 * @param fd Дескриптор сокета или последовательного порта.
 * @return handle_t
 */
handle_t socket_handle_type(int fd);

// ---Функции задержки---------------------------------------------------------------------------------

/**
 * @brief Переход в режим ожидания на указанное количество секунд
 * 
 * @param seconds 
 */
void sleep(unsigned int seconds);

/**
 * @brief Переход в режим ожидания на указанное количество микросекунд
 * 
 * @param usec 
 */
void usleep(unsigned long usec);

/**
 * @brief Переход в режим ожидания на указанное количество миллисекунд
 * 
 * @param msec 
 */
void msleep(unsigned long msec);

__END_DECLS

#endif
