/**
 * @file http.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Разбор строки URL
 * @version 0.1
 * @date 2024-05-19
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef URL_H
#define URL_H

#include <cdefs.h>

__BEGIN_DECLS

/**
 * @brief Структура - результат разбора строки URL
 */
typedef struct url_struct {
    char *uri;                  ///< URI
    char *scheme;               ///< схема: http, https
    char *host;                 ///< хост
    unsigned short int port;    ///< порт
    char *path;                 ///< путь
    char *query;                ///< запрос
    char *fragment;             ///< фрагмент
    char *username;             ///< имя пользователя
    char *password;             ///< пароль
} url_t;

/**
 * @brief Разбор строки URL в соответствии с RFC 1738/3986.
 * При успешном разборе строки функция выделяет память для структуры url_t, заполняет её поля и возвращает указатель на структуру.
 * 
 * Память должна быть в последствии освобождена функцией HttpFreeURL.
 * При ошибке разбора функция возвращает NULL.
 * @example Пример URL
 * http://login:password@www.test.ru:8080/path1/path2?param=val
 * http://              - протокол, обязателен. Поддерживаются: http, https, mqtt, mqtts, telnet, ssh, smtp, ftp
 * login:password@      - имя пользователя и пароль для авторизации, опционально
 * www.test.ru          - адрес сервера, обязателен
 * :8080                - номер TCP-порта, опционально, при отсутствии используется 80 для http и 443 для https
 * /path1/path2         - путь к ресурсу, опционально
 * ?param=val           - параметры запроса, опционально
 * @param url Указатель на строку URL.
 * @return url_t* Указатель на структуру или NULL при ошибке.
 */
url_t *HttpParseURL(const char *url);

/**
 * @brief Освобождение памяти, занимаемой структурой parsed_url.
 * @param purl Указатель на структуру URL
 */
void HttpFreeURL(url_t *purl);

__END_DECLS

#endif
