/**
 * @file wchar.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы со строками wchar_t.
 * @version 0.1
 * @date 2024-05-11
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef UTFSTR_H
#define UTFSTR_H

#include <cdefs.h>
#include <stdint.h>
#include <stddef.h>
#include <string.h>

__BEGIN_DECLS

/**
 * @brief Возвращает количество символов в строке UTF32.
 * @param str 
 * @return size_t 
 */
size_t wcslen(const wchar_t* str);

/**
 * @brief Копирование строки UTF32.
 * @param dest Указатель на буфер назначения.
 * @param src Указатель на исходную строку.
 * @return wchar_t* 
 */
wchar_t* wcscpy(wchar_t* dest, const wchar_t* src);

/**
 * @brief Объединение строк UTF32.
 * @param dest Указатель на строку назначения.
 * @param src Указатель на исходную строку.
 * @return wchar_t* 
 */
wchar_t* wcscat(wchar_t* dest, const wchar_t* src);

/**
 * @brief Сравнение строк UTF32 с ограничением длины и учётом регистра символов.
 * Возвращает значение 0 (строки одинаковые), <0 - строка 1 меньше или 1 - строка 1 больше.
 * @param wcs1 Указатель на строку 1.
 * @param wcs2 Указатель на строку 2.
 * @param ztCount Максимальное количество сравниваемых символов.
 * @return int 0-строки одинаковые, >0-строка 1 больше, <0-строка 1 меньше.
 */
int wcsncmp(const wchar_t* wcs1, const wchar_t* wcs2, size_t ztCount);

/**
 * @brief Сравнение строк UTF32 с учётом регистра символов.
 * Возвращает значение 0 (строки одинаковые), <0 - строка 1 меньше или 1 - строка 1 больше.
 * @param wcs1 Указатель на строку 1.
 * @param wcs2 Указатель на строку 2.
 * @return int 0-строки одинаковые, >0-строка 1 больше, <0-строка 1 меньше.
 */
int wcscmp(const wchar_t* wcs1, const wchar_t* wcs2);

/**
 * @brief Перевод символов строки UTF32 в верхний регистр.
 * Функция переводит символы строки pwcs в верхний регистр.
 * @param pwcs Указатель на строку UTF32.
 * @return wchar_t* Указатель на строку pwcs.
 */
wchar_t* wcsupr(wchar_t* pwcs);

/**
 * @brief Перевод символов строки UTF32 в нижний регистр.
 * Функция переводит символы строки pwcs в нижний регистр.
 * @param pwcs Указатель на строку UTF32.
 * @return wchar_t* Указатель на строку pwcs.
 */
wchar_t* wcslwr(wchar_t* pwcs);

/**
 * @brief Сравнение строк UTF32 с ограничением длины и без учёта регистра символов.
 * Возвращает значение 0 (строки одинаковые), <0 - строка 1 меньше или 1 - строка 1 больше.
 * @param wcs1 Указатель на строку 1.
 * @param wcs2 Указатель на строку 2.
 * @param ztCount Максимальное количество сравниваемых символов.
 * @return int 0-строки одинаковые, >0-строка 1 больше, <0-строка 1 меньше.
 */
int wcsncasecmp(const wchar_t* wcs1, const wchar_t* wcs2, size_t ztCount);

/**
 * @brief Сравнение строк UTF32 без учёта регистра символов.
 * Возвращает значение 0 (строки одинаковые), <0 - строка 1 меньше или 1 - строка 1 больше.
 * @param wcs1 Указатель на строку 1.
 * @param wcs2 Указатель на строку 2.
 * @return int 0-строки одинаковые, >0-строка 1 больше, <0-строка 1 меньше.
 */
int wcscasecmp(const wchar_t* wcs1, const wchar_t* wcs2);

/**
 * @brief Поиск без учёта регистра символов UTF32 подстроки wcs2 в UTF32 строке wcs1.
 * Если подстрока найдена, то функция возвращает указатель на начало найденного вхождения в строке wcs1.
 * Если подстрока не найдена, то функция возвращает NULL.
 * @param wcs1 Указатель на строку 1.
 * @param wcs2 Указатель на строку 2.
 * @return wchar_t* 
 */
wchar_t* wcscasestr(const wchar_t* wcs1, const wchar_t* wcs2);

// -- UTF 8 ----------------------------------------------------------------------------------

/**
 * @brief Возвращает длину строки UTF8 в символах.
 * @param pmbc Указатель на строку.
 * @return size_t 
 */
size_t mbslen(const char* pmbc);

/**
 * @brief Функция преобразует строку UTF32 в строку UTF8.
 * Функция выделяет память для строки UTF8, выполняет преобразование и возвращает указатель на строку UTF8.
 * Выделенная память должна быть в последствии освобождена функцией free.
 * @param pwcs Указатель на строку UTF32.
 * @return char* 
 */
char* wcstombs(const wchar_t* pwcs);

/**
 * @brief Функция преобразует строку UTF8 в строку UTF32.
 * Функция выделяет память для строки UTF32, выполняет преобразование и возвращает указатель на строку UTF32.
 * Выделенная память должна быть в последствии освобождена функцией free.
 * @param pmbc Указатель на строку UTF8.
 * @return wchar_t* 
 */
wchar_t* mbstowcs(const char* pmbc);

/**
 * @brief Преобразование символов строки UTF8 в верхний регистр.
 * Функция выделяет память для новой строки UTF8, выполняет преобразование и возвращает указатель на строку UTF8.
 * Выделенная память должна быть в последствии освобождена функцией free.
 * @param pmbc Указатель на строку UTF8.
 * @return char* Указатель на новую строку - результат преобразования регистра символов.
 */
char* mbsmakeupr(const char* pmbc);

/**
 * @brief Преобразование символов строки UTF8 в нижний регистр.
 * Функция выделяет память для новой строки UTF8, выполняет преобразование и возвращает указатель на строку UTF8.
 * Выделенная память должна быть в последствии освобождена функцией free.
 * @param pmbc Указатель на строку UTF8.
 * @return char* Указатель на новую строку - результат преобразования регистра символов.
 */
char* mbsmakelwr(const char* pmbc);

/**
 * @brief Регистронезависимвое сравнение двух строк UTF8 с ограничением длины.
 * При ошибке функция возвращает значение -1.
 * @param pmbs1 Указатель на строку 1.
 * @param pmbs2 Указатель на строку 2.
 * @param ztCount Максимальная длина строк.
 * @return int 0-строки одинаковые, >0-строка 1 больше, <0-строка 1 меньше.
 */
int mbsncasecmp(const char* pmbs1, const char* pmbs2, size_t ztCount);

/**
 * @brief Регистронезависимвое сравнение двух строк UTF8.
 * При ошибке функция возвращает значение -1.
 * @param pmbs1 Указатель на строку 1.
 * @param pmbs2 Указатель на строку 2.
 * @return int 0-строки одинаковые, >0-строка 1 больше, <0-строка 1 меньше.
 */
int mbscasecmp(const char* pmbs1, const char* pmbs2);

/**
 * @brief Поиск без учёта регистра символов UTF8 подстроки pmbs2 в UTF32 строке pmbs1.
 * Если подстрока найдена, то функция возвращает указатель на начало найденного вхождения в строке pmbs1.
 * Если подстрока не найдена, то функция возвращает NULL.
 * @param pmbs1 Указатель на строку 1.
 * @param pmbs2 Указатель на строку 2.
 * @return char* 
 */
char* mbscasestr(const char* pmbs1, const char* pmbs2);

__END_DECLS

#endif
