/**
 * @file wctype.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с символами wchar_t.
 * @version 0.1
 * @date 2024-05-14
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef WCTYPE_H
#define WCTYPE_H

#include <cdefs.h>
#include <stddef.h>

//
// Поддерживаемые кодовые страницы UTF для towupper() и towlower()
//
#define UTF_SUPPORT_LATIN       0           // Латиница: Итальянский, Румынский, Молдавский, Древнеримский и т.п.
#define UTF_SUPPORT_FWLATIN     0           // Моноширийная латиница
#define UTF_SUPPORT_GREEK       0           // Греческий, Древнегреческий
#define UTF_SUPPORT_COPTIC      0           // Коптский
#define UTF_SUPPORT_CYRILLIC    0           // Кирилица: Белорусский, Украинский, Сербский, Болгарский и т.п.
#define UTF_SUPPORT_CYRILLICR   1           // Кирилица: Русский
#define UTF_SUPPORT_OLDCYRILLIC 0           // Старорусский, Церковнославянский
#define UTF_SUPPORT_ARMENIAN    0           // Армянский
#define UTF_SUPPORT_GEORGIAN    0           // Грузинский
#define UTF_SUPPORT_CHEROKEE    0           // Письменность индейцев Чероки
#define UTF_SUPPORT_GLAGOLITIC  0           // Глаголица
#define UTF_SUPPORT_DESERET     0           // Письменность DESERET (Дезеретский алфавит, придумали мормоны)
#define UTF_SUPPORT_OSAGE       0           // Письменность OSAGE (мёртвый сиуанский язык индейцев осейджи, близкий к языку дхегиха)
#define UTF_SUPPORT_HUNGARIAN   0           // Венгерский
#define UTF_SUPPORT_WARANG_CITI 0           // Письменность WARANG CITI (письменность индийских штатов Джаркханд, Бихар и Орисса)
#define UTF_SUPPORT_MEDEFAIDRIN 0           // Письменность MEDEFAIDRIN (письменность религиозной секты "Obɛri ɔkaimɛ")
#define UTF_SUPPORT_ADLAM       0           // Письменность ADLAM (язык нигеро-конголезской языковой макросемьи)

#if UTF_SUPPORT_CYRILLIC == 1
    #define UTF_SUPPORT_CYRILLICR   1
#endif

__BEGIN_DECLS

/**
 * @brief Перевод "широкого" символа в верхний регистр.
 * @param wc 
 * @return wchar_t 
 */
wchar_t towupper(wchar_t wc);

/**
 * @brief Перевод "широкого" символа в нижний регистр.
 * @param wc 
 * @return wchar_t 
 */
wchar_t towlower(wchar_t wc);

/**
 * @brief Функция возвращает ненулевое значение, если c является широким символом, принадлежащим классу широких символов "alpha".
 * В других случаях возвращается ноль.
 * @param c 
 * @return int 
 */
int iswalpha(wchar_t c);

/**
 * @brief Функция возвращает ненулевое значение, если c является цифрой.
 * @param c 
 * @return int 
 */
int iswdigit(wchar_t c);

/**
 * @brief Функция проверяет, является ли символ широким текстовым символом.
 * @param c 
 * @return int 
 */
int iswalnum(wchar_t c);

/**
 * @brief Функция проверяет, является ли символ "пустым" широким символом.
 * @param c 
 * @return int 
 */
int iswblank(wchar_t c);

/**
 * @brief Функция проверяет, является ли символ широким управляющим символом.
 * @param c 
 * @return int 
 */
int iswcntrl(wchar_t c);

/**
 * @brief Функция проверяет, является ли символ широким символом в нижнем регистре.
 * @param c 
 * @return int 
 */
int iswlower(wchar_t c);

/**
 * @brief Функция проверяет, является ли символ широким символом в верхнем регистре.
 * @param c 
 * @return int 
 */
int iswupper(wchar_t c);

/**
 * @brief Функция проверяет, является ли символ печатаемым широким символом.
 * @param c 
 * @return int 
 */
int iswprint(wchar_t c);

/**
 * @brief Функция проверяет, является ли символ широким графическим символом.
 * @param c 
 * @return int 
 */
int iswgraph(wchar_t c);

/**
 * @brief Функция проверяет, является ли символ "пустым" широким символом.
 * @param c 
 * @return int 
 */
int iswspace(wchar_t c);

/**
 * @brief Функция проверяет, является ли данный символ широким символом пунктуации или символьным широким символом.
 * @param c 
 * @return int 
 */
int iswpunct(wchar_t c);

/**
 * @brief Функция проверяет, является ли символ широким символом в виде шестнадцатеричного числа.
 * @param c 
 * @return int 
 */
int iswxdigit(wchar_t c);

__END_DECLS

#endif
