/**
 * @file weather.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Получение данных о фактической погоде.
 * Для получения информации о фактической погоде используются данные предоставляемые 
 * проектом "OpenWeather" ©2012—2024 OpenWeather ® All rights reserved.
 * Работоспособность сервиса и достоверность предоставляемых данных не гарантируется.
 * @version 0.1
 * @date 2024-05-26
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef WEATHER_H
#define WEATHER_H

#include <cdefs.h>

__BEGIN_DECLS

// Значения кода общего описания погоды: weather
//
// 2xx: Гроза
//  200  Гроза с небольшим дождём
//  201  Гроза с дождём
//  202  Гроза с сильным дождём
//  210  Небольшая гроза
//  211  Гроза
//  212  Сильная гроза
//  221  Очень сильная гроза
//  230  Гроза с моросящим дождём
//  231  Гроза с моросящим дождём
//  232  Гроза с сильным моросящим дождём
//
// 3xx  Моросящий дождь
//  300  Слабый моросящий дождь
//  301  Моросящий дождь
//  302  Сильный моросящий дождь
//  310  Слабый моросящий дождь
//  311  Моросящий дождь
//  312  Сильный моросящий дождь
//  313  Сильный ливень
//  314  Сильный ливень
//  321  Мелкий дождь

// 5хх  Дождь
//  500  Небольшой дождь
//  501  Умеренный дождь
//  502  Сильный дождь
//  503  Очень сильный дождь
//  504  Очень сильный дождь
//  511  Ледяной дождь
//  520  Лёгкий дождь
//  521  Ливень
//  522  Сильный ливень
//  531  Очень сильный ливень

// 6xx  Снег
//  600  Небольшой снег
//  601  Снег
//  602  Сильный снегопад
//  611  Мокрый снег
//  612  Небольшой дождь со снегом
//  613  Дождь со снегом
//  615  Небольшой дождь со снегом
//  616  Снег с дождём
//  620  Небольшой дождь со снегом
//  621  Дождь со снегом
//  622  Очень сильный снегопад

// 7xx  Атмосфера
//  701  Туман
//  711  Смог
//  721  Дымка
//  731  Песчаная буря
//  741  Туман
//  751  Песок
//  761  Пыль
//  762  Вулканический пепел
//  771  Шквал
//  781  Торнадо

// 8xx  Облачность
//  800  Ясно
//  801  Небольшая облачность 11-25%
//  802  Рассеянная облачность 25-50%
//  803  Разорванные облака 51-84%
//  804  Облачность 85-100%

// Числовые идентификаторы мест.
// Полный список числовых идентификаторов может быть получен по адресу http://bulk.openweathermap.org/sample.
#define WID_MOSCOW_RU           524901
#define WID_SPETERSBURG         498817
#define WID_MURMANSK_RU         524305
#define WID_YEKATERINBURG_RU    1486209
#define WID_YAKUTSK_RU          2013159
#define WID_PERM_RU             511196
//
#define WID_MINSK_BY            625144

typedef struct {
    int updated;                        ///< время фактичекого обновления данных по часам кэш-сервера
    //
    int weather;                        ///< код общего описание погоды
    float temperature;                  ///< температура, С
    float humidity;                     ///< относительная влажность, %
    float pressure;                     ///< атмосферное давление, гПа
    float wind_speed;                   ///< скорость ветра, м/с
    float wind_dir;                     ///< направление ветра
    float cloud;                        ///< облачность, %
    float visibility;                   ///< видимиость, м
} weather_data_t;

/**
 * @brief Запрос данных о фактической погоде для указанного места.
 * Для указания места, для которого выполняется запрос погодных данных, используется числовой идентификатор.
 * Полный список числовых идентификаторов может быть получен по адресу http://bulk.openweathermap.org/sample.
 * При успешном выполнении функция заполняет поля структуры weather_data и возвращает значение 1.
 * При ошибке функция возвращает значение 0.
 * @param place_id Числовой идентификатор места.
 * @param weather_data Указатель на структуру данных.
 * @return int 1-успех,0-ошибка.
 */
int weather(int place_id, weather_data_t *weather_data);

/**
 * @brief Возвращает указатель на строку описания погодных условий.
 * При некорректном значении кода возвращает значение NULL.
 * @param id Код описания погодных условий.
 * @return char* Указатель на строку или NULL.
 */
const char *weather_text(int id);

__END_DECLS

#endif
