/**
 * @file lanmon.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции клиента сервера LanMon.
 * @version 0.1
 * @date 2024-04-25
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <satthreads.h>
#include <syscallp.h>
#include <satsock.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <iconv.h>

#include <lanmon.h>

// #define LMDEBUG

#ifdef LMDEBUG
    #include <stdio.h>
    #define TRACE(fmt, ...) \
                do { printf(fmt, ##__VA_ARGS__); } while (0)
#else
    #define TRACE(fmt, ...)
#endif

static lm_channel_t *LMChannelList      = NULL;
static size_t        LMChannelListCount = 0;

/**
 * @brief Признак запущенного потока приёма переменных
 */
static volatile int lm_thread_running   = 0;

/**
 * @brief Признак необходимости остановки потока приёма переменных
 */
static volatile int lm_thread_terminate = 0;

static char *lm_host        = NULL;
static char *lm_login       = NULL;
static char *lm_password    = NULL;
static unsigned short int lm_port;
static unsigned int lm_timeout;

/**
 * @brief Идентификатор потока приёма переменных
 */
static thrd_t lm_thread;

/**
 * @brief Состояние потока
 */
typedef enum lm_thr_state {
    LMTS_WAITING,
    LMTS_START_CONNECT,
    LMTS_WAIT_CONNECTION,
    LMTS_CONNECTED
} lm_thr_state_t;

static int LmOpen(char *host, unsigned short int port, unsigned int timeout, char *login, char *password);
static int LmChanRegister(char *chname, lm_type_t type, char *unit, char *cmnts, int ctrl);
static int LmSendChan(lm_channel_t * channel);
static void LmClose(void);
static int LmState(void);

/**
 * @brief Функция потока обработки подключения к серверу LM.
 * @param p Функция - обработчик получения команды управления от сервера.
 * @return int 
 */
static int lm_thread_func(void *p) {
    //
    lmcontrol_func cb = p;
    lm_thr_state_t thr_state = LMTS_START_CONNECT;
    int waiting = 30;
    //
    while(!lm_thread_terminate) {
        msleep(100);
        //
        switch (thr_state) {
            case LMTS_WAITING:
                //
                // ожидание перед началом подключения
                if(waiting > 0) waiting--;
                if(waiting == 0) thr_state = LMTS_START_CONNECT;
                break;
            case LMTS_START_CONNECT:
                //
                // имя хоста ещё не установлено
                if(!lm_host) continue;
                //
                if(!LmOpen(lm_host, lm_port, lm_timeout, lm_login, lm_password)) {
                    // попытка запуска не удалась, ждём 5 секунд
                    waiting = 50;
                    thr_state = LMTS_WAITING;
                } else thr_state = LMTS_WAIT_CONNECTION;
                break;
            case LMTS_WAIT_CONNECTION: {
                    // ждём установки соединения
                    int lmstate = LmState();
                    if(lmstate == lmsOK) {
                        // соединение успешно установлено
                        // сбрасываем флаги активности и прочего...
                        for(int i=0; i < LMChannelListCount; i++) {
                            lm_channel_t *channel = &LMChannelList[i];
                            //
                            channel->num = 0;                                   // идентификатор на сервере
                            channel->state.active = 0;                          // канал больше не активен
                            channel->state.registered = 0;                      // не зарегистрирован
                            channel->state.need_registration = 1;               // требует регистрации
                            channel->state.control = 0;                         // нет управления
                            channel->state.changed = channel->state.valid;      // если значение канала было установлено, то ставим признак изменения
                        }
                        // устанавливаем список каналов
                        __internal_syscall_invoke2(SYSCALL_LMSetChannelList, (long)LMChannelList, LMChannelListCount);
                        thr_state = LMTS_CONNECTED;
                    } else {
                        if(lmstate == lmsDISCONNECTED) {
                            // соединение не установлено
                            LmClose();
                            // делаем повторную попытку через 10 секунд
                            waiting = 100;
                            thr_state = LMTS_WAITING;
                        }
                    }
                }
                break;
            case LMTS_CONNECTED: {
                    // соединение установлено, проверка на обрыв
                    if(LMConnected() == 0) {
                        // соединение было оборвано, закрываем соединение
                        LmClose();
                        // делаем повторную попытку через 10 секунд
                        waiting = 100;
                        thr_state = LMTS_WAITING;
                    }
                    // соединение точно установлено
                    for(int i=0; i < LMChannelListCount; i++) {
                        // перебор каналов
                        lm_channel_t *channel = &LMChannelList[i];
                        //
                        if(channel->state.need_registration) {
                            if(LmChanRegister(channel->name, channel->type, channel->unit, channel->cmnts, channel->state.rw)) {
                                channel->state.need_registration = 0;
                            } else {
                                // ошибка регистрации канала
                                TRACE("reg_e:%s\n", channel->name);
                            }
                        } else
                        if(channel->state.registered) {
                            // канал зарегистрирован
                            if(channel->state.control) {
                                // была принята команда управления
                                channel->quality = lmqOK;
                                channel->state.valid = 1;
                                if(cb) {
                                    cb(i, channel);
                                } else {
                                    // если нет обработчика, то ставим признак изменения для передачи обратно на сервер
                                    channel->state.changed = 1;
                                }
                                channel->state.control = 0;
                            }
                            //
                            if(channel->state.changed && channel->state.active) {
                                // канал изменён и активен
                                if(LmSendChan(channel)) {
                                    // канал успешно отправлен
                                    channel->state.changed = 0;
                                } else {
                                    // ошибка отправки канала
                                    TRACE("snd_e:%s\n", channel->name);
                                }
                            }
                        }
                    }
                }
                break;
        }
    }
    //
    // останавливаем работу
    LmClose();
    //
    // освобождаем память строк единиц измерения и коментариев
    for(int i=0; i < LMChannelListCount; i++) {
        lm_channel_t *channel = &LMChannelList[i];
        //
        if(channel->unit) free(channel->unit);
        if(channel->cmnts) free(channel->cmnts);
    }
    //
    // освобождаем память списка каналов
    LMChannelListCount = 0;
    free(LMChannelList);
    LMChannelList = NULL;
    //
    if(lm_password) free(lm_password);
    if(lm_login) free(lm_login);
    if(lm_host) free(lm_host);
    //
    lm_thread_running = 0;
    return 0;
}

/**
 * @brief Функция начинает работу встроенного клиента сервера LanMon.
 * Функция запускает отдельный поток, который устанавливает и поддерживает соединение с сервером.
 * Для завершения работы с сервром необходимо использовать функцию LMStop().
 * Система поддерживает только одно соединение с сервером.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0. Ошибка может быть вызвана некорректным
 * значением параметров, нехваткой системных ресурсов или попыткой повтороного запуска.
 * @param host Имя или IP адрес сервера.
 * @param port Номер TCP-порта сервера.
 * @param timeout Таймаут на операции обмена с сервером, мс
 * @param login Имя учётной записи на сервере.
 * @param password Пароль учётной записи на сервере.
 * @param cb Функция - обработчик получения команды управления от сервера.
 * @return int 1-успех, 0-ошибка.
 */
int LMStart(char *host, unsigned short int port, unsigned int timeout, char *login, char *password, lmcontrol_func cb) {
    // проверка попытки повторного запуска
    if(lm_thread_running) return 0;
    // проверка корректности параметров
    if(!host) return 0;
    if(!login) return 0;
    if(!password) return 0;
    if(!port) port = 3000;
    //
    lm_host = NULL;
    lm_login = NULL;
    lm_password = NULL;
    lm_timeout = timeout;
    lm_port = port;
    //
    // запускаем поток
    lm_thread_terminate = 0;
    if(thrd_create(&lm_thread, lm_thread_func, cb) != thrd_success) return 0;
    lm_thread_running = 1;
    //
    lm_password = strdup(password);
    lm_login = strdup(login);
    lm_host = strdup(host);
    //
    return 1;
}

/**
 * @brief Функция останавливает работу встроенного клиента сервера LanMon и освобождает все выделенные ресурсы.
 */
void LMStop(void) {
    if(!lm_thread_running) return;
    //
    // останавливаем поток
    lm_thread_terminate = 1;
    // ожидаем завершения его работы
    thrd_join(lm_thread, NULL);
}

#define ABS_DIF(a,b) ((a>b)?(a-b):(b-a))

static int filter_int(lm_channel_t *channel, int oldv, int newv) {
    if(channel->state.filter == lmfAbsolute)
        return (ABS_DIF(newv, oldv) > channel->filter)?1:0;            // фильтрация абсолютного изменения значения
    else
        return (newv == oldv)?0:1;                                  // фильтрации нет
}

static int filter_float(lm_channel_t *channel, float oldv, float newv) {
    switch (channel->state.filter) {
        case lmfAbsolute:
            // абсолютная фильтрация
            if(ABS_DIF(newv, oldv) > channel->filter) return 1; else return 0;
        case lmfPercent:
            // относительная фильтрация
            if(ABS_DIF(newv, oldv) > oldv * channel->filter) return 1; else return 0;
        default:
            // фильтрации нет
            return (newv != oldv)?1:0;
    }
}

static int filter_double(lm_channel_t *channel, double oldv, double newv) {
    switch (channel->state.filter) {
        case lmfAbsolute:
            // абсолютная фильтрация
            if(ABS_DIF(newv, oldv) > channel->filter) return 1; else return 0;
        case lmfPercent:
            // относительная фильтрация
            if(ABS_DIF(newv, oldv) > oldv * channel->filter) return 1; else return 0;
        default:
            // фильтрации нет
            return (newv != oldv)?1:0;
    }
}

static int LMCompareChannelValue(lm_channel_t *channel, lm_value_t *value) {
    if(channel->quality != lmqOK) return 1;
    if(channel->state.valid == 0) return 1;
    //
    switch (channel->type) {
        case lmtI2:     return filter_int(channel, channel->value.i2, value->i2);
        case lmtI4:     return filter_int(channel, channel->value.i4, value->i4);
        case lmtR4:     return filter_float(channel, channel->value.r4, value->r4);
        case lmtR8:     return filter_double(channel, channel->value.r8, value->r8);
        case lmtDATE:   return filter_double(channel, channel->value.datetime, value->datetime);
        case lmtBOOL:   return (channel->value.boolean == value->boolean)?0:1;
        case lmtI1:     return filter_int(channel, channel->value.i1, value->i1);
        case lmtUI1:    return filter_int(channel, channel->value.ui1, value->ui1);
        case lmtUI2:    return filter_int(channel, channel->value.ui2, value->ui2);
        case lmtUI4:    return filter_int(channel, channel->value.ui4, value->ui4);
        default:        return 0;
    }
}

static int LMValidType(lm_type_t type) {
    switch (type) {
        case lmtI2:
        case lmtI4:
        case lmtR4:
        case lmtR8:
        case lmtDATE:
        case lmtBOOL:
        case lmtI1:
        case lmtUI1:
        case lmtUI2:
        case lmtUI4:
            return 1;
        default:
            return 0;
    }
}

char *copy_attr_str(char *str) {
    if(!str) return NULL;
    //
    int len = strlen(str);
    char *buf = alloca(len+1);
    if(utf8_to_cp1251(str, buf, len+1)) {
        return strdup(buf);
    } else {
        return NULL;
    }
}

/**
 * @brief Добавление нового канала.
 * При успешном выполнении функция возвращает числовой идентификатор канала: число большее или равное 0.
 * При ошибке функция возвращает значение -1. Ошибка может быть вызвана некорректным значением параметров или
 * попыткой создания канала с уже существующим именем.
 * @param chname Имя канала.
 * @param type Тип канала.
 * @param unit Единицы измерения.
 * @param cmnts Текстовый коментарий.
 * @param ctrl Признак возможности управления: 1-возможность записи, 0-только чтение.
 * @return int Идентификатор канала или ошибка (-1).
 */
int LMAddChannel(char *chname, lm_type_t type, char *unit, char *cmnts, int rw) {
    if(!lm_thread_running) return -1;
    if(!LMValidType(type)) return -1;
    if(!chname) return -1;
    //
    // преобразуем имя канала в CP1251
    char name[LM_MAX_NAME_SIZE];
    if(!utf8_to_cp1251(chname, name, LM_MAX_NAME_SIZE)) return -1;
    //
    for(int i=0; i < LMChannelListCount; i++) {
        // проверка имени канала
        if(strcmp(name, LMChannelList[i].name) == 0) return -1;
    }
    // дубликатов нет, создаём новый канал
    lm_channel_t channel;
    memset(&channel, 0, sizeof(lm_channel_t));
    //
    // копируем имя
    memmove(channel.name, name, LM_MAX_NAME_SIZE);
    //
    channel.unit =  copy_attr_str(unit);
    channel.cmnts = copy_attr_str(cmnts);
    //
    channel.num = 0;
    channel.quality = lmqUndefined;
    channel.type = type;
    channel.state.need_registration = 1;        // канал требует регистрации
    channel.state.registered = 0;               // канал зарегистрирован на сервере
    channel.state.changed = 0;                  // канал изменён клиентом
    channel.state.control = 0;                  // канал изменён сервером
    channel.state.active = 0;                   // канал активен
    channel.state.valid = 0;                    // значение хотя-бы один раз было установлено
    channel.state.rw = rw;                      // канал поддерживает запись
    //
    // добавляем в список каналов
    LMChannelListCount++;
    LMChannelList = realloc(LMChannelList, LMChannelListCount*sizeof(lm_channel_t));
    memmove(&LMChannelList[LMChannelListCount-1], &channel, sizeof(lm_channel_t));
    //
    __internal_syscall_invoke2(SYSCALL_LMSetChannelList, (long)LMChannelList, LMChannelListCount);
    //
    // возвращаем индекс нового канала в списке
    return LMChannelListCount-1;
}

/**
 * @brief Получение описания канала.
 * При успешном выполнении функция возвращает значение указателя (lm_channel) на канал с числовым идентификатором client_id.
 * При отсутствии запрошенного канала функция возвращает значение NULL.
 * @param client_id Числовой идентификатор канала.
 * @return lm_channel_t* Указатель на описание канала или NULL.
 */
lm_channel_t *LMGetChannelInfo(int client_id) {
    if((client_id < 0) || (client_id >= LMChannelListCount)) return NULL;
    //
    return &LMChannelList[client_id];
}

/**
 * @brief Передача значения канала на сервер.
 * @param client_id Числовой идентификатор канала.
 * @param value Значение канала.
 * @return int 2-значение отфильтровано, 1-успех, 0-ошибка.
 */
int LMSendValue(int client_id, lm_value_t *value) {
    if(!value) return 0;
    lm_channel_t *channel = LMGetChannelInfo(client_id);
    if(!channel) return 0;
    //
    // сравнение с предыдущим значением
    if(!LMCompareChannelValue(channel, value)) return 2;
    //
    // значения не совпадают, переносим значение в список
    memmove(&channel->value, value, sizeof(lm_value_t));
    channel->quality = lmqOK;
    channel->state.valid = 1;
    channel->state.changed = 1;
    channel->state.control = 0;
    //
    return 1;
}

/**
 * @brief Установка свойства Quality для указанного канала.
 * Функция позволяет установить значение Quality отличное от lmqOK, для установки значения lmqOK необходимо  использовать
 * функцию LMSendValue. При успешном выполнении функция возвращает значение 1.
 * Если предыдущее значение свойства совпадает с quality, то функция возвращает значение 2.
 * Если канал с числовым идентификатором client_id не найден или (quality == lmqOK), то функция возвращает значение 0 (ошибка).
 * @param client_id Числовой идентификатор канала.
 * @param quality Значение Quality (элемент перечисления lm_quality).
 * @return int 1-успех, 2-значение отфильтровано, 0-ошибка.
 */
int LMSetQuality(int client_id, lm_quality_t quality) {
    if((quality < lmqOff) || (quality > lmqRegFailure)) return 0;
    lm_channel_t *channel = LMGetChannelInfo(client_id);
    if(!channel) return 0;
    //
    if((channel->quality == quality) && (channel->state.valid == 1)) return 2;
    //
    channel->quality = quality;
    channel->state.valid = 1;
    channel->state.changed = 1;
    channel->state.control = 0;
    //
    return 1;
}

/**
 * @brief Установка значения фильтра на относительное изменение значения канала.
 * Если значение канала изменилось меньше, чем на указанную относительную величину, то оно передавться на сервер не будет.
 * Фильтр действует только для каналов с типами lmtR4 и lmtR8.
 * Если канал с указанным client_id не найден или его тип не соотвествует lmtR4 или lmtR8, то функция 
 * вернёт значение 0, при успешном выполнении: 1.
 * @param client_id Числовой идентификатор канала.
 * @param value Значение фильтра, проценты
 * @return int 1-успех, 0-ошибка
 */
int LMSetFilterPercent(int client_id, float value) {
    lm_channel_t *channel = LMGetChannelInfo(client_id);
    if(!channel) return 0;
    if((channel->type != lmtR4) && (channel->type != lmtR8)) return 0;
    //
    channel->filter = value / 100.0f;
    channel->state.filter = lmfPercent;
    return 1;
}

/**
 * @brief Установка значения фильтра на абсолютное изменение значения канала.
 * Если значение канала изменилось меньше, чем на указанную абсолютную величину, то оно передавться на сервер не будет.
 * Если канал с указанным client_id не найден, то функция вернёт значение 0, при успешном выполнении: 1.
 * @param client_id Числовой идентификатор канала.
 * @param value Значение фильтра, абсолютная величина изменения.
 * @return int 1-успех, 0-ошибка
 */
int LMSetFilterAbsolute(int client_id, float value) {
    lm_channel_t *channel = LMGetChannelInfo(client_id);
    if(!channel) return 0;
    //
    channel->filter = value;
    channel->state.filter = lmfAbsolute;
    return 1;
}

/**
 * @brief Отключение фильтрации значений канала.
 * Функция отключает предварительно установленную фильтрацию.
 * Если канал с указанным client_id не найден, то функция вернёт значение 0, при успешном выполнении: 1.
 * @param client_id Числовой идентификатор канала.
 * @return int 1-успех, 0-ошибка
 */
int LMSetFilterNone(int client_id) {
    lm_channel_t *channel = LMGetChannelInfo(client_id);
    if(!channel) return 0;
    //
    channel->state.filter = lmfNone;
    return 1;
}

/**
 * @brief Получение статуса подключения к серверу LanMon.
 * При активном подключении к серверу функция возвращает значение 1, в противном случае: 0.
 * @return int 1-есть подключение, 0-нет подключения.
 */
inline int LMConnected(void) {
    return __internal_syscall_invoke1(SYSCALL_LmConnected, 0);
}

/**
 * @brief Регистрация нового канала на сервере.
 * Функция LmChanRegister позволяет выполнить регистрацию канала на сервере и получить номер канала и активность канала на сервере.
 * @param chname Имя канала (непустая строка длиной не более 63 символов).
 * @param type Тип создаваемого канала.
 * @param unit Строка описывающая единицы измерения переменной значения канала.
 * @param cmnts Текстовая строка описания канала.
 * @param ctrl Разрешение управления каналом: 1-разрешено, 0-запрещено.
 * @return int 1-успех, 0-ошибка.
 */
static inline int LmChanRegister(char *chname, lm_type_t type, char *unit, char *cmnts, int ctrl) {
    return __internal_syscall(SYSCALL_LmChanRegister, (long)chname, type, (long)unit, (long)cmnts, ctrl, 0);
}

/**
 * @brief Запись числового значения канала на сервер по номеру канала или по имени канала.
 * Канал идентифицируется по ненулевому номеру в структуре channel.
 * Если номер канала в структуре нулевой, то канал идентифицируется по имени в структуре channel.
 * @param channel Указатель на описание канала.
 * @return int 1-успех, 0-ошибка.
 */
static int LmSendChan(lm_channel_t * channel) {
    return __internal_syscall_invoke1(SYSCALL_LmSendChan, (long)channel);
}

/**
 * @brief Функция начинает работу встроенного клиента сервера LanMon в неблокирующем (асинхронном) режиме.
 * Функция иницирует подключение к серверу. После вызова этой функции встроенный клиент самостоятельно 
 * поддерживает связь с сервером, выполняет все необходимые действия по авторизации на сервере и
 * переподключение в случае утери связи.
 * Текущее состояние подключения можно узнать при помощи функций LMConnected и LmState.
 * При успешном выполнении функция возвращает 1, при ошибке: 0.
 * @param host Имя или IP адрес сервера.
 * @param port Номер TCP-порта сервера.
 * @param timeout Таймаут на операции обмена с сервером, мс
 * @param login Имя учётной записи на сервере.
 * @param password Пароль учётной записи на сервере.
 * @return int 1-успех, 0-ошибка.
 */
static int LmOpen(char *host, unsigned short int port, unsigned int timeout, char *login, char *password) {
    if(!host) return 0;
    if(!login) return 0;
    if(!password) return 0;
    if(!port) port = 3000;
    //
    // проверка статуса Ethernet
    if(EthernetLink() < ETH_LINK_10_HALF) return 0;
    //
    return __internal_syscall(SYSCALL_LmOpen, (long)host, port, timeout, (long)login, (long)password, 0);
}

/**
 * @brief Завершение работы встроенного клиента LanMon.
 * Если имеется подключение к серверу, то оно будет завершено.
 */
static inline void LmClose(void) {
    __internal_syscall_invoke1(SYSCALL_LmClose, 0);
}

/**
 * @brief Получение состояния клиента сервера LanMon.
 * Функция возвращает текущее состояние клиента сервера LanMon - одно из значений перечисления lm_state_enum.
 * @return int Текущее состояние клиента.
 */
static inline int LmState(void) {
    return __internal_syscall_invoke1(SYSCALL_LmState, 0);
}
