/**
 * @file merc230.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Чтение данных счётчика электрической энергии Меркурий 230.
 * @version 0.1
 * @date 2024-03-30
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * @example Пример чтения текущих данных:
 * 
 * merc230_current_t current_data;
 * //
 * // открываем порт
 * int handle = SerialOpen(1, B9600, PARNONE, 1);
 * //
 * int res = Merc230_ReadCurrentData(handle, 0, &current_data);
 * if(res) {
 *   printf("Data read successfully\n");
 * } else {
 *   printf("Can''t read data\n");
 * }
 * //
 * SerialClose(handle);
 * 
 */

#include <stdint.h>
#include <satcrc.h>
#include <fastmem.h>
#include <math.h>

#include <merc230.h>

#define MERC_TIMEOUT    200

/**
 * @brief Открыть канал связи
 * 
 * @param handle 
 * @param addr 
 * @return int 
 */
static int merc230open(int handle, uint8_t addr) {
    uint8_t rbuf[16];
    uint8_t req[16];
    uint16_t crc;
    //
    req[0] = addr;                      // адрес
    req[1] = 1;                         // код запроса
    req[2] = 1;                         // уровень доступа
    //
    req[3] = 1;                         // пароль 1
    req[4] = 1;                         // пароль 2
    req[5] = 1;                         // пароль 3
    req[6] = 1;                         // пароль 4
    req[7] = 1;                         // пароль 5
    req[8] = 1;                         // пароль 6
    //
    crc = crc_modbus(req, 9);           // crc
    req[9]  = crc & 0xff;
    req[10] = (crc >> 8) & 0xff;
    //
    // передача запроса
    purge(handle);
    write(handle, req, 11);
    //
    // приём ответа
    int res = read_block(handle, rbuf, 4, MERC_TIMEOUT);
    //
    if(res > 0) return 1; else return res;
}

/**
 * @brief Чтение вспомогательных параметров
 * 
 * @param handle Дескриптор последовательного порта.
 * @param addr Сетевой адрес счётчика: 1..240, 0 - широковещательный адрес.
 * @param bwri Код параметра
 * @param alen Ожидаемый размер данных
 * @param answer Указатель на буфер принятых данных
 * @return int 
 */
static int ReadAuxParameters(int handle, uint8_t addr, uint8_t bwri, int alen, uint8_t *answer) {
    uint8_t rbuf[16];
    uint8_t req[8];
    uint16_t crc;
    //
    req[0] = addr;                      // адрес
    req[1] = 8;                         // код запроса
    req[2] = 0x16;                      // номер параметра
    req[3] = bwri;                      // bwri
    //
    crc = crc_modbus(req, 4);           // crc
    req[4] = crc & 0xff;
    req[5] = (crc >> 8) & 0xff;
    //
    // передача запроса
    purge(handle);
    write(handle, (char *)req, 6);
    //
    // приём ответа
    int res = read_block(handle, rbuf, alen + 3, MERC_TIMEOUT);
    if(res <= 0) return res;
    //
    // проверка CRC
    if(crc_modbus(rbuf, alen + 3)) return 0;
    //
    // копирование данных
    fast_memmove(answer, &rbuf[1], alen);
    //
    return 1;
}

static int ReadEnergy(int handle, uint8_t addr, uint8_t tarif, uint8_t *answer) {
    uint8_t rbuf[24];
    uint8_t req[8];
    uint16_t crc;
    //
    req[0] = addr;                      // адрес
    req[1] = 5;                         // код запроса
    req[2] = 0;                         // номер массива (от сброса) + номер месяца (0)
    req[3] = tarif;                     // tarif
    //
    crc = crc_modbus(req, 4);           // crc
    req[4] = crc & 0xff;
    req[5] = (crc >> 8) & 0xff;
    //
    // передача запроса
    purge(handle);
    write(handle, req, 6);
    //
    // приём ответа
    int res = read_block(handle, (char *)rbuf, 19, MERC_TIMEOUT);
    if(res <= 0) return res;
    //
    // проверка CRC
    if(crc_modbus(rbuf, 19)) return 0;
    //
    // копирование данных
    fast_memmove(answer, &rbuf[1], 16);
    //
    return 1;
}

static uint32_t value4(uint8_t *data) {
    uint32_t tmp = (uint32_t)(data[1])*256*256*256 + (uint32_t)(data[0])*65536 + (uint32_t)(data[3])*256 + data[2];
    if(tmp == 0xffffffff) return 0;
    return tmp;
}

static int value3(uint8_t *data) {
    int tmp = (uint32_t)(data[0] & 0x3f)*65536 + (uint32_t)(data[2])*256 + data[1];
    return tmp;
}

/**
 * @brief Чтение текущих данных счётчика электрической энергии Меркурий 230.
 * Для чтения данных используется дескриптор последовательного порта или сетевого сокета.
 * Функция возвращает 1 при успешном чтении данных. При ошибке функция возвращает:
 * 0  - счётчик не отвечает на запрос;
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param handle Дескриптор последовательного порта или сетевого сокета.
 * @param addr Сетевой адрес счётчика: 1..240, 0 - широковещательный адрес.
 * @param data Указатель на структуру текущих данных.
 * @return int 
 */
int Merc230_ReadCurrentData(int handle, uint8_t addr, merc230_current_t *data) {
    if(socket_handle_type(handle) == HDL_INVALID) return -1;
    if(!data) return 0;
    //
    uint8_t answer[32];
    int res;
    //
    res = merc230open(handle, addr);
    if(res <= 0) return res;
    //
    // фазное напряжение
    if(!ReadAuxParameters(handle, addr, 0x11, 9, answer)) return 0;
    for(int i=0; i<3; i++) {
        data->U[i] = value3(&answer[i*3]) / 100.0f;
    }
    //
    // ток
    if(!ReadAuxParameters(handle, addr, 0x21, 9, answer)) return 0;
    for(int i=0; i<3; i++) {
        data->I[i] = value3(&answer[i*3]) / 1000.0f;
    }
    //
    // cosF
    if(!ReadAuxParameters(handle, addr, 0x30, 12, answer)) return 0;
    for(int i=0; i<4; i++) {
        data->cosf[i] = value3(&answer[i*3]) / 1000.0f;
        if(data->cosf[i] > 1.0f) data->cosf[i] = 1.0f;
    }
    // меняем местами cosF
    float tmp = data->cosf[0];
    data->cosf[0] = data->cosf[1];
    data->cosf[1] = data->cosf[2];
    data->cosf[2] = data->cosf[3];
    data->cosf[3] = tmp;
    //
    // значения мощностей рассчитываем, т.к. протокол не может передать значения > 41.9 кВт
    // пофазный расчёт P, Q, S
    data->S[FASE_SUMMA] = 0;
    data->P[FASE_SUMMA] = 0;
    data->Q[FASE_SUMMA] = 0;
    //
    for(int i=0; i<3; i++) {
        data->S[i] = data->U[i] * data->I[i];                                   // полная мощность
        data->P[i] = data->S[i] * data->cosf[i];                                // активная мощность
        data->Q[i] = data->S[i] * sqrtf(1.0f - data->cosf[i]*data->cosf[i]);    // реактивная мощность
        //
        data->S[FASE_SUMMA] += data->S[i];
        data->P[FASE_SUMMA] += data->P[i];
        data->Q[FASE_SUMMA] += data->Q[i];
    }
    //
    return 1;
}

/**
 * @brief Чтение накопленных данных счётчика электрической энергии Меркурий 230.
 * Для чтения данных используется дескриптор последовательного порта или сетевого сокета.
 * Функция выполняет чтение накопленных данных для указанного количества тарифов.
 * Функция возвращает 1 при успешном чтении данных. При ошибке функция возвращает:
 * 0  - счётчик не отвечает на запрос;
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param handle Дескриптор последовательного порта или сетевого сокета.
 * @param addr Сетевой адрес счётчика: 1..240, 0 - широковещательный адрес.
 * @param tarif Количество читаемых тарифов: 0..4, 0 - читать только суммарные данные.
 * @param data Указатель на структуру накопленных данных.
 * @return int 
 */
int Merc230_ReadTotalData(int handle, uint8_t addr, uint8_t tarif, merc230_total_t *data) {
    if(socket_handle_type(handle) == HDL_INVALID) return -1;
    if(!data) return 0;
    //
    uint8_t answer[32];
    int res;
    //
    res = merc230open(handle, addr);
    if(res <= 0) return res;
    //
    // чтение данных по тарифам
    if(tarif > 4) tarif = 4;
    for(int ct=0; ct <= tarif; ct++) {
        res = ReadEnergy(handle, addr, ct, answer);
        if(res <= 0) return res;
        //
        for(int i=0; i<4; i++) {
            double tmp = value4(&answer[i*4]) / 1000.0;
            //
            switch (ct) {
                case 0:
                    data->summa[i] = tmp;
                    break;
                case 1:
                    data->tarif1[i] = tmp;
                    break;
                case 2:
                    data->tarif2[i] = tmp;
                    break;
                case 3:
                    data->tarif3[i] = tmp;
                    break;
                case 4:
                    data->tarif4[i] = tmp;
                    break;
            }
        }
    }
    return 1;
}
