/**
 * @file pgsql.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с сервером PostgreSQL.
 * @version 0.1
 * @date 2024-04-25
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include <syscallp.h>

#include <stdlib.h>
#include <pgsql.h>

#define CELLBUFSIZE   48

/**
 * @brief Установка максимальной продолжительности подключения и выполнение SQL запроса.
 * @param timeout Таймаут, мс.
 */
inline void PgSetTimeout(unsigned int timeout) {
    __internal_syscall_invoke1(SYSCALL_PgSetTimeout, timeout);
}

/**
 * @brief Установка кодировки символов для клиента СУБД.
 * Список поддерживаемых кодировок можно узнать из документации СУБД: https://www.postgresql.org/docs/current/multibyte.html.
 * Пример: "UTF8", "WIN1251", "WIN866".
 * По умаолчанию используется кодировка сервера.
 * @param encoding Строка с наименованием кодировки.
 */
inline void PgSetEncoding(char *encoding) {
    if(!encoding) return;
    __internal_syscall_invoke1(SYSCALL_PgSetEncoding, (long)encoding);
}

/**
 * @brief Установка буфера, используемого для ответов СУБД.
 * @param buf Указатель на буфер в котором будут размещён ответ сервера.
 * @param bufsize Размер буфера buf.
 */
void PgSetBuffer(void *buf, unsigned int bufsize) {
    if(!buf) return;
    if(!bufsize) return;
    //
    __internal_syscall_invoke2(SYSCALL_PgSetBuffer, (long)buf, bufsize);
}

/**
 * @brief Функция запускает процесс подключения к СУБД PostgreSQL.
 * При успешном начале подключения к серверу функция возвращает значение 1. При ошибке функция возвращает 0.
 * Ошибка может быть вызвана некорректным значением параметров, отсутствием свободные TCP сокетов или слишком маленьким размером буфера buf.
 * @param host Адрес сервера: имя или IP-адрес.
 * @param port Номер порта TCP.
 * @param database Наименование базы данных.
 * @param login Имя пользователя для подключения к СУБД.
 * @param password Пароль для подключения к СУБД.
 * @return int 1-успех, 0-ошибка.
 */
int PgOpen(char *host, unsigned short int port, char *database, char *login, char *password) {
    if(!host) return 0;
    if(!database) return 0;
    if(!password) return 0;
    if(!login) return 0;
    //
    return __internal_syscall(SYSCALL_PgOpen, (long)host, port, (long)database, (long)login, (long)password, 0);
}

/**
 * @brief Функция выполняет подключение к СУБД PostgreSQL в синхронном (блокирующем) режиме.
 * При успешном подключении к СУБД функция возвращает значение 1. При ошибке функция возвращает 0.
 * Ошибка может быть вызвана некорректным значением параметров, недоступностью СУБД, отсутствием свободные TCP сокетов или 
 * слишком маленьким размером буфера buf.
 * @param host Адрес сервера: имя или IP-адрес.
 * @param port Номер порта TCP.
 * @param database Наименование базы данных.
 * @param login Имя пользователя для подключения к СУБД.
 * @param password Пароль для подключения к СУБД.
 * @return int 1-успех, 0-ошибка.
 */
int PgOpenBlock(char *host, unsigned short int port, char *database, char *login, char *password) {
    int res = PgOpen(host, port, database, login, password);
    if(!res) return res;
    //
    while (1) {
        res = PgState();
        //
        switch (res) {
            case pgsOK:
                return 1;
            case pgsNONE:
            case pgsTCP_ERROR:
            case pgsSQL_TIMEOUT:
            case pgsDISCONNECTED:
            case pgsSQL_BUFSIZE:
                return 0;
            default:;
        }
    }
}

/**
 * @brief Функция выполняет отключение от СУБД PostgreSQL.
 * Вызов функции безопасен, в том числе, если подключение к серверу отсутствует.
 */
inline void PgClose(void) {
    __internal_syscall_invoke1(SYSCALL_PgClose, 0);
}

/**
 * @brief Возвращает состояние подключение к СУБД PostgreSQL.
 * @return int 0-подключение отсутствует, 1-подключение активно.
 */
inline int PgConnected(void) {
    return __internal_syscall_invoke1(SYSCALL_PgConnected, 0);
}

/**
 * @brief Функция заполняет переданный буфер buf пятибайтовой строкой с кодом ошибки выполнение SQL запроса.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * @param buf Указатель на буфер (минимум 5 байт).
 * @return int 1-успех, 0-ошибка.
 */
int PgErrorCode(char *buf) {
    if(!buf) return 0;
    //
    return __internal_syscall_invoke1(SYSCALL_PgErrorCode, (long)buf);
}

/**
 * @brief Поучение текста ошибки выполнения запроса.
 * Функция помещает в буфер text сообщение об ошибке вызванной выполнением последнего запроса.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * @param text Указатель на буфер.
 * @param textsize Размер буфера.
 * @return int 1 - успех, 0 - ошибка.
 */
int PgErrorText(char *text, unsigned int textsize) {
    if(!text) return 0;
    if(!textsize) return 0;
    //
    return __internal_syscall_invoke2(SYSCALL_PgErrorText, (long)text, textsize);
}

/**
 * @brief Начать выполнение SQL запроса.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * Ошибка может быть вызвана отсутствием подключения к СУБД, повторным вызовом функции (запрос уже выполняется) или некоректным значением аргумента.
 * @param request Строка SQL-запроса.
 * @return int 1-успех, 0-ошибка.
 */
inline int PgSQL(char *request) {
    if(!request) return 0;
    //
    return __internal_syscall_invoke1(SYSCALL_PgSql, (long)request);
}

/**
 * @brief Начать выполнение SQL запроса в блокирующем (синхронном) режиме.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * Ошибка может быть вызвана отсутствием подключения к СУБД, повторным вызовом функции (запрос уже выполняется) или некоректным значением аргумента.
 * @param request Строка SQL-запроса.
 * @return int 1-успех, 0-ошибка.
 */
int PgSQLBlock(char *request) {
    int res = PgSQL(request);
    if(!res) return 0;
    //
    while (1) {
        res = PgState();
        //
        switch (res) {
            case pgsSQL_OK:
                return 1;
            case pgsNONE:
            case pgsTCP_ERROR:
            case pgsSQL_TIMEOUT:
            case pgsDISCONNECTED:
            case pgsSQL_BUFSIZE:
                return 0;
            default:;
        }
    }
}

/**
 * @brief Получение состояния клиента СУБД PostgreSQL.
 * Результат функции может принимать значения присутствующие в перечислении pg_state_enum.
 * @return int Состояние клиентского подключения.
 */
inline int PgState(void) {
    return __internal_syscall_invoke1(SYSCALL_PgState, 0);
}

/**
 * @brief Получение длины ответа СУБД на запрос.
 * Значением можно воспользоваться при возникновении ошибки "Сервер выполнил запрос, но переданный буфер слишком мал для приёма ответа сервера" (pgsSQL_BUFSIZE).
 * Следует проверить значение длины ответа и увеличить размер буфера buf в функции PgSetBuffer или изменить SQL запрос.
 * @return unsigned int 
 */
inline unsigned int PgDataLen(void) {
    return __internal_syscall_invoke1(SYSCALL_PgDataLen, 0);
}

/**
 * @brief Получение количества колонок в ответе СУБД на SQL запрос, выполненный при помощи функции PgSql.
 * Если при выполнении SQL запроса СУБД вернула пустой ответ или произошла ошибка, то возвращается значение 0.
 * @return unsigned int Количество колонок в ответе СУБД.
 */
inline unsigned int PgColCount(void) {
    return __internal_syscall_invoke1(SYSCALL_PgColCount, 0);
}

/**
 * @brief Получение количества строк в ответе СУБД на SQL запрос, выполненный при помощи функции PgSql.
 * Если при выполнении SQL запроса СУБД вернула пустой ответ или произошла ошибка, то возвращается значение 0.
 * @return unsigned int Количество строк в ответе СУБД.
 */
inline unsigned int PgRowCount(void) {
    return __internal_syscall_invoke1(SYSCALL_PgRowCount, 0);
}

/**
 * @brief Получение имени колонки в ответе СУБД на SQL запрос, выполненный при помощи функции PgSql.
 * При успешном выполнении функция помещает в буфер name имя запрошенной колонки и возвращает знаечение 1.
 * При ошибке функция возвращает 0.
 * @param col Номер колонки для которой запрашивается имя: 0..PgColCount()-1.
 * @param name Буфер для имени колонки.
 * @param namesize Размер буфера name.
 * @return int 1-успех, 0-ошибка.
 */
inline int PgColName(unsigned int col, char *name, unsigned int namesize) {
    if(!name) return 0;
    //
    return __internal_syscall(SYSCALL_PgColName, col, (long)name, namesize, 0, 0, 0);
}

/**
 * @brief Получение текстового значения ячейки с номером колонки col и номером строки row из ответа СУБД.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * @param col Номер колонки: 0..PgColCount()-1
 * @param row Номер строки: 0..PgRowCount()-1
 * @param cell Указатель на буфер, куда будет записан текст значения указанной ячейки.
 * @param cellsize Размер буфера cell.
 * @return int 1-успех, 0-ошибка.
 */
inline int PgCell(unsigned int col, unsigned int row, char *cell, unsigned int cellsize) {
    return __internal_syscall(SYSCALL_PgCell, col, row, (long)cell, cellsize, 0, 0);
}

/**
 * @brief Получение числового (int) значения ячейки с номером колонки col и номером строки row из ответа СУБД.
 * При успешном выполнении функция помещает числовое значение в value и возвращает значение 1, при ошибке - 0.
 * @param col Номер колонки: 0..PgColCount()-1
 * @param row Номер строки: 0..PgRowCount()-1
 * @param value Указатель на значение.
 * @return int 1-успех, 0-ошибка.
 */
int PgCellAsInteger(unsigned int col, unsigned int row, int *value) {
    char buf[CELLBUFSIZE];
    int res = PgCell(col, row, buf, CELLBUFSIZE);
    if(!res) return 0;
    //
    *value = atoi(buf);
    return res;
}

/**
 * @brief Получение числового (double) значения ячейки с номером колонки col и номером строки row из ответа СУБД.
 * При успешном выполнении функция помещает числовое значение в value и возвращает значение 1, при ошибке - 0.
 * @param col Номер колонки: 0..PgColCount()-1
 * @param row Номер строки: 0..PgRowCount()-1
 * @param value Указатель на значение.
 * @return int 1-успех, 0-ошибка.
 */
int PgCellAsDouble(unsigned int col, unsigned int row, double *value) {
    char buf[CELLBUFSIZE];
    int res = PgCell(col, row, buf, CELLBUFSIZE);
    if(!res) return 0;
    //
    *value = atof(buf);
    return res;
}
