/**
 * @file satsun.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Астрономические функции.
 * @version 0.1
 * @date 2024-03-17
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include <math.h>
#include <stdbool.h>

#include <satsun.h>

static bool ArcError;

#define RADIANS(x) ((x) * 0.0174532925f)
#define DEGREES(x) ((x) * 57.2957795f)

#define BYTE    uint8_t
#define WORD    uint16_t
#define UINT    unsigned int

/**
 * @brief Расчёт угла возвышения солнца над горизонтом
 * This is a function that calculates the sun position (altitude) at a given location (latitude and longitude) and time (Unix timestamp).
 * It uses the solar coordinate algorithm from the U.S. Naval Observatory Astronomical Applications Department.
 * @param lat Широта
 * @param lon Долгота
 * @param unix Текущее время
 * @return float Угол над горизонтом в градусах. Положительное значение - солнце над горизонтом, отрицательное значение - солнце под горизонтом.
 */
float SunPosition(float lat, float lon, time_t unix) {
    lat = RADIANS(lat);
    float hours = (unix % 86400ul) / 86400.0f;                                                                                      // decimal hours
    float JulCent = (unix / 86400.0f - 10957.5f) / 36525.0f;                                                                        // Julian Century
    //
    float GeomMeanLong = RADIANS(fmodf(280.46646f + JulCent * (36000.76983f + JulCent * 0.0003032f), 360));                         // Geom Mean Long Sun
    float GeomMeanAnom = RADIANS(357.52911f + JulCent * (36000.0f - 0.0001537f * JulCent));                                         // Geom Mean Anom Sun
    //
    float EccEart = 0.016708634f - JulCent * (0.000042037f + 0.0000001267f * JulCent);                                              // Eccent Earth Orbit
    float SunEqCtr = RADIANS(sinf(GeomMeanAnom) * (1.914602f - JulCent * (0.004817f + 0.000014f * JulCent)) + sinf(2 * GeomMeanAnom) * (0.019993f - 0.000101f * JulCent) + sinf(3 * GeomMeanAnom) * 0.000289f); // Sun Eq of Ctr
    float SunApp = GeomMeanLong + SunEqCtr - 0.0001f - 0.00008f * sinf(RADIANS(125.04f - 1934.136f * JulCent));                     // Sun App Long
    float MeanOblq = 23 + (26 + ((21.448f - JulCent * (46.815f + JulCent * (0.00059f - JulCent * 0.001813f)))) / 60.0f) / 60.0f;    // Mean Obliq Ecliptic
    float ObliqCor = RADIANS(MeanOblq + 0.00256f * cosf(RADIANS(125.04f - 1934.136f * JulCent)));                                   // Obliq Corr
    float decl = asinf(sinf(ObliqCor) * sinf(SunApp));                                                                              // Sun Declin
    float y = tanf(ObliqCor / 2) * tanf(ObliqCor / 2);
    float eqTime = 4 * DEGREES(y * sinf(2 * GeomMeanLong) - 2 * EccEart * sinf(GeomMeanAnom) + 4 * EccEart * y * sinf(GeomMeanAnom) * cosf(2 * GeomMeanLong) - 0.5f * y * y * sinf(4 * GeomMeanLong) - 1.25f * EccEart * EccEart * sinf(2 * GeomMeanAnom));   // Eq of Time (minutes)
    float hrAngl = fmodf(hours * 1440 + eqTime + 4 * lon, 1440) / 4;                                                                // True Solar Time (min)
    hrAngl = hrAngl + (hrAngl < 0 ? 180 : -180);                                                                                    // Hour Angle
    float zen = acosf(sinf(lat) * sinf(decl) + cosf(lat) * cosf(decl) * cosf(RADIANS(hrAngl)));                                     // Zenith
    //
    return 90.0f - DEGREES(zen);
}

//---------------------------------------------------------------------------

static float calcMeanObliquityOfEcliptic(float t) {
    float seconds;
    seconds= 21.448f - t*(46.8150f + t*(0.00059f- t*(0.001813f)));
    return 23.0f + (26.0f + (seconds/60.0f))/60.0f;
}

static float calcGeomMeanLongSun(float t) {
    float L;
    L= 280.46646f + t * (36000.76983f + 0.0003032f * t);
    while(L > 360)L-=360.0f;
    while(L < 0)L+=360.0f;
    return L;
}

static float calcObliquityCorrection(float t) {
    float e0, omega;
    e0= calcMeanObliquityOfEcliptic(t);
    omega= 125.04f - 1934.136f * t;
    return e0 + 0.00256f * cosf(RADIANS(omega));
}

static float calcEccentricityEarthOrbit(float t) {
    return 0.016708634f - t * (0.000042037f + 0.0000001267f * t);
}

static float calcGeomMeanAnomalySun(float t) {
  return 357.52911f + t * (35999.05029f - 0.0001537f * t);
}

static float calcEquationOfTime(float t) {
    float epsilon, l0, e, m, y, sin2l0, sinm, cos2l0, sin4l0, sin2m, Etime;
    //
    epsilon=calcObliquityCorrection(t);
    l0=calcGeomMeanLongSun(t);
    e=calcEccentricityEarthOrbit(t);
    m=calcGeomMeanAnomalySun(t);
    y=tanf(RADIANS(epsilon)/2.0);
    y=y * y;
    sin2l0= sinf(2.0f * RADIANS(l0));
    sinm=sinf(RADIANS(m));
    cos2l0=cosf(2.0f * RADIANS(l0));
    sin4l0=sinf(4.0f * RADIANS(l0));
    sin2m=sinf(2.0f * RADIANS(m));
    Etime=y * sin2l0 - 2.0f * e * sinm + 4.0f * e * y * sinm * cos2l0- 0.5f * y * y * sin4l0 - 1.25f * e * e * sin2m;
    return DEGREES(Etime)*4.0f;	// in minutes of time
}

static float calcTimeJulianCent(float jd) {
    return ( jd - 2451545.0f)/36525.0f;
}

static float calcSunEqOfCenter(float t) {
    float mrad, sinm, sin2m, sin3m;
    //
    mrad= RADIANS(calcGeomMeanAnomalySun(t));
    sinm= sinf(mrad);
    sin2m= sinf(mrad+mrad);
    sin3m= sinf(mrad+mrad+mrad);
    return sinm * (1.914602f - t * (0.004817f + 0.000014f * t)) + sin2m * (0.019993f - 0.000101f * t) + sin3m * 0.000289f;
}

static float calcSunTrueLong(float t) {
    return calcGeomMeanLongSun(t) + calcSunEqOfCenter(t);
}

static float calcSunApparentLong(float t) {
    return calcSunTrueLong(t)-0.00569f-0.00478f*sinf(RADIANS(125.04f-1934.136f*t));
}

static float calcSunDeclination(float t) {
    float e, lambda, sint;
    //
    e= calcObliquityCorrection(t);
    lambda= calcSunApparentLong(t);
    sint= sinf(RADIANS(e)) * sinf(RADIANS(lambda));
    if(sint > 1.0 || sint < -1.0) {
        ArcError=true;
        return 0;
    }
    return DEGREES(asin(sint));
}

static float calcHourAngle(float lat,float solarDec) {
    float latRad, sdRad;
    latRad= RADIANS(lat);
    sdRad= RADIANS(solarDec);
    float val=cosf(RADIANS(90.833f))/(cosf(latRad)*cosf(sdRad))-tanf(latRad)*tanf(sdRad);
    if(val>1.0 || val<-1.0) {
        ArcError=true;
        return 0;
    }
    return (acos(val));
}

static float calcJD(WORD year,BYTE month,BYTE day) {
    int A,B;

    if(month <= 2) {
        year--;
        month+=12;
    }
    A = floorf(year/100.0f);
    B = 2 - A + floorf(A/4);
    return floor(365.25f*(year + 4716)) + floorf(30.6001f*(month+1))+day+B-1524.5f;
}

// static float calcJDFromJulianCent(float t) {
//     return t * 36525.0f + 2451545.0f;
// }

static float calcSunriseSetUTC(bool rise,float JD,float latitude,float longitude) {
    float t, eqTime, solarDec, hourAngle, delta, timeDiff, timeUTC;

    t = calcTimeJulianCent(JD);
    eqTime= calcEquationOfTime(t);
    solarDec= calcSunDeclination(t);
    if(ArcError)return 0;
    hourAngle = calcHourAngle(latitude, solarDec);
    if(ArcError)return 0;
    if(!rise)hourAngle=-hourAngle;
    delta= longitude + DEGREES(hourAngle);
    timeDiff= 4 * delta;	                // в минутах времени
    timeUTC= 720 - timeDiff - eqTime;       // в минутах
    return timeUTC;
}

//---------------------------------------------------------------------------

/**
 * @brief Получение значения времени восхода солнца в определённой географической точке в указанный день
 * 
 * @param latitude Широта географической точки
 * @param longitude Долгота географической точки
 * @param day День месяца (1-31)
 * @param month Номер месяца(1-12)
 * @param year Год (2022 ...)
 * @return unsigned int Количество секунд от начала суток или 0 при ошибке
 */
unsigned int TimeOfSunrise(float latitude,float longitude, uint8_t day, uint8_t month, uint16_t year) {
    unsigned int utc;
    ArcError=false;
    utc=roundf(calcSunriseSetUTC(true,calcJD(year, month, day),latitude,longitude)*60);
    if(ArcError)return 0;
    return utc;
}

/**
 * @brief Получение значения времени захода солнца в определённой географической точке в указанный день
 * 
 * @param latitude Широта географической точки
 * @param longitude Долгота географической точки
 * @param day День месяца (1-31)
 * @param month Номер месяца(1-12)
 * @param year Год (2022 ...)
 * @return unsigned int Количество секунд от начала суток или 0 при ошибке
 */
unsigned int  TimeOfSunset(float latitude,float longitude, uint8_t day, uint8_t month, uint16_t year) {
    unsigned int utc;
    ArcError=false;
    utc=roundf(calcSunriseSetUTC(false,calcJD(year, month, day),latitude,longitude)*60);
    if(ArcError)return 0;
    return utc;
}
