/**
 * @file weather.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Получение данных о фактической погоде.
 * Для получения информации о фактической погоде используются данные предоставляемые 
 * проектом "OpenWeather" ©2012—2024 OpenWeather ® All rights reserved.
 * Работоспособность сервиса и достоверность предоставляемых данных не гарантируется.
 * @version 0.1
 * @date 2024-05-26
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <stdio.h>
#include <cjson.h>
#include <http.h>

#include <weather.h>

#define MAX_URL_SIZE    64

/**
 * @brief Запрос данных о фактической погоде для указанного места.
 * Для указания места, для которого выполняется запрос погодных данных, используется числовой идентификатор.
 * Полный список числовых идентификаторов может быть получен по адресу http://bulk.openweathermap.org/sample.
 * При успешном выполнении функция заполняет поля структуры weather_data и возвращает значение 1.
 * При ошибке функция возвращает значение 0.
 * @param place_id Числовой идентификатор места.
 * @param weather_data Указатель на структуру данных.
 * @return int 1-успех,0-ошибка.
 */
int weather(int place_id, weather_data_t *weather_data) {
    if(!weather_data) return 0;
    //
    char url[MAX_URL_SIZE];
    http_response_t resp;
    //
    // запрос
    snprintf(url, MAX_URL_SIZE, "http://www.mnppsaturn.ru/weather/?city=%d", place_id);
    http_result_t res = HttpGET(url, NULL, &resp);
    if(res != HTTP_RES_OK) return 0;
    // проверка статуса
    if(resp.status_code != 200) {
        HttpFreeResponse(&resp);
        return 0;
    }
    //
    // разбор JSON
    cJSON *json = cJSON_Parse(resp.body);
    // ответ HTTP больше не нужен
    HttpFreeResponse(&resp);
    //
    if(!json) {
        // ошибка разбора JSON
        return 0;
    }
    if(cJSON_ByPath(json, "error")) {
        // запрос вернул ошибку
        cJSON_Delete(json);
        return 0;
    }
    //
    // заполняем структуру данных
    weather_data->updated =     cJSON_GetIntByPath(json, "upd", 0);
    weather_data->weather =     cJSON_GetIntByPath(json, "id", 0);
    weather_data->temperature = cJSON_GetFloatByPath(json, "temp", 0);
    weather_data->humidity =    cJSON_GetFloatByPath(json, "humidity", 0);
    weather_data->pressure =    cJSON_GetFloatByPath(json, "pressure", 0);
    weather_data->wind_speed =  cJSON_GetFloatByPath(json, "wspeed", 0);
    weather_data->wind_dir =    cJSON_GetFloatByPath(json, "wdir", 0);
    weather_data->cloud =       cJSON_GetFloatByPath(json, "cloud", 0);
    weather_data->visibility =  cJSON_GetFloatByPath(json, "vis", 0);
    //
    cJSON_Delete(json);
    //
    return 1;
}
