/**
 * @file weather_text.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Получение данных о фактической погоде.
 * Для получения информации о фактической погоде используются данные предоставляемые 
 * проектом "OpenWeather" ©2012—2024 OpenWeather ® All rights reserved.
 * Работоспособность сервиса и достоверность предоставляемых данных не гарантируется.
 * @version 0.1
 * @date 2024-05-26
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <stddef.h>

#include <weather.h>

static const char wid_2xx[] = "Гроза";
static const char wid_200[] = "Гроза с небольшим дождём";
static const char wid_201[] = "Гроза с дождём";
static const char wid_202[] = "Гроза с сильным дождём";
static const char wid_210[] = "Небольшая гроза";
static const char wid_212[] = "Сильная гроза";
static const char wid_221[] = "Очень сильная гроза";
static const char wid_230[] = "Гроза с моросящим дождём";

static const char wid_3xx[] = "Моросящий дождь";
static const char wid_300[] = "Слабый моросящий дождь";
static const char wid_302[] = "Сильный моросящий дождь";
static const char wid_313[] = "Сильный ливень";
static const char wid_321[] = "Мелкий дождь";

static const char wid_5xx[] = "Дождь";
static const char wid_500[] = "Небольшой дождь";
static const char wid_501[] = "Умеренный дождь";
static const char wid_502[] = "Сильный дождь";
static const char wid_503[] = "Очень сильный дождь";
static const char wid_511[] = "Ледяной дождь";
static const char wid_520[] = "Лёгкий дождь";
static const char wid_521[] = "Ливень";
static const char wid_522[] = "Сильный ливень";
static const char wid_531[] = "Очень сильный ливень";

static const char wid_6xx[] = "Снег";
static const char wid_600[] = "Небольшой снег";
static const char wid_602[] = "Сильный снег";
static const char wid_611[] = "Мокрый снег";
static const char wid_612[] = "Небольшой снег с дождём";
static const char wid_613[] = "Снег с дождём";
static const char wid_622[] = "Метель";

static const char wid_7xx[] = "Плохая видимость";
static const char wid_701[] = "Туман";
static const char wid_711[] = "Смог";
static const char wid_721[] = "Дымка";
static const char wid_731[] = "Песчаная буря";
static const char wid_751[] = "Ветер с песком";
static const char wid_761[] = "Пыль";
static const char wid_762[] = "Вулканический пепел";
static const char wid_771[] = "Шквал";
static const char wid_781[] = "Торнадо";

static const char wid_8xx[] = "Облачно";
static const char wid_800[] = "Ясно";
static const char wid_801[] = "Небольшая облачность";
static const char wid_802[] = "Умеренная облачность";
static const char wid_803[] = "Облачно с прояснениями";
static const char wid_804[] = "Сплошная облачность";

/**
 * @brief Возвращает указатель на строку описания погодных условий.
 * При некорректном значении кода возвращает значение NULL.
 * @param id Код описания погодных условий.
 * @return char* Указатель на строку или NULL.
 */
const char *weather_text(int id) {

    if((id >= 200) && (id < 300)) {
        switch (id) {
            case 200:   return wid_200;
            case 201:   return wid_201;
            case 202:   return wid_202;
            case 210:   return wid_210;
            case 212:   return wid_212;
            case 221:   return wid_221;
            case 230:
            case 231:
            case 232:   return wid_230;
            default:    return wid_2xx;
        }
    } else
    if((id >= 300) && (id < 400)) {
        switch(id) {
            case 300:
            case 310:   return wid_300;
            case 302:
            case 312:   return wid_302;
            case 313:
            case 314:   return wid_313;
            case 321:   return wid_321;
            default:    return wid_3xx;
        }
    } else
    if((id >= 500) && (id < 600)) {
        switch(id) {
            case 500:   return wid_500;
            case 501:   return wid_501;
            case 502:   return wid_502;
            case 503:
            case 504:   return wid_503;
            case 511:   return wid_511;
            case 520:   return wid_520;
            case 521:   return wid_521;
            case 522:   return wid_522;
            case 531:   return wid_531;
            default:    return wid_5xx;
        }
    } else
    if((id >= 600) && (id < 700)) {
        switch(id) {
            case 600:   return wid_600;
            case 602:   return wid_602;
            case 611:   return wid_611;
            case 612:
            case 615:
            case 620:   return wid_612;
            case 613:
            case 616:
            case 621:   return wid_613;
            case 622:   return wid_622;
            default:    return wid_6xx;
        }
    } else
    if((id >= 700) && (id < 800)) {
        switch(id) {
            case 701:
            case 741:   return wid_701;
            case 711:   return wid_711;
            case 721:   return wid_721;
            case 731:   return wid_731;
            case 751:   return wid_751;
            case 761:   return wid_761;
            case 762:   return wid_762;
            case 771:   return wid_771;
            case 781:   return wid_781;
            default:    return wid_7xx;
        }
    } else
    if((id >= 800) && (id < 900)) {
        switch(id) {
            case 800:   return wid_800;
            case 801:   return wid_801;
            case 802:   return wid_802;
            case 803:   return wid_803;
            case 804:   return wid_804;
            default:    return wid_8xx;
        }
    } else {
        // неопределённая погода
        return NULL;
    }
}
