/**
 * @file cjson_del.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с JSON.
 * @version 0.1
 * @date 2024-05-16
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 * Библиотека разработана на основе оригинальных исходных текстов библиотеки cJSON:
 * @copyright Copyright (c) 2009-2017 Dave Gamble and cJSON contributors
 * @copyright Copyright (c) 2015-2016 The Khronos Group Inc.
 * @copyright Copyright (c) 2015-2016 Valve Corporation
 * @copyright Copyright (c) 2015-2016 LunarG, Inc.
 * @license MIT
 */

#include <stdlib.h>
#include <string.h>

#include <cjson.h>

/**
 * @brief Извлечение существующего элемента массива.
 * Функция извлекает из массива элемент с индексом which. Элементы с большими индексами сдвигаются "влево".
 * Функция не освобождает память занимаемую извлекаемым элементом.
 * Функция возвращает указатель на извлечённый элемент.
 * Если массив не содержит элемент с указанным идексом, то функция возвращает NULL.
 * @param array Указатель на структуру cJSON (массив).
 * @param which Индекс элемента массива.
 * @return cJSON* 
 */
cJSON *cJSON_DetachItemFromArray(cJSON *array, int which) {
    cJSON *c = array->child;
    while (c && which > 0) c = c->next, which--;
    if (!c) return 0;
    if (c->prev) c->prev->next = c->next;
    if (c->next) c->next->prev = c->prev;
    if (c == array->child) array->child = c->next;
    c->prev = c->next = 0;
    return c;
}

/**
 * @brief Удаление существующего элемента массива.
 * Функция удаляет из массива элемент с индексом which. Элементы с большими индексами сдвигаются "влево".
 * Функция освобождает память занимаемую удаляемым элементом.
 * Если массив не содержит элемент с указанным идексом, то функция ничего не дедает.
 * @param array Указатель на структуру cJSON (массив).
 * @param which Индекс элемента массива.
 */
void cJSON_DeleteItemFromArray(cJSON *array, int which) {
    cJSON_Delete(cJSON_DetachItemFromArray(array, which));
}

/**
 * @brief Извлечение существующего элемента объекта.
 * Функция извлекает из объекта элемент с именем name.
 * Функция не освобождает память занимаемую извлекаемым элементом.
 * Функция возвращает указатель на извлечённый элемент.
 * Если объект не содержит элемент с указанным именем, то функция возвращает NULL.
 * @param array Указатель на структуру cJSON (массив).
 * @param name Имя извлекаемого элемента
 * @return cJSON* 
 */
cJSON *cJSON_DetachItemFromObject(cJSON *object, const char *name) {
    int i = 0;
    cJSON *c = object->child;
    while (c && strcmp(c->string, name)) i++, c = c->next;
    if (c) return cJSON_DetachItemFromArray(object, i);
    return 0;
}

/**
 * @brief Удаление существующего элемента объекта.
 * Функция удаляет из объекта элемент с именем name и освобождает занимаемую им память.
 * Если объект не содержит элемент с указанным именем, то функция ничего не делает.
 * @param array Указатель на структуру cJSON (массив).
 * @param name Имя извлекаемого элемента
 */
void cJSON_DeleteItemFromObject(cJSON *object, const char *name) {
    cJSON_Delete(cJSON_DetachItemFromObject(object, name));
}
