/**
 * @file cjson_ins.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с JSON.
 * @version 0.1
 * @date 2024-05-16
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 * Библиотека разработана на основе оригинальных исходных текстов библиотеки cJSON:
 * @copyright Copyright (c) 2009-2017 Dave Gamble and cJSON contributors
 * @copyright Copyright (c) 2015-2016 The Khronos Group Inc.
 * @copyright Copyright (c) 2015-2016 Valve Corporation
 * @copyright Copyright (c) 2015-2016 LunarG, Inc.
 * @license MIT
 */

#include <string.h>

#include <cjson.h>

/**
 * @brief Вставка нового элемента в массив JSON.
 * Функция вставляет новый элемент в массив в позицию, указанную в параметре which. Уже имеющиеся
 * элементы массива расположенные после вставляемого сдвигаются вправо.
 * @param array Указатель на структуру cJSON (массив).
 * @param which Позиция вствки нового элемента.
 * @param newitem Вставляемый элемент массива.
 */
void cJSON_InsertItemInArray(cJSON *array, int which, cJSON *newitem) {
    cJSON *c = array->child;
    while (c && which > 0) c = c->next, which--;
    if (!c) {
        cJSON_AddItemToArray(array, newitem);
        return;
    }
    newitem->next = c;
    newitem->prev = c->prev;
    c->prev = newitem;
    if (c == array->child)
        array->child = newitem;
    else
        newitem->prev->next = newitem;
}

/**
 * @brief Замена существующего элемента в массиве JSON.
 * Функция выполняет замену существующего элемента массива в позиции which на новый элемент newitem.
 * Если массив не содержит элемента в позиции which, то функция ничего не делает.
 * Память занимаемая заменяемым элеметом освобождается.
 * @param array Указатель на структуру cJSON (массив).
 * @param which Позиция заменяемого элемента.
 * @param newitem Новый элемент массива.
 */
void cJSON_ReplaceItemInArray(cJSON *array, int which, cJSON *newitem) {
    cJSON *c = array->child;
    while (c && which > 0) c = c->next, which--;
    if (!c) return;
    newitem->next = c->next;
    newitem->prev = c->prev;
    if (newitem->next) newitem->next->prev = newitem;
    if (c == array->child)
        array->child = newitem;
    else
        newitem->prev->next = newitem;
    c->next = c->prev = 0;
    cJSON_Delete(c);
}

/**
 * @brief Замена существующего элемента в объекте JSON.
 * Функция выполняет замену существующего элемента объекта с именем name на новый элемент newitem.
 * Если объект не содержит элемента с именем name, то функция ничего не делает.
 * Память занимаемая заменяемым элеметом освобождается.
 * @param object Указатель на структуру cJSON (объект).
 * @param name Имя заменяемого элемента.
 * @param newitem Новый элемент массива.
 */
void cJSON_ReplaceItemInObject(cJSON *object, const char *name, cJSON *newitem) {
    int i = 0;
    cJSON *c = object->child;
    while (c && strcmp(c->string, name)) i++, c = c->next;
    if (c) {
        newitem->string = strdup(name);
        cJSON_ReplaceItemInArray(object, i, newitem);
    }
}
