/**
 * @file cjson_min.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с JSON.
 * @version 0.1
 * @date 2024-05-16
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 * Библиотека разработана на основе оригинальных исходных текстов библиотеки cJSON:
 * @copyright Copyright (c) 2009-2017 Dave Gamble and cJSON contributors
 * @copyright Copyright (c) 2015-2016 The Khronos Group Inc.
 * @copyright Copyright (c) 2015-2016 Valve Corporation
 * @copyright Copyright (c) 2015-2016 LunarG, Inc.
 * @license MIT
 */

#include <stdlib.h>
#include <string.h>

#include <cjson.h>

#define ELEM_TYPE(item)     ((item->type) & 255)

static cJSON *find_by_name(cJSON *item, char *name) {
    // TODO:
    
    int type = ELEM_TYPE(item);
    //
    if(type == cJSON_Array) {
        return cJSON_GetArrayItem(item, atoi(name));
    } else
    if(type == cJSON_Object) {
        return cJSON_GetObjectItem(item, name);
    } else return NULL;
}

/**
 * @brief Поиск элемента JSON по текcтовому пути.
 * Путь представляет собой строку, вида: "xxx/yyy/nnn", где xxx,yyy,nnn наименования элементов объекта или индексы элементов массива.
 * Функция возвращает указатель на элемент по указанному пути. Если элемент отсутствует, то функция возвращает NULL.
 * @param item Элемент с которого начинается поиск.
 * @param path Путь.
 * @return cJSON* 
 */
cJSON *cJSON_ByPath(cJSON *item, char *path) {
    cJSON *jv;
    //
    if(!path) return item;
    int path_len = strlen(path);
    if(path_len == 0) return item;
    //
    char *delim_pos = strchr(path, '/');
    if(delim_pos) {
        // в пути есть ещё разделитель
        int name_len = delim_pos - path;
        char *name = alloca(name_len + 1);
        memcpy(name, path, name_len);
        name[name_len] = 0;
        //
        jv = find_by_name(item, name);
    } else {
        jv = find_by_name(item, path);
    }
    if(jv) {
        if(delim_pos) {
            // если у пути ещё есть остаток, то выполняем рекурсивный поиск по остатку
            return cJSON_ByPath(jv, delim_pos + 1);
        } else {
            // если остатка нет, то возвращаем найденный узел
            return jv;
        }
    } else {
        // узел не найден
        return NULL;
    }
}

/**
 * @brief Получение значения типа "int" для элемента JSON путь к которому указан в параметре path.
 * Если указанный элемент присутствует и его тип соответствует требуемому, то функция возвращает его значение.
 * В противном случае функция возвращает значение defval.
 * @param item Указатель на структуру cJSON.
 * @param path Путь к элементу JSON.
 * @param defval Значение по умолчанию.
 * @return int 
 */
int cJSON_GetIntByPath(cJSON *item, char *path, int defval) {
    cJSON *js = cJSON_ByPath(item, path);
    if(!js) return defval;
    //
    if(ELEM_TYPE(js) == cJSON_Number) return js->valueint; else return defval;
}

/**
 * @brief Получение значения типа "bool" для элемента JSON путь к которому указан в параметре path.
 * Если указанный элемент присутствует и его тип соответствует требуемому, то функция возвращает его значение.
 * В противном случае функция возвращает значение defval.
 * @param item Указатель на структуру cJSON.
 * @param path Путь к элементу JSON.
 * @param defval Значение по умолчанию.
 * @return int 
 */
int cJSON_GetBoolByPath(cJSON *item, char *path, int defval) {
    cJSON *js = cJSON_ByPath(item, path);
    if(!js) return defval;
    //
    if(ELEM_TYPE(js) == cJSON_True) return 1;
    else
    if(ELEM_TYPE(js) == cJSON_False) return 0;
    else
    return defval;
}

/**
 * @brief Получение значения типа "double" для элемента JSON путь к которому указан в параметре path.
 * Если указанный элемент присутствует и его тип соответствует требуемому, то функция возвращает его значение.
 * В противном случае функция возвращает значение defval.
 * @param item Указатель на структуру cJSON.
 * @param path Путь к элементу JSON.
 * @param defval Значение по умолчанию.
 * @return int 
 */
double cJSON_GetDoubleByPath(cJSON *item, char *path, double defval) {
    cJSON *js = cJSON_ByPath(item, path);
    if(!js) return defval;
    //
    if(ELEM_TYPE(js) == cJSON_Number) return js->valuedouble; else return defval;
}

/**
 * @brief Получение значения типа "float" для элемента JSON путь к которому указан в параметре path.
 * Если указанный элемент присутствует и его тип соответствует требуемому, то функция возвращает его значение.
 * В противном случае функция возвращает значение defval.
 * Если параметр path пустая строка или NULL, то функция возвращает значение элемента item.
 * @param item Указатель на структуру cJSON.
 * @param path Путь к элементу JSON или NULL.
 * @param defval Значение по умолчанию.
 * @return int 
 */
float cJSON_GetFloatByPath(cJSON *item, char *path, float defval) {
    cJSON *js = cJSON_ByPath(item, path);
    if(!js) return defval;
    //
    if(ELEM_TYPE(js) == cJSON_Number) return js->valuedouble; else return defval;
}

/**
 * @brief Получение значения типа "string" для элемента JSON путь к которому указан в параметре path.
 * Если указанный элемент присутствует и его тип соответствует требуемому, то функция возвращает его значение.
 * В противном случае функция возвращает значение defval.
 * @param item Указатель на структуру cJSON.
 * @param path Путь к элементу JSON.
 * @param defval Значение по умолчанию.
 * @return char*
 */
char *cJSON_GetStringByPath(cJSON *item, char *path, char *defval) {
    cJSON *js = cJSON_ByPath(item, path);
    if(!js) return defval;
    //
    if(ELEM_TYPE(js) == cJSON_String) return js->valuestring; else return defval;
}
