/**
 * @file crc8.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции вычисления контрольной суммы.
 * @version 0.1
 * @date 2024-03-30
 * 
 * @copyright Copyright (c) 1999-2018 Lammert Bies
 * https://github.com/lammertb/libcrc
 * 
 * This file is licensed under the MIT License as stated below
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 */

#include <stdio.h>
#include <satcrc.h>

static uint8_t sht75_crc_table[] = {
    0,   49,  98,  83,  196, 245, 166, 151, 185, 136, 219, 234, 125, 76,  31,  46,
    67,  114, 33,  16,  135, 182, 229, 212, 250, 203, 152, 169, 62,  15,  92,  109,
    134, 183, 228, 213, 66,  115, 32,  17,  63,  14,  93,  108, 251, 202, 153, 168,
    197, 244, 167, 150, 1,   48,  99,  82,  124, 77,  30,  47,  184, 137, 218, 235,
    61,  12,  95,  110, 249, 200, 155, 170, 132, 181, 230, 215, 64,  113, 34,  19,
    126, 79,  28,  45,  186, 139, 216, 233, 199, 246, 165, 148, 3,   50,  97,  80,
    187, 138, 217, 232, 127, 78,  29,  44,  2,   51,  96,  81,  198, 247, 164, 149,
    248, 201, 154, 171, 60,  13,  94,  111, 65,  112, 35,  18,  133, 180, 231, 214,
    122, 75,  24,  41,  190, 143, 220, 237, 195, 242, 161, 144, 7,   54,  101, 84,
    57,  8,   91,  106, 253, 204, 159, 174, 128, 177, 226, 211, 68,  117, 38,  23,
    252, 205, 158, 175, 56,  9,   90,  107, 69,  116, 39,  22,  129, 176, 227, 210,
    191, 142, 221, 236, 123, 74,  25,  40,  6,   55,  100, 85,  194, 243, 160, 145,
    71,  118, 37,  20,  131, 178, 225, 208, 254, 207, 156, 173, 58,  11,  88,  105,
    4,   53,  102, 87,  192, 241, 162, 147, 189, 140, 223, 238, 121, 72,  27,  42,
    193, 240, 163, 146, 5,   52,  103, 86,  120, 73,  26,  43,  188, 141, 222, 239,
    130, 179, 224, 209, 70,  119, 36,  21,  59,  10,  89,  104, 255, 206, 157, 172
};

/**
 * @brief Расчёт значения контрольной суммы crc8
 * 
 * @param input_str Указатель на массив данных
 * @param num_bytes Длина массива
 * @return uint8_t 
 */
uint8_t crc_8( const unsigned char *input_str, size_t num_bytes ) {

    size_t a;
    uint8_t crc;
    const unsigned char *ptr;

    crc = CRC_START_8;
    ptr = input_str;

    if ( ptr != NULL ) for (a=0; a<num_bytes; a++) {

        crc = sht75_crc_table[(*ptr++) ^ crc];
    }

    return crc;

}  /* crc_8 */

/**
 * @brief Побайтное вычисление контрольной суммы по алгоритму crc8.
 * 
 * @param crc Текущее значение контрольной суммы
 * @param c Новый байт
 * @return uint8_t Новое значение контрольной суммы
 */
uint8_t update_crc_8( unsigned char crc, unsigned char val ) {

    return sht75_crc_table[val ^ crc];

}  /* update_crc_8 */

/**
 * @brief Вычисление контрольной суммы, используемой в протоколе NMEA.
 * Процедура не пытается проверить саму строку.
 * Начальный символ '$' будет проигнорирован, поскольку этот символ является 
 * частью предложения NMEA, но не частью вычисления контрольной суммы.
 * Вычисление останавливается, когда обнаружен CR, LF, символ '*' или конец строки.
 * Поскольку проверка синтаксиса NMEA не задействована, функция всегда возвращается с 
 * успехом, если в качестве параметра не указан нулевой указатель. 
 * Возвращаемое значение - это указатель на буфер результатов, предоставленный вызывающим 
 * приложением, или NULL в случае ошибки.
 * @param input_str Проверяемая строка
 * @param result Результат вычисления
 * @return unsigned char* 
 */
unsigned char * checksum_NMEA( const unsigned char *input_str, unsigned char *result ) {

    const unsigned char *ptr;
    unsigned char checksum;

    if ( input_str == NULL ) return NULL;
    if ( result    == NULL ) return NULL;

    checksum = 0;
    ptr      = (const unsigned char *) input_str;

    if ( *ptr == '$' ) ptr++;

    while ( *ptr  &&  *ptr != '\r'  &&  *ptr != '\n'  &&  *ptr != '*' ) checksum ^= *ptr++;

    snprintf( (char *) result, 3, "%02hhX", checksum );

    return result; 

}  /* checksum_NMEA */
