/**
 * @file base64.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Кодирование/декодирование по алгоритму BASE64.
 * @version 0.1
 * @date 2024-04-25
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <stdint.h>
#include <stddef.h>

#include <base64.h>

#define NEWLINE_INVL 76

#define BYTE uint8_t

// Note: To change the charset to a URL encoding, replace the '+' and '/' with '*' and '-'
static const BYTE charset[]={"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"};

static BYTE revchar(char ch) {
    if (ch >= 'A' && ch <= 'Z')
        ch -= 'A';
    else if (ch >= 'a' && ch <='z')
        ch = ch - 'a' + 26;
    else if (ch >= '0' && ch <='9')
        ch = ch - '0' + 52;
    else if (ch == '+')
        ch = 62;
    else if (ch == '/')
        ch = 63;

    return(ch);
}

/**
 * @brief Кодирование BASE64
 * Возвращает размер выходных данных. Если вызывается c out = NULL, то просто возвращается размер,
 * который был бы у выходных данных (без завершающего 0).
 * @param in Входные данные.
 * @param out Выходные данные.
 * @param len Размер входных данных.
 * @param newline_flag 
 * @return size_t 
 */
size_t base64_encode(const BYTE in[], BYTE out[], size_t len, int newline_flag) {
    size_t idx, idx2, blks, blk_ceiling, left_over, newline_count = 0;
    //
    blks = (len / 3);
    left_over = len % 3;
    //
    if(!out) {
        idx2 = blks * 4;
        if(left_over) idx2 += 4;
        if(newline_flag) idx2 += len / 57;   // (NEWLINE_INVL / 4) * 3 = 57. One newline per 57 input bytes.
        //
        return idx2;
    }
    // Since 3 input bytes = 4 output bytes, determine out how many even sets of
    // 3 bytes the input has.
    blk_ceiling = blks * 3;
    for (idx = 0, idx2 = 0; idx < blk_ceiling; idx += 3, idx2 += 4) {
        out[idx2]     = charset[in[idx] >> 2];
        out[idx2 + 1] = charset[((in[idx] & 0x03) << 4) | (in[idx + 1] >> 4)];
        out[idx2 + 2] = charset[((in[idx + 1] & 0x0f) << 2) | (in[idx + 2] >> 6)];
        out[idx2 + 3] = charset[in[idx + 2] & 0x3F];
        // The offical standard requires a newline every 76 characters.
        // (Eg, first newline is character 77 of the output.)
        if (((idx2 - newline_count + 4) % NEWLINE_INVL == 0) && newline_flag) {
            out[idx2 + 4] = '\n';
            idx2++;
            newline_count++;
        }
    }
    if (left_over == 1) {
        out[idx2]     = charset[in[idx] >> 2];
        out[idx2 + 1] = charset[(in[idx] & 0x03) << 4];
        out[idx2 + 2] = '=';
        out[idx2 + 3] = '=';
        idx2 += 4;
    }
    else
    if (left_over == 2) {
        out[idx2]     = charset[in[idx] >> 2];
        out[idx2 + 1] = charset[((in[idx] & 0x03) << 4) | (in[idx + 1] >> 4)];
        out[idx2 + 2] = charset[(in[idx + 1] & 0x0F) << 2];
        out[idx2 + 3] = '=';
        idx2 += 4;
    }
    //
    return(idx2);
}

/**
 * @brief Декодирование BASE64
 * Возвращает размер выходных данных. Если вызывается c out = NULL, то просто возвращается размер,
 * который был бы у выходных данных (без завершающего 0).
 * @param in Входные данные.
 * @param out Выходные данные.
 * @param len Размер входных данных.
 * @return size_t 
 */
size_t base64_decode(const BYTE in[], BYTE out[], size_t len) {
    size_t idx, idx2, blks, blk_ceiling, left_over;

    if (in[len - 1] == '=')
        len--;
    if (in[len - 1] == '=')
        len--;

    blks = len / 4;
    left_over = len % 4;

    if (out == NULL) {
        if (len >= 77 && in[NEWLINE_INVL] == '\n')   // Verify that newlines where used.
            len -= len / (NEWLINE_INVL + 1);
        blks = len / 4;
        left_over = len % 4;

        idx = blks * 3;
        if (left_over == 2)
            idx ++;
        else if (left_over == 3)
            idx += 2;
    }
    else {
        blk_ceiling = blks * 4;
        for (idx = 0, idx2 = 0; idx2 < blk_ceiling; idx += 3, idx2 += 4) {
            if (in[idx2] == '\n')
                idx2++;
            out[idx]     = (revchar(in[idx2]) << 2) | ((revchar(in[idx2 + 1]) & 0x30) >> 4);
            out[idx + 1] = (revchar(in[idx2 + 1]) << 4) | (revchar(in[idx2 + 2]) >> 2);
            out[idx + 2] = (revchar(in[idx2 + 2]) << 6) | revchar(in[idx2 + 3]);
        }

        if (left_over == 2) {
            out[idx]     = (revchar(in[idx2]) << 2) | ((revchar(in[idx2 + 1]) & 0x30) >> 4);
            idx++;
        }
        else if (left_over == 3) {
            out[idx]     = (revchar(in[idx2]) << 2) | ((revchar(in[idx2 + 1]) & 0x30) >> 4);
            out[idx + 1] = (revchar(in[idx2 + 1]) << 4) | (revchar(in[idx2 + 2]) >> 2);
            idx += 2;
        }
    }

    return(idx);
}

#ifdef TEST_BASE64

#include <string.h>

int test_base64() {
    uint8_t text[3][1024] = {{"fo"},
                        {"foobar"},
                        {"Man is distinguished, not only by his reason, but by this singular passion from other animals, which is a lust of the mind, that by a perseverance of delight in the continued and indefatigable generation of knowledge, exceeds the short vehemence of any carnal pleasure."}};
    uint8_t code[3][1024] = {{"Zm8="},
                        {"Zm9vYmFy"},
                        {"TWFuIGlzIGRpc3Rpbmd1aXNoZWQsIG5vdCBvbmx5IGJ5IGhpcyByZWFzb24sIGJ1dCBieSB0aGlz\nIHNpbmd1bGFyIHBhc3Npb24gZnJvbSBvdGhlciBhbmltYWxzLCB3aGljaCBpcyBhIGx1c3Qgb2Yg\ndGhlIG1pbmQsIHRoYXQgYnkgYSBwZXJzZXZlcmFuY2Ugb2YgZGVsaWdodCBpbiB0aGUgY29udGlu\ndWVkIGFuZCBpbmRlZmF0aWdhYmxlIGdlbmVyYXRpb24gb2Yga25vd2xlZGdlLCBleGNlZWRzIHRo\nZSBzaG9ydCB2ZWhlbWVuY2Ugb2YgYW55IGNhcm5hbCBwbGVhc3VyZS4="}};
    uint8_t buf[1024];
    size_t buf_len;
    int pass = 1;
    int idx;

    for (idx = 0; idx < 3; idx++) {
        buf_len = base64_encode(text[idx], buf, strlen((char *)text[idx]), 1);
        pass = pass && ((buf_len == strlen((char *)code[idx])) && (buf_len == base64_encode(text[idx], NULL, strlen((char *)text[idx]), 1)));
        pass = pass && !strcmp((char *)code[idx], (char *)buf);

        memset(buf, 0, sizeof(buf));
        buf_len = base64_decode(code[idx], buf, strlen((char *)code[idx]));
        pass = pass && ((buf_len == strlen((char *)text[idx])) && (buf_len == base64_decode(code[idx], NULL, strlen((char *)code[idx]))));
        pass = pass && !strcmp((char *)text[idx], (char *)buf);
    }

    return pass;
}

#endif
