/**
 * @file tls_common.h
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * TLS: Структуры данных и константы.
 * @version 0.1
 * @date 2024-04-25
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#ifndef TLS_COMMON_H
#define TLS_COMMON_H

#include <cdefs.h>
#include <stdint.h>
#include <stddef.h>

__BEGIN_DECLS

// TLS 1.3: Supported SSL Ciphers:
//  TLS_AES_256_GCM_SHA384              {0x13, 0x02}
//  TLS_CHACHA20_POLY1305_SHA256        {0x13, 0x03}
//  TLS_AES_128_GCM_SHA256              {0x13, 0x01}
//  TLS_AES_128_CCM_8_SHA256            {0x13, 0x05} +
//  TLS_AES_128_CCM_SHA256              {0x13, 0x04} +


struct ProtocolVersion {
    uint8_t major;
    uint8_t minor;
} __attribute__((packed));

#define PROTOCOL_VERSION_TLS_1_0                {3,1}
#define PROTOCOL_VERSION_TLS_1_1                {3,2}
#define PROTOCOL_VERSION_TLS_1_2                {3,3}
#define PROTOCOL_VERSION_TLS_1_3                {3,4}

struct Random {
    uint32_t    gmt_unix_time;
    uint8_t     random_bytes[28];
} __attribute__((packed));

typedef uint16_t    ciphersuite_t;

// TLS_NULL_WITH_NULL_NULL is specified and is the initial state of a TLS connection during the first handshake on that channel, but MUST
// NOT be negotiated, as it provides no more protection than an unsecured connection.
#define TLS_NULL_WITH_NULL_NULL                 {0x00, 0x00}

// The following CipherSuite definitions require that the server provide an RSA certificate that can be used for key exchange. The server may
// request any signature-capable certificate in the certificate request message.
//
// Следующие определения CipherSuite требуют, чтобы сервер предоставил сертификат RSA, который может использоваться для обмена ключами.
// Сервер может запросить любой сертификат, поддерживающий подпись, в сообщении с запросом сертификата.
#define TLS_RSA_WITH_NULL_MD5                   {0x00, 0x01}
#define TLS_RSA_WITH_NULL_SHA                   {0x00, 0x02}
#define TLS_RSA_WITH_NULL_SHA256                {0x00, 0x3B}
#define TLS_RSA_WITH_RC4_128_MD5                {0x00, 0x04}
#define TLS_RSA_WITH_RC4_128_SHA                {0x00, 0x05}
#define TLS_RSA_WITH_3DES_EDE_CBC_SHA           {0x00, 0x0A}
#define TLS_RSA_WITH_AES_128_CBC_SHA            {0x00, 0x2F}
#define TLS_RSA_WITH_AES_256_CBC_SHA            {0x00, 0x35}
#define TLS_RSA_WITH_AES_128_CBC_SHA256         {0x00, 0x3C}
#define TLS_RSA_WITH_AES_256_CBC_SHA256         {0x00, 0x3D}

// The following cipher suite definitions are used for serverauthenticated (and optionally client-authenticated) Diffie-Hellman.
// DH denotes cipher suites in which the server’s certificate contains the Diffie-Hellman parameters signed by the certificate authority (CA).
// DHE denotes ephemeral Diffie-Hellman, where the Diffie-Hellman parameters are signed by a signature-capable certificate, which has
// been signed by the CA. The signing algorithm used by the server is specified after the DHE component of the CipherSuite name.
// The server can request any signature-capable certificate from the client for client authentication, or it may request a Diffie-Hellman
// certificate. Any Diffie-Hellman certificate provided by the client must use the parameters (group and generator) described by the server.
//
// Следующие определения набора шифров используются для проверки подлинности сервера (и, возможно, клиента) по методу Диффи-Хеллмана.
// DH обозначает наборы шифров, в которых сертификат сервера содержит параметры Диффи-Хеллмана, подписанные центром сертификации (CA).
// Обозначает эфемерный сертификат Диффи-Хеллмана, где параметры Диффи-Хеллмана подписаны сертификатом с поддержкой подписи, который был
// подписан центром сертификации. Алгоритм подписи, используемый сервером, указан после компонента имени набора шифров. 
// Сервер может запросить у клиента любой сертификат с поддержкой подписи для проверки подлинности клиента или сертификат Диффи-Хеллмана.
// Любой сертификат Диффи-Хеллмана, предоставляемый клиентом, должен использовать параметры (группу и генератор), описанные сервером.
#define TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA        {0x00, 0x0D}
#define TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA        {0x00, 0x10}
#define TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA       {0x00, 0x13}
#define TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA       {0x00, 0x16}
#define TLS_DH_DSS_WITH_AES_128_CBC_SHA         {0x00, 0x30}
#define TLS_DH_RSA_WITH_AES_128_CBC_SHA         {0x00, 0x31}
#define TLS_DHE_DSS_WITH_AES_128_CBC_SHA        {0x00, 0x32}
#define TLS_DHE_RSA_WITH_AES_128_CBC_SHA        {0x00, 0x33}
#define TLS_DH_DSS_WITH_AES_256_CBC_SHA         {0x00, 0x36}
#define TLS_DH_RSA_WITH_AES_256_CBC_SHA         {0x00, 0x37}
#define TLS_DHE_DSS_WITH_AES_256_CBC_SHA        {0x00, 0x38}
#define TLS_DHE_RSA_WITH_AES_256_CBC_SHA        {0x00, 0x39}
#define TLS_DH_DSS_WITH_AES_128_CBC_SHA256      {0x00, 0x3E}
#define TLS_DH_RSA_WITH_AES_128_CBC_SHA256      {0x00, 0x3F}
#define TLS_DHE_DSS_WITH_AES_128_CBC_SHA256     {0x00, 0x40}
#define TLS_DHE_RSA_WITH_AES_128_CBC_SHA256     {0x00, 0x67}
#define TLS_DH_DSS_WITH_AES_256_CBC_SHA256      {0x00, 0x68}
#define TLS_DH_RSA_WITH_AES_256_CBC_SHA256      {0x00, 0x69}
#define TLS_DHE_DSS_WITH_AES_256_CBC_SHA256     {0x00, 0x6A}
#define TLS_DHE_RSA_WITH_AES_256_CBC_SHA256     {0x00, 0x6B}

// The following cipher suites are used for completely anonymous Diffie-Hellman communications in which neither party is authenticated.
// Note that this mode is vulnerable to man-in-themiddle attacks. Using this mode therefore is of limited use: These cipher suites MUST NOT 
// be used by TLS 1.2 implementations unless the application layer has specifically requested to allow anonymous key exchange.
// (Anonymous key exchange may sometimes be acceptable, for example, to support opportunistic encryption when no set-up for authentication is in place,
// or when TLS is used as part of more complex security protocols that have other means to ensure authentication)
//
// Следующие наборы шифров используются для полностью анонимной связи по методу Диффи-Хеллмана, при которой ни одна из сторон не проходит проверку подлинности.
// Обратите внимание, что этот режим уязвим для атак типа "человек посередине". Таким образом, использование этого режима имеет ограниченное применение:
// эти наборы шифров НЕ должны использоваться реализациями TLS 1.2, если только прикладной уровень специально не запросил разрешение использования анонимного ключа обмен.
// (Анонимный обмен ключами иногда может быть приемлемым, например, для поддержки случайного шифрования, когда нет настройки для аутентификации, 
// или когда TLS используется как часть более сложных протоколов безопасности, которые имеют другие средства обеспечения аутентификации.)
//
// Обратите внимание, что использование неанонимного обмена ключами без фактической проверки обмена ключами по сути эквивалентно анонимному обмену ключами, 
// и применяются те же меры предосторожности.
// Хотя неанонимный обмен ключами, как правило, требует более высоких вычислительных и коммуникационных затрат, чем анонимный обмен ключами, 
// в интересах совместимости может быть полезно не отключать неанонимный обмен ключами, когда прикладной уровень разрешает анонимный обмен ключами.
#define TLS_DH_anon_WITH_RC4_128_MD5            {0x00, 0x18}
#define TLS_DH_anon_WITH_3DES_EDE_CBC_SHA       {0x00, 0x1B}
#define TLS_DH_anon_WITH_AES_128_CBC_SHA        {0x00, 0x34}
#define TLS_DH_anon_WITH_AES_256_CBC_SHA        {0x00, 0x3A}
#define TLS_DH_anon_WITH_AES_128_CBC_SHA256     {0x00, 0x6C}
#define TLS_DH_anon_WITH_AES_256_CBC_SHA256     {0x00, 0x6D}


#define CIPHER_SUITES_COUNT                     86

struct ClientHello {
    uint8_t                 content_type;                       // Handshake:       22              1
    struct ProtocolVersion  content_version;                    // TLS 1.0:         0x0301          2
    uint16_t                content_length;                     // 302                              2
    //
    uint8_t                 handshake_type;                     // Client Hello:    1               1
    uint8_t                 length[3];                          // 298: 0x00012A                    3
    struct ProtocolVersion  version;                            // TLS 1.2:         0x0303          2
    struct Random           random;                             //                                  32
    uint8_t                 session_id_length;                  // 0                                1
    uint16_t                cipher_suites_length;               // 172=CIPHER_SUITES_COUNT*2        2
    ciphersuite_t           cipher_suites[CIPHER_SUITES_COUNT]; //                                  172
    uint8_t                 compression_methods_length;         // 1                                1
    uint8_t                 compression_method;                 // 0                                1
    uint16_t                extensions_length;                  // 85                               1
                                                                //                                  85

} __attribute__((packed));

__END_DECLS

#endif
