/**
 * @file httpreq.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции клиента HTTP.
 * @version 0.1
 * @date 2024-05-19
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <satsock.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <base64.h>
#include <stdio.h>
#include <time.h>

#include <http.h>

// максимальный размер заголовков запроса
#define MAX_REQ_HEADERS_LEN 1024

unsigned int http_timeout          = 4000;
unsigned int http_max_response_len = 16384;

static http_result_t http_req(char *http_headers, url_t *purl, http_response_t *hresp) {
    int sock;
    int tmpres;
    //
    hresp->response = NULL;
    hresp->status_code = 0;
    //
    // создаём сокет
    sock = CreateSocket(SOCK_STREAM);
    if(sock == 0) {
        return HTTP_RES_BAD_SOCKET;
    }
    // подключаемся
    if(!SocketConnectBlock(sock, purl->host, purl->port, http_timeout)) {
        CloseSocket(&sock);
        return HTTP_RES_NO_CONNECTION;
    }
    // посылаем заголовки
    int sent = 0;
    int http_headers_len = strlen(http_headers);
    while(sent < http_headers_len) {
        tmpres = SocketWrite(sock, http_headers+sent, http_headers_len-sent);
        if(tmpres == -1) {
            CloseSocket(&sock);
            return HTTP_RES_BAD_SOCKET;
        }
        sent += tmpres;
    }
    //
    // приём ответа
    char *response = NULL;
    int response_len = 0;
    char *buf = malloc(BUFSIZ);
    unsigned int timer = GetTickCount();
    while(1) {
        int rec_size = SocketRead(sock, buf, BUFSIZ);
        // -1 - некоррекнтый сокет
        // -2 - соединение оборвано
        if(rec_size < 0) break;
        //
        if(rec_size > 0) {
            // проверка наличия "HTTP" в начале ответа
            if((response_len == 0) && (rec_size >= 4)) {
                if(memcmp(buf, "HTTP", 4)) {
                    // пришла какая-то чушь
                    free(buf);
                    CloseSocket(&sock);
                    return HTTP_RES_BAD_RESP_PROTO;
                }
            }
            // проверка на переполнение
            if(response_len + rec_size > http_max_response_len) {
                // превышение максимального размера ответа
                free(buf);
                free(response);
                CloseSocket(&sock);
                return HTTP_RES_RESP_TOO_LARGE;
            }
            // +1 - это для завершающего нуля
            response = realloc(response, response_len + rec_size + 1);
            memmove(&response[response_len], buf, rec_size);
            response_len += rec_size;
        } else {
            // время чтения не истекло?
            if((GetTickCount() - timer) >= http_timeout) break;
            // ничего не принято, надо подождать...
            msleep(50);
        }
    }
    free(buf);
    CloseSocket(&sock);
    //
    if(response_len == 0) {
        return HTTP_RES_EMPTY_RESP;             // ничего не принято
    }
    response[response_len] = 0;
    //
    // инициализация hresp
    hresp->response = response;
    hresp->response_len = response_len;
    //
    // HTTP/1.0 200 OK
    // HTTP/1.1 200 OK
    // HTTP/1.1 403 Forbidden
    //
    // ищем тело ответа
    hresp->body = strstr(response, "\r\n\r\n");
    if(hresp->body) {
        // найдено
        // первый символ '\r' это конец заголовков, пишем туда 0
        *(hresp->body) = 0;
        // смещаем указатель на 4
        hresp->body += 4;
    }
    // ищем пробел перед кодом возврата
    char *temp = strchr(response, ' ');
    if(!temp) {
        // нет пробела? странно...
        HttpFreeResponse(hresp);
        return HTTP_RES_BAD_RESP_FORMAT;
    }
    // переходим к первой цифре кода
    temp++;
    hresp->status_code = atoi(temp);
    // ищем пробел после кода возврата
    temp = strchr(temp, ' ');
    if(!temp) {
        // нет пробела? странно...
        HttpFreeResponse(hresp);
        return HTTP_RES_BAD_RESP_FORMAT;
    }
    // переходим к первому символу статуса
    temp++;
    hresp->status_text = temp;
    // ищем конец строки статуса
    temp = strstr(temp, "\r\n");
    if(temp) {
        // конец строки найден, меняем '\r' на 0
        *temp = 0;
    } else {
        // конец строки не найден, очень странно
        // заголовков похоже дальше тоже нет
        hresp->status_text = NULL;
        return HTTP_RES_OK;
    }
    // переходим к началу заголовков
    temp += 2;
    hresp->response_headers = temp;
    //
    return HTTP_RES_OK;
}

// добавление заголовка авторизации
static int add_http_auth(char *http_headers, url_t *purl) {
    //
    // буфер для пары "username:password"
    int upwd_len = strlen(purl->username) + strlen(purl->password) + 1;
    if(upwd_len > 128) {
        free(http_headers);
        HttpFreeURL(purl);
        return 0;
    }
    char *upwd = malloc(upwd_len + 1);
    sprintf(upwd, "%s:%s", purl->username, purl->password);
    //
    // вычисляем размер в BASE64
    int base64_len = base64_encode((uint8_t *)upwd, NULL, upwd_len, 0);
    char *base64 = malloc(base64_len + 1);
    // кодируем BASE64
    base64_encode((uint8_t *)upwd, (uint8_t *)base64, upwd_len, 0);
    //
    strlcat(http_headers, "Authorization: Basic ", MAX_REQ_HEADERS_LEN);
    strlcat(http_headers, base64, MAX_REQ_HEADERS_LEN);
    strlcat(http_headers, "\r\n", MAX_REQ_HEADERS_LEN);
    //
    free(base64);
    free(upwd);
    //
    return 1;
}

/**
 * @brief Выполнение HTTP запроса GET
 * Функция возвращает результат выполнения запроса: элемент перечисления http_result_t.
 * При успешном выполнении функция заполняет элементы структуры http_response_t.
 * Память занимаемая структурой должны быть освобождена функцией HttpResponseFree.
 * @param url Строка URL.
 * @param custom_headers Пользовательские заголовки.
 * @return http_result_t Результат выполнения запроса.
 */
http_result_t HttpGET(char *url, char *custom_headers, http_response_t *hresp) {
    //
    // разбор URL
    url_t *purl = HttpParseURL(url);
    if(!purl) return HTTP_RES_BAD_URL;
    //
    char *http_headers = malloc(MAX_REQ_HEADERS_LEN);
    if(purl->path) {
        if(purl->query) {
            snprintf(http_headers, MAX_REQ_HEADERS_LEN, "GET /%s?%s HTTP/1.1\r\nHost:%s", purl->path, purl->query, purl->host);
        } else {
            snprintf(http_headers, MAX_REQ_HEADERS_LEN, "GET /%s HTTP/1.1\r\nHost:%s", purl->path, purl->host);
        }
    } else {
        if(purl->query) {
            snprintf(http_headers, MAX_REQ_HEADERS_LEN, "GET /?%s HTTP/1.1\r\nHost:%s", purl->query, purl->host);
        } else {
            snprintf(http_headers, MAX_REQ_HEADERS_LEN, "GET / HTTP/1.1\r\nHost:%s", purl->host);
        }
    }
    strlcat(http_headers, "\r\nConnection:close\r\n", MAX_REQ_HEADERS_LEN);
    //
    // добавляем логин/пароль, если они есть
    if(purl->username) {
        if(!add_http_auth(http_headers, purl)) {
            return HTTP_RES_AUTH_TOO_LARGE;
        }
    }
    //
    // добавляем заголовки пользователя
    if(custom_headers) {
        strlcat(http_headers, custom_headers, MAX_REQ_HEADERS_LEN);
    }
    //
    // разделитель перед телом запроса
    if(strlcat(http_headers, "\r\n", MAX_REQ_HEADERS_LEN) >= MAX_REQ_HEADERS_LEN) {
        // переполнение буфера заголовков
        free(http_headers);
        HttpFreeURL(purl);
        return HTTP_RES_HEADERS_TOO_LARGE;
    }
    // это может и не нужно?
    http_headers = realloc(http_headers, strlen(http_headers) + 1);
    //
    // выполняем запрос
    http_result_t req_res = http_req(http_headers, purl, hresp);
    //
    free(http_headers);
    HttpFreeURL(purl);
    //
    return req_res;
}

/**
 * @brief Выполнение HTTP запроса POST.
 * Функция возвращает результат выполнения запроса: элемент перечисления http_result_t.
 * При успешном выполнении функция возвращает HTTP_RES_OK и заполняет элементы структуры hresp.
 * Пользователь должен дополнительно проверить статус выполнения запроса (поле hresp->status_code).
 * Память занимаемая структурой в последствии должна быть освобождена функцией HttpFreeResponse.
 * Формат строки url должен соответствовать описанному в функции HttpParseURL.
 * @param url Строка URL.
 * @param custom_headers Пользовательские заголовки (могут быть NULL)
 * @param post_data Записываемые данные POST.
 * @param post_type Тип данных POST: httpdUSER | httpdJSON | httpdHTML | httpdTEXT | httpdFORMURLENCODED | httpdXML
 * @param hresp Указатель на структуру ответа.
 * @return http_result_t Результат выполнения запроса.
 */
http_result_t HttpPOST(char *url, char *custom_headers, char *post_data, http_content_type_t post_type, http_response_t *hresp) {
    //
    if(!post_data) return HTTP_RES_EMPTY_POST_DATA;
    unsigned int post_data_len = strlen(post_data);
    // разбор URL
    url_t *purl = HttpParseURL(url);
    if(!purl) return HTTP_RES_BAD_URL;
    //
    unsigned int post_headers_len = MAX_REQ_HEADERS_LEN + post_data_len;
    char *http_headers = malloc(post_headers_len);
    if(purl->path) {
        if(purl->query) {
            snprintf(http_headers, post_headers_len, "POST /%s?%s HTTP/1.1\r\nHost:%s\r\nContent-Length:%u", purl->path, purl->query, purl->host, post_data_len);
        } else {
            snprintf(http_headers, post_headers_len, "POST /%s HTTP/1.1\r\nHost:%s\r\nContent-Length:%u", purl->path, purl->host, post_data_len);
        }
    } else {
        if(purl->query) {
            snprintf(http_headers, post_headers_len, "POST /?%s HTTP/1.1\r\nHost:%s\r\nContent-Length:%u", purl->query, purl->host, post_data_len);
        } else {
            snprintf(http_headers, post_headers_len, "POST / HTTP/1.1\r\nHost:%s\r\nContent-Length:%u", purl->host, post_data_len);
        }
    }
    //
    strlcat(http_headers, "\r\nConnection:close\r\n", post_headers_len);
    //
    // добавляем заголовок типа данных
    switch (post_type) {
        case httpdJSON:
            strlcat(http_headers, "Content-Type:application/json\r\n", post_headers_len);
            break;
        case httpdHTML:
            strlcat(http_headers, "Content-Type:text/html\r\n", post_headers_len);
            break;
        case httpdTEXT:
            strlcat(http_headers, "Content-Type:text/plain\r\n", post_headers_len);
            break;
        case httpdFORMURLENCODED:
            strlcat(http_headers, "Content-Type:application/x-www-form-urlencoded\r\n", post_headers_len);
            break;
        case httpdXML:
            strlcat(http_headers, "Content-Type:text/xml\r\n", post_headers_len);
            break;
        default:
            // httpdUSER: заголовок типа должен добавить пользователь!
            break;
    }
    //
    // добавляем логин/пароль, если они есть
    if(purl->username) {
        if(!add_http_auth(http_headers, purl)) {
            return HTTP_RES_AUTH_TOO_LARGE;
        }
    }
    //
    // добавляем заголовки пользователя
    if(custom_headers) {
        strlcat(http_headers, custom_headers, post_headers_len);
    }
    //
    // добавляем разделитель и данные POST
    strlcat(http_headers, "\r\n", post_headers_len);
    if(strlcat(http_headers, post_data, post_headers_len) >= post_headers_len) {
        // переполнение буфера заголовков
        free(http_headers);
        HttpFreeURL(purl);
        return HTTP_RES_HEADERS_TOO_LARGE;
    }
    //
    // это может и не нужно?
    http_headers = realloc(http_headers, strlen(http_headers) + 1);
    //
    // выполняем запрос
    http_result_t req_res = http_req(http_headers, purl, hresp);
    //
    free(http_headers);
    HttpFreeURL(purl);
    //
    return req_res;
}

/**
 * @brief Освобождение памяти занимаемой ответом HTTP.
 * @param hresp Указатель на структуру ответа.
 */
void HttpFreeResponse(http_response_t *hresp) {
    if(hresp) {
        if(hresp->response) {
            free(hresp->response);
            hresp->response_len = 0;
            hresp->response = NULL;
        }
    }
}
