/**
 * @file httpurl.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции клиента HTTP. Разбор URL.
 * @version 0.1
 * @date 2024-05-19
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <http.h>

struct url_scheme {
    const char      *name;
    unsigned short  port;
};

static const struct url_scheme url_schemes[] = {
    {"ftp",         21},
    {"ssh",         22},
    {"telnet",      23},
    {"smtp",        25},
    {"http",        80},
    {"imap",        143},
    {"https",       443},
    {"imaps",       993},
    {"mqtt",        1883},
    {"mqtts",       8883}
};

static unsigned short port_by_scheme(char *name, unsigned short int def_port) {
    for(int i=0; i < (sizeof(url_schemes) / sizeof((url_schemes)[0])); i++) {
        if(strcmp(name, url_schemes[i].name) == 0) {
            return url_schemes[i].port;
        }
    }
    //
    return def_port;
}

static int is_scheme_char(int c) {
    return ((!isalpha(c)) && ('+' != c) && ('-' != c) && ('.' != c)) ? 0 : 1;
}

/**
 * @brief Освобождение памяти, занимаемой структурой parsed_url.
 * @param purl Указатель на структуру URL
 */
void HttpFreeURL(url_t *purl) {
    if (!purl) return;
    //
    if (purl->scheme)   free(purl->scheme);
    if (purl->host)     free(purl->host);
    if (purl->path)     free(purl->path);
    if (purl->query)    free(purl->query);
    if (purl->fragment) free(purl->fragment);
    if (purl->username) free(purl->username);
    if (purl->password) free(purl->password);
    free(purl);
}

/**
 * @brief Разбор строки URL в соответствии с RFC 1738/3986.
 * При успешном разборе строки функция выделяет память для структуры url_t, заполняет её поля и возвращает указатель на структуру.
 * При ошибке разбора функция возвращает NULL.
 * @param url Указатель на строку URL.
 * @return url_t* 
 */
url_t *HttpParseURL(const char *url) {
    url_t *purl;
    const char *tmpstr;
    const char *curstr;
    int userpass_flag;
    int bracket_flag;
    int len;
    int i;
    //
    purl = (url_t*)malloc(sizeof(url_t));
    if (!purl) return NULL;
    memset(purl, 0, sizeof(url_t));
    //
    curstr = url;
    //
    // Чтение схемы:
    // <scheme>:<scheme-specific-part>
    // <scheme> := [a-z\+\-\.]+
    //
    tmpstr = strchr(curstr, ':');
    if (!tmpstr) {
        HttpFreeURL(purl);
        return NULL;
    }
    //
    len = tmpstr - curstr;
    for ( i = 0; i < len; i++ ) {
        if (!is_scheme_char(curstr[i])) {
            // неверный формат схемы
            HttpFreeURL(purl);
            return NULL;
        }
    }
    purl->scheme = malloc(len + 1);
    if (!purl->scheme ) {
        HttpFreeURL(purl);
        return NULL;
    }
    // копирование имени схемы в стуктуру, заодно переводим в нижний регистр
    for ( i = 0; i < len; i++ ) {
        purl->scheme[i] = tolower(curstr[i]);
    }
    purl->scheme[len] = 0;
    //
    // Пропуск ':'
    tmpstr++;
    curstr = tmpstr;
    //
    // //<user>:<password>@<host>:<port>/<url-path>
    //
    // Пропуск "//"
    for ( i = 0; i < 2; i++ ) {
        if ( *curstr != '/' ) {
            HttpFreeURL(purl);
            return NULL;
        }
        curstr++;
    }
    // Проверка наличия имени и пароля
    userpass_flag = 0;
    tmpstr = curstr;
    while (*tmpstr ) {
        if (*tmpstr == '@') {
            // имя и пароль имеются
            userpass_flag = 1;
            break;
        } 
        else if ( '/' == *tmpstr ) {
            // имени и пароля нет
            userpass_flag = 0;
            break;
        }
        tmpstr++;
    }
    // Разбор имени и пароля
    tmpstr = curstr;
    if ( userpass_flag ) {
        // имя
        while ( (*tmpstr) && (*tmpstr != ':') && (*tmpstr != '@')) tmpstr++;
        len = tmpstr - curstr;
        purl->username = malloc(len + 1);
        if (!purl->username) {
            HttpFreeURL(purl);
            return NULL;
        }
        strncpy(purl->username, curstr, len);
        purl->username[len] = '\0';
        // а имеется ли пароль?
        curstr = tmpstr;
        if (*curstr == ':') {
            // пропуск ":"
            curstr++;
            // пароль
            tmpstr = curstr;
            while ((*tmpstr) && (*tmpstr != '@')) tmpstr++;
            len = tmpstr - curstr;
            purl->password = malloc(len + 1);
            if (!purl->password) {
                HttpFreeURL(purl);
                return NULL;
            }
            strncpy(purl->password, curstr, len);
            purl->password[len] = 0;
            curstr = tmpstr;
        }
        // пропуск '@'
        if ( *curstr != '@') {
            HttpFreeURL(purl);
            return NULL;
        }
        curstr++;
    }
    if (*curstr == '[') bracket_flag = 1; else bracket_flag = 0;
    //
    // разделители перед хостом
    tmpstr = curstr;
    while (*tmpstr) {
        if (bracket_flag && (*tmpstr == ']')) {
            // конец IPv6
            tmpstr++;
            break;
        } else 
        if ( !bracket_flag && ((*tmpstr == ':') || (*tmpstr == '/'))) {
            // порт указан
            break;
        }
        tmpstr++;
    }
    len = tmpstr - curstr;
    purl->host = malloc(len + 1);
    if ((!purl->host) || (len <= 0)) {
        HttpFreeURL(purl);
        return NULL;
    }
    strncpy(purl->host, curstr, len);
    purl->host[len] = 0;
    curstr = tmpstr;
    // порт был?
    if ( *curstr == ':') {
        curstr++;
        tmpstr = curstr;
        while ( *tmpstr && (*tmpstr != '/')) tmpstr++;
        len = tmpstr - curstr;
        //
        purl->port = 0;
        for(i=0; i<len; i++) {
            if(!isdigit(curstr[i])) {
                HttpFreeURL(purl);
                return NULL;
            }
            purl->port = purl->port*10 + (curstr[i] - '0');
        }
        curstr = tmpstr;
    } else {
        // если номер порта не указан, то ставим по умолчанию
        purl->port = port_by_scheme(purl->scheme, 80);
    }
    //
    purl->uri = (char*)url;
    if (*curstr == '\0') return purl;

    // пропуск '/'
    if (*curstr != '/') {
        HttpFreeURL(purl);
        return NULL;
    }
    curstr++;
    //
    // разбор пути
    tmpstr = curstr;
    while ((*tmpstr != 0) && (*tmpstr != '#')  && (*tmpstr != '?')) {
        tmpstr++;
    }
    len = tmpstr - curstr;
    purl->path = malloc(len + 1);
    if (!purl->path) {
        HttpFreeURL(purl);
        return NULL;
    }
    strncpy(purl->path, curstr, len);
    purl->path[len] = 0;
    curstr = tmpstr;
    //
    // параметры запроса имеются?
    if (*curstr == '?') {
        curstr++;
        tmpstr = curstr;
        while ((*tmpstr != 0) && (*tmpstr != '#')) {
            tmpstr++;
        }
        len = tmpstr - curstr;
        purl->query = malloc(len + 1);
        if (!purl->query) {
            HttpFreeURL(purl);
            return NULL;
        }
        strncpy(purl->query, curstr, len);
        purl->query[len] = 0;
        curstr = tmpstr;
    }
    //
    // фрагмент имеется?
    if (*curstr == '#') {
        curstr++;
        tmpstr = curstr;
        while (*tmpstr) tmpstr++;
        len = tmpstr - curstr;
        purl->fragment = malloc(len + 1);
        if (!purl->fragment) {
            HttpFreeURL(purl);
            return NULL;
        }
        strncpy(purl->fragment, curstr, len);
        purl->fragment[len] = 0;
        curstr = tmpstr;
    }
    return purl;
}
