/**
 * @file isocptoutf8.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функция табличного преобразования кодировки в UTF8.
 * @version 0.1
 * @date 2024-05-07
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <string.h>
#include <iconv.h>

/**
 * @brief Перекодирование символов строки в кодировку UTF8 при помощи таблицы преобразование.
 * При успешном выполнении функция преобразует кодировку символов строки src в кодировку UTF8 и размещает её в буфере dest.
 * Функция возвращает количество байт занятых в буфере dest (включая завершающий символ 0).
 * При ошибке функция возвращает значение 0. Ошибка может быть вызвана недостаточным размером буфера dest или некорректными значениями параметров.
 * @param cptable Таблица преобразования.
 * @param src Строка в исходной кодировке.
 * @param dest Буфер для строки в кодировке UTF8.
 * @param maxdestlen Размер буфера для строки UTF8.
 * @return int 0-ошибка, >0-количество занятых байт.
 */
int char_string_to_UTF8(const unsigned short *cptable, char *src, char *dest, int maxdestlen) {
    int srcindex, destindex;
    unsigned long int wc;
    //
    if(!src) return 0;
    if(!dest) return 0;
    if(!cptable) return 0;
    //
    int srclen = strlen(src);
    if(!srclen) {
        if(!maxdestlen) return 0;
        dest[0] = 0;
        return 1;
    }
    //
    srcindex = 0;
    destindex = 0;
    //
    while(srcindex < srclen) {
        wc = src[srcindex++];
        if(wc >= 0x80) wc = cptable[wc - 0x80];
        //
        if(wc < 0x080) {
            if(destindex + 1 >= maxdestlen) return 0;
            dest[destindex++] = (unsigned char)wc;
        }
        else if(wc < 0x800) {
            if(destindex + 2 >= maxdestlen) return 0;
            dest[destindex++] = 0xc0 | (wc >> 6);
            dest[destindex++] = 0x80 | (wc & 0x3f);
        } else {
            if(destindex + 3 >= maxdestlen) return 0;
            dest[destindex++] = 0xe0 | (wc >> 12);
            dest[destindex++] = 0x80 | ((wc >> 6) & 0x3f);
            dest[destindex++] = 0x80 | (wc & 0x3f);
        }
    }
    dest[destindex++] = 0;
    //
    return destindex;
}
