/**
 * @file fpde64.c
 * @brief Эмуляция плавающей точки двойной точности
 * @version 0.1
 * @date 2024-03-09
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */


#include <syscallp.h>

// ---------------------------------------------------------------------------------------------------------
// эмуляция double


//                             6666555555555544444444443333333333222222222211111111110000000000
//                             3210987654321098765432109876543210987654321098765432109876543210
#define DOUBLE_ZERO          0b0000000000000000000000000000000000000000000000000000000000000000
#define DOUBLE_NZERO         0b1000000000000000000000000000000000000000000000000000000000000000
#define DOUBLE_SIGNBIT       0b1000000000000000000000000000000000000000000000000000000000000000
#define DOUBLE_MEANINGBITS   0b0111111111111111111111111111111111111111111111111111111111111111
#define DOUBLE_EXPONENT      0b0111111111110000000000000000000000000000000000000000000000000000
#define DOUBLE_MANTISSA      0b0000000000001111111111111111111111111111111111111111111111111111
#define DOUBLE_MSBMANTISSA   0b0000000000001000000000000000000000000000000000000000000000000000

#define DOUBLE_IS_NAN(x)     ((((x) & DOUBLE_EXPONENT) == DOUBLE_EXPONENT) && ((x) & DOUBLE_MANTISSA))

// These functions return a nonzero value if either argument is NaN, or if a and b are unequal.
//
// int __nesf2 (float a, float b)
// int __netf2 (long double a, long double b)
int __nedf2 (double a, double b) {
    FP_UNION_D_T dua = {.d = a};
    if(DOUBLE_IS_NAN(dua.u64)) return 1;
    //
    FP_UNION_D_T dub = {.d = b};
    if(DOUBLE_IS_NAN(dub.u64)) return 1;
    //
    return (dua.u64 == dub.u64)?0:1;
}

// These functions return zero if neither argument is NaN, and a and b are equal.
// int __eqsf2 (float a, float b)
// int __eqtf2 (long double a, long double b)
int __eqdf2 (double a, double b) {
    FP_UNION_D_T dua = {.d = a};
    if(DOUBLE_IS_NAN(dua.u64)) return 1;
    //
    FP_UNION_D_T dub = {.d = b};
    if(DOUBLE_IS_NAN(dub.u64)) return 1;
    //
    return (dua.u64 == dub.u64)?0:1;
}

// These functions return a value greater than zero if neither argument is NaN, and a is strictly greater than b.
//
// int __gtsf2 (float a, float b)
// int __gttf2 (long double a, long double b)
int __gtdf2 (double a, double b) {
    FP_UNION_D_T dua = {.d = a};
    if(DOUBLE_IS_NAN(dua.u64)) return 0;
    //
    FP_UNION_D_T dub = {.d = b};
    if(DOUBLE_IS_NAN(dub.u64)) return 0;
    //
    return __internal_syscall(SYSCALL___gtdf2, (long)dua.u32.lo, (long)dua.u32.hi, (long)dub.u32.lo, (long)dub.u32.hi, 0, 0);
}


// These functions return a value greater than or equal to zero if neither argument is NaN, and a is greater than or equal to b.
// int __gesf2 (float a, float b)
// int __getf2 (long double a, long double b)
int __gedf2 (double a, double b) {
    FP_UNION_D_T dua = {.d = a};
    if(DOUBLE_IS_NAN(dua.u64)) return -1;
    //
    FP_UNION_D_T dub = {.d = b};
    if(DOUBLE_IS_NAN(dub.u64)) return -1;
    //
    if(dua.u64 == dub.u64) return 0;
    //
    long res = __internal_syscall(SYSCALL___gtdf2, (long)dua.u32.lo, (long)dua.u32.hi, (long)dub.u32.lo, (long)dub.u32.hi, 0, 0);
    return (res == 0)?-1:res;
}

// These functions return a value less than zero if neither argument is NaN, and a is strictly less than b.
//
// int __ltsf2 (float a, float b)
// int __lttf2 (long double a, long double b)
int __ltdf2 (double a, double b) {
    FP_UNION_D_T dua = {.d = a};
    if(DOUBLE_IS_NAN(dua.u64)) return 0;
    //
    FP_UNION_D_T dub = {.d = b};
    if(DOUBLE_IS_NAN(dub.u64)) return 0;
    //
    long res = __internal_syscall(SYSCALL___gtdf2, (long)dub.u32.lo, (long)dub.u32.hi, (long)dua.u32.lo, (long)dua.u32.hi, 0, 0);
    return (res == 1)?-1:0;
}

// These functions return a value less than or equal to zero if neither argument is NaN, and a is less than or equal to b.
// int __lesf2 (float a, float b)
// int __letf2 (long double a, long double b)
int __ledf2 (double a, double b) {
    FP_UNION_D_T dua = {.d = a};
    if(DOUBLE_IS_NAN(dua.u64)) return 1;
    //
    FP_UNION_D_T dub = {.d = b};
    if(DOUBLE_IS_NAN(dub.u64)) return 1;
    //
    if(dua.u64 == dub.u64) return 0;
    //
    long res = __internal_syscall(SYSCALL___gtdf2, (long)dub.u32.lo, (long)dub.u32.hi, (long)dua.u32.lo, (long)dua.u32.hi, 0, 0);
    return (res == 1)?0:1;
}

// These functions convert a to a signed integer, rounding toward zero.
//
// int __fixsfsi (float a)
// int __fixtfsi (long double a)
// int __fixxfsi (long double a)
int __fixdfsi (double a) {
    FP_UNION_D_T dua = {.d = a};
    //
    return __internal_syscall_invoke2(SYSCALL___fixdfsi, (long)dua.u32.lo, (long)dua.u32.hi);
}

// These functions convert a to an unsigned integer, rounding toward zero. Negative values all become zero.
//
// unsigned int __fixunssfsi (float a)
// unsigned int __fixunstfsi (long double a)
// unsigned int __fixunsxfsi (long double a)
unsigned int __fixunsdfsi (double a) {
    FP_UNION_D_T dua = {.d = a};
    //
    if (dua.u64 & DOUBLE_SIGNBIT) return 0;
    return (unsigned)__internal_syscall_invoke2(SYSCALL___fixunsdfsi, (long)dua.u32.lo, (long)dua.u32.hi);
}

// These functions return the difference between b and a; that is, a - b.
//
// float __subsf3 (float a, float b)
// long double __subtf3 (long double a, long double b)
// long double __subxf3 (long double a, long double b)
double __subdf3 (double a, double b) {
    FP_UNION_D_T dua = {.d = a};
    FP_UNION_D_T dub = {.d = b};
    FP_UNION_D_T dur;
    // 
    // результат (double) получаем через указатель
    __internal_syscall(SYSCALL___subdf3, (long)dua.u32.lo, (long)dua.u32.hi, (long)dub.u32.lo, (long)dub.u32.hi, 0, (long)&dur);
    //
    return dur.d;
}

// These functions return the sum of a and b.
//
// float __addsf3 (float a, float b)
// long double __addtf3 (long double a, long double b)
// long double __addxf3 (long double a, long double b)
double __adddf3 (double a, double b) {
    FP_UNION_D_T dua = {.d = a};
    FP_UNION_D_T dub = {.d = b};
    FP_UNION_D_T dur;
    // 
    // результат (double) получаем через указатель
    __internal_syscall(SYSCALL___adddf3, (long)dua.u32.lo, (long)dua.u32.hi, (long)dub.u32.lo, (long)dub.u32.hi, 0, (long)&dur);
    //
    return dur.d;
}

// These functions return the product of a and b.
//
// float __mulsf3 (float a, float b)
// long double __multf3 (long double a, long double b)
// long double __mulxf3 (long double a, long double b)
double __muldf3 (double a, double b) {
    FP_UNION_D_T dua = {.d = a};
    FP_UNION_D_T dub = {.d = b};
    FP_UNION_D_T dur;
    // 
    // результат (double) получаем через указатель
    __internal_syscall(SYSCALL___mulsf3, (long)dua.u32.lo, (long)dua.u32.hi, (long)dub.u32.lo, (long)dub.u32.hi, 0, (long)&dur);
    //
    return dur.d;
}

// These functions return the quotient of a and b; that is, a / b.
//
// float __divsf3 (float a, float b)
// long double __divtf3 (long double a, long double b)
// long double __divxf3 (long double a, long double b)
double __divdf3 (double a, double b) {
    FP_UNION_D_T dua = {.d = a};
    FP_UNION_D_T dub = {.d = b};
    FP_UNION_D_T dur;
    // 
    // результат (double) получаем через указатель
    __internal_syscall(SYSCALL___divdf3, (long)dua.u32.lo, (long)dua.u32.hi, (long)dub.u32.lo, (long)dub.u32.hi, 0, (long)&dur);
    //
    return dur.d;
}

// These functions convert i, an unsigned integer, to floating point.
//
// float __floatunsisf (unsigned int i)
// long double __floatunsitf (unsigned int i)
// long double __floatunsixf (unsigned int i)
double __floatunsidf (unsigned int i) {
    FP_UNION_D_T dur;
    // результат (double) получаем через указатель
    __internal_syscall(SYSCALL___floatunsidf, (long)i, 0, 0, 0, 0, (long)&dur);
    //
    return dur.d;
}

// These functions convert i, a signed integer, to floating point.
//
// float __floatsisf (int i)
// long double __floatsitf (int i)
// long double __floatsixf (int i)
double __floatsidf (int i) {
    FP_UNION_D_T dur;
    // результат (double) получаем через указатель
    __internal_syscall(SYSCALL___floatsidf, (long)i, 0, 0, 0, 0, (long)&dur);
    //
    return dur.d;
}

// These functions extend a to the wider mode of their return type.
//
// long double __extendsftf2 (float a)
// long double __extendsfxf2 (float a)
// long double __extenddftf2 (double a)
// long double __extenddfxf2 (double a)
double __extendsfdf2 (float a) {
    FP_UNION_S_T sua = {.f = a};
    FP_UNION_D_T dur;
    // результат (double) получаем через указатель
    __internal_syscall(SYSCALL___extendsfdf2, sua.s32, 0, 0, 0, 0, (long)&dur);
    //
    return dur.d;
}

// These functions truncate a to the narrower mode of their return type, rounding toward zero.
// double __truncxfdf2 (long double a)
// double __trunctfdf2 (long double a)
// float __truncxfsf2 (long double a)
// float __trunctfsf2 (long double a)
float __truncdfsf2 (double a) {
    FP_UNION_D_T dua = {.d = a};
    FP_UNION_S_T sur;
    //
    // результат (float) получаем через указатель
    __internal_syscall(SYSCALL___truncdfsf2, (long)dua.u32.lo, (long)dua.u32.hi, 0, 0, 0, (long)&sur);
    //
    return sur.f;
}

/**
 * @brief Эмуляция функции с одним аргументом типа плавающая точка двойной точности
 * 
 * @param n Номер SYSCALL
 * @param x Аргумент 1
 * @return double 
 */
double __internal_dfp_invoke1(long n, double x) {
    FP_UNION_D_T dux = {.d = x};
    FP_UNION_D_T dur;
    // 
    // результат (double) получаем через указатель
    __internal_syscall(n, (long)dux.u32.lo, (long)dux.u32.hi, 0, 0, 0, (long)&dur);
    //
    return dur.d;
}

/**
 * @brief Эмуляция функции с двумя аргументами типа плавающая точка двойной точности
 * 
 * @param n Номер SYSCALL
 * @param x Аргумент 1
 * @return double 
 */
double __internal_dfp_invoke2(long n, double x, double y) {
    FP_UNION_D_T dux = {.d = x};
    FP_UNION_D_T duy = {.d = y};
    FP_UNION_D_T dur;
    // 
    // результат (double) получаем через указатель
    __internal_syscall(n, (long)dux.u32.lo, (long)dux.u32.hi, (long)duy.u32.lo, (long)duy.u32.hi, 0, (long)&dur);
    //
    return dur.d;
}
