/**
 * @file modbus.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции поддержки работы с протоколом обмена Modbus.
 * Реализация соответствует спецификациям:
 * "MODBUS Application Protocol Specification V1.1b3"
 * "MODBUS Messaging on TCP/IP Implementation Guide V1.0b"
 * @version 0.1
 * @date 2024-04-01
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */


#include <satcrc.h>
#include <unistd.h>
#include <fastmem.h>

#include <modbus.h>

/**
 * @brief Таймаут операций обмена данными по протоколу Mosbus-RTU в режиме Master, мс
 */
int ModbusRTUTimeout = 500;

/**
 * @brief Таймаут операций обмена данными по протоколу Mosbus-TCP в режиме Master, мс
 */
int ModbusTCPTimeout = 500;

/**
 * @brief Текущее значение идентификатора транзакции Modbus-TCP
 */
static uint16_t ModbusTCPTranId = 0;

/* Коды функций Modbus */
#define MODBUS_FC_READ_COILS                0x01
#define MODBUS_FC_READ_DISCRETE_INPUTS      0x02
#define MODBUS_FC_READ_HOLDING_REGISTERS    0x03
#define MODBUS_FC_READ_INPUT_REGISTERS      0x04
#define MODBUS_FC_WRITE_SINGLE_COIL         0x05
#define MODBUS_FC_WRITE_SINGLE_REGISTER     0x06
#define MODBUS_FC_READ_EXCEPTION_STATUS     0x07
#define MODBUS_FC_WRITE_MULTIPLE_COILS      0x0F
#define MODBUS_FC_WRITE_MULTIPLE_REGISTERS  0x10
#define MODBUS_FC_REPORT_SLAVE_ID           0x11
#define MODBUS_FC_MASK_WRITE_REGISTER       0x16
#define MODBUS_FC_WRITE_AND_READ_REGISTERS  0x17

/* Ограничения (Modbus_Application_Protocol_V1_1b.pdf):
 *  Страница 12: Quantity of Coils to read (2 bytes): 1 to 2000 (0x7D0)
 *  Страница 29: Quantity of Coils to write (2 bytes): 1 to 1968 (0x7B0)
 *  Страница 15: Quantity of Registers to read (2 bytes): 1 to 125 (0x7D)
 *  Страница 31: Quantity of Registers to write (2 bytes) 1 to 123 (0x7B)
 *  Страница 38: Quantity of Registers to write in R/W registers (2 bytes) 1 to 121 (0x79)
 */
#define MODBUS_MAX_READ_BITS                2000
#define MODBUS_MAX_WRITE_BITS               1968

#define MODBUS_MAX_READ_REGISTERS           125
#define MODBUS_MAX_WRITE_REGISTERS          123
#define MODBUS_MAX_RW_READ_REGISTERS        125
#define MODBUS_MAX_RW_WRITE_REGISTERS       121

#define MODBUS_TCP_MBAP_LEN                 7

#define RTU_OVER_TCP_FLAG                   0x00010000
#define RTU_OVER_TCP_FLAG                   0x00010000

/**
 * @brief Установка варианта протокола Modbus.
 * Функция используется только для дескрипторов сетевых (TCP или UDP) сокетов и позволяет установить
 * вариант используемого протокола Modbus.
 * При указании значения варианта протокола MDBS_PROTO_TCP будет использован протокол "Modbus TCP", этот вариант используется по
 * умолчанию и его можно не указывать.
 * При указании варианта протокола MDBS_PROTO_RTU будет использован протокол "Modbus RTU over TCP" или "Modbus RTU over UDP" в
 * завимимости от типа сетевого сокета.
 * Вариант MDBS_PROTO_RTU может быть использован при подключении устройства с протоколом "Modbus RTU" через "прозрачный"
 * преобразователь интерфейса.
 * Функция возвращает 1 при успешном выполнении или 0 при ошибке.
 * Ошибкой является некорректное значение дескриптора или несовместимая комбинация типа дескриптора и указанного варианта протокола.
 * @param handle Указатель на дескриптор сетевого сокета.
 * @param proto Вариант протокола.
 * @return int 
 */
int ModbusSetProto(int *handle, mdbs_proto_t proto) {
    if(socket_handle_type(*handle) != HDL_NETWORK) return 0;
    //
    switch (proto) {
        case MDBS_PROTO_RTU_OVER_TCP:
            *handle = *handle | RTU_OVER_TCP_FLAG;
            return 1;
        case MDBS_PROTO_TCP:
            *handle = *handle & (~RTU_OVER_TCP_FLAG);
            return 1;
        default:
            return 0;
    }
};

/**
 * @brief Запись/чтение PDU Modbus RTU
 */
static mdbs_result_t pdu_rtu_data_exchange(int handle, uint8_t adr, uint16_t wlen, uint16_t rlen, uint8_t *pdu) {
    uint8_t adu[256];
    uint16_t crc;
    int res;
    //
    adu[0] = adr;                                   // адрес ADU
    fast_memmove(&adu[1], pdu, wlen);               // содержимое PDU
    //
    crc = crc_modbus(adu, wlen + 1);                // CRC
    adu[wlen + 1] = crc & 0xff;
    adu[wlen + 2] = (crc >> 8) & 0xff;
    //
    purge(handle);                                  // передача запроса
    write(handle, adu, wlen + 3);
    //
    // приём заголовка 5 байт: адрес, код функции
    res = read_block(handle, adu, 5, ModbusRTUTimeout);
    if(res == 0) return MDBS_DEVICE_NOT_RESPOND;
    if(res < 0) return res;
    //
    // проверка адреса
    if(adr != adu[0]) return MDBS_MISMATCH_ADDR;
    //
    // проверка кода ошибки выполнения
    if(adu[1] & 0x80) {
        // ответ с кодом ошибки, проверяем CRC
        if(crc_modbus(adu, 5)) return MDBS_INVALID_CRC;
        // возвращаем код ошибки
        return adu[2];
    }
    // ошибок нет, дочитываем остаток
    res = read_block(handle, &adu[5], rlen - 2, ModbusRTUTimeout);
    if(res == 0) return MDBS_DEVICE_NOT_RESPOND;
    if(res < 0) return res;
    //
    // проверяем CRC
    if(crc_modbus(adu, rlen + 3)) return MDBS_INVALID_CRC;
    //
    fast_memmove(pdu, &adu[1], rlen);
    //
    return MDBS_OK;
}

/**
 * @brief Запись/чтение PDU Modbus TCP
 */
static mdbs_result_t pdu_tcp_data_exchange(int handle, uint8_t adr, uint16_t wlen, uint16_t rlen, uint8_t *pdu) {
    uint8_t adu[264];
    //
    int res;
    uint16_t tranid   = ModbusTCPTranId++;
    uint8_t tranid_hi = (tranid >> 8) & 0xff;
    uint8_t tranid_lo = tranid & 0xff;
    //
    adu[0] = tranid_hi;                                 // Transaction Id
    adu[1] = tranid_lo;
    //
    adu[2] = 0;                                         // Protocol Id
    adu[3] = 0;
    //
    uint16_t wlenp1 = wlen + 1;
    adu[4] = (wlenp1 >> 8) & 0xff;                      // Number of following bytes
    adu[5] = wlenp1 & 0xff;
    //
    adu[6] = adr;                                       // Unit Identifier (slave adr)
    //
    fast_memmove(&adu[MODBUS_TCP_MBAP_LEN], pdu, wlen); // перенос PDU
    //
    // очистка приёмного буфера и передача запроса
    purge(handle);
    write(handle, adu, wlen + MODBUS_TCP_MBAP_LEN);
    //
    // приём ответа
    res = read_block(handle, adu, MODBUS_TCP_MBAP_LEN + 2, ModbusTCPTimeout);
    if(res == 0) return MDBS_DEVICE_NOT_RESPOND;
    if(res < 0) return res;
    //
    if((adu[0] != tranid_hi) || (adu[1] != tranid_lo)) return MDBS_MISMATCH_TRANSACTION_ID;
    if((adu[2] != 0) || (adu[3] != 0)) return MDBS_MISMATCH_TRANSACTION_ID;
    //
    // проверка кода ошибки выполнения
    if(adu[7] & 0x80) {
        // ответ с кодом ошибки, возвращаем код ошибки
        return adu[8];
    }
    // дочитываем ответ
    res = read_block(handle, &adu[9], rlen - 2, ModbusTCPTimeout);
    if(res == 0) return MDBS_DEVICE_NOT_RESPOND;
    if(res < 0) return res;
    //
    fast_memmove(pdu, &adu[7], rlen);
    //
    return MDBS_OK;
}

/**
 * @brief Запись/чтение PDU Modbus RTU/TCP
 */
static mdbs_result_t pdu_data_exchange(int handle, uint8_t adr, uint16_t wlen, uint16_t rlen, uint8_t *pdu) {
    //
    if((socket_handle_type(handle) == HDL_SERIAL) || (handle & RTU_OVER_TCP_FLAG)) {
        // подключение через последовательный порт или через сетевой сокет в режиме "RTU over TCP"
        return pdu_rtu_data_exchange(handle, adr, wlen, rlen, pdu);
    } else {
        // подключение через сетевой сокет
        return pdu_tcp_data_exchange(handle, adr, wlen, rlen, pdu);
    }
}

/**
 * @brief Чтение регистров
 */
static mdbs_result_t internal_read_registers(int handle, uint8_t adr, uint8_t fc, uint16_t startadr, uint16_t cnt, uint16_t *reg) {
    if(socket_handle_type(handle) == HDL_INVALID) return MDBS_BAD_PARAMS;
    if((cnt == 0)||(cnt > MODBUS_MAX_READ_REGISTERS)) return MDBS_BAD_PARAMS;
    if(!reg) return MDBS_BAD_PARAMS;
    //
    uint8_t pdu[256];
    mdbs_result_t rc;
    //
    pdu[0] = fc;
    pdu[1] = (startadr >> 8) & 0xff;
    pdu[2] = startadr & 0xff;
    pdu[3] = 0;
    pdu[4] = cnt;
    //
    rc = pdu_data_exchange(handle, adr, 5, cnt*2 + 2, pdu);
    if(rc != MDBS_OK) return rc;
    //
    // перенос результата с изменение порядка байт: (Hi)(Lo) -> (Lo)(Hi)
    for(int i=0; i < cnt; i++) {
        reg[i] = ((uint16_t)pdu[i*2 + 2] << 8) | pdu[i*2 + 3];
    }
    //
    return MDBS_OK;
}

/**
 * @brief Чтение битовых флагов
 */
static mdbs_result_t internal_read_flags(int handle, uint8_t adr, uint8_t fc, uint16_t startadr, uint16_t cnt, unsigned char *data) {
    if(socket_handle_type(handle) == HDL_INVALID) return MDBS_BAD_PARAMS;
    if((cnt == 0)||(cnt > MODBUS_MAX_READ_BITS)) return MDBS_BAD_PARAMS;
    if(!data) return MDBS_BAD_PARAMS;
    //
    uint8_t pdu[256];
    mdbs_result_t rc;
    //
    pdu[0] = fc;
    pdu[1] = (startadr >> 8) & 0xff;
    pdu[2] = startadr & 0xff;
    pdu[3] = (cnt >> 8) & 0xff;
    pdu[4] = cnt & 0xff;
    //
    if(cnt % 8) cnt = cnt/8 + 1; else cnt = cnt/8;          // размер битового массива ответа
    rc = pdu_data_exchange(handle, adr, 5, cnt + 2, pdu);
    if(rc != MDBS_OK) return rc;
    //
    // перенос результата
    fast_memmove(data, &pdu[2], cnt);
    //
    return MDBS_OK;
}

/**
 * @brief Чтение значений регистров хранения (Holding Registers).
 * Функция выполняет чтение значений регистров "Holding Registers" начиная с адреса regadr в количестве cnt.
 * При успешном выполнении прочитанные данные помещаются в массив по указателю reg. Размер массива
 * должен быть не менее cnt*2 байт. Порядок байт в прочитанных регистрах: "Little Endian".
 * При успешном выполнении функция возвращает значение MDBS_OK, иначе - код ошибки.
 * @param handle Дескриптор последовательного порта или соединения TCP.
 * @param adr Адрес устройства Modbus.
 * @param startadr Адрес первого читаемого регистра (0..65535).
 * @param cnt Количество читаемых регистров (1..125).
 * @param reg Указатель на массив данных.
 * @return mdbs_result_t Код завершения.
 */
mdbs_result_t ModbusReadHoldingRegisters(int handle, uint8_t adr, uint16_t startadr, uint16_t cnt, uint16_t *reg) {
    return internal_read_registers(handle, adr, MODBUS_FC_READ_HOLDING_REGISTERS, startadr, cnt, reg);
}

/**
 * @brief Чтение значений входных регистров (Input Registers).
 * Функция выполняет чтение значений регистров "Input Registers" начиная с адреса regadr в количестве cnt.
 * При успешном выполнении прочитанные данные помещаются в массив по указателю reg. Размер массива
 * должен быть не менее cnt*2 байт. Порядок байт в прочитанных регистрах: "Little Endian".
 * При успешном выполнении функция возвращает значение MDBS_OK, иначе - код ошибки.
 * @param handle Дескриптор последовательного порта или соединения TCP.
 * @param adr Адрес устройства Modbus.
 * @param startadr Адрес первого читаемого регистра (0..65535).
 * @param cnt Количество читаемых регистров (1..125).
 * @param reg Указатель на массив данных.
 * @return mdbs_result_t Код завершения.
 */
mdbs_result_t ModbusReadInputRegisters(int handle, uint8_t adr, uint16_t startadr, uint16_t cnt, uint16_t *reg) {
    return internal_read_registers(handle, adr, MODBUS_FC_READ_INPUT_REGISTERS, startadr, cnt, reg);
}

/**
 * @brief Чтение значений дискретных входов (Discrete Inputs).
 * Функция выполняет чтение состояния до 2000 последовательно расположенных дискретных входов.
 * При успешном выполнении функция возвращает значение MDBS_OK, прочитанные данные помещаются в массив по указателю data.
 * Размер массива data должен быть не менее чем (cnt/8+1) байт.
 * В случае ошибки функция возврвращает её код.
 * @param handle Дескриптор последовательного порта или соединения TCP.
 * @param adr Адрес устройства Modbus.
 * @param startadr Адрес первого читаемого дискретного входа (0..65535).
 * @param cnt Количество читаемых входов (1..2000)
 * @param data Указатель на массив данных.
 * @return mdbs_result_t 
 */
mdbs_result_t ModbusReadDiscreteInputs(int handle, uint8_t adr, uint16_t startadr, uint16_t cnt, unsigned char *data) {
    return internal_read_flags(handle, adr, MODBUS_FC_READ_DISCRETE_INPUTS, startadr, cnt, data);
}

/**
 * @brief Чтение значений дискретных выходов (Coils).
 * Функция выполняет чтение состояния до 2000 последовательно расположенных дискретных выходов (coils).
 * При успешном выполнении функция возвращает значение MDBS_OK, прочитанные данные помещаются в массив по указателю data.
 * Размер массива data должен быть не менее чем (cnt/8+1) байт.
 * В случае ошибки функция возврвращает её код.
 * @param handle Дескриптор последовательного порта или соединения TCP.
 * @param adr Адрес устройства Modbus.
 * @param startadr Адрес первого читаемого дискретного выхода (0..65535).
 * @param cnt Количество читаемых выходов (1..2000)
 * @param data Указатель на массив данных.
 * @return mdbs_result_t 
 */
mdbs_result_t ModbusReadCoils(int handle, uint8_t adr, uint16_t startadr, uint16_t cnt, unsigned char *data) {
    return internal_read_flags(handle, adr, MODBUS_FC_READ_COILS, startadr, cnt, data);
}

/**
 * @brief Запись значения одного регистра хранения (Holding Register).
 * Функция выполняет запись значения value в регистр с адресом startadr.
 * Для записи используется функция 0x06.
 * Перед записью функция изменяет порядок байтов значения value.
 * При успешном выполнении функция возвращает значение MDBS_OK, иначе - код ошибки.
 * @param handle Дескриптор последовательного порта или соединения TCP.
 * @param adr Адрес устройства Modbus.
 * @param startadr Адрес регистра записываемого регистра (0..65535).
 * @param value Записываемое значение.
 * @return mdbs_result_t 
 */
mdbs_result_t ModbusWriteSingleRegister(int handle, uint8_t adr, uint16_t startadr, uint16_t value) {
    if(socket_handle_type(handle) == HDL_INVALID) return MDBS_BAD_PARAMS;
    //
    uint8_t pdu[8];
    mdbs_result_t rc;
    //
    pdu[0] = MODBUS_FC_WRITE_SINGLE_REGISTER;
    pdu[1] = (startadr >> 8) & 0xff;
    pdu[2] = startadr & 0xff;
    pdu[3] = (value >> 8) & 0xff;
    pdu[4] = value & 0xff;
    //
    rc = pdu_data_exchange(handle, adr, 5, 5, pdu);
    if(rc != MDBS_OK) return rc;
    //
    return MDBS_OK;
}

/**
 * @brief Запись значений нескольких регистров хранения (Holding Registers).
 * Функция выполняет запись значений из массива data в регистры хранения, начиная с адреса startadr.
 * Для записи используется функция 0x10.
 * При записи функция изменяет порядок байтов записываемых значений.
 * При успешном выполнении функция возвращает значение MDBS_OK, иначе - код ошибки.
 * @param handle Дескриптор последовательного порта или соединения TCP.
 * @param adr Адрес устройства Modbus.
 * @param startadr Начальный адрес регистра (0..65535).
 * @param cnt Количество записываемых регистров (1..123).
 * @param data Указатель на массив записываемых данных.
 * @return mdbs_result_t 
 */
mdbs_result_t ModbusWriteRegisters(int handle, uint8_t adr, uint16_t startadr, uint16_t cnt, uint16_t *data) {
    if(socket_handle_type(handle) == HDL_INVALID) return MDBS_BAD_PARAMS;
    if((cnt == 0)||(cnt > MODBUS_MAX_WRITE_REGISTERS)) return MDBS_BAD_PARAMS;
    if(!data) return MDBS_BAD_PARAMS;
    //
    uint8_t pdu[256];
    mdbs_result_t rc;
    //
    pdu[0] = MODBUS_FC_WRITE_MULTIPLE_REGISTERS;
    pdu[1] = (startadr >> 8) & 0xff;        // hi(startadr)
    pdu[2] = startadr & 0xff;               // lo(startadr)
    pdu[3] = 0;                             // hi(cnt)
    pdu[4] = cnt;                           // lo(cnt)
    pdu[5] = cnt * 2;                       // количество байт
    // копирование данных с перестановкой порядка байт
    for(int i=0; i < cnt; i++) {
        pdu[i*2 + 6] = (data[i] >> 8) & 0xff;
        pdu[i*2 + 7] = data[i] & 0xff;
    }
    //
    rc = pdu_data_exchange(handle, adr, cnt*2 + 6, 5, pdu);
    if(rc != MDBS_OK) return rc;
    //
    return MDBS_OK;
}

/**
 * @brief Запись значения одного дискретного выхода (coil).
 * Для записи используется функция 0x05.
 * При успешном выполнении функция возвращает значение MDBS_OK, иначе - код ошибки.
 * @param handle Дескриптор последовательного порта или соединения TCP.
 * @param adr Адрес устройства Modbus.
 * @param startadr Начальный адрес дискретного выхода (0..65535).
 * @param value Значение: 0 - выключить, !0 - включить.
 * @return mdbs_result_t 
 */
mdbs_result_t ModbusWriteSingleCoil(int handle, uint8_t adr, uint16_t startadr, uint16_t value) {
    if(socket_handle_type(handle) == HDL_INVALID) return MDBS_BAD_PARAMS;
    //
    uint8_t pdu[8];
    mdbs_result_t rc;
    //
    pdu[0] = MODBUS_FC_WRITE_SINGLE_COIL;
    pdu[1] = (startadr >> 8) & 0xff;
    pdu[2] = startadr & 0xff;
    pdu[3] = value?0xff:0;
    pdu[4] = 0;
    //
    rc = pdu_data_exchange(handle, adr, 5, 5, pdu);
    if(rc != MDBS_OK) return rc;
    //
    return MDBS_OK;
}

/**
 * @brief Запись значений дискретных выходов (coils).
 * Каждый дискретный выход может находиться в состоянии "1" (включен) или "0" (выключен).
 * Записываемые состояния определяются содержимым массива coils. Логическая "1" в битовой
 * позиции массива coils требует выключение соответствующего дискретного выхода.
 * Для записи используется функция 0x0f.
 * @param handle Дескриптор последовательного порта или соединения TCP.
 * @param adr Адрес устройства Modbus.
 * @param startadr Начальный адрес дискретных выходов (0..65535).
 * @param cnt Количество записываемых дискретных выходов (1..1968)
 * @param coils Указатель на массив значений дискретных выходов
 * @return mdbs_result_t 
 */
mdbs_result_t ModbusWriteCoils(int handle, uint8_t adr, uint16_t startadr, uint16_t cnt, uint8_t *coils) {
    if(socket_handle_type(handle) == HDL_INVALID) return MDBS_BAD_PARAMS;
    if((cnt == 0)||(cnt > MODBUS_MAX_WRITE_BITS)) return MDBS_BAD_PARAMS;
    if(!coils) return MDBS_BAD_PARAMS;
    //
    uint8_t pdu[256];
    mdbs_result_t rc;
    //
    pdu[0] = MODBUS_FC_WRITE_MULTIPLE_COILS;
    pdu[1] = (startadr >> 8) & 0xff;
    pdu[2] = startadr & 0xff;
    pdu[3] = (cnt >> 8) & 0xff;
    pdu[4] = cnt & 0xff;
    if(cnt % 8) cnt = cnt/8 + 1; else cnt = cnt/8;
    pdu[5] = cnt;   // max 246
    fast_memmove(&pdu[6], coils, cnt);
    //
    rc = pdu_data_exchange(handle, adr, cnt + 6, 5, pdu);
    if(rc != MDBS_OK) return rc;
    //
    return MDBS_OK;
}

/**
 * @brief Запись и последующее чтение значений нескольких регистров хранения (Holding Registers).
 * Функция выполняет запись wcnt значений из массива wdata в регистры хранения, начиная с адреса wadr.
 * При записи функция изменяет порядок байтов записываемых значений.
 * Функция выполняет чтение значений регистров "Holding Registers" начиная с адреса radr в количестве rcnt.
 * При успешном выполнении прочитанные данные помещаются в массив по указателю rdata. Размер массива
 * должен быть не менее cnt*2 байт. Порядок байт в прочитанных регистрах: "Little Endian".
 * Для чтения и записи используется функция 0x17.
 * При успешном выполнении функция возвращает значение MDBS_OK, иначе - код ошибки.
 * @param handle Дескриптор последовательного порта или соединения TCP.
 * @param adr Адрес устройства Modbus.
 * @param radr Начальный адрес регистра чтения (0..65535).
 * @param rcnt Количество читаемых регистров (1..125).
 * @param rdata Указатель на массив прочитанных данных.
 * @param wadr Начальный адрес регистра запис (0..65535).
 * @param wcnt Количество записываемых регистров (1..125).
 * @param wdata Указатель на массив записываемых данных.
 * @return mdbs_result_t 
 */
mdbs_result_t ModbusReadWriteRegisters(int handle, uint8_t adr, uint16_t radr, uint16_t rcnt, uint16_t *rdata, uint16_t wadr, uint16_t wcnt, uint16_t *wdata) {
    if(socket_handle_type(handle) == HDL_INVALID) return MDBS_BAD_PARAMS;
    if((rcnt == 0)||(rcnt > MODBUS_MAX_RW_READ_REGISTERS)) return MDBS_BAD_PARAMS;
    if((wcnt == 0)||(rcnt > MODBUS_MAX_RW_WRITE_REGISTERS)) return MDBS_BAD_PARAMS;
    if(!rdata) return MDBS_BAD_PARAMS;
    if(!wdata) return MDBS_BAD_PARAMS;
    //
    uint8_t pdu[256];
    mdbs_result_t rc;
    //
    pdu[0] = MODBUS_FC_WRITE_AND_READ_REGISTERS;
    pdu[1] = (radr >> 8) & 0xff;                    // адрес чтения
    pdu[2] = radr & 0xff;                           //
    pdu[3] = 0;                                     // количество регистров чтения
    pdu[4] = rcnt;                                  //
    pdu[5] = (wadr >> 8) & 0xff;                    // адрес записи
    pdu[6] = wadr & 0xff;                           //
    pdu[7] = 0;                                     // количество регистров записи
    pdu[8] = wcnt;                                  //
    pdu[9] = wcnt*2;                                // количество записываемых байт
    // копирование записываемых данных с перестановкой порядка байт
    for(int i=0; i < wcnt; i++) {
        pdu[i*2 + 10] = (wdata[i] >> 8) & 0xff;
        pdu[i*2 + 11] = wdata[i] & 0xff;
    }
    //
    rc = pdu_data_exchange(handle, adr, wcnt*2 + 10, rcnt*2 + 2, pdu);
    if(rc != MDBS_OK) return rc;
    //
    // перенос результата с изменение порядка байт: (Hi)(Lo) -> (Lo)(Hi)
    for(int i=0; i < rcnt; i++) {
        rdata[i] = ((uint16_t)pdu[i*2 + 2] << 8) | pdu[i*2 + 3];
    }
    //
    return MDBS_OK;
}
