/**
 * @file modbus_data.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции преобразования типов для Modbus.
 * @version 0.1
 * @date 2024-04-02
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <stdint.h>
#include <modbus.h>

union modbus_float_u {
    float f;
    //
    struct {
        uint8_t a;
        uint8_t b;
        uint8_t c;
        uint8_t d;
    };
};

/**
 * @brief Преобразование значения 2-х последовательных регистров Modbus (4 байта) в значение float
 * без преобразования (ABCD)
 * @param src Указатель на первый регистр
 * @return float 
 */
float ModbusGetFloatABCD(const uint16_t *src) {
    uint8_t *in = (uint8_t *) src;
    union modbus_float_u mf;
    //
    mf.a = in[0];
    mf.b = in[1];
    mf.c = in[2];
    mf.d = in[3];
    //
    return mf.f;
}

/**
 * @brief Преобразование значения 2-х последовательных регистров Modbus (4 байта) в значение float
 * в обратном порядке (DCBA)
 * @param src Указатель на первый регистр
 * @return float 
 */
float ModbusGetFloatDCBA(const uint16_t *src) {
    uint8_t *in = (uint8_t *) src;
    union modbus_float_u mf;
    //
    mf.d = in[0];
    mf.c = in[1];
    mf.b = in[2];
    mf.a = in[3];
    //
    return mf.f;
}

/**
 * @brief Преобразование значения 2-х последовательных регистров Modbus (4 байта) в значение float
 * с изменённым порядком байт в регистрах (BADC)
 * @param src 
 * @return float 
 */
float ModbusGetFloatBADC(const uint16_t *src) {
    uint8_t *in = (uint8_t *) src;
    union modbus_float_u mf;
    //
    mf.b = in[0];
    mf.a = in[1];
    mf.d = in[2];
    mf.c = in[3];
    //
    return mf.f;
}

/**
 * @brief Преобразование значения 2-х последовательных регистров Modbus (4 байта) в значение float
 * с изменённым порядком регистров (CDAB)
 * @param src 
 * @return float 
 */
float ModbusGetFloatCDAB(const uint16_t *src) {
    uint8_t *in = (uint8_t *) src;
    union modbus_float_u mf;
    //
    mf.c = in[0];
    mf.d = in[1];
    mf.a = in[2];
    mf.b = in[3];
    //
    return mf.f;
}

/**
 * @brief Преобразование значения float в два последовательных регистра Modbus (4 байта)
 * без изменения порядка (ABCD)
 * @param f 
 * @param dest 
 */
void ModbusSetFloatABCD(float f, uint16_t *dest) {
    uint8_t *out = (uint8_t *) dest;
    union modbus_float_u mf;
    mf.f = f;
    //
    out[0] = mf.a;
    out[1] = mf.b;
    out[2] = mf.c;
    out[3] = mf.d;
}

/**
 * @brief Преобразование значения float в два последовательных регистра Modbus (4 байта)
 * в обратном порядке (DCBA)
 * @param f 
 * @param dest 
 */
void ModbusSetFloatDCBA(float f, uint16_t *dest) {
    uint8_t *out = (uint8_t *) dest;
    union modbus_float_u mf;
    mf.f = f;
    //
    out[0] = mf.d;
    out[1] = mf.c;
    out[2] = mf.b;
    out[3] = mf.a;
}

/**
 * @brief Преобразование значения float в два последовательных регистра Modbus (4 байта)
 * с изменённым порядком байт в регистрах (BADC)
 * @param f 
 * @param dest 
 */
void ModbusSetFloatBADC(float f, uint16_t *dest) {
    uint8_t *out = (uint8_t *) dest;
    union modbus_float_u mf;
    mf.f = f;
    //
    out[0] = mf.b;
    out[1] = mf.a;
    out[2] = mf.d;
    out[3] = mf.c;
}

/**
 * @brief Преобразование значения float в два последовательных регистра Modbus (4 байта)
 * с изменённым порядком регистров (CDAB)
 * @param f 
 * @param dest 
 */
void ModbusSetFloatCDAB(float f, uint16_t *dest) {
    uint8_t *out = (uint8_t *) dest;
    union modbus_float_u mf;
    mf.f = f;
    //
    out[0] = mf.c;
    out[1] = mf.d;
    out[2] = mf.a;
    out[3] = mf.b;
}
