/**
 * @file modbus_serv.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции Modbus Server.
 *
 * @version 0.1
 * @date 2024-04-03
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <stdint.h>
#include <satser.h>
#include <syscallp.h>

#include <modbus.h>

/**
 * @brief Начало работы сервера MODBUS RTU. Для работы сервера используется указанный последовательный порт.
 * Последовательный порт должен быть предварительно открыт функцией SerialOpen().
 * В качестве параметров передаются адреса и количество регистров хранения и входных регистров используемых для
 * работы сервера.
 * При успешном выполнении функция возвращает значение 1, при ошибке - значение 0.
 * @param handle Идентификатор последовательного порта.
 * @param hld Указатель на область регистров хранения (Holding Registers).
 * @param hldcnt Количество регистров хранения
 * @param inp Указатель на область входных регистров (Input Registers)
 * @param inpcnt Количество входных регистров
 * @return int Результат выполнения: 1 - успех, 0 - ошибка
 */
int StartModbusRTUServer(int handle, uint16_t *hld, uint16_t hldcnt, uint16_t *inp, uint16_t inpcnt) {
    if(!SerialHandleIsValid(handle)) return 0;
    //
    return __internal_syscall(SYSCALL_StartModbusRTUServer, handle, (long)hld, hldcnt, (long)inp, inpcnt, 0);
}

/**
 * @brief Функция возвращает количество корректных запросов к серверу MODBUS RTU работающему на указанном последовательном порту.
 * @param handle Идентификатор последовательного порта.
 * @return unsigned int 
 */
unsigned int ModbusRTUCount(int handle) {
    if(!SerialHandleIsValid(handle)) return 0;
    //
    return __internal_syscall_invoke1(SYSCALL_ModbusRTUCount, handle);
}

/**
 * @brief Завершение работы сервера MODBUS RTU использующего указанный последовательный порт.
 * @param handle Идентификатор последовательного порта.
 */
void StopModbusRTUServer(int handle) {
    if(!SerialHandleIsValid(handle)) return;
    //
    __internal_syscall_invoke1(SYSCALL_StopModbusRTUServer, handle);
}

/**
 * @brief Начало работы сервера MODBUS TCP.
 * В качестве параметров передаются адреса и количество регистров хранения и входных регистров используемых для
 * работы сервера.
 * При успешном выполнении функция возвращает значение 1, при ошибке - значение 0.
 * @param hld Указатель на область регистров хранения (Holding Registers).
 * @param hldcnt Количество регистров хранения
 * @param inp Указатель на область входных регистров (Input Registers)
 * @param inpcnt Количество входных регистров
 * @return int Результат выполнения: 1 - успех, 0 - ошибка
 */
int StartModbusTCPServer(uint16_t *hld, uint16_t hldcnt, uint16_t *inp, uint16_t inpcnt) {
    return __internal_syscall(SYSCALL_StartModbusTCPServer, (long)hld, hldcnt, (long)inp, inpcnt, 0, 0);
}

/**
 * @brief Функция возвращает количество корректных запросов к серверу MODBUS TCP
 * @return unsigned int 
 */
unsigned int ModbusTCPCount() {
    return __internal_syscall_invoke1(SYSCALL_ModbusTCPCount, 0);
}

/**
 * @brief Завершение работы сервера MODBUS TCP.
 */
void StopModbusTCPServer(void) {
    __internal_syscall_invoke1(SYSCALL_StopModbusTCPServer, 0);
}
