/**
 * @file menu.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Системное меню.
 * @version 0.1
 * @date 2024-03-28
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include <stdlib.h>
#include <string.h>
#include <syscallp.h>
#include <satmenu.h>

/**
 * @brief Начало системного меню
 */
sysmenu_item_t *sysmenu_head = NULL;

/**
 * @brief Начало описаний сигналов
 */
signame_item_t *signame_head = NULL;

/**
 * @brief Возвращает указатель на последний элемент меню
 * 
 * @return sysmenu_item_t* 
 */
static sysmenu_item_t *get_last_item() {
    sysmenu_item_t *mi = sysmenu_head;
    //
    while(mi) {
        if(!mi->next) return mi;
        mi = mi->next;
    }
    return mi;
}

/**
 * @brief Обновление системного меню
 */
static void sysmenu_update() {
    __internal_syscall_invoke1(SYSCALL_SystemMenuUpd, (long)sysmenu_head);
}

/**
 * @brief Отображение системного меню.
 * При вызове функции будет показано системное меню контроллера.
 */
void SystemMenu(void) {
    __internal_syscall_invoke1(SYSCALL_SystemMenu, 0);
}

/**
 * @brief Проверка статуса системного меню.
 * Если системное меню отображается, то функция возвращает 1, в противном случае - 0.
 * @return int 
 */
int SystemMenuVisible(void) {
    return __internal_syscall_invoke1(SYSCALL_SystemMenuVis, 0);
}

/**
 * @brief Добавление пункта в системное меню.
 * Функция добавляет в системное меню пункт связанный с переменной типа "int".
 * @param var Указатель на переменную связанную с пунктом меню
 * @param title Имя пункта меню
 * @param min Минимальное значение переменной
 * @param max Максимальное значение переменной
 * @return sysmenu_item_t* Указатель на новый пункт меню
 */
sysmenu_item_t *AddSystemMenuIntItem(const char *title, int *var, int min, int max) {
    sysmenu_item_t *lastItem = get_last_item();
    sysmenu_item_t *newItem = malloc(sizeof(sysmenu_item_t));
    if(!newItem) return NULL;
    //
    newItem->type      = 0;
    newItem->precision = 1;
    if(CPU_MEM_IS_RAM(title)) {
        newItem->title  = strdup(title);
    } else {
        newItem->title  = (char *)title;
    }
    newItem->var    = (sysmenu_value_t *)var;
    newItem->min.i  = min;
    newItem->max.i  = max;
    newItem->next   = NULL;
    newItem->prev   = lastItem;
    //
    if(lastItem) lastItem->next = newItem; else sysmenu_head = newItem;
    sysmenu_update();
    //
    return newItem;
}

/**
 * @brief Добавление пункта в системное меню.
 * Функция добавляет в системное меню пункт связанный с переменной типа "float".
 * @param var Указатель на переменную связанную с пунктом меню
 * @param title Имя пункта меню
 * @param min Минимальное значение переменной
 * @param max Максимальное значение переменной
 * @param precision Количество знаков после запятой при отображении значения
 * @return sysmenu_item_t* Указатель на новый пункт меню
 */
sysmenu_item_t *AddSystemMenuFloatItem(const char *title, float *var, float min, float max, int precision) {
    sysmenu_item_t *lastItem = get_last_item();
    sysmenu_item_t *newItem = malloc(sizeof(sysmenu_item_t));
    if(!newItem) return NULL;
    //
    newItem->type = 1;
    newItem->precision = precision;
    if(CPU_MEM_IS_RAM(title)) {
        newItem->title  = strdup(title);
    } else {
        newItem->title  = (char *)title;
    }
    newItem->var    = (sysmenu_value_t *)var;
    newItem->min.f  = min;
    newItem->max.f  = max;
    newItem->next   = NULL;
    newItem->prev   = lastItem;
    //
    if(lastItem) lastItem->next = newItem; else sysmenu_head = newItem;
    sysmenu_update();
    //
    return newItem;
}

/**
 * @brief Добавление пункта в системное меню.
 * Функция добавляет в системное меню пункт связанный с переменной типа "int", которая может иметь значения 0 (нет) или 1 (да).
 * @param var Указатель на переменную связанную с пунктом меню
 * @param title Имя пункта меню
 * @return sysmenu_item_t* Указатель на новый пункт меню
 */
sysmenu_item_t *AddSystemMenuBoolItem(const char *title, int *var) {
    sysmenu_item_t *lastItem = get_last_item();
    sysmenu_item_t *newItem = malloc(sizeof(sysmenu_item_t));
    if(!newItem) return NULL;
    //
    newItem->type   = 2;
    if(CPU_MEM_IS_RAM(title)) {
        newItem->title  = strdup(title);
    } else {
        newItem->title  = (char *)title;
    }
    newItem->var    = (sysmenu_value_t *)var;
    newItem->min.i  = 0;
    newItem->max.i  = 1;
    newItem->next   = NULL;
    newItem->prev   = lastItem;
    //
    if(lastItem) lastItem->next = newItem; else sysmenu_head = newItem;
    sysmenu_update();
    //
    return newItem;
}

/**
 * @brief Удаление всех пунктов из системного меню.
 */
void ClearSystemMenu(void) {
    sysmenu_item_t *next;
    sysmenu_item_t *tmp = sysmenu_head;
    //
    while(tmp) {
        next = tmp->next;
        if(CPU_MEM_IS_RAM(tmp->title)) free(tmp->title);
        free(tmp);
        tmp = next;
    }
    sysmenu_head = NULL;
    sysmenu_update();
}

/**
 * @brief Удаление пункта из системного меню.
 * 
 * @param item Указатель на пункт меню.
 */
void DeleteSystemMenuItem(sysmenu_item_t *item) {
    if(!item) return;
    //
    if(item->prev) item->prev->next = item->next;
    if(item->next) item->next->prev = item->prev;
    //
    // удаление первого пункта меню
    if(!item->prev) sysmenu_head = item->next;
    //
    if(CPU_MEM_IS_RAM(item->title)) free(item->title);
    free(item);
    //
    sysmenu_update();
}

/**
 * @brief Поиск элемента описания имени сигнала
 */
static signame_item_t *find_signame(int type, int index) {
    signame_item_t *cur = signame_head;
    //
    while (cur) {
        if((cur->type == type) && (cur->index == index)) return cur;
        cur = cur->next;
    }
    return NULL;
}

/**
 * @brief Функция позволяет задать имена сигналов, отображаемые в системных меню.
 * @param type Тип сигнала: 0-дискретные входы,1-дискретные выходы,2-аналоговые входы,3-аналоговые выходы,4-датчики температуры
 * @param index Номер сигнала
 * @param name Наименование сигнала
 * @return signame_item_t* Указатель на новый пункт описания имени сигнала
 */
signame_item_t *SetSigName(int type, int index, const char *name) {
    signame_item_t *item;
    //
    item = find_signame(type, index);
    if(item) {
        // описание уже есть
        if(CPU_MEM_IS_RAM(item->name)) free(item->name);
        if(CPU_MEM_IS_RAM(name)) {
            item->name = strdup(name);
        } else {
            item->name = (char *)name;
        }
        return item;
    } 
    // такого описания пока нет
    if(!name) return NULL;
    //
    item = malloc(sizeof(signame_item_t));
    if(!item) return NULL;
    item->type = type;
    item->index = index;
    if(CPU_MEM_IS_RAM(name)) {
        item->name = strdup(name);
    } else {
        item->name = (char *)name;
    }
    item->next = signame_head;
    //
    signame_head = item;
    __internal_syscall_invoke1(SYSCALL_SetSigName, (long)signame_head);
    //
    return item;
}
