/**
 * @file netvars.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с сетевыми переменными.
 * @version 0.1
 * @date 2024-04-21
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <satthreads.h>
#include <satparams.h>
#include <satsock.h>
#include <stdlib.h>
#include <string.h>
#include <hash.h>

#include <netvars.h>

/**
 * @brief UDP-порт используемый для приёма и отправки сетевых переменных.
 */
uint16_t NetVarsPort = 3301;

/**
 * @brief Идентификатор группы сетевых переменных.
 */
uint16_t NetVarsGroup = 1;

/**
 * @brief Счётчик принятых пакетов, содержащих сетевые переменные.
 */
unsigned int NetVarsPacketCount;

#define UNKNOWN_SERIAL  0xffffffff

/**
 * @brief Серийный номер контроллера
 */
static uint32_t NetVarsSerialNumber = UNKNOWN_SERIAL;

/**
 * @brief Подпись
 */
static const char NetVarsSecret[16]   = "z49rNFLJmKV2Vd8m";

#define NET_VARS_SIGNATURE  0x5654454e
#define NV_RECBUFSIZE       1024

/**
 * @brief Заголовок пакета сетевых переменных.
 */
struct nv_header {
    uint32_t    signature;                      // NETV
    uint16_t    size;                           // размер пакета
    uint16_t    group;                          // группа
    uint32_t    sender;                         // серийный номер отправителя
    uint32_t    rnd;                            // случайное число
} __attribute__((packed));
typedef struct nv_header    nv_header_t;

/**
 * @brief Сокет, используемый при работе с сетевыми переменными.
 */
static int nvsocket = 0;

/**
 * @brief Признак запущенного потока приёма переменных
 */
static volatile int nv_thread_running   = 0;

/**
 * @brief Признак необходимости остановки потока приёма переменных
 */
static volatile int nv_thread_terminate = 0;

/**
 * @brief Идентификатор потока приёма переменных
 */
static thrd_t nv_thread;

// Сохранённые значения параметров вызова NetVarsStart
static volatile int32_t    *nv_def_nums;
static volatile int32_t    *nv_def_vals;
static volatile uint32_t   *nv_def_sns;
static volatile uint16_t    nv_def_cnt;

// формат пакета:
// 'NETV'     - сигнатура                                           4 байта     0
// Size       - размер всего пакета с переменными включая сигнатуру 2 байта     4
// Group      - идентификатор группы                                2 байта     6
// SN         - серийный номер отправителя                          4 байта     8
// RND        - случайное число                                     4 байта     12

// Var1Number - номер переменной 1                                  4 байта     16+8*cnt
// Var1Value  - значение переменной 1                               4 байта     20+8*cnt
// Var2Number - номер переменной 2                                  4 байта     24+8*cnt
// Var2Value  - значение переменной 2                               4 байта     28+8*cnt
// ...
// ...
// MD5        - дайджест MD5 блока от сигнатуры (включительно) до   16 байт     16+8*cnt
//              MD5 (не включая) с дописанным в конце паролем

static int nv_thread_func(void *p) {
    //
    MD5Context ctx;
    //
    // выделение памяти для приёмого буфера
    uint8_t *nv_buf = malloc(NV_RECBUFSIZE);
    if(!nv_buf) return 0;
    nv_header_t *header = (nv_header_t *)nv_buf;
    //
    netvars_func cb = p;
    union {
        uint8_t     byte[4];
        uint32_t    val;
    } ser;
    //
    while(!nv_thread_terminate) {
        int len = SocketReadLength(nvsocket);
        if(len <= 0) {
            thrd_sleep(100);
            continue;
        }
        if(len > NV_RECBUFSIZE) {
            // переполнение буфера
            SocketClear(nvsocket);
            thrd_sleep(100);
            continue;
        }
        // читаем принятое
        SocketRead(nvsocket, nv_buf, len);
        //
        // проверка минимальной длины
        if(len < 40) continue;
        //
        // проверка сигнатуры
        if(header->signature != NET_VARS_SIGNATURE) continue;
        //
        // проверка длины
        if(header->size != len) continue;
        //
        // проверка MD5
        md5Init(&ctx);
        md5Update(&ctx, nv_buf, len - 16);
        md5Update(&ctx, (uint8_t *)NetVarsSecret, 16);
        md5Finalize(&ctx);
        if(memcmp(ctx.digest, &nv_buf[len-16], 16)) continue;
        //
        NetVarsPacketCount++;
        //
        ser.val = header->sender;
        cb(ser.byte[3] * 100000u + ser.byte[2]*1000u + ser.byte[1]*256u + ser.byte[0], header->group, (len-32)/8, (net_var_t *)&nv_buf[16]);
    }
    // освобождаем память
    free(nv_buf);
    //
    nv_thread_running = 0;
    return 0;
}

static int nv_init(void) {
    nvsocket = CreateSocket(SOCK_DGRAM);
    if(!nvsocket) return 0;
    if(!SocketBind(nvsocket, NetVarsPort)) return 0;
    int res = SocketConnectBlock(nvsocket, INADDR_BROADCAST_S, NetVarsPort, 200);
    return res;
}

/**
 * @brief Передать значение одной или нескольких сетевых переменных.
 * При отправке используются значения глобальных переменных NetVarsPort и NetVarsGroup.
 * Функция возвращает 1 при успешной отправке или 0 при ошибке.
 * @param nums Массив идентификаторов (номеров) сетевых переменных.
 * @param vals Массив значений сетевых переменных.
 * @param cnt Количество передаваемых переменных.
 * @return int 
 */
int NetVarsSend(int *nums, int *vals, unsigned short cnt) {
    if(!nums) return 0;
    if(!vals) return 0;
    if(!cnt) return 1;
    //
    if(!SocketHandleIsValid(nvsocket)) {
        if(!nv_init()) return 0;
    }
    //
    uint16_t bufSize = 32 + 8*cnt;
    uint8_t *buf = alloca(bufSize);
    nv_header_t *header = (nv_header_t *)buf;
    //
    header->signature = NET_VARS_SIGNATURE;
    header->size = bufSize;
    header->group = NetVarsGroup;
    if(NetVarsSerialNumber == UNKNOWN_SERIAL) GetPLCParameter(PARAM_ID_NVSERIAL, &NetVarsSerialNumber);
    header->sender = NetVarsSerialNumber;
    header->rnd = rand();
    //
    for(int i=0; i<cnt; i++) {
        memcpy(&buf[16 + i*8], &nums[i], 4);
        memcpy(&buf[20 + i*8], &vals[i], 4);
    }
    //
    // memcpy(&buf[16 + cnt*8], NetVarsSecret, 16);    // secret salt
    //
    MD5Context ctx;                                 // md5
    md5Init(&ctx);
    md5Update(&ctx, buf, bufSize-16);
    md5Update(&ctx, (uint8_t *)NetVarsSecret, 16);
    md5Finalize(&ctx);
    memcpy(&buf[16 + cnt*8], ctx.digest, 16);
    //
    struct sockaddr_in addr;
    addr.sin_addr.S_un.S_addr = INADDR_BROADCAST;
    addr.sin_port = NetVarsPort;
    //
    int res = SocketWriteTo(nvsocket, buf, bufSize, &addr);
    if(res > 0) return 1;
    return 0;
}

/**
 * @brief Запуск работы модуля приёма сетевых переменных.
 * Функция запускает выполнение отдельного потока приёма сетевых переменных.
 * При приёме одной или нескольких сетевых переменных производится вызов функции cb, которая должна выполнить
 * соответствующую обработку полученных значений. Вызов функции выполняется в контексте потока обрабатывающего приём сетевых переменных.
 * При приёме используется значение глобальной переменной NetVarsPort (номер UDP порта).
 * Функция обнуляет значение количества принятых сетевых пакетов NetVarsPacketCount.
 * Функция возвращает 1 при успешном выполнении или 0 при ошибке.
 * @param cb Функция - обработчик получения сетевых переменных.
 * @return int  1 - успех, 0 - ошибка.
 */
int NetVarsStartEx(netvars_func cb) {
    if(!cb) return 0;
    if(nv_thread_running) return 0;                 // поток уже запущен
    //
    // создаём сокет
    if(!SocketHandleIsValid(nvsocket)) {
        if(!nv_init()) return 0;
    }
    // очищаем приёмный буфер
    SocketClear(nvsocket);
    //
    // запускаем поток
    nv_thread_running = 1;
    nv_thread_terminate = 0;
    if(thrd_create(&nv_thread, nv_thread_func, cb) != thrd_success) return 0;
    NetVarsPacketCount = 0;
    //
    return 1;
}

/**
 * @brief Стандартный обработчик приёма сетевых переменных.
 * Используется совместно с вызовом NetVarsStart.
 */
static void def_nvcb(int32_t sender, uint16_t group, uint16_t cnt, net_var_t *vars) {
    if(group != NetVarsGroup) return;
    //
    // перебор принятых переменных
    for(int i=0; i < cnt; i++) {
        // поиск в сохранённых аргументах NetVarsStart
        for(int j=0; j < nv_def_cnt; j++) {
            if(vars[i].id == nv_def_nums[j]) {
                // переменная найдена
                nv_def_vals[j] = vars[i].value;
                if(nv_def_sns) nv_def_sns[j] = sender;
                //
                // считаем, что дубликатов в массиве nv_def_nums быть не должно
                // переходим к следующей принятой переменной
                break;
            }
        }
    }
}

/**
 * @brief Запуск работы модуля приёма сетевых переменных.
 * Функция запускает выполнение отдельного потока приёма сетевых переменных. Идентификаторы принимаемых переменных
 * должны быть указаны в массиве nums. При приёме сетевой переменной её значение помещается в массив vals.
 * Также, если sns!=NULL, то серийный номер отправителя помещается в массив sns.
 * Массивы nums, vals и sns (если он используется) должны иметь глобальную область видимости или выделены функцией malloc.
 * При приёме используются значения глобальных переменных NetVarsPort (номер UDP порта) и NetVarsGroup (номер группы переменных).
 * Функция обнуляет значение количества принятых сетевых пакетов NetVarsPacketCount.
 * Функция возвращает 1 при успешном выполнении или 0 при ошибке.
 * @param nums Массив идентификаторов (номеров) сетевых переменных.
 * @param vals Массив значений сетевых переменных.
 * @param sns Массив серийных номеров отправителей (может быть NULL).
 * @param cnt Количество сетевых переменных.
 * @return int 1 - успех, 0 - ошибка.
 */
int NetVarsStart(int *nums, int *vals, unsigned int *sns, unsigned short cnt) {
    if(nv_thread_running) return 0;
    if(cnt == 0) return 0;
    if(!nums) return 0;
    if(!vals) return 0;
    //
    nv_def_nums = nums;
    nv_def_vals = vals;
    nv_def_sns  = sns;
    nv_def_cnt  = cnt;
    //
    int res = NetVarsStartEx(def_nvcb);
    return res;
}

/**
 * @brief Завершение работы модуля приёма сетевых переменных.
 * Функция останавливает и дожидается завершения работы потока приёма сетевых 
 * переменных и закрывает используемый при работе модуля сетевой сокет.
 */
void NetVarsStop(void) {
    if(nv_thread_running) {
        // останавливаем поток
        nv_thread_terminate = 1;
        // ожидаем завершения его работы
        thrd_join(nv_thread, NULL);
        // закрываем сокет
        CloseSocket(&nvsocket);
    }
}
