/**
 * @file satmem.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с BACKUPRAM и EPROM.
 * @version 0.1
 * @date 2024-03-14
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include <stdint.h>
#include <syscallp.h>

#include <satmem.h>

/**
 * @brief Чтение байта данных из энергонезависимой памяти BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти (0...4095)
 * @return uint8_t
 */
uint8_t BackupRamRead8(uint16_t adr) {
    if(adr > 4095) return 0;
    return __internal_syscall_invoke1(SYSCALL_BackupRamRead8, adr) & 0xff;
}

/**
 * @brief Чтение uint16_t данных из энергонезависимой памяти BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти (0...4094)
 * @return uint16_t
 */
uint16_t BackupRamRead16(uint16_t adr) {
    if(adr > 4094) return 0;
    return ((uint16_t)BackupRamRead8(adr) | ((uint16_t)BackupRamRead8(adr+1) << 8));
}

/**
 * @brief Чтение uint32_t данных из энергонезависимой памяти BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти (0...4092)
 * @return uint32_t 
 */
uint32_t BackupRamRead32(uint16_t adr) {
    if(adr > 4092) return 0;
    return ((uint32_t)BackupRamRead16(adr) | ((uint32_t)BackupRamRead16(adr+2) << 16));
}

/**
 * @brief Чтение массива данных из энергонезависимой памяти BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти (0...4095)
 * @param data Указатель на массив данных
 * @param len Количество читаемых байт
 */
void BackupRamRead(uint16_t adr, unsigned char * data, size_t len) {
    while(len--) {
        *data = BackupRamRead8(adr++);
        data++;
    }
}

/**
 * @brief Сохранение байта данных в энергонезависимой памяти BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти (0...4095)
 * @param val Записываемое значение
 */
void BackupRamWrite8(uint16_t adr, uint8_t val) {
    if(adr > 4095) return;
    __internal_syscall_invoke2(SYSCALL_BackupRamWrite8, adr, val);
}

/**
 * @brief Запись uint16_t в энергонезависимую память BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти (0...4094)
 * @param val Записываемое значение
 */
void BackupRamWrite16(uint16_t adr, uint16_t val) {
    BackupRamWrite8(adr, val & 0xff);
    BackupRamWrite8(adr+1, (val >> 8) & 0xff);
}

/**
 * @brief Запись uint32_t в энергонезависимую память BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти (0...4092)
 * @param val Записываемое значение
 */
void BackupRamWrite32(uint16_t adr, uint32_t val) {
    BackupRamWrite16(adr, val & 0x0000ffff);
    BackupRamWrite16(adr+2, (val >> 16) & 0x0000ffff);
}

/**
 * @brief Запись массива данных в энергонезависимую память BACKUPRAM.
 * 
 * @param adr Адрес в энергонезависимой памяти (0...4095)
 * @param data Указатель на массив данных
 * @param len Количество записываемых байт
 */
void BackupRamWrite(uint16_t adr, unsigned char * data, size_t len) {
    while(len--) {
        BackupRamWrite8(adr++, *data);
        data++;
    }
}

/**
 * @brief Чтение данных из энергонезависимой памяти EPROM.
 * 
 * @param adr Адрес (0..511)
 * @return uint16_t 
 */
uint16_t ReadEprom(uint16_t adr) {
    if(adr > 511) return 0;
    return __internal_syscall_invoke1(SYSCALL_ReadEprom, adr) & 0xffff;
}

/**
 * @brief Сохранение данных в энергонезависимой памяти EPROM.
 * 
 * @param adr Адрес (0..511)
 * @param data Записываемые данные
 */
void WriteEprom(uint16_t adr, uint16_t data) {
    if(adr > 511) return;
    __internal_syscall_invoke2(SYSCALL_WriteEprom, adr, data);
}
