/**
 * @file satscr.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с экраном контроллера.
 * @version 0.2
 * @date 2024-03-13
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include "stdio.h"
#include "string.h"
#include "syscallp.h"

#include "satscr.h"

/**
 * @brief Структура описания экрана
 */
struct _SATPLC_SCREEN    _sat_screen;

/**
 * @brief Вывод на дисплей целого числа (только значащие цифры).
 * @param val Значение числа
 */
void PrintInt(int val) {
    printf("%d", val);
}

/**
 * @brief Вывод на дисплей целого числа заданным количеством цифр (заполнение нулями слева).
 * 
 * @param val Значение числа
 * @param len Количество цифр
 */
void PrintIntLen(int val,unsigned char len) {
    printf("%0*d", len, val);
}

/**
 * @brief Установка курсора (указателя рисования и вывода текста) в положение x, y на дисплее.
 * 
 * @param x 
 * @param y 
 */
void GotoXY(short int x,short int y) {
    _sat_screen.X = x;
    _sat_screen.Y = y;
}

/**
 * @brief Возвращает указатель на описание файла с именем name в файловой системе filesystem.
 * Если файл найден, то функция возвращает указатель на структуру описания файла. Если файл не найден, то функция
 * возвращает NULL.
 * @param filesystem Указатель на файловую систему
 * @param name Имя файла
 * @return flash_file* 
 */
flash_file *OpenFile(const flash_file *filesystem, const char *name) {
    if(!filesystem) return NULL;
    if(!name) return NULL;
    //
    int i=0;
    while(filesystem[i].Name) {
        if(!strcmp(filesystem[i].Name, name)) return (flash_file *)(&filesystem[i]);
        i++;
    }
    return NULL;
}

/**
 * @brief Рисование картинки в заданном месте дисплея. Картинка может рисоваться с прозрачным цветом.
 * 
 * @param x Координата X
 * @param y Координата Y
 * @param id Указатель на Идентификатор картинки
 * @param transparent Использовать прозрачность
 */
void DrawPicture(int x, int y, flash_file *id, char transparent) {
    if(!id) return;
    __internal_syscall(SYSCALL_DrawPicture, x, y, (long)id, transparent, 0, 0);
}

/**
 * @brief Очистка экрана и заливка цветом кисти.
 */
void ClrScr(void) {
    __internal_syscall_invoke1(SYSCALL_ClrScr, 0);
}

/**
 * @brief Установка цвета пикселя дисплея
 * 
 * @param x Координата X
 * @param y Координата Y
 * @param color Цвет пикселя
 */
void SetPixel(int x, int y, int color) {
    __internal_syscall(SYSCALL_SetPixel, x, y, color, 0, 0, 0);
}

/**
 * @brief Рисование линии из позиции курсора в позицию x,y при помощи текущей настройки пера.
 * 
 * @param x Координата X
 * @param y Координата Y
 */
void LineTo(int x, int y) {
    __internal_syscall_invoke2(SYSCALL_LineTo, x, y);
}

/**
 * @brief Рисование прямоугольника. Контур обводится пером, а внутренняя часть заполняется цветом кисти.
 * 
 * @param left Координата левой границы
 * @param top Координата верхней границы
 * @param right Координата правой границы
 * @param bottom Координата нижней границы
 */
void Rectangle(int left, int top, int right, int bottom) {
    __internal_syscall(SYSCALL_Rectangle, left, top, right, bottom, 0, 0);
}

/**
 * @brief Рисование прямоугольника, заполненного цветом кисти. Перо не используется.
 * @param left Координата левой границы
 * @param top Координата верхней границы
 * @param right Координата правой границы
 * @param bottom Координата нижней границы
 */
void RectangleBrush(int left, int top, int right, int bottom) {
    __internal_syscall(SYSCALL_RectangleBrush, left, top, right, bottom, 0, 0);
}

/**
 * @brief Рисование контура прямоугольника при помощи пера. Внутренняя часть не заполняется. Кисть не используется.
 * @param left Координата левой границы
 * @param top Координата верхней границы
 * @param right Координата правой границы
 * @param bottom Координата нижней границы
 */
void RectanglePen(int left, int top, int right, int bottom) {
    __internal_syscall(SYSCALL_RectanglePen, left, top, right, bottom, 0, 0);
}

/**
 * @brief Рисование эллипса или круга. Контур обводится пером, а внутренняя часть заполняется цветом кисти.
 * 
 * @param left Координата левой границы
 * @param top Координата верхней границы
 * @param right Координата правой границы
 * @param bottom Координата нижней границы
 */
void Ellipse(int left, int top, int right, int bottom) {
    __internal_syscall(SYSCALL_Ellipse, left, top, right, bottom, 0, 0);
}

/**
 * @brief Рисование эллипса или круга пером, внутренняя часть не заполняется. Кисть не используется.
 * 
 * @param left Координата левой границы
 * @param top Координата верхней границы
 * @param right Координата правой границы
 * @param bottom Координата нижней границы
 */
void EllipsePen(int left, int top, int right, int bottom) {
    __internal_syscall(SYSCALL_EllipsePen, left, top, right, bottom, 0, 0);
}

/**
 * @brief Завершение рисования экрана (копирование видеообласти).
 */
void ShowScr(void) {
    __internal_syscall_invoke1(SYSCALL_ShowScr, 0);
}
