/**
 * @file satsock.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с сетью: клиент TCP/IP, UDP/IP.
 * @version 0.1
 * @date 2024-04-04
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <syscallp.h>

#include <satsock.h>

#define SOCKETMASK  0x0000ffff

/**
 * @brief Создание сетевого сокета указанного типа.
 * Функция создаёт сетевой сокет указанного типа: SOCK_STREAM (TCP/IP) или SOCK_DGRAM (UDP/IP).
 * При успешном выполнении возвращает дескриптор созданного сокета, при ошибке - значение 0.
 * Ошибка может быть вызвана тем, что указан неправильный тип или в системе нет свободных сокетов.
 * @param style SOCK_STREAM | SOCK_DGRAM
 * @return int Дескриптор сокета или 0 при ошибке.
 */
int CreateSocket(socket_style_t style) {
    if((style != SOCK_STREAM) && (style != SOCK_DGRAM)) return 0;
    //
    return __internal_syscall_invoke1(SYSCALL_CreateSocket, style);
}

/**
 * @brief Функция закрывает сетевое соединение (если оно установлено) и освобождает ресурсы системы
 * связанные с указанным сокетом.
 * @param socket Указатель на дескриптор сокета.
 */
void CloseSocket(int *socket) {
    if(!SocketHandleIsValid(*socket)) return;
    //
    __internal_syscall_invoke1(SYSCALL_CloseSocket, (*socket) & SOCKETMASK);
    *socket = 0;
}

/**
 * @brief Проверка дескриптора сокета на корректность.
 * Возвращает 1, если handle явлается дескриптором сетевого сокета или 0 в противном случае.
 * @param socket Дескриптор сокета.
 * @return int 
 */
int SocketHandleIsValid(int socket) {
    int style = (socket >> 8) & 0xff;
    if(((style == SOCK_STREAM) || (style == SOCK_DGRAM)) && (socket & 0xff)) return 1;
    return 0;
}

/**
 * @brief Привязка сетевого сокета к локальному TCP или UDP порту.
 * Функция делает попытку привязки сокета к локальному порту с указанным номером.
 * Если функцию не использовать, то привязка выполнится автоматически к порту со случайным номером.
 * Если функция используется, то она должна быть вызвана до начала установки соединения.
 * При успешном выполнении привязки функция возвращает значение 1, при ошибке - 0.
 * Ошибка может быть вызвана некорретным значением сокета или тем, что указанный порт уже
 * используется в системе.
 * @param socket Дескриптор сокета.
 * @param port Номер TCP или UDP порта.
 * @return int При успешном выполнении 1 или 0 при ошибке.
 */
int SocketBind(int socket, uint16_t port) {
    if(!SocketHandleIsValid(socket)) return 0;
    //
    return __internal_syscall_invoke2(SYSCALL_SocketBind, socket & SOCKETMASK, port);
}

/**
 * @brief Установка соединения с хостом с указанным IP-адресом и портом.
 * Для сокетов с типом SOCK_STREAM функция начинает асинхронную процедуру установки соединения с указанным хостом.
 * Параметр timeout содержит максимальное время (мс) в течении которого будет выполняться попытка установки соединения.
 * Процесс установки соединения необходимо контролировать вызовами функции SocketState().
 * Для сокетов с типом SOCK_DGRAM функция заполняет внутренние структуры указанными значениями адреса и номера порта.
 * В обоих случаях функция возвращает значение без задержки.
 * Сокет должен быть предварительно создан функцией CreateSocket().
 * Функция возвращает 1, если процесс подключения успешно запущен или 0 в случае ошибки.
 * Ошибка может быть вызвана некорректными значениями параметров или нехваткой системных ресурсов.
 * @param socket Дескриптор сокета.
 * @param host Строка именем или IP-адресом хоста.
 * @param port Номер порта (0..65535).
 * @param timeout Максимальное время для установки соединения, мс.
 * @return int При успешном выполнении 1 или 0 при ошибке.
 */
int SocketConnect(int socket, char *host, uint16_t port, uint32_t timeout) {
    if(!SocketHandleIsValid(socket)) return 0;
    if(!host) return 0;
    //
    return __internal_syscall(SYSCALL_SocketConnect, socket & SOCKETMASK, (long)host, port, timeout, 0, 0);
}

/**
 * @brief Установка соединения с хостом с указанным IP-адресом и портом.
 * Для сокетов с типом SOCK_STREAM функция выполняет процедуру установки соединения с указанным хостом.
 * Параметр timeout содержит максимальное время (мс) в течении которого будет выполняться попытка установки соединения.
 * Для сокетов с типом SOCK_DGRAM функция заполняет внутренние структуры указанными значениями адреса и номера порта.
 * Сокет должен быть предварительно создан функцией CreateSocket().
 * Функция возвращает 1, если подключение успешно выполнено или 0 в случае ошибки.
 * Ошибка может быть вызвана некорректными значениями параметров, нехваткой системных ресурсов или невозможностью
 * подключения к указанному хосту.
 * @param socket Дескриптор сокета.
 * @param host Строка именем или IP-адресом хоста.
 * @param port Номер порта (0..65535).
 * @param timeout Максимальное время для установки соединения, мс.
 * @return int При успешном выполнении 1 или 0 при ошибке.
 */
int SocketConnectBlock(int socket, char *host, uint16_t port, uint32_t timeout) {
    if(!SocketHandleIsValid(socket)) return 0;
    if(SocketState(socket) == sockOK) return 1;
    //
    // начинаем процесс подключения
    int res = SocketConnect(socket, host, port, timeout);
    if(!res) return 0;
    //
    while(1) {
        res = SocketState(socket);
        if(res == sockOK) return 1;
        if(res > sockOK)  return 0;
        // ожидание
        msleep(10);
    };
}

/**
 * @brief Проверка статуса установки соединения с хостом.
 * Функция возвращает значение из перечисления socket_state_enum соответствующее текущему состоянию указанного сокета.
 * 
 * @param socket Дескриптор сокета.
 * @return int 
 */
int SocketState(int socket) {
    if(!SocketHandleIsValid(socket)) return sockBADSOCKET;
    //
    return __internal_syscall_invoke1(SYSCALL_SocketState, socket & SOCKETMASK);
}

/**
 * @brief Проверка статуса TCP-соединения.
 * Для сокетов с типом SOCK_STREAM функция проверяет текущий статус TCP-соединения.
 * Функция возвращает 1,если соединение установлено или 0, если соединение отсутствует.
 * Для сокетов с типом SOCK_DGRAM функция всегда возвращает 1.
 * Если параметр socket содержит некорректное значение дескриптора, то функция возвращает 0.
 * @param socket Дескриптор сокета.
 * @return int При установленном соединении: 1, при отсутствии соединения или ошибке: 0.
 */
int SocketConnected(int socket) {
    if(SocketState(socket) == sockOK) return 1;
    return 0;
}

/**
 * @brief Запись данных в указанный сокет.
 * Используемый сокет должен быть предварительно создан функцией CreateSocket и назначен
 * адрес удалённого хоста (для SOCK_DGRAM) или установлено соединение (для SOCK_STREAM)
 * функцией SocketConnect.
 * При успешном выполнении функция возвращает количество записанных байт данных, при ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * Ошибка может быть вызвана некорректным значением параметров или отсутствием соединения (для SOCK_STREAM).
 * @param socket Дескриптор сокета.
 * @param data Указатель на массив данных.
 * @param len Количество записываемых данных.
 * @return int Количество записанных данных или отрицательное значение при ошибке.
 */
int SocketWrite(int socket, void *data, size_t len) {
    if(!SocketHandleIsValid(socket)) return -1;
    //
    return __internal_syscall(SYSCALL_SocketWrite, socket & SOCKETMASK, (long)data, len, 0, 0, 0);
}

/**
 * @brief Передача данных на указанный адрес.
 * Используется только с сокетами типа SOCK_DGRAM.
 * При успешном выполнении функция возвращает количество записанных байт данных, при ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета.
 * @param data Указатель на массив данных.
 * @param len Количество записываемых данных.
 * @param addr Указатель на структуру описывающую адрес назначения.
 * @return int 
 */
int SocketWriteTo(int socket, void *data, size_t len, struct sockaddr_in *addr) {
    if(!SocketHandleIsValid(socket)) return -1;
    //
    return __internal_syscall(SYSCALL_SocketSendTo, socket & SOCKETMASK, (long)data, len, (long)addr, 0, 0);
}

/**
 * @brief Возвращает значение количества принятых и доступных для чтения данных.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета.
 * @return int Количество доступных для чтения байт в приёмном буфере или отрицательное значение при ошибке.
 */
int SocketReadLength(int socket) {
    if(!SocketHandleIsValid(socket)) return -1;
    //
    return __internal_syscall_invoke1(SYSCALL_SocketReadLength, socket & SOCKETMASK);
}

/**
 * @brief Прочитать не более size байт данных из приёмного буфера сокета.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета.
 * @param data Указатель на массив данных.
 * @param size Количество читаймых байт данных.
 * @return int Количество прочитанных данных или отрицательное значение при ошибке.
 */
int SocketRead(int socket, void *data, size_t size) {
    if(!SocketHandleIsValid(socket)) return -1;
    //
    return __internal_syscall(SYSCALL_SocketRead, socket & SOCKETMASK, (long)data, size, 0, 0, 0);
}

/**
 * @brief Прочитать не более size байт данных из приёмного буфера сокета.
 * При использовании с сокетами типа SOCK_DGRAM функция присваивает адрес отправителя значению параметра addr.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета.
 * @param data Указатель на массив данных.
 * @param size Количество читаймых байт данных.
 * @return int Количество прочитанных данных или отрицательное значение при ошибке.
 */
int SocketReadFrom(int socket, void *data, size_t size, struct sockaddr_in *addr) {
    if(!SocketHandleIsValid(socket)) return -1;
    //
    if(!addr) return __internal_syscall(SYSCALL_SocketRead, socket & SOCKETMASK, (long)data, size, 0, 0, 0);
    return __internal_syscall(SYSCALL_SocketReadFrom, socket & SOCKETMASK, (long)data, size, (long)addr, 0, 0);
}

/**
 * @brief Прочитать блок данных указанной длины, в течении не более чем указанного времени.
 * Чтение выполняется в течении не более чем timeout мс.
 * При успешном выполнении возвращает 1, если данные не приняты - 0.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета
 * @param data Указатель на массив данных
 * @param len Количество читаемых данных
 * @param timeout Максимальное время чтения данных, мс
 * @return int 
 */
int SocketReadBlock(int socket, void *data, size_t len, int timeout) {
    int index = 0;
    //
    while(1) {
        int rc = SocketReadLength(socket);
        if(rc < 0) return rc;
        if(rc > 0) {
            if(rc > len) rc = len;
            //
            int res = SocketRead(socket, data + index, rc);
            if(res < 0) return res;
            //
            len = len - rc;
            if(len == 0) return 1;
            index = index + rc;
        } else {
            if(timeout <= 0) return 0;
            timeout = timeout - 10;
            msleep(10);
        }
    }
}

/**
 * @brief Очистка приёмного буфера сокета.
 * @param socket Дескриптор сокета.
 */
void SocketClear(int socket) {
    if(!SocketHandleIsValid(socket)) return;
    //
    __internal_syscall_invoke1(SYSCALL_SocketClear, socket & SOCKETMASK);
}

/**
 * @brief Установка размера приёмного буфера.
 * По умолчанию размер приёмного буфера 2048 байт. Размер приёмного буфера не может превышать 32768.
 * При успешном выполнении функция возвращает значение 1, при ошибке - 0.
 * @param socket Дескриптор сокета.
 * @param size Новый размер приёмного буфера.
 * @return int 
 */
int SocketSetBufSize(int socket, size_t size) {
    if(!SocketHandleIsValid(socket)) return 0;
    //
    return __internal_syscall_invoke2(SYSCALL_SocketSetBufSize, socket & SOCKETMASK, size);
}

/**
 * @brief Получить значение адреса локального сокета.
 * При отсутствии ошибок функция заполняет поля структуры addr и возвращает значение 1.
 * При ошибке функция возвращает значение 0.
 * @param socket Дескриптор сокета.
 * @param addr Указатель на структуру адреса сокета.
 * @return int 
 */
int SocketGetLocalAddr(int socket, struct sockaddr_in *addr) {
    if(!SocketHandleIsValid(socket)) return 0;
    //
    return __internal_syscall_invoke2(SYSCALL_SocketGetLocalAddr, socket & SOCKETMASK, (long)addr);
}

/**
 * @brief Получить значение адреса удалённого сокета.
 * При отсутствии ошибок функция заполняет поля структуры addr и возвращает значение 1.
 * При ошибке функция возвращает значение 0.
 * @param socket Дескриптор сокета.
 * @param addr Указатель на структуру адреса сокета.
 * @return int 
 */
int SocketGetRemoteAddr(int socket, struct sockaddr_in *addr) {
    if(!SocketHandleIsValid(socket)) return 0;
    //
    return __internal_syscall_invoke2(SYSCALL_SocketGetRemoteAddr, socket & SOCKETMASK, (long)addr);
}

/**
 * @brief Эта функция преобразует интернет-адрес IPv4 из строки со стандартной записью в виде цифр и точек в 
 * двоичные данные и сохраняет его в структуре in_addr, на которую указывает addr.
 * Функция возвращает ненулевое значение, если адрес корректный или 0 в случае ошибки.
 * @param cp Указатель на строку с IPv4 адресом
 * @param addr Указатель на структуру in_addr.
 * @return int 
 */
int inet_atoh(const char *cp, struct in_addr *addr) {
    if(!addr) return 0;
    if(!cp) return 0;
    //
    uint32_t parts[4];
    uint32_t val = 0;
    const char *c;
    char *endptr;
    int gotend, n;

    c = (const char *)cp;
    n = 0;

    /*
    * Run through the string, grabbing numbers until
    * the end of the string, or some error
    */
    gotend = 0;
    while (!gotend) {
        uint32_t l;
        //
        l = strtoul(c, &endptr, 0);
        if (l == ULONG_MAX || (l == 0 && endptr == c)) return (0);
        val = l;
        /*
        * If the whole string is invalid, endptr will equal
        * c.. this way we can make sure someone hasn't
        * gone '.12' or something which would get past
        * the next check.
        */
        if (endptr == c) return (0);
        parts[n] = val;
        c = endptr;

        /* Check the next character past the previous number's end */
        switch (*c) {
        case '.' :
            // Make sure we only do 3 dots
            if (n == 3) return (0);
            n++;
            c++;
            break;
        case '\0':
            gotend = 1;
            break;
        default:
            if (isspace((unsigned char)*c)) {
                gotend = 1;
                break;
            } else {
                /* Invalid character, then fail. */
                return (0);
            }
        }
    }
    /* Concoct the address according to the number of parts specified. */
    switch (n) {
    case 0:
        break;
    case 1:         // a.b -- 8.24 bits
        if (val > 0xffffff || parts[0] > 0xff) return (0);
        val |= parts[0] << 24;
        break;

    case 2:         // a.b.c -- 8.8.16 bits
        if (val > 0xffff || parts[0] > 0xff || parts[1] > 0xff)
            return (0);
        val |= (parts[0] << 24) | (parts[1] << 16);
        break;

    case 3:         // a.b.c.d -- 8.8.8.8 bits
        if (val > 0xff || parts[0] > 0xff || parts[1] > 0xff || parts[2] > 0xff) return (0);
        val |= (parts[0] << 24) | (parts[1] << 16) | (parts[2] << 8);
        break;
    }
    //
    addr->s_addr = htonl(val);
    return (1);
}

/**
 * @brief Функция преобразует адрес addr в строку стандартной формы в виде цифр и точек.
 * Функция помещает строку по адресу, указанному в параметре buf.
 * Размер буфера buf должен быть не меньше чем MAX_IPv4_STR_LEN байт!
 * @param addr Указатель на структуру in_addr.
 * @param buf Указатель на буфер строки с адресом.
 * @return const char* 
 */
const char *inet_htop(struct in_addr *addr, char *buf) {
    if(!buf) return NULL;
    //
    sprintf(buf, "%d.%d.%d.%d", addr->S_un.b[0], addr->S_un.b[1], addr->S_un.b[2], addr->S_un.b[3]);
    return buf;
}

/**
 * @brief Преобразует беззнаковое короткое целое число hostshort из порядка байтов хоста в сетевой порядок байтов.
 * @param v 
 * @return uint16_t 
 */
uint16_t htons(uint16_t v) {
    return (v >> 8) | (v << 8);
}

/**
 * @brief Преобразует целое число без знака v из порядка байтов хоста в сетевой порядок байтов.
 * @param v 
 * @return uint32_t 
 */
uint32_t htonl(uint32_t v) {
    return htons(v >> 16) | (htons((uint16_t) v) << 16);
}

/**
 * @brief Получение текущего состояния интерфейса Ethernet.
 * Функция возвращает значение из перечисления eth_state_enum.
 * @return int 
 */
int EthernetLink(void) {
    return __internal_syscall_invoke1(SYSCALL_EthernetLink, 0);
}

/**
 * @brief Разрешение или запрет перезапуска интерфейса Ethernet при отсутствии подключения.
 * В случае разрешения (enabled=1) и отсутствия линка периодически будет выполняться инициализация интерфейса ethernet,
 * которая выполняется несколько секунд, в течение которых все функции контроллера приостанавливаются.
 * Если программе не требуется работа с IP сетью, то рекомендуется установить переменную в значение 0.
 * По умолчанию при старте программы перезапуск интерфейса ethernet запрещён.
 * @param enabled Запрет перезапуска: 0, разрешение 1.
 */
void EthernetRestart(int enabled) {
    __internal_syscall_invoke1(SYSCALL_EthernetRestart, enabled);
}

/**
 * @brief Возвращает количество байт утерянных из-за переполнения приёмного буфера соединения.
 * Вызов функции обнуляет счётчик потерянных байт.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param socket Дескриптор сокета.
 * @return int Количество потерянных байт или отрицательное число при ошибке.
 */
int SocketLost(int socket) {
    if(!SocketHandleIsValid(socket)) return -1;
    //
    return __internal_syscall_invoke1(SYSCALL_SocketLost, socket & SOCKETMASK);
}
