/**
 * @file serial.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Работа с последовательным портом и Mosbus RTU.
 * @version 0.1
 * @date 2024-03-28
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include <stddef.h>
#include <unistd.h>
#include <syscallp.h>

#include <satser.h>

#define PURGE_BUF_SIZE  64

/**
 * @brief Открыть последовательный порт
 * При успешном выполнении функция возвращает дескриптор: положительное число - идентификатор последовательного порта, который затем необходимо
 * использовать в функциях чтения и записи данных.
 * При ошибке функция возвращает значение меньшее или равное 0.
 * @param port Номер последовательного порта: 1,2,...
 * @param baudrate Скорость обмена: 110,300,600,1200,2400,4800,9600,19200,38400,56000,57600,115200
 * @param parity Контроль чётности: 0-None, 1-Odd, 2-Even
 * @param stopbits Количество стоп-бит: 1,2
 * @return int 
 */
int SerialOpen(int port, int baudrate, int parity, int stopbits) {
    return __internal_syscall(SYSCALL_SerialOpen, port, baudrate, parity, stopbits, 0, 0);
}

/**
 * @brief Проверка состояния последовательного порта.
 * Функция возвращает значение: 0 - порт закрыт, 1 - порт открыт.
 * @param port Номер последовательного порта: 1,2,...
 * @return int 
 */
int SerialIsOpen(int port) {
    return __internal_syscall_invoke1(SYSCALL_SerialIsOpen, port);
}

/**
 * @brief Проверка дескриптора порта на корректность.
 * Возвращает 1, если handle явлается дескриптором открытого последовательного порта.
 * @param handle Дескриптор последовательного порта
 * @return int 
 */
int SerialHandleIsValid(int handle) {
    if(handle <= 0) return 0;
    if(handle > 255) return 0;
    return 1;
}

/**
 * @brief Закрыть последовательный порт
 * 
 * @param handle Дескриптор последовательного порта
 */
void SerialClose(int *handle) {
    __internal_syscall_invoke1(SYSCALL_SerialClose, *handle);
    *handle = 0;
}

/**
 * @brief Запись данных в последовательный порт
 * При успешном выполнении функция возвращает значение 1, при ошибке - значение 0
 * @param handle Дескриптор порта
 * @param data Указатель на массив данных
 * @param len Количество записываемых данных
 * @return int 
 */
int SerialWrite(int handle, char *data, size_t len) {
    return __internal_syscall(SYSCALL_SerialWrite, handle, (long)data, len, 0, 0, 0);
}

/**
 * @brief Получить количество байтов, доступных для чтения из приёмного буфера последовательного порта.
 * 
 * @param handle Дескриптор порта
 * @return int 
 */
int SerialReadLength(int handle) {
    return __internal_syscall_invoke1(SYSCALL_SerialReadLen, handle);
}

/**
 * @brief Прочитать данные из приёмного буфера последовательного порта.
 * Возвращаемое значение: количество прочитанных данных.
 * @param handle Дескриптор порта
 * @param data Указатель на массив данных
 * @param len Количество читаемых данных
 * @return short int 
 */
int SerialRead(int handle, char *data, size_t len) {
    return __internal_syscall(SYSCALL_SerialRead, handle, (long)data, len, 0, 0, 0);
}

/**
 * @brief Очистка приёмного буфера последовательного порта.
 * 
 * @param handle  Дескриптор порта
 */
void SerialPurge(int handle) {
    char buf[PURGE_BUF_SIZE];
    //
    while(SerialReadLength(handle) > 0) SerialRead(handle, buf, PURGE_BUF_SIZE);
}

/**
 * @brief Прочитать блок данных, принятых через последовательный порт, в течении указанного времени.
 * При успешном выполнении возвращает 1, если данные не приняты - 0.
 * @param handle Дескриптор порта
 * @param data Указатель на массив данных
 * @param len Количество читаемых данных
 * @param timeout Максимальное время чтения данных, мс
 * @return int 
 */
int SerialReadBlock(int handle, char *data, size_t len, int timeout) {
    int index = 0;
    //
    while(1) {
        int rc = SerialReadLength(handle);
        if(rc > 0) {
            if(rc > len) rc = len;
            SerialRead(handle, data + index, rc);
            len = len - rc;
            if(len == 0) return 1;
            index = index + rc;
        } else {
            if(timeout <= 0) return 0;
            timeout = timeout - 10;
            msleep(10);
        }
    }
}

/**
 * @brief Возвращает количество байт утерянных из-за переполнения приёмного буфера соединения.
 * Вызов функции обнуляет счётчик потерянных байт.
 * @param handle Дескриптор порта
 * @return int Количество потерянных байт или отрицательное число при ошибке.
 */
int SerialLost(int handle) {
    return __internal_syscall_invoke1(SYSCALL_SerialLost, handle);
}
