/**
 * @file unistd.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Унифицированный обмен данными через сокеты и последовательный порт.
 * @version 0.1
 * @date 2024-04-05
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include <satser.h>
#include <satsock.h>
#include <satthreads.h>
#include <unistd.h>

/**
 * @brief Прочитать полученные данные из приёмного буфера.
 * Функция возвращает значение количества прочитанных данных или код ошибки:
 * -1 - некорректное значение дескриптора;
 * -2 - соединение не установлено или оборвано.
 * @param fd Дескриптор сокета или последовательного порта
 * @param buf Указатель на массив данных
 * @param count Количество прочитанных данных
 * @return int 
 */
int read(int fd, void *buf, size_t count) {
    switch (socket_handle_type(fd)) {
        case HDL_SERIAL:
            return SerialRead(fd, buf, count);
        case HDL_NETWORK:
            return SocketRead(fd, buf, count);
        default:
            return -1;
    }
}

/**
 * @brief Возвращает значение количества принятых и доступных для чтения данных.
 * При ошибке функция возвращает:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param fd Дескриптор сокета или последовательного порта
 * @return int Количество доступных для чтения байт в приёмном буфере или отрицательное значение при ошибке.
 */
int read_length(int fd) {
    switch (socket_handle_type(fd)) {
        case HDL_SERIAL:
            return SerialReadLength(fd);
        case HDL_NETWORK:
            return SocketReadLength(fd);
        default:
            return -1;
    }
}

/**
 * @brief Прочитать блок данных указанной длины, в течении не более чем указанного времени.
 * Чтение выполняется в течении не более чем timeout мс.
 * При успешном выполнении возвращает 1, если данные не приняты - 0.
 * При ошибке функция возвращает код:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param fd Дескриптор сокета или последовательного порта.
 * @param data Указатель на массив данных.
 * @param count Количество читаемых данных.
 * @param timeout Максимальное время чтения данных, мс.
 * @return int 
 */
int read_block(int fd, void *data, size_t count, int timeout) {
    switch (socket_handle_type(fd)) {
        case HDL_SERIAL:
            return SerialReadBlock(fd, data, count, timeout);
        case HDL_NETWORK:
            return SocketReadBlock(fd, data, count, timeout);
        default:
            return -1;
    }
}

/**
 * @brief Очистка приёмного буфера сокета или последовательного порта.
 * @param fd Дескриптор сокета или последовательного порта.
 */
void purge(int fd) {
    switch (socket_handle_type(fd)) {
        case HDL_SERIAL:
            SerialPurge(fd);
            break;
        case HDL_NETWORK:
            SocketClear(fd);
            break;
        default:;
    }
}

/**
 * @brief Запись данных в последовательный порт или сокет.
 * При успешном выполнении функция возвращает количество записанных данных, при ошибке код:
 * -1 - некорректное значение дескриптора сокета;
 * -2 - соединение не установлено или оборвано.
 * @param fd Дескриптор сокета или последовательного порта.
 * @param buf Указатель на массив данных.
 * @param count Количество записываемых данных.
 * @return int 
 */
int write(int fd, void *buf, size_t count) {
    switch (socket_handle_type(fd)) {
        case HDL_SERIAL:
            if(SerialWrite(fd, buf, count)) return count;
            return 0;
        case HDL_NETWORK:
            return SocketWrite(fd, buf, count);
        default:
            return -1;
    }
}

/**
 * @brief Возвращает количество байт утерянных из-за переполнения приёмного буфера соединения.
 * Вызов функции обнуляет счётчик потерянных байт. При ошибке функция возвращает отрицательное значение:
 * -1 - некорректное значение дескриптора;
 * -2 - соединение не установлено или оборвано.
 * @param fd Дескриптор сокета или последовательного порта
 * @return int Количество потерянных байт или отрицательное число при ошибке.
 */
int lost_bytes(int fd) {
    switch (socket_handle_type(fd)) {
        case HDL_SERIAL:
            return SerialLost(fd);
        case HDL_NETWORK:
            return SocketLost(fd);
        default:
            return -1;
    }
}

/**
 * @brief Функция закрывает и освобождает ресурсы занятые сокетом или последовательным портом.
 * @param fd Указатель на дескриптор сокета или последовательного порта.
 */
void close(int *fd) {
    switch (socket_handle_type(*fd)) {
        case HDL_SERIAL:
            SerialClose(fd);
            break;
        case HDL_NETWORK:
            CloseSocket(fd);
            break;
        default:;
    }
}

/**
 * @brief Проверка дескриптора сокета или последовательного порта.
 * Функция возвращает результат:
 * HDL_SERIAL  - fd: корректный дескриптор последовательного порта;
 * HDL_NETWORK - fd: корректный дескриптор сетевого сокета.
 * В ином случае функция возвращает HDL_INVALID.
 * @param fd Дескриптор сокета или последовательного порта.
 * @return handle_t
 */
handle_t socket_handle_type(int fd) {
    if(SerialHandleIsValid(fd)) return HDL_SERIAL;
    if(SocketHandleIsValid(fd)) return HDL_NETWORK;
    //
    return HDL_INVALID;
}

/**
 * @brief Переход в режим ожидания на указанное количество секунд
 * 
 * @param seconds 
 */
void sleep(unsigned int seconds) {
    msleep(seconds * 1000);
}

/**
 * @brief Переход в режим ожидания на указанное количество микросекунд
 * 
 * @param usec 
 */
void usleep(unsigned long usec) {
    msleep(usec / 1000);
}

/**
 * @brief Переход в режим ожидания на указанное количество миллисекунд
 * Выполнение неосновных потоков на время ожидания приостанавливается.
 * @param msec 
 */
void msleep(unsigned long msec) {
    // получаем идентификатор текущего потока
    thrd_t t = thrd_current();
    if(t == 0) {
        // для основного потока задержка циклом
        uint32_t t = GetTickCount();
        while((GetTickCount() - t) < msec);
    } else {
        // для неосновных потоков перевод в состояние SLEEP
        thrd_sleep(msec);
    }
}
