/**
 * @file malloc.c
 * @brief Менеджер динамической памяти
 * @version 0.1
 * @date 2024-03-09
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */


#include <limits.h>
#include <string.h>
#include <satthreads.h>
#include <syscallp.h>
#include <stdlib.h>


// |<-managed_memory_start           last_valid_address->|
// +------+----------+------+----------+------+----------+
// | MCB1 | ........ | MCB2 | ........ | MCBn | ........ |
// +------+----------+------+----------+------+----------+
// | <- MCB1.size -> | <- MCB2.size -> | <- MCBn.size -> |

/**
 * @brief Минимальный размер выделяемого блока памяти
 */
#define MIN_MALLOC_SIZE 4

/**
 * @brief Структура MCB, идет ПЕРЕД выделяемыми областями памяти
 */
struct MCB {
    unsigned int size : 31;
    unsigned int is_available : 1;
};

/**
 * @brief Размер структуры MCB
 */
#define SIZE_OF_MCB sizeof(struct MCB)

static int      mem_initialized = 0;
static void     *managed_memory_start;
static void     *last_valid_address;

/**
 * @brief Циклические блокировки
 */
static spin_t   lock_free;
static spin_t   lock_malloc;
static spin_t   lock_realloc;

#if MALLOC_DEBUG == 1

/**
 * @brief Вывод дампа кучи
 */
void _heap_dump(void) {
    void    *current_location;
    struct  MCB *current_location_mcb;
    int     blockn;
    //
    current_location = managed_memory_start;
    blockn = 1;
    //
    printf("! N     MCB         MEM    size  used\n");
    printf("+--- ---------- ---------- ----- ----\n");

    while (current_location != last_valid_address) {
        //
        current_location_mcb = (struct MCB *)current_location;
        printf("!%3d 0x%.8x 0x%.8x %.5u %s\n", blockn, (long)current_location, (long)current_location + SIZE_OF_MCB, current_location_mcb->size, current_location_mcb->is_available?"No":"Yes");
        current_location = current_location + current_location_mcb->size;
        blockn++;
    }
    printf("+------------------------------------\n");
}
#endif

/**
 * @brief Запрос дополнительной памяти
 * 
 * @param size Размер памяти
 * @return void* Адрес начала нового блока или NULL
 */
static void *sbrk(int increment) {
    // аналогичный вызов имеется в Linux и FreeBSD
    return (void *)__internal_syscall_invoke1(SYSCALL_SBRK, increment);
}

/**
 * @brief Инициализация менеджера памяти
 */
static void malloc_init() {
    // получаем начальный адрес heap
    last_valid_address = sbrk(0);
    // пока память не выделялась, устанавливаем начальный адрес менеджера памяти
    managed_memory_start = last_valid_address;
    //
    mem_initialized = 1;
}

/**
 * @brief Объединение свободных блоков памяти
 */
static void malloc_coalesce(void) {
    void    *next_location;
    void    *current_location;
    struct  MCB *next_location_mcb;
    struct  MCB *current_location_mcb;
    //
    current_location = managed_memory_start;
    //
    while (current_location != last_valid_address) {
        //
        current_location_mcb = (struct MCB *)current_location;
        // проверка на нулевой блок
        // такого вообще-то не может быть, но во избежании зацикливания...
        if(current_location_mcb->size == 0) return;
        //
        if(current_location_mcb->is_available) {
            // блок свободен, смотрим на следующий по порядку
            next_location = current_location + current_location_mcb->size;
            //
            if(next_location == last_valid_address) return;                 // этот блок был последний
            next_location_mcb = (struct MCB *)next_location;
            if(next_location_mcb->is_available) {
                // следующий блок есть и он тоже пустой
                // объединяем текущий и следующий блоки
                current_location_mcb->size = current_location_mcb->size + next_location_mcb->size;
                // не изменяем значение current_location: вдруг дальше есть ещё один свободный блок?
            } else {
                // следующий блок есть, но он занят
                // продолжаем проход по списку
                current_location = next_location;
            }
        } else {
            // текущий блок был занят, переходим к следующему
            current_location = current_location + current_location_mcb->size;
        }
    }
}

/**
 * @brief Освобождение динамической памяти ранее выделенной функциями malloc(), realloc(), calloc(), strdup() и подобными
 * 
 * @param mem Адрес освобождаемой памяти
 */
static void _free(void *mem) {
    struct MCB *mcb;
    //
    if(!mem) return;
    // ставим признак свободного блока
    mcb = mem - SIZE_OF_MCB;
    mcb->is_available = 1;
    // объединяем свободные блоки
    malloc_coalesce();
    //
    return;
}

/**
 * @brief Поиск свободного блока памяти с минимальным размером, который больше или равен необходимому
 * 
 * @param size Необходимый размер блока памяти
 * @return void* Найденный блок памяти или NULL
 */
static void *find_min_mem_block(size_t size) {
    //
    struct MCB *mcb;
    size_t min_size = UINT_MAX;
    void *current_location = managed_memory_start;
    void *memory_location  = NULL;
    //
    while (current_location != last_valid_address) {
        mcb = (struct MCB *)current_location;
        if (mcb->is_available) {
            // блок свободен
            if(mcb->size >= size) {
                // размер подходящий
                if((memory_location == NULL) || (mcb->size < min_size)) {
                    // если у блока размер меньше, чем у предыдущего найденного, запоминаем его
                    memory_location = current_location;
                    min_size = mcb->size;
                }
            }
        }
        // переходим к следующему блоку
        current_location = current_location + mcb->size;
    }
    return memory_location;
}

/**
 * @brief Выделение динамической памяти
 * 
 * @param size Требуемый размер
 * @return void* Адрес выделенный памяти или NULL при ошибке
 */
static void *_malloc(size_t size) {
    //
    void *memory_location;
    struct MCB *current_location_mcb;
    struct MCB *new_location_mcb;
    //
    // проверка запрошенного объёма
    if(size == 0) return NULL;
    //
    // если менеджер не инициализирован делаем это
    if (!mem_initialized) malloc_init();
    //
    // какая-то ошибка инициализации, такого не может быть, но вдруг...
    if(managed_memory_start == 0) return NULL;
    //
    // минимальный объём выделяемой памяти MIN_MALLOC_SIZE байта
    size = (size + 3) & 0xFFFFFFFC;
    //
    // необходимо выделить запрошенный объём + размер структуры MCB
    size = size + SIZE_OF_MCB;
    //
    // ищем минимально необходимый блок памяти
    memory_location = find_min_mem_block(size);
    //
    if(memory_location) {
        // чего-то нашли
        current_location_mcb = (struct MCB *)memory_location;
        // блок больше не свободный
        current_location_mcb->is_available = 0;
        //
        // если найденный блок больше или равен необходимого на (sizeof(MCB) + MIN_MALLOC_SIZE), (MIN_MALLOC_SIZE - минимальный объём выделяемой памяти)
        if(current_location_mcb->size >= (size + SIZE_OF_MCB + MIN_MALLOC_SIZE)) {
            //
            // то разбиваем его на две части, вторую часть делаем свободной
            new_location_mcb = memory_location + size;
            new_location_mcb->size = current_location_mcb->size - size;
            new_location_mcb->is_available = 1;
            //
            // обрезаем размер текущего блока
            current_location_mcb->size = size;
            //
            // процедура разбиения свободного блока НЕ может вызвать сегментирования памяти,т.к. следующий блок явно был занят
            // иначе он был-бы объединён с текущим блоком при предыдущем освобождении памяти
        }
    } else {
        // ничего не нашли, запрашиваем новый объём памяти в ОС (в нашем случае у эмулятора)
        // если места не хватает, то sbrk() аварийно завершит выполнение
        if(sbrk(size) == NULL) return NULL;
        //
        memory_location = last_valid_address;
        //
        last_valid_address = last_valid_address + size;
        // инициализация нового MCB
        current_location_mcb = memory_location;
        current_location_mcb->is_available = 0;
        current_location_mcb->size = size;
    }
    //
    // теперь memory_location однозначно содержит адрес найденной или выделенной памяти (включая MCB)
    // устанавливаем адрес ПОСЛЕ блока MCB
    memory_location = memory_location + SIZE_OF_MCB;
    //
    return memory_location;
}

/**
 * @brief Изменение размера блока памяти ранее выделенного функциями malloc() или calloc()
 * 
 * @param mem Указатель на ранее выделенный блок памяти
 * @param newsize Новый размер
 * @return void* Указатель на новый блок памяти
 */
static void *_realloc(void *mem, size_t len) {
    struct MCB *mcb;
    size_t new_size;
    //
    // простое выделение памяти
    if(!mem) return malloc(len);
    //
    // освобождение памяти
    if(len == 0) {
        free(mem);
        return NULL;
    }
    // выясняем размер старого блока
    mcb = mem - SIZE_OF_MCB;
    new_size = ((len + 3) & 0xFFFFFFFC) + SIZE_OF_MCB;
    //
    if(new_size <= mcb->size) {
        // новый размер меньше или равен текущему
        if(mcb->size >= (new_size + SIZE_OF_MCB + MIN_MALLOC_SIZE)) {
            // если существующий блок больше или равен необходимого на (sizeof(MCB) + MIN_MALLOC_SIZE),
            // то разбиваем его на две части, вторую часть делаем свободной
            struct MCB *new_mcb;
            //
            new_mcb = (void *)mcb + new_size;
            new_mcb->size = mcb->size - new_size;
            new_mcb->is_available = 1;
            //
            // обрезаем размер текущего блока
            mcb->size = new_size;
            //
            // разбиение занятого блока может вызвать сегментирование памяти, т.к. следующий блок теоретически может быть свободным
            // запускаем очистку
            malloc_coalesce();
        }
        return mem;
    }
    // новый размер больше текущего, выделяем новый блок
    // планы на будущее: можно проверить нет-ли свободного места в следующем блоке...
    void *newmem = malloc(len);
    //
    // копируем в него содержимое старого блока
    memmove(newmem, mem, mcb->size - SIZE_OF_MCB);
    //
    // освобождаем старый
    free(mem);
    //
    return newmem;
}

/**
 * @brief Выделяет массив в памяти и инициализирует его элементы значением 0
 * 
 * @param num Количество выделяемых элементов
 * @param size Размер одного элемента
 * @return void* Адрес выделенный памяти или NULL при ошибке
 */
void *calloc(size_t num, size_t size) {
    size = size * num;
    //
    void *newmem = malloc(size);
    if(!newmem) return NULL;
    //
    memset(newmem, 0, size);
    //
    return newmem;
}


/**
 * @brief Освобождение динамической памяти ранее выделенной функциями malloc(), realloc(), calloc(), strdup() и подобными
 * 
 * @param mem Адрес освобождаемой памяти
 */
void free(void *mem) {
    //
    // вход в критическую секцию
    spin_lock(&lock_free);
    _free(mem);
    //
    // выход из критической секции
    spin_unlock(&lock_free);
}

/**
 * @brief Выделение динамической памяти
 * 
 * @param size Требуемый размер
 * @return void* Адрес выделенный памяти или NULL при ошибке
 */
void *malloc(size_t size) {
    void *p;
    //
    // вход в критическую секцию
    spin_lock(&lock_malloc);
    p = _malloc(size);
    //
    // выход из критической секции
    spin_unlock(&lock_malloc);
    //
    return p;
}

/**
 * @brief Изменение размера блока памяти ранее выделенного функциями malloc() или calloc()
 * 
 * @param mem Указатель на ранее выделенный блок памяти
 * @param newsize Новый размер
 * @return void* Указатель на новый блок памяти
 */
void *realloc(void *mem, size_t len) {
    void *p;
    //
    // вход в критическую секцию
    spin_lock(&lock_realloc);
    p = _realloc(mem, len);
    //
    // выход из критической секции
    spin_unlock(&lock_realloc);
    //
    return p;
}
