/**
 * @file strtol.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC"
 * @version 0.1
 * @date 2024-04-19
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include <ctype.h>
#include <limits.h>
#include <stdlib.h>

/**
 * @brief Преобразование строки в длинное целое число
 * Функция конвертирует начальную часть строки nptr в длинное целое в соответствии с указанным base, 
 * которое должно находиться в диапазоне от 2-х до 36-х включительно или быть равным нулю.
 * @param nptr Указатель на строку
 * @param endptr Указатель на первое некорректное значение в строке
 * @param base Основание системы счисления
 * @return long int 
 */
long int strtol(const char *nptr, char **endptr, int base) {
    const char *s;
    long acc, cutoff;
    int neg, any, cutlim;
    int c;
    //
    s = nptr;
    do {
        c = (unsigned char) *s++;
    } while (isspace(c));
    if (c == '-') {
        neg = 1;
        c = *s++;
    } else {
        neg = 0;
        if (c == '+') c = *s++;
    }
    if ((base == 0 || base == 16) && c == '0' && (*s == 'x' || *s == 'X')) {
        c = s[1];
        s += 2;
        base = 16;
    }
    if (base == 0) base = c == '0' ? 8 : 10;

    cutoff = neg ? LONG_MIN : LONG_MAX;
    cutlim = (int)(cutoff % base);
    cutoff /= base;
    if (neg) {
        if (cutlim > 0) {
            cutlim -= base;
            cutoff += 1;
        }
        cutlim = -cutlim;
    }
    for (acc = 0, any = 0;; c = (unsigned char) *s++) {
        if (isdigit(c)) c -= '0';
        else if (isalpha(c)) c -= isupper(c) ? 'A' - 10 : 'a' - 10;
        else break;
        if (c >= base) break;
        if (any < 0) continue;
        if (neg) {
            if (acc < cutoff || (acc == cutoff && c > cutlim)) {
                any = -1;
                acc = LONG_MIN;
                errno = ERANGE;
            } else {
                any = 1;
                acc *= base;
                acc -= c;
            }
        } else {
            if (acc > cutoff || (acc == cutoff && c > cutlim)) {
                any = -1;
                acc = LONG_MAX;
                errno = ERANGE;
            } else {
                any = 1;
                acc *= base;
                acc += c;
            }
        }
    }
    if(endptr) *endptr = (char*)(any ? s - 1 : nptr);
    //
    return acc;
}
