/**
 * @file strtoul.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC"
 * @version 0.1
 * @date 2024-04-19
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include <ctype.h>
#include <limits.h>
#include <stdlib.h>

/**
 * @brief Преобразование строки в длинное целое число без знака.
 * Функция конвертирует начальную часть строки nptr в длинное целое в соответствии с указанным base, 
 * которое должно находиться в диапазоне от 2-х до 36-х включительно или быть равным нулю.
 * При ошибках устанавливает значение переменной errno.
 * @param nptr Указатель на строку
 * @param endptr Указатель на первое некорректное значение в строке
 * @param base Основание системы счисления
 * @return unsigned long int
 */
unsigned long strtoul(const char *nptr, char **endptr, int base) {
    const char *s = nptr;
    unsigned long acc;
    int c;
    unsigned long cutoff;
    int neg = 0, any, cutlim;
    do {
        c = *s++;
    } while (isspace(c));
    //
    if (c == '-') {
        neg = 1;
        c = *s++;
    } else if (c == '+') c = *s++;
    //
    if ((base == 0 || base == 16) && c == '0' && (*s == 'x' || *s == 'X')) {
        c = s[1];
        s += 2;
        base = 16;
    }
    if (base == 0) base = c == '0' ? 8 : 10;
    //
    cutoff = (unsigned long)ULONG_MAX / (unsigned long)base;
    cutlim = (unsigned long)ULONG_MAX % (unsigned long)base;
    //
    for (acc = 0, any = 0;; c = *s++) {
        if (isdigit(c)) c -= '0'; else
        if (isalpha(c)) c -= isupper(c) ? 'A' - 10 : 'a' - 10; else break;
        if (c >= base) break;
        if ((any < 0) || (acc > cutoff) || ((acc == cutoff) && (c > cutlim))) any = -1; else {
            any = 1;
            acc *= base;
            acc += c;
        }
    }
    if (any < 0) {
        acc = ULONG_MAX;
        errno = ERANGE;
    } else if (neg) acc = -acc;
    //
    if (endptr != 0) *endptr = (char *)(any ? s - 1 : nptr);
    return (acc);
}
