/**
 * @file satcthread.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Управление потоками "ISO C threads".
 * @version 0.1
 * @date 2024-03-19
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include <stdlib.h>
#include <syscallp.h>

#include <satthreads.h>

/**
 * @brief Создание нового потока
 * 
 * @param thr 
 * @param func Функция потока
 * @param arg Аргументы для функции потока
 * @return int 
 */
int thrd_create(thrd_t *thr, thrd_start_t func, void *arg) {
    if(!thr) return thrd_error;
    if(!func) return thrd_error;
    //
    // выделяем пямять для стека
    void *stack = malloc(THREAD_STACK_SIZE);
    //
    // уведомляем эмулятор:
    *thr = __internal_syscall(SYSCALL_thrd_create,
        (long)func,                                     // A0 - функция потока
        (long)arg,                                      // A1 - указатель на параметры функции потока
        (long)stack,                                    // A2 - указатель на стек
        THREAD_STACK_SIZE,                              // A3 - размер стека
        (long)thrd_exit,                                // A4 - указатель на функцию завершения потока
        0                                               // A5
    );
    if(*thr < 0) {
        // ошибка при создании потока, освобождаем память стека
        free(stack);
        return thrd_error;
    }
    // поток успешно создан
    return thrd_success;
}

/**
 * @brief Функция возвращает идентификатор потока из которого она вызвана.
 * 
 * @return thrd_t 
 */
thrd_t thrd_current(void) {
    return __internal_syscall_invoke1(SYSCALL_thrd_current, 0);
}

/**
 * @brief Проверка того, что lhs и rhs ссылаются на один и тот же поток.
 * Если lhs и rhs являются разными потоками, то функция возвращает 0, иначе - ненулевое значение.
 * @param lhs Идентификатор потока 1
 * @param rhs Идентификатор потока 2
 * @return int 
 */
int thrd_equal(thrd_t lhs, thrd_t rhs) {
    if((lhs <= 0) || (rhs <= 0)) return 0;
    return (lhs == rhs)?1:0;
}

/**
 * @brief Приостановка выполнения потока на указанное количество миллисекунд.
 * Обратите внимание: в отличии от стандарта, функция имеет только один параметр msec.
 * Функция не может быть использована для приостановки выполнения основного потока,
 * в этом случае она сразу возвращает управление без задержки.
 * @param msec Величина задержки в миллисекундах
 * @return int Остаток времени
 */
int thrd_sleep(uint32_t msec) {
    int res = __internal_syscall_invoke1(SYSCALL_thrd_sleep, msec);
    return res;
}

/**
 * @brief Передаёт управление другому потоку.
 */
void thrd_yield(void) {
    __internal_syscall_invoke1(SYSCALL_thrd_yield, 0);
}

/**
 * @brief Завершение выполнения текущего потока.
 * Функция не возвращает управление.
 * @param res Код возврата
 */
void thrd_exit(int res) {
    //
    // получение указателя на стек
    void *stack = (void *)__internal_syscall_invoke1(SYSCALL_thrd_stack, 0);
    //
    // освобождение памяти стека
    if(stack) free(stack);
    //
    // удаление потока
    __internal_syscall_invoke1(SYSCALL_thrd_exit, res);
    //
    // следующая строка никогда не выполняется
    while(1);
}

/**
 * @brief Принудительное завершение выполнения потока thr.
 * Вызов освобождает память выделенную для стека завершаемого потока.
 * Если завершаемый поток в ходе своей работы запрашивал выделение памяти (malloc) или других ресурсов, то их освобождение не выполняется!
 * Вызов выполненный в отношении потока, который уже ранее завершил выполнение приведёт к возврату значения thrd_error.
 * Функция не может быть выполнена в отношении основного потока (вернёт значение thrd_error).
 * Функция не может быть выполнена в отношении текущего потока (вернёт значение thrd_error), используйте для этого thrd_exit().
 * Функция может вернуть значения thrd_success или thrd_error.
 * 
 * @param thr Идентификатор завершаемого потока
 * @return int 
 */
int thrd_detach(thrd_t thr) {
    //
    // получение указателя на стек
    void *stack = (void *)__internal_syscall_invoke1(SYSCALL_thrd_detach, thr);
    //
    if(stack) {
        // если указатель получен, то поток существует
        // освобождаем память
        free(stack);
        return thrd_success;
    } else {
        // получен указатель NULL: такого потока нет или какая-то ошибка
        return thrd_error;
    }
}

/**
 * @brief Блокирует выполнение текущего потока до тех пор, пока выполнение потока thr не будет завершено.
 * Если значение res не равно NULL, то результирующий код потока thr помещается по указателю res.
 * Функция возвращает значение, когда поток thr завершит свою работу.
 * Функция может вернуть значения thrd_success или thrd_error.
 * @param thr Указатель на поток, завершения работы которого ожидается.
 * @param res Результат завершения потока.
 * @return int 
 */
int thrd_join(thrd_t thr, int *res) {
    // основной поток никогда не завершается, его ожидание приведёт к блокировке
    if(thr == 0) return thrd_error;
    //
    thrd_t curt = thrd_current();
    // нельзя выполнить объединение с самим собой
    if(thr == curt) return thrd_error;
    //
    if(curt == 0) {
        // текущий поток - основной
        // основной поток никогда не переходит в SLEEP
        // ожидание выполняем пустым циклом
        while(1) {
            // проверяем состояние ожидаемого потока
            int r = __internal_syscall_invoke2(SYSCALL_thrd_state, thr, (long)res);
            // ошибка
            if(r < 0) return thrd_error;
            // если поток завершён, то возвращаем управление
            if((r == thrd_stopped) || ((r == thrd_finished))) return thrd_success;
            // передаём управление другим потокам
            thrd_yield();
        }
    } else {
        // текущий поток - не основной
        int r = __internal_syscall_invoke1(SYSCALL_thrd_join, thr);
        if(res) *res = r;
        return thrd_success;
    }
}

/**
 * @brief Вызывает функцию func ровно один раз, даже если она вызывается из нескольких потоков.
 * Завершение функции func синхронизируется - со всеми предыдущими или последующими вызовами call_once с той же переменной флага.
 * @param flag Указатель на флаг.
 * @param func Указтель на функцию.
 */
void call_once(once_flag *flag, void (*func) (void)) {
    spin_lock(flag);
    func();
    spin_unlock(flag);
}

/**
 * @brief Инициализация циклической блокировки
 * @example
 * spin_t crit_section;
 * spin_init(&crit_section);
 * ...
 * // начало критической секции
 * spin_lock(&crit_section);
 * ...
 * ...
 * // конец критической секции
 * spin_unlock(&crit_section);
 * 
 * @param p Указатель на идентификатор блокировки
 */
void spin_init(spin_t *p) {
    *p = 0;
}

/**
 * @brief Вход в циклическую блокировку (критическая секция).
 * Спин-блокировки являются аналогами мьютексов, позволяющими тратить меньше времени на процедуру блокировки потока,
 * поскольку не требуется переводить поток в заблокированное состояние.
 * @param p Указатель на идентификатор блоировки
 */
void spin_lock(spin_t *p) {
    __critical_section_enter(p);
}

/**
 * @brief Выход из циклической блокировки (критическая секция).
 * 
 * @param p Указатель на идентификатор блокировки
 */
void spin_unlock(spin_t *p) {
    __critical_section_leave(p);
}

/**
 * @brief Инициализация нового мьютекса указанного типа.
 * В качестве типа создаваемого мьютекса могут быть указаны следующие значения: mtx_plain, mtx_timed, mtx_plain | mtx_recursive, mtx_timed | mtx_recursive.
 * Отличий в реализации между mtx_plain и mtx_timed нет.
 * Функция может вернуть значения thrd_success или thrd_error.
 * @param mutex Указатель на идентификатор создаваемого мьютекса
 * @param type Тип создаваемого мьютекса
 * @return int 
 */
int	mtx_init(mtx_t *mutex, int type) {
    if(!mutex) return thrd_error;
    //
    switch (type) {
        case mtx_plain | mtx_recursive:
        case mtx_timed | mtx_recursive:
            mutex->type = mtx_plain | mtx_recursive;
            break;
        case mtx_plain:             // No difference between both in standard
        case mtx_timed:
            mutex->type = mtx_plain;
            break;
        default:
            // некорректный тип мьютекса
            return thrd_error;
    }
    //
    mutex->state = mtx_unlocked;
    mutex->owner = -1;
    //
    return thrd_success;
}

/**
 * @brief Уничтожение мьютекса
 * Если есть какие-либо потоки, ожидающие мьютекса, поведение не определено.
 * @param mutex Указатель на идентификатор мьютекса
 */
void mtx_destroy(mtx_t *mutex) {
    if(!mutex) return;
    //
    mutex->type = 0;
    mutex->state = mtx_unlocked;
}

/**
 * @brief Блокировка указанного мьютекса.
 * Блокирует текущий поток до тех пор, пока указанный мьютекс является заблокированным.
 * Если текущий поток уже заблокировал мьютекс и мьютекс является рекурсивным, то функция возвращает thrd_success.
 * Если текущий поток уже заблокировал мьютекс и мьютекс не является рекурсивным, то функция возвращает thrd_error.
 * Функция может вернуть значения thrd_success или thrd_error.
 * @param mutex Указатель на идентификатор мьютекса
 * @return int 
 */
int	mtx_lock(mtx_t *mutex) {
    return mtx_timedlock(mutex, 0);
}

/**
 * @brief Блокировка указанного мьютекса.
 * Блокирует текущий поток до тех пор, пока указанный мьютекс является заблокированным или текущеее время (GMT) превысило значение timespec.
 * Если текущий поток уже заблокировал мьютекс и мьютекс является рекурсивным, то функция возвращает thrd_success.
 * Если текущий поток уже заблокировал мьютекс и мьютекс не является рекурсивным, то функция возвращает thrd_error.
 * Если текущее время превысило значение timespec, а ресурс мьютекс остался заблокированным, то функция возвращает значение thrd_timedout.
 * Обратите внимание: параметр timespec содержит абсолютное время, если вам необходим интервал, то его необходимо вычислить самостоятельно.
 * В отличии от стандарта ANSI C тип параметра timespec объявлен как time_t.
 * Функция может вернуть значения thrd_success, thrd_timedout или thrd_error.
 * @param mutex Указатель на идентификатор мьютекса
 * @param timespec Абсолютное время GMT
 * @return int 
 */
int	mtx_timedlock(mtx_t *mutex, const time_t timespec) {
    long res;
    //
    if(!mutex) return thrd_error;
    //
    res = __atomic_swap(mutex, mtx_locked);
    // 0 - успешно
    // 1 - уже ранее был заблокирован
    if(res == mtx_unlocked) {
        // мьютекс успешно заблокирован
        mutex->owner = thrd_current();
        return thrd_success;
    }
    // мьютекс был заблокирован, 
    if(thrd_current() == mutex->owner) {
        // мьютекс заблокирован текущим потоком
        // если мьютекс рекурсивный, то возвращаем thrd_success
        if(mutex->type == (mtx_plain | mtx_recursive)) return thrd_success;
        // иначе возвращаем thrd_error
        return thrd_error;
    }
    // мьютекс был заблокирован другим потоком
    // ждём освобождения
    while (1) {
        if(__atomic_swap(mutex, mtx_locked) == mtx_unlocked) {
            // освободился, блокируем
            mutex->owner = thrd_current();
            return thrd_success;
        }
        if(timespec) {
            // время ожидания истекло
            if(time(NULL) > timespec) return thrd_timedout;
        }
        // передаём управление другому потоку
        // для неосновного потока можно организовать перевод в состояние sleep...
        thrd_yield();
    }
}

/**
 * @brief Попытка блокировки указанного мьютекса без блокировки текущего потока.
 * Если указанный мьютекс заблокирован, то функция возвращает значение thrd_busy.
 * В противном случе функция блокирует мьютекс и возвращает значение thrd_success.
 * Функция может вернуть значения thrd_success, thrd_busy или thrd_error.
 * @param mutex Указатель на идентификатор мьютекса
 * @return int 
 */
int	mtx_trylock(mtx_t *mutex) {
    long res;
    //
    if(!mutex) return thrd_error;
    //
    res = __atomic_swap(mutex, mtx_locked);
    // 0 - успешно
    // 1 - уже ранее был заблокирован
    if(res == mtx_unlocked) {
        // мьютекс успешно заблокирован
        mutex->owner = thrd_current();
        return thrd_success;
    }
    return thrd_busy;
}

/**
 * @brief Функция разблокирует указанный мьютекс.
 * Если мьютекс не находился в заблокированном состоянии функция возвращает thrd_error.
 * Функция может вернуть значения thrd_success или thrd_error.
 * @param mutex Указатель на идентификатор мьютекса
 * @return int 
 */
int	mtx_unlock(mtx_t *mutex) {
    long res;
    //
    if(!mutex) return thrd_error;
    //
    res = __atomic_swap(mutex, mtx_unlocked);
    //
    // если мьютекс не находился в заблокированном состоянии возвращаем thrd_error
    return (res == mtx_unlocked)?thrd_error:thrd_success;
}
