/**
 * @file gmttime.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с временем.
 * @version 0.1
 * @date 2024-03-10
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include "stddef.h"
#include "time.h"

/* Converts a Unix timestamp (number of seconds since the beginning of 1970
 * CE) to a Gregorian civil date-time tuple in GMT (UTC) time zone.
 *
 * This conforms to C89 (and C99...) and POSIX.
 *
 * This implementation works, and doesn't overflow for any sizeof(time_t).
 * It doesn't check for overflow/underflow in tm->tm_year output. Other than
 * that, it never overflows or underflows. It assumes that that time_t is
 * signed.
 *
 * This implements the inverse of the POSIX formula
 * (http://pubs.opengroup.org/onlinepubs/9699919799/basedefs/V1_chap04.html#tag_04_15)
 * for all time_t values, no matter the size, as long as tm->tm_year doesn't
 * overflow or underflow. The formula is: 
 * tm_sec + tm_min*60 + tm_hour*3600 + tm_yday*86400 + (tm_year-70)*31536000 + ((tm_year-69)/4)*86400 - ((tm_year-1)/100)*86400 + ((tm_year+299)/400)*86400
 * 
 * https://github.com/pts/minilibc686/blob/master/fyi/c_gmtime.c
 * 
 */
struct tm *gmtime_r(const time_t *timep, struct tm *tm) {
    const time_t ts = *timep;
    //
    time_t t = ts / 86400;
    unsigned hms = ts % 86400;                  // -86399 <= hms <= 86399. This needs sizeof(int) >= 4.
    time_t c;
#if SIZE_OF_TIME_T > 4
    time_t f;
#endif
    unsigned yday;                              // 0 <= yday <= 426. Also fits to an `unsigned short', but `int' is faster.
    unsigned a;                                 // 0 <= a <= 2133. Also fits to an `unsigned short', but `int' is faster.
    //
    if ((int)hms < 0) {
        // Fix quotient and negative remainder if ts was negative (i.e. before year 1970 CE).
        --t; hms += 86400; 
    }
    // Now: -24856 <= t <= 24855.
    tm->tm_sec  = hms % 60;
    hms /= 60;
    tm->tm_min  = hms % 60;
    tm->tm_hour = hms / 60;

#if SIZE_OF_TIME_T > 4
//  if (sizeof(time_t) > 4) {                   // Optimization. For int32_t, this would keep t intact, so we won't have to do it. This produces unreachable code.
        f = (t + 4) % 7;
        if (f < 0) f += 7;                      // Fix negative remainder if (t + 4) was negative.
        /* Now 0 <= f <= 6. */
        tm->tm_wday = f;
        c = (t << 2) + 102032;
        f = c / 146097;
        if (c % 146097 < 0) --f;                // Fix negative remainder if c was negative.
        --f;
        t += f;
        f >>= 2;
        t -= f;
        f = (t << 2) + 102035;
        c = f / 1461;
        if (f % 1461 < 0) --c;                  // Fix negative remainder if f was negative.
#else   //    } else {
        tm->tm_wday = (t + 24861) % 7;  /* t + 24861 >= 0. */
        /* Now: -24856 <= t <= 24855. */
        c = ((t << 2) + 102035) / 1461;
#endif

//    }
    yday = t - 365 * c - (c >> 2) + 25568;
    // Now: 0 <= yday <= 425.
    a = yday * 5 + 8;
    // Now: 8 <= a <= 2133.
    tm->tm_mon = a / 153;
    a %= 153;  /* No need to fix if a < 0, because a cannot be negative here. */
    // Now: 2 <= tm->tm_mon <= 13.
    // Now: 0 <= a <= 152.
    tm->tm_mday = 1 + a / 5;  /* No need to fix if a < 0, because a cannot be negative here. */
    // Now: 1 <= tm->tm_mday <= 31.
    if (tm->tm_mon >= 12) {
        tm->tm_mon -= 12;
        // Now: 0 <= tm->tm_mon <= 1.
        ++c;
        yday -= 366;
    } else {
        /* Check for leap year (in c). */
        /* Now: 2 <= tm->tm_mon <= 11. */
        /* 1903: not leap; 1904: leap, 1900: not leap; 2000: leap */
        /* With sizeof(time_t) == 4, we have 1901 <= year <= 2038; of these
        * years only 2000 is divisble by 100, and that's a leap year, no we
        * optimize the check to `(c & 3) == 0' only.
        */
        if (!((c & 3) == 0 && (sizeof(time_t) <= 4 || c % 100 != 0 || (c + 300) % 400 == 0))) --yday;  /* These `== 0' comparisons work even if c < 0. */
    }
    tm->tm_year = c;  /* This assignment may overflow or underflow, we don't check it. Example: time_t is a huge int64_t, tm->tm_year is int32_t. */
    /* Now: 0 <= tm->tm_mon <= 11. */
    /* Now: 0 <= yday <= 365. */
    tm->tm_yday = yday;
    tm->tm_isdst = 0;
    return tm;
}

/**
 * @brief Преобразует календарное время timep в локальное представление времени, определенное во всеобщем скоординированном времени (UTC).
 * 
 * @param timep Указатель на значение текущего календарного времени
 * @return struct tm* 
 */
struct tm *gmtime(const time_t *timep) {
    static struct tm tm;
    //
    return gmtime_r(timep, &tm);
}

/**
 * @brief Получение текущего календарного времени (GMT).
 * 
 * @param hour Час (0..23)
 * @param min Минута (0..59)
 * @param sec Секунда (0..60)
 */
void GetTime(unsigned char *hour, unsigned char *min, unsigned char *sec) {
    time_t tt = time(NULL);
    struct tm *tm = gmtime(&tt);
    //
    *hour   = tm->tm_hour;
    *min    = tm->tm_min;
    *sec    = tm->tm_sec;
}

/**
 * @brief Получение текущей календарной даты (GMT).
 * 
 * @param date День месяца (1..31)
 * @param month Месяц (1..12)
 * @param year Год (2024..)
 */
void GetDate(unsigned char *date, unsigned char *month, unsigned short *year) {
    time_t tt = time(NULL);
    struct tm *tm = gmtime(&tt);
    //
    *date   = tm->tm_mday;
    *month  = tm->tm_mon + 1;
    *year   = tm->tm_year + 1900;
}

/**
 * @brief Получение текущего календарного времени и текущей календарной даты (GMT).
 * 
 * @param date День месяца (1..31)
 * @param month Месяц (1..12)
 * @param year Год (2024..)
 * @param hour Час (0..23)
 * @param min min Минута (0..59)
 * @param sec sec Секунда (0..59)
 */
void GetDateTime(unsigned char * date,unsigned char * month,unsigned short * year,unsigned char * hour,unsigned char * min,unsigned char * sec) {
    time_t tt = time(NULL);
    struct tm *tm = gmtime(&tt);
    //
    *date   = tm->tm_mday;
    *month  = tm->tm_mon + 1;
    *year   = tm->tm_year + 1900;
    *hour   = tm->tm_hour;
    *min    = tm->tm_min;
    *sec    = tm->tm_sec;
}

/**
 * @brief Получение значения секунд из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char SecOfTime(time_t t) {
    struct tm *tm = gmtime(&t);
    //
    return tm->tm_sec;
}

/**
 * @brief Получение значения минут из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char MinOfTime(time_t t) {
    struct tm *tm = gmtime(&t);
    //
    return tm->tm_min;
}

/**
 * @brief Получение значения часа из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char HourOfTime(time_t t) {
    struct tm *tm = gmtime(&t);
    //
    return tm->tm_hour;
}

/**
 * @brief Получение значения номера дня месяца Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char DayOfTime(time_t t) {
    struct tm *tm = gmtime(&t);
    //
    return tm->tm_mday;
}

/**
 * @brief Получение значения номера месяца из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char MonthOfTime(time_t t) {
    struct tm *tm = gmtime(&t);
    //
    return tm->tm_mon + 1;
}

/**
 * @brief Получение значения номера года из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned short int 
 */
unsigned short int YearOfTime(time_t t) {
    struct tm *tm = gmtime(&t);
    //
    return tm->tm_year + 1900;
}

/**
 * @brief Получение значения номера дня недели из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char WeekDayOfTime(time_t t) {
    struct tm *tm = gmtime(&t);
    //
    return tm->tm_wday;
}

/**
 * @brief Получение значения номера дня года из Unix Timestamp (GMT).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned short int 
 */
unsigned short int YearDayOfTime(time_t t) {
    struct tm *tm = gmtime(&t);
    //
    return tm->tm_yday;
}

