/**
 * @file localtime.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с временем.
 * @version 0.1
 * @date 2024-03-10
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include "stddef.h"
#include "time.h"

/**
 * @brief Преобразует календарное время timep в локальное представление времени, определенное в локальной временной зоне.
 * 
 * @param timep Указатель на значение текущего календарного времени
 * @return struct tm* 
 */
struct tm *localtime_r(const time_t *timep, struct tm *result) {
    time_t loc_time;
    //
    tzset();
    loc_time = *timep - timezone;
    return gmtime_r(&loc_time, result);
}

/**
 * @brief Преобразует календарное время timep в локальное представление времени, определенное в локальной временной зоне.
 * 
 * @param timep Указатель на значение текущего календарного времени
 * @return struct tm* 
 */
struct tm *localtime(const time_t *timep) {
    static struct tm tm;
    //
    return localtime_r(timep, &tm);
}

/**
 * @brief Получение текущего календарного времени (локальная временная зона).
 * 
 * @param hour Час (0..23)
 * @param min Минута (0..59)
 * @param sec Секунда (0..60)
 */
void GetLocalTime(unsigned char *hour, unsigned char *min, unsigned char *sec) {
    time_t tt = time(NULL);
    struct tm *tm = localtime(&tt);
    //
    *hour   = tm->tm_hour;
    *min    = tm->tm_min;
    *sec    = tm->tm_sec;
}

/**
 * @brief Получение текущей календарной даты (локальная временная зона).
 * 
 * @param date День месяца (1..31)
 * @param month Месяц (1..12)
 * @param year Год (2024..)
 */
void GetLocalDate(unsigned char *date, unsigned char *month, unsigned short *year) {
    time_t tt = time(NULL);
    struct tm *tm = localtime(&tt);
    //
    *date   = tm->tm_mday;
    *month  = tm->tm_mon + 1;
    *year   = tm->tm_year + 1900;
}

/**
 * @brief Получение текущего календарного времени и текущей календарной даты (локальная временная зона).
 * 
 * @param date День месяца (1..31)
 * @param month Месяц (1..12)
 * @param year Год (2024..)
 * @param hour Час (0..23)
 * @param min min Минута (0..59)
 * @param sec sec Секунда (0..59)
 */
void GetLocalDateTime(unsigned char * date,unsigned char * month,unsigned short * year,unsigned char * hour,unsigned char * min,unsigned char * sec) {
    time_t tt = time(NULL);
    struct tm *tm = localtime(&tt);
    //
    *date   = tm->tm_mday;
    *month  = tm->tm_mon + 1;
    *year   = tm->tm_year + 1900;
    *hour   = tm->tm_hour;
    *min    = tm->tm_min;
    *sec    = tm->tm_sec;
}

/**
 * @brief Получение значения секунд из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char SecOfLocalTime(time_t t) {
    struct tm *tm = localtime(&t);
    //
    return tm->tm_sec;
}

/**
 * @brief Получение значения минут из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char MinOfLocalTime(time_t t) {
    struct tm *tm = localtime(&t);
    //
    return tm->tm_min;
}

/**
 * @brief Получение значения часа из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char HourOfLocalTime(time_t t) {
    struct tm *tm = localtime(&t);
    //
    return tm->tm_hour;
}

/**
 * @brief Получение значения номера дня месяца Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char DayOfLocalTime(time_t t) {
    struct tm *tm = localtime(&t);
    //
    return tm->tm_mday;
}

/**
 * @brief Получение значения номера месяца из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char MonthOfLocalTime(time_t t) {
    struct tm *tm = localtime(&t);
    //
    return tm->tm_mon + 1;
}

/**
 * @brief Получение значения номера года из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned short int 
 */
unsigned short int YearOfLocalTime(time_t t) {
    struct tm *tm = localtime(&t);
    //
    return tm->tm_year + 1900;
}

/**
 * @brief Получение значения номера дня недели из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned char 
 */
unsigned char WeekDayOfLocalTime(time_t t) {
    struct tm *tm = localtime(&t);
    //
    return tm->tm_wday;
}

/**
 * @brief Получение значения номера дня года из Unix Timestamp (локальная временная зона).
 * 
 * @param t Количество секунд с 01-01-1970 (GMT)
 * @return unsigned short int 
 */
unsigned short int YearDayOfLocalTime(time_t t) {
    struct tm *tm = localtime(&t);
    //
    return tm->tm_yday;
}

