/**
 * @file mktime.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы с временем.
 * @version 0.1
 * @date 2024-03-10
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 * 
 */

#include "time.h"

static int get_yday(int mon, int day) {
    //                           Янв Фвр Мрт Апр Май Июн Июл Авг Снт Окт Нбр Дек
    //                           31  28  31  30  31   30   31   31   30   31   30   31
    static const int days[13] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};
    return days[mon] + day;
}

#define SEC_PER_DAY 86400

/**
 * @brief // Преобразует местное время, заданное структурой *tm, в календарное
 * 
 * @param tp Местное время
 * @return time_t 
 */
time_t mktime(struct tm *tm) {
    time_t t;
    //
    int yday = get_yday(tm->tm_mon, tm->tm_mday);
    //
    // t = tm->tm_sec + tm->tm_min*60 + tm->tm_hour*3600 + tm->tm_yday*SEC_PER_DAY + (tm->tm_year-70)*31536000 + ((tm->tm_year-69)/4)*SEC_PER_DAY - ((tm->tm_year-1)/100)*SEC_PER_DAY + ((tm->tm_year+299)/400)*SEC_PER_DAY;
    t = tm->tm_sec + tm->tm_min*60 + tm->tm_hour*3600 + yday*SEC_PER_DAY + (tm->tm_year-70)*31536000 + ((tm->tm_year-69)/4)*SEC_PER_DAY - ((tm->tm_year-1)/100)*SEC_PER_DAY + ((tm->tm_year+299)/400)*SEC_PER_DAY;
    //
    return t;
}

/**
 * @brief Получение значения Unix Timestamp из заданных значений времени и даты (GMT).
 * 
 * @param hour Час (0-23)
 * @param min Минута (0-59)
 * @param sec Секунда (0-59)
 * @param day День месяца (1-31)
 * @param month Номер месяца(1-12)
 * @param year Год (1970...)
 * @return time_t Количество секунд с 01-01-1970 (GMT)
 */
time_t MkTime(unsigned char hour,unsigned char min,unsigned char sec,unsigned char day,unsigned char month,unsigned short int year) {
    struct tm tm;
    //
    tm.tm_hour = hour;
    tm.tm_min  = min;
    tm.tm_sec  = sec;
    tm.tm_mday = day;
    tm.tm_mon  = month - 1;
    tm.tm_year = year - 1900;
    //
    return mktime(&tm);
}

/**
 * @brief Получение значения Unix Timestamp из заданных значений времени и даты (локальная временная зона).
 * 
 * @param hour Час (0-23)
 * @param min Минута (0-59)
 * @param sec Секунда (0-59)
 * @param day День месяца (1-31)
 * @param month Номер месяца(1-12)
 * @param year Год (1970...)
 * @return time_t Количество секунд с 01-01-1970 (GMT)
 */
time_t MkLocalTime(unsigned char hour,unsigned char min,unsigned char sec,unsigned char day,unsigned char month,unsigned short int year) {
    time_t tt;
    //
    tzset();
    tt = MkTime(hour, min, sec, day, month, year) + timezone;
    return tt;
}
