/**
 * @file wcstombs.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы со строками в кодировках UTF8/16/32.
 * @version 0.1
 * @date 2024-05-11
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <stdlib.h>

#include <wchar.h>

/**
 * @brief Возвращает длину строки UTF8, которая необходима для преобразования в неё строки UTF32.
 * @param pwcs Указатель на строку UTF32.
 * @return size_t 
 */
static size_t StrLenUtf32AsUtf8(const wchar_t* pwcs) {
    size_t ztUtf32 = 0;
    size_t ztNumUtf8Chars = 0;
    while (pwcs[ztUtf32] != 0) {
        if (pwcs[ztUtf32] >= 0x10000u) ztNumUtf8Chars += 3;
        else 
        if (pwcs[ztUtf32] >= 0x800u) ztNumUtf8Chars += 2;
        else 
        if (pwcs[ztUtf32] >= 0x80u) ztNumUtf8Chars += 1;
        //
        ztNumUtf8Chars++;
        ztUtf32++;
    }
    return ztNumUtf8Chars;
}

/**
 * @brief Функция преобразует строку UTF32 в строку UTF8.
 * Функция выделяет память для строки UTF8, выполняет преобразование и возвращает указатель на строку UTF8.
 * Выделенная память должна быть в последствии освобождена функцией free.
 * @param pwcs Указатель на строку UTF32.
 * @return char* 
 */
char* wcstombs(const wchar_t* pwcs) {
    size_t ztUtf32 = 0;
    char* pmbc = calloc(StrLenUtf32AsUtf8(pwcs) + 1, sizeof(char));
    if (pmbc) {
        // https://datatracker.ietf.org/doc/html/rfc3629#page-4
        while (pwcs[ztUtf32] != 0) {
            if (pwcs[ztUtf32] <= 0x7Fu) {
                pmbc[strlen((const char*)pmbc)] = (char)pwcs[ztUtf32];
            }
            else if (pwcs[ztUtf32] <= 0x7FFu) {
                pmbc[strlen((const char*)pmbc)] = (char)(0xC0u | (pwcs[ztUtf32] >> 6));             /* 110xxxxx */
                pmbc[strlen((const char*)pmbc)] = (char)(0x80u | (pwcs[ztUtf32] & 0x3Fu));          /* 10xxxxxx */
            }
            else if (pwcs[ztUtf32] <= 0xFFFFu) {
                pmbc[strlen((const char*)pmbc)] = (char)(0xE0u | (pwcs[ztUtf32] >> 12));            /* 1110xxxx */
                pmbc[strlen((const char*)pmbc)] = (char)(0x80u | ((pwcs[ztUtf32] >> 6) & 0x3Fu));   /* 10xxxxxx */
                pmbc[strlen((const char*)pmbc)] = (char)(0x80u | (pwcs[ztUtf32] & 0x3Fu));          /* 10xxxxxx */
            }
            else if (pwcs[ztUtf32] <= 0x10FFFFu) {
                pmbc[strlen((const char*)pmbc)] = (char)(0xF0u | (pwcs[ztUtf32] >> 18));            /* 11110xxx */
                pmbc[strlen((const char*)pmbc)] = (char)(0x80u | ((pwcs[ztUtf32] >> 12) & 0x3Fu));  /* 10xxxxxx */
                pmbc[strlen((const char*)pmbc)] = (char)(0x80u | ((pwcs[ztUtf32] >> 6) & 0x3Fu));   /* 10xxxxxx */
                pmbc[strlen((const char*)pmbc)] = (char)(0x80u | (pwcs[ztUtf32] & 0x3Fu));          /* 10xxxxxx */
            }
            ztUtf32++;
        }
        return pmbc;
    }
    return NULL;
}
