/**
 * @file mbstowcs.c
 * @brief Стандартная библиотека контроллера "Сатурн-PLC".
 * Функции работы со строками в кодировках UTF8/16/32.
 * @version 0.1
 * @date 2024-05-11
 * 
 * @copyright Copyright (C) 2024, ООО "МНПП Сатурн"
 */

#include <stdlib.h>

#include <wchar.h>

/**
 * @brief Функция преобразует строку UTF8 в строку UTF32.
 * Функция выделяет память для строки UTF32, выполняет преобразование и возвращает указатель на строку UTF32.
 * Выделенная память должна быть в последствии освобождена функцией free.
 * @param pmbc Указатель на строку UTF8.
 * @return wchar_t* 
 */
wchar_t* mbstowcs(const char* pmbc) {
    size_t ztUtf8 = 0;
    size_t ztUtf32 = 0;
    wchar_t* pwcs = calloc(mbslen(pmbc) + 1, sizeof(wchar_t));
    if (pwcs) {
        // https://datatracker.ietf.org/doc/html/rfc3629#page-4
        while (pmbc[ztUtf8] != 0) {
            if ((pmbc[ztUtf8] & 0b10000000) == 0) {
                // 1 byte code point, ASCII
                pwcs[ztUtf32] = (pmbc[ztUtf8] & 0b01111111);
                ztUtf8 += 1;
            }
            else if ((pmbc[ztUtf8] & 0b11100000) == 0b11000000) {
                // 2 byte code point
                pwcs[ztUtf32] = (pmbc[ztUtf8] & 0b00011111) << 6 | (pmbc[ztUtf8 + 1] & 0b00111111);
                if (pwcs[ztUtf32] < 0x80u)            // Not a valid result, Wrong encoding
                    pwcs[ztUtf32] = 0;                // Out of UTF8 bound, skip data  
                else if (pwcs[ztUtf32] > 0x7ffu)      // Not a valid result, Wrong encoding
                    pwcs[ztUtf32] = 0;                // Out of UTF8 bound, skip data 
                ztUtf8 += 2;
            }
            else if ((pmbc[ztUtf8] & 0b11110000) == 0b11100000) {
                // 3 byte code point
                pwcs[ztUtf32] = (pmbc[ztUtf8] & 0b00001111) << 12 | (pmbc[ztUtf8 + 1] & 0b00111111) << 6 | (pmbc[ztUtf8 + 2] & 0b00111111);
                if (pwcs[ztUtf32] < 0x800u)           // Not a valid result, Wrong encoding
                    pwcs[ztUtf32] = 0;                // Out of UTF8 bound, skip data  
                else if (pwcs[ztUtf32] > 0xffffu)     // Not a valid result, Wrong encoding
                    pwcs[ztUtf32] = 0;                // Out of UTF8 bound, skip data  
                ztUtf8 += 3;
            }
            else {
                // 4 byte code point
                if (pmbc[ztUtf8] <= 0xf4u)             // RFC3629 UTF8 do not allow larger values
                    pwcs[ztUtf32] = (pmbc[ztUtf8] & 0b00000111) << 18 | (pmbc[ztUtf8 + 1] & 0b00111111) << 12 | (pmbc[ztUtf8 + 2] & 0b00111111) << 6 | (pmbc[ztUtf8 + 3] & 0b00111111);
                else                                    // Not a valid result, Wrong encoding
                    pwcs[ztUtf32] = 0;                // Out of UTF8 bound, skip data 
                if (pwcs[ztUtf32] < 0x10000u)         // Not a valid result, Wrong encoding
                    pwcs[ztUtf32] = 0;                // Out of UTF8 bound, skip data  
                else if (pwcs[ztUtf32] > 0x10FFFFu)   // Not a valid result, Wrong encoding 
                    pwcs[ztUtf32] = 0;                // Out of UTF8 bound, skip data  
                ztUtf8 += 4;
            }
            if (pwcs[ztUtf32]) // Only step when valid data
                ztUtf32++;
        }
        return pwcs;
    }
    return NULL;
}
